/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Création du fichier : 4 janvier 1999

   Module de gestion d'une structure chimique
   Chargement des données structurales d'un fichier de la Protein Data Bank
   Créé à partir des spécifications décrites sur le site officiel :

http://www.rcsb.org/pdb/docs/format/pdbguide2.2/guide2.2_frame.html

*/

#undef __STRICT_ANSI__

#include <math.h> /* Pour log10 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h> /* Pour toupper() */

#include "systeme.h"

#include "atomes.h"
#include "cristallo.h"
#include "fichiers.h"
#include "pdb.h"

/* Les textes en différentes langues */

#include "lit_pdb.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int   PDB_MOL_chaine (PDB_Chaine *chaine, PDB_Chaine *premiere, boolean fondre_chaines, int ref_code, const char *titre);
static char *PDB_creer_titre(char *base, char *code, int chaine, unsigned long modele);
static char *PDB_propose_chaine(char *ligne);
static void  PDB_termine    (PDB_Chaine *premiere_chaine, PDB_hetero *premier_groupe);
static void  PDB_annule     (PDB_Chaine *premiere_chaine, PDB_hetero *premier_groupe);

#define PDB_LIGNE_OK     0
#define PDB_LIGNE_ERREUR 1
#define PDB_LIGNE_FINIR  2

static char *PDB_lit_chaine(FILE *fichier, char *ligne, PDB_infos *infos_pdb, PDB_Chaine *chaine);
static char *PDB_lit_residu(FILE *fichier, char *ligne, PDB_infos *infos_pdb,
			    PDB_Chaine *chaine, int num_residu, Fragment *residu);
static char *PDB_lit_atome (FILE *fichier, char *ligne, PDB_infos *infos_pdb,
			    PDB_Chaine *chaine, int num_residu, Fragment *residu, int num_atome);

static int quel_aa(char *symbole, int *nbr_H, int *nbr_l, char **nom);

static int lit_pdb(FILE *fichier, const char *nom_fichier, PDB_infos *infos);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

int MOL_Import_PDB(const char *nom_fichier, struct infos_PDB *options)
{
  int retour;
  FILE *fichier;

  if ( ( NULL == nom_fichier ) || ( NULL == options ) ) return -1;

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "rb" );
  if ( NULL == fichier ) {
    (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier , NULL, NULL);
    return -1;
  }

  /* Si une base de nom est fournie, on en fait une copie pour faciliter le travail */
  if ( options->base_nom != NULL ) {
    if ( options->base_nom[0] != 0 ) { /* Le nom fourni n'est pas vide */
      options->base_nom = strdup( options->base_nom );
    } else {
      options->base_nom = NULL; /* Nom vide : pas de nom ! */
    }
  }

  /* Log dans l'historique */
  HIST_Titre(1, "Import d'une structure PDB", "E/S");
  HIST_Texte(1, 0, "Fichier : %s\n", nom_fichier );

  retour = lit_pdb( fichier, nom_fichier, options );
  (void) fclose( fichier ); /* Plus l'usage */

  if ( retour != 0 ) { /* Problème au chargement */
    (void) fprintf(stderr, "charge_pdb : erreur au chargement [code %d]\n", retour);
  }

  return retour;
}

/* -------- Lecture proprement dite du fichier ------------------------------\
 |   Renvoie 0 si tout va bien, <0 en cas d'erreur                           |
 |                                                                           |
 | fichier        = le fichier dans lequel il faut lire                      |
 | nom_fichier    = le nom de ce fichier (pour les alertes)                  |
 | tous_modeles   = VRAI si l'on veut charger tous les modèles du fichier    |
 | fondre_chaines = VRAI si l'on veut fondre toutes les chaînes en une seule |
 | nom_base       = utilisé pour construire le nom des molécules chargées    |
 \-------------------------------------------------------------------------- */

static int lit_pdb(FILE *fichier, const char *nom_fichier, PDB_infos *infos)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};
  char *ligne;

  /* Vérification de l'en-tête : doit commencer par HEADER */
  ligne = lit_ligne(fichier, 80); /* Une ligne PDB doit faire 80 octets... */
  if ( NULL == ligne ) {
    (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, "Premi"e_GRAVE"re ligne illisible");
    return -2;
  }
  strncpy(commande, ligne, 6);

  if ( strcmp(commande, "HEADER") != 0 ) {
    int retour = BI_ERR_alerte(NULL, &PDB_err_HEADER, VRAI);
    if (3 == retour) /* On annule le chargement */
      return -1;

    if (2 == retour) { /* On ignore toutes les lignes jusqu'au HEADER */
      do {
	free(ligne);
	ligne = lit_ligne(fichier, 80);
	if (NULL == ligne) {
	  (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
	  return -2;
	}
	strncpy(commande, ligne, 6);
      } while ( strcmp(commande, "HEADER") != 0 );
    }
  }

  do {
    ligne = PDB_lit_Titre(fichier, ligne, infos); /* Section « Title » doit être la première */
    if (ligne == NULL) {
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Primaire(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      PDB_annule( infos->premiere_chaine, infos->premier_groupe );
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Heterogene(fichier, ligne, infos); /* Section sur la structure primaire */
    if ( ligne == NULL ) {
      PDB_annule( infos->premiere_chaine, infos->premier_groupe );
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Secondaire(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      PDB_annule( infos->premiere_chaine, infos->premier_groupe );
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Liaisons(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      PDB_annule( infos->premiere_chaine, infos->premier_groupe );
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Divers(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      PDB_annule( infos->premiere_chaine, infos->premier_groupe );
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Cristallo(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      PDB_annule( infos->premiere_chaine, infos-> premier_groupe);
      if ( infos->code_retour != PDB_LIGNE_FINIR ) /* Pas de message déjà affiché */
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      return -2;
    }
    ligne = PDB_lit_Coordonnees(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      if ( infos->code_retour == PDB_LIGNE_ERREUR ) { /* Erreur au chargement, on annule tout */
	PDB_annule( infos->premiere_chaine, infos->premier_groupe );
	return -2;
      }
    }
    ligne = PDB_lit_TableLiaisons(fichier, ligne, infos); /* Section sur la structure primaire */
    if (ligne == NULL) {
      if ( infos->code_retour == PDB_LIGNE_ERREUR ) { /* Erreur au chargement, on annule tout */
	PDB_annule( infos->premiere_chaine, infos->premier_groupe );
	return -2;
      }
    }
    ligne = PDB_lit_Fin(fichier, ligne, infos); /* Section sur la structure primaire */
    if ( ligne != NULL ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Lignes superflues après la fin de fichier [%s] !\n",
		     __FILE__, __FUNCTION__, __LINE__, ligne);
#endif
      if ( 0 == *ligne ) { /* Fin de fichier atteinte de façon imprévue ! */
	infos->code_retour = PDB_LIGNE_FINIR;
	(void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
	free( ligne );
	ligne = NULL;
      } else {
	(void) BI_ERR_alerte(NULL, &PDB_err_inconnu, VRAI, commande);
	free(ligne);
	ligne = lit_ligne(fichier, 80);
	if (ligne == NULL) {
	  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
	}
      }
    } else {
      /* ligne = NULL : on termine la construction si la dernière ligne était OK... Mais la commande END manquait (déjà affiché ?) */
      if ( PDB_LIGNE_OK == infos->code_retour ) {
	infos->code_retour = PDB_LIGNE_FINIR;
/* 	(void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande); */
      }
    }
  } while (ligne != NULL);

  free( infos->base_nom ); infos->base_nom = NULL;

  if ( infos->code_retour == PDB_LIGNE_FINIR ) {
    PDB_termine( infos->premiere_chaine, infos->premier_groupe );
  } else {
    PDB_annule( infos->premiere_chaine, infos->premier_groupe );
  }

  return 0;
}

/* -------- Lecture de la section Title -------- */

char *PDB_lit_Titre(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp(commande, "AUTHOR") == 0 ) { /* Auteur(s) */
      /*      printf("Auteur : %s", ligne + 10); */
    } else if ( strcmp(commande, "CAVEAT") == 0 ) {
      size_t j;

      j = strlen(ligne) - 1;
      while ( isspace(ligne[j]) ) {
	ligne[j] = 0;
	j--;
      }
      (void) BI_ERR_alerte(NULL, &PDB_err_CAVEAT, VRAI, ligne + 19); /* Avant : code... */
    } else if ( strcmp(commande, "COMPND") == 0 ) { /* Composé. Si pas encore de nom, utilisé pour ça */
#ifdef _DEVELOPPE_
      printf("COMPND / Composé : %s", ligne + 10);
#endif
      if (infos_pdb->base_nom == NULL) {
	infos_pdb->base_nom = PDB_propose_chaine(ligne + 10);
      }
    } else if ( strcmp(commande, "EXPDTA") == 0 ) { /* Technique de détermination de la structure */
      /* Liste de techniques séparée par ; - commence en + 10 */
      /* IL FAUDRAIT DÉDUIRE LE NOMBRE DE MODÈLES */
    } else if ( strcmp(commande, "FTNOTE") == 0 ) { /* Note - Plus supporté ? Pas dans la doc... */
      /*      printf("Note : %s", contenu); */
    } else if ( strcmp(commande, "HEADER") == 0 ) {
      int i;

      if (infos_pdb->base_nom == NULL) { /* Pas de nom proposé => On utilise le nom du fichier */
	infos_pdb->base_nom = PDB_propose_chaine(ligne + 10);
	if (infos_pdb->base_nom != NULL) { /* On a bien créé quelque chose */
	  size_t j;

	  infos_pdb->base_nom[40] = 0;
	  j = strlen(infos_pdb->base_nom);
	  while ( infos_pdb->base_nom[j - 1] == ' ') { /* On fait sauter les éventuels espaces finaux */
	    infos_pdb->base_nom[j - 1] = 0;
	    j--;
	  }
	  if ( infos_pdb->base_nom[0] == 0) {
	    free(infos_pdb->base_nom);
	    infos_pdb->base_nom = NULL;
	  }
	}
      }

      /* Lecture du code : colonne */
      if ( isalnum( ligne[62] ) ) { /* Code PDB : chiffre ou lettres */
	for (i = 0 ; i < 4 ; i++)
	  infos_pdb->code[i] = ligne[62 + i];
	fprintf(stderr, "Code de la protéine : %s\n", infos_pdb->code);
      } else {
	fprintf(stderr, "PDB_lit_Titre : ligne HEADER sans code pour le composé !\n");
      }
    } else if ( strcmp(commande, "JRNL  ") == 0 ) {
      /*      printf("Journal : %s", ligne + 12); */
    } else if ( strcmp(commande, "KEYWDS") == 0 ) { /* Une série de mot-clefs */
      /*      printf("Les mot-clefs : %s", ligne + 10); */
    } else if ( strcmp(commande, "OBSLTE") == 0 ) {
      char nv_code[5];
      int choix;
      /* Premier code de remplacement [colonnes 32-35] */
      strncpy(nv_code, ligne + 31, 4);
      nv_code[5] = 0;
      choix = BI_ERR_alerte(NULL, &PDB_err_OBSLTE, VRAI, nv_code);
      if (choix == 2) {
	infos_pdb->code_retour = PDB_LIGNE_FINIR;
	free(ligne);
	return NULL;
      }
    } else if ( strcmp(commande, "REMARK") == 0 ) {
      int code_remarque;
      float format;

/*      if ( *(ligne + 7) == ' ' ) {
	fprintf(stderr, "PDB_lit_Titre : remarque sans code [%s], donne %d\n", ligne, atoi(ligne + 7));
      } */
      code_remarque = atoi(ligne + 7);

      switch (code_remarque) {
      case 1: break; /* Publis */
      case 2: break; /* Infos sur la résolution */
      case 3: break; /* Programme de résolution */
      case 4: /* Format de fichier */
	free(ligne); /* Première ligne vide... */
	ligne = lit_ligne(fichier, 80);
	if ( ligne == NULL ) return NULL;
	if ( strncmp(ligne, "REMARK", 6) != 0 ) {
	  fprintf(stderr, "LIT_PDB : remarque 4 / format incorrect\n");
	  return ligne;
	}
	format = atof( ligne + 40 );
	printf("Format de fichier : %f\n", format);
	break;

      case 5: /* Avertissement */
	free(ligne); /* Première ligne vide */
	ignore_ligne(fichier); /* Deuxième ligne : texte "WARNING" */
	ligne = lit_ligne(fichier, 80);
	code_remarque = atoi(ligne + 7);
	if (code_remarque == 5) {
	  while (code_remarque == 5) {
	    /* A FAIRE */
	    break;
	  }
	} else {
	  fprintf(stderr, "PDB : ligne d'avertissement sans avertissement !\n");
	}
	break;

	/* Infos diverses sur les ADN/ARN : on ignore */
      case 100: break; /* Code NDB */
      case 101: break; /* Nucléotides substitués */
      case 102: break; /* Appariements inhabituels */
      case 103: break; /* Appariements inhabituels de l'inosine */
      case 104: break; /* Inosines */
      case 105: break; /* Nomenclature */
      case 106: break; /* Appariements classiques */

	/* Infos sur les conditions expérimentales */
      case 200: break; /* Source de RX, cristal,... */
      case 205: break; /* Diffraction de fibres */
      case 210: break; /* RMN : conditions expérimentales */
      case 215: break; /* RMN */
      case 217: break; /* RMN du solide */
      case 220: break; /* Modèle théorique */
      case 225: break; /* Modèle théorique */
      case 230: break; /* Diffraction des neutrons */
      case 240: break; /* Diffraction d'électrons */
      case 250: break; /* Autre type d'expérience */
      case 280: break; /* Caractéristiques du cristal */
      case 285: break; /* Paramètres du cristal */
      case 290: break; /* Symétries cristallines */
      case 295: break; /* Autres symétries */

	/* Informations sur la molécule */
      case 300: break;
      case 350: break; /* Obtention de la molécule */
      case 375: break; /* Positions spéciales */
      case 400: break;
      case 450: break;
      case 460: break;
      case 470: /* Atomes manquants */
	sauter_lignes(fichier, 4); /* Ne servent à rien */
	fscanf(fichier, "%*s %d", &code_remarque);
	while (code_remarque == 470) {
	  char code_aa[4] = {0, 0, 0,0}, code_chaine;
	  unsigned int num_aa;
	  
	  ignore_car(fichier, 5); /* Normalement, que des espaces */
	  code_aa[0] = fgetc(fichier); /* Code du résidu */
	  code_aa[1] = fgetc(fichier);
	  code_aa[2] = fgetc(fichier);
	  if (fscanf(fichier, "%c %u", &code_chaine, &num_aa) != 2) {
	    (void) SYS_ERREUR(ERREUR_LECTURE, NULL, fichier, "Remarque code 470 incorrecte");
	    break;
	  } else {
	    (void) BI_ERR_alerte(NULL, &PDB_err_REMARK470, VRAI, num_aa, code_chaine, code_aa);
	  }
	  ignore_ligne(fichier);
	  fscanf(fichier, "%*s %d", &code_remarque);	  
	}
	ignore_ligne(fichier);
	break;

      case 500: break; /* Géométrie */
      case 525: break; /* Solvant */
      case 550: break; /* Lié à SEGID */
      case 600: break; /* Infos sur les hétérogènes */
      case 650: break; /* Infos sur les hélices */
      case 700: break; /* Infos sur les feuillets */
      case 750: break; /* Infos sur les coudes */
      case 800: break; /* Infos sur le site actif */
      case 850: break; /* Corrections avant publication */
      case 860: break; /* Corrections après publication */
      case 900: break; /* Fichiers PDB associés */
      case 999: break; /* Infos sur la séquence */

      case 0: /* Certains fichiers contiennent REMARK XXX ..., XXX = code du groupe hétérogène
	         (Fichiers du site HIC-Up) ==> on en déduit un code, s'il n'y en a pas */
	if ( (infos_pdb->code[0] == 0) && ( *(ligne + 7) >= 'A') ) {
	  unsigned int i;

	  for (i = 0 ; i < 3 ; i++)
	    infos_pdb->code[i] = ligne[7 + i];
	  infos_pdb->code[3] = 0;
#ifdef _DEVELOPPE_
	  printf("Code du composé : %s\n", infos_pdb->code);
#endif
	}
	break;

      default:
	if ( (code_remarque >= 6) && (code_remarque < 100) )
	  fprintf(stderr, "PDB : Remarque non-codifiée : code %d\n", code_remarque);
	else
	  fprintf(stderr, "PDB : code de remarque inconnu : %d\n", code_remarque);
      }
    } else if ( strcmp(commande, "REVDAT") == 0 ) { /* Dates de modifications */
      /*      printf("Date de sortie : %s", contenu + 7); */
    } else if ( strcmp(commande, "SOURCE") == 0 ) { /* Source de la molécule */
      /*      printf("Origine : %s", ligne + 10); */
    } else if ( strcmp(commande, "SPRSDE") == 0 ) { /* Remplace... */
    } else if ( strcmp(commande, "TITLE ") == 0 ) {
#ifdef _DEVELOPPE_
      printf("Titre de ce fichier : %s\n", ligne + 10);
#endif
      if ( infos_pdb->base_nom == NULL ) { /* Pas de titre défini */
	infos_pdb->base_nom = PDB_propose_chaine(ligne + 10);
      }
      HIST_Texte(4, 2, "TITLE : %s\n", ligne + 10);
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Titre : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  fprintf(stderr, "PDB_lit_Titre : fin de fichier atteinte !\n");
  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section Structure primaire -------- */

char *PDB_lit_Primaire(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  boolean avec_SEQRES = VRAI;
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp(commande, "DBREF ") == 0 ) { /* Références vers d'autres bases de données */
    } else if ( strcmp(commande, "FTNOTE") == 0 ) { /* Note - Plus supporté ? Pas dans la doc... */
      /*      printf("Note : %s", contenu); */
    } else if ( strcmp(commande, "MODRES") == 0 ) { /* Résidus modifiés */
    } else if ( strcmp(commande, "SEQADV") == 0 ) { /* Conflits de séquence */
      avec_SEQRES = FAUX; /* Car problèmes de numérotation... */
    } else if ( strcmp(commande, "SEQRES") == 0 ) { /* Séquence des AA */
      unsigned int num_aa;
      int i, num_ligne;
      char code_chaine;
      char *code_aa;
      PDB_Chaine *chaine;
      Fragment *fragment = NULL;

      if (avec_SEQRES == FAUX) { /* On ignore, car SEQADV indique des problèmes de numérotation */
	free(ligne);
	ligne = lit_ligne(fichier, 80);
	if (ligne == NULL) {
	  infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
	  return NULL;
	}
	continue;
      }
      
      num_ligne = atoi( ligne + 8 );
      code_chaine = ligne[11];
      if ( (infos_pdb->ref_type == 0) || (infos_pdb->ref_type == code_chaine) ) { /* On charge cette chaîne */
	chaine = trouver_chaine(&(infos_pdb->premiere_chaine), (int) code_chaine);
	if (NULL == chaine) {
	  infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	  free(ligne);
	  return NULL;
	}
	if (1 == num_ligne) { /* On commence une nouvelle chaîne */
	  char *titre;

	  if ( chaine->nbr_residus == 0 ) {
	    chaine->nbr_residus = atoi(ligne + 13);
	    printf("La séquence de la chaîne %c contient %ld acides aminés\n", chaine->code, chaine->nbr_residus);
	  }
	  if (chaine->chaine == NULL) { /* Créer la molécule */
	    titre = PDB_creer_titre(infos_pdb->base_nom, infos_pdb->code, code_chaine, 0);
	    PDB_MOL_chaine(chaine, infos_pdb->premiere_chaine, ( infos_pdb->drapeaux & PDB_OPTIONS_UNE_MOLECULE ), infos_pdb->ref_type, titre);
	    free(titre);
	    fragment = NULL;
	  }
	}
	/* On se met sur le dernier fragment, pour ajouter les nouveaux après */
	if ( chaine->chaine->fragments != NULL )
	  for ( fragment = chaine->chaine->fragments ; fragment->suivant != NULL ;
		fragment = fragment->suivant );
	else
	  fragment = NULL;
      /* Remarque : il y a au plus 13 résidus par ligne */
#ifdef _DEVELOPPE_
      /*      printf("Séquence : %s\n", ligne + 19); */
#endif
	if (num_ligne > 0) { /* Séquence inconnue si num_ligne = 0 */
	  code_aa = ligne + 19;
	  for (i = 0, num_aa = (num_ligne - 1) * 13 + i + 1  ;
	       i < 13 ; i++, code_aa += 4, num_aa++) {
	    char *nom_aa = NULL, *titre;
	    int nbr_atomes, nbr_liaisons;
	
	    if ( num_aa > chaine->nbr_residus ) break; /* Séquence finie, normalement*/
	
	    code_aa[ 3 ] = 0;
	    nbr_atomes = quel_aa( code_aa, NULL, &nbr_liaisons, &nom_aa );
	    if ( nbr_atomes <= 0 ) {
#ifdef _DEVELOPPE_
	      (void) fprintf( stderr, "%s,%s l.%u : Acide aminé inconnu ! Code = %s\n",
			      __FILE__, __FUNCTION__, __LINE__, code_aa );
#endif
	      nbr_atomes = 1; /* Arbitrairement */
	    }

	    titre = TXT_genere_chaine( "%s %u", nom_aa, num_aa );
	    fragment = FRG_creer(chaine->chaine, fragment, titre, nbr_atomes, NULL, NULL);
	    free( titre ); titre = NULL;
	    if ( NULL == fragment ) {
	      infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	      free( ligne ); ligne = NULL;
	      return NULL;
	    }
	    fragment->drapeaux.aamine = VRAI; /* C'est un acide aminé */
	    fragment->numero = num_aa;
	    if ( 1 == num_aa )
	      chaine->prem_fragment = fragment; /* Premier fragment de la séquence */
	  }
	}
      }
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Primaire : commande non reconnue [%s]\n", commande);
#endif
      return ligne;
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);

  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section Groupes prosthétiques -------- */

char *PDB_lit_Heterogene(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp(commande, "FORMUL") == 0 ) { /* Formule d'un groupe hétéro */
      /*      printf("Formule des hétéro atomes : %s", contenu); */
    } else if ( strcmp(commande, "FTNOTE") == 0 ) { /* Note - Plus supporté ? Pas dans la doc... */
      /*      printf("Note : %s", contenu); */
    } else if ( strcmp(commande, "HET   ") == 0 ) {
      char code[4] = {0, 0, 0, 0};
      char code_chaine;
      unsigned int nbr_atomes, num_residu;
      Fragment *residu = NULL;
      PDB_hetero *groupe;
      PDB_Chaine *chaine;

      code[0] = ligne[7]; code[1] = ligne[8]; code[2] = ligne[9];
      code_chaine = ligne[12];
      nbr_atomes = atoi(ligne + 20);
      num_residu = atoi(ligne + 13);

      groupe = trouver_hetero(&(infos_pdb->premier_groupe), code);
      if (groupe == NULL) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free(ligne);
	return NULL;
      }
#ifdef _DEVELOPPE_
      printf("Groupe HET, code %s : %u atomes dans la chaîne %c\n", groupe->code, nbr_atomes, code_chaine);
#endif
      if (groupe->nbr_atomes == 0) {
	groupe->nbr_atomes = nbr_atomes;
      } else {
	if (nbr_atomes != groupe->nbr_atomes) {
	  fprintf(stderr, "PDB : commande HET, deux groupes de même code avec un nombre d'atomes différent !\n");
	}
      }
      if ( NULL == groupe->nom ) {
	if ( strlen( ligne ) > 70 )
	  ligne[ 69 ] = 0;
	if ( strlen( ligne ) > 32 )
	  groupe->nom = PDB_propose_chaine(ligne + 31);
      }

      /* On tache d'ajouter le fragment correspondant dans la chaîne */
      chaine = trouver_chaine(&(infos_pdb->premiere_chaine), (int) code_chaine);
      if (chaine == NULL) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free(ligne);
	return NULL;
      }
      if ( NULL == chaine->chaine ) {
	char *titre;

	titre = PDB_creer_titre(infos_pdb->base_nom, infos_pdb->code, code_chaine, 0);
	PDB_MOL_chaine(chaine, infos_pdb->premiere_chaine, ( infos_pdb->drapeaux & PDB_OPTIONS_UNE_MOLECULE ), infos_pdb->ref_type, titre);
	free( titre ); titre = NULL;
      }
      if ( NULL == chaine->chaine ) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free( ligne ); ligne = NULL;
	return NULL;
      }
      if (chaine->chaine->fragments != NULL)
	for (residu = chaine->chaine->fragments ; residu->suivant ; residu = residu->suivant);
      else
	residu = NULL;
      residu = FRG_creer(chaine->chaine, residu, code, groupe->nbr_atomes, NULL, NULL);
      if ( NULL == residu ) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free( ligne ); ligne = NULL;
	return NULL;
      }
      residu->numero = num_residu;
      if (chaine->prem_fragment == NULL)
	chaine->prem_fragment = residu;
    } else if ( strcmp(commande, "HETNAM") == 0 ) { /* Nom d'un groupe hétéro */
      PDB_hetero *groupe;
      char code[4] = {0, 0, 0, 0};

      code[0] = ligne[11]; code[1] = ligne[12]; code[2] = ligne[13]; /* On lit le code */
      groupe = trouver_hetero(&(infos_pdb->premier_groupe), code);
      if (groupe == NULL) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free(ligne);
	return NULL;
      }
      if (groupe->nom == NULL) { /* Pas encore de nom */
	groupe->nom = PDB_propose_chaine(ligne + 15);
      }
    } else if ( strcmp(commande, "HETSYN") == 0 ) { /* Synonymes pour un groupe hétéro */
      PDB_hetero *groupe;
      char code[4] = {0, 0, 0, 0};

      code[0] = ligne[11]; code[1] = ligne[12]; code[2] = ligne[13]; /* On lit le code */
      groupe = trouver_hetero(&(infos_pdb->premier_groupe), code);
      if (groupe == NULL) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free(ligne);
	return NULL;
      }
      free(groupe->nom); /* Nom du synonyme souvent plus joli ! */
      groupe->nom = PDB_propose_chaine(ligne + 15);
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Heterogene : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section Structure secondaire -------- */

char *PDB_lit_Secondaire(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

#ifdef _DEVELOPPE_
  printf("PDB_lit_Secondaire : fichier = %p, ligne = %s, infos_pdb = %p\n", (void *) fichier, ligne, (void *) infos_pdb);
#endif

  do {
    strncpy(commande, ligne, 6);

    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );

    if ( strcmp(commande, "HELIX ") == 0 ) { /* Les hélices dans la protéine */
      int code_chaine __attribute__((__unused__)), classe __attribute__((__unused__));
      unsigned int aa_debut __attribute__((__unused__)), aa_fin __attribute__((__unused__));

      code_chaine = (int) ligne[19];
      aa_debut = atoi(ligne + 21);
      aa_fin = atoi(ligne + 33);
      classe = atoi(ligne + 38);
      /*      printf("Hélice chaîne %c : aa %u à %u, type %u\n", code_chaine, aa_debut, aa_fin, classe); */
    } else if ( strcmp(commande, "SHEET ") == 0 ) { /* Feuillets dans la protéine */
#ifdef _DEVELOPPE_
/*      printf("Feuillet bêta : %s", contenu); */
#endif
    } else if ( strcmp(commande, "TURN  ") == 0 ) { /* Coudes dans la protéine */
#ifdef _DEVELOPPE_
      /*      printf("Coude bêta : %s", contenu); */
#endif
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Secondaire : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section « infos sur les liaisons » -------- */

char *PDB_lit_Liaisons(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

#ifdef _DEVELOPPE_
  printf("PDB_lit_Liaisons : fichier = %p, ligne = %s, infos_pdb = %p", (void *) fichier, ligne, (void *) infos_pdb);
#endif

  do {
    strncpy(commande, ligne, 6);

    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );

    if ( strcmp(commande, "CISPEP") == 0 ) { /* liaisons peptidiques en cis */
    } else if ( strcmp(commande, "HYDBND") == 0 ) { /* Liaison H */
    } else if ( strcmp(commande, "LINK  ") == 0 ) { /* Liaison spéciale entre résidus */
    } else if ( strcmp(commande, "SLTBRG") == 0 ) { /* Pont salin */
    } else if ( strcmp(commande, "SSBOND") == 0 ) { /* Pont disulfure [repris dans les CONECT] */
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Liaisons : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section « divers » -------- */

char *PDB_lit_Divers(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

#ifdef _DEVELOPPE_
  printf("PDB_lit_Divers : fichier = %p, ligne = %s, infos_pdb = %p\n", (void *) fichier, ligne, (void *) infos_pdb);
#endif

  do {
    strncpy(commande, ligne, 6);

    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );

    if ( strcmp(commande, "SITE  ") == 0 ) { /* Site métallique */
#ifdef _DEVELOPPE_
      printf("Site : %s\n", ligne + 7);
#endif
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Divers : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section « Cristallo » -------- */

char *PDB_lit_Cristallo(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp(commande, "CRYST1") == 0 ) { /* Paramètres de maille */
      float a, b, c, alpha, beta, gamma;

      ligne[54] = 0; gamma = atof( ligne + 47 );
      ligne[47] = 0; beta  = atof( ligne + 40 );
      ligne[40] = 0; alpha = atof( ligne + 33 );
      ligne[33] = 0; c = atof( ligne + 24 );
      ligne[24] = 0; b = atof( ligne + 15 );
      ligne[15] = 0; a = atof( ligne + 6 );
      if (a > 1.0) { /* Il y a bien des paramètres de maille */
	PDB_Chaine *chaine;
	for (chaine = infos_pdb->premiere_chaine ; chaine ; chaine = chaine->suivante) {
	  /* Remarque : la maille pourrait déjà exister car mémorisation du groupe d'espace, mais pas encore des paramètres ! */
	  if ( chaine->chaine != NULL ) {
	    if ( NULL == chaine->chaine->maille ) {
	      chaine->chaine->maille = CRIS_CreerMaille();
	    }
	    if ( NULL != chaine->chaine->maille ) {
	      chaine->chaine->maille->a = a;
	      chaine->chaine->maille->b = b;
	      chaine->chaine->maille->c = c;
	      chaine->chaine->maille->alpha = alpha * M_PI / 180.0; /* Codés en degrés dans le fichier PDB... */
	      chaine->chaine->maille->beta  = beta  * M_PI / 180.0; /* ... mais en radians en interne */
	      chaine->chaine->maille->gamma = gamma * M_PI / 180.0;
	    }
	  }
	}
      }
    } else if ( strcmp(commande, "MTRIX1") == 0 ) { /* Opérations de symétrie */
    } else if ( strcmp(commande, "MTRIX2") == 0 ) { /* autres que le groupe d'espace */
    } else if ( strcmp(commande, "MTRIX3") == 0 ) { /* Sans intérêt actuellement */
    } else if ( strcmp(commande, "ORIGX1") == 0 ) { /* Transformation (x,y,z) PDB */
    } else if ( strcmp(commande, "ORIGX2") == 0 ) { /* --> (x, y, z) originales */
    } else if ( strcmp(commande, "ORIGX3") == 0 ) { /* Sans intérêt ici... */
    } else if ( strcmp(commande, "SCALE1") == 0 ) { /* Coordonnées PDB orthogonales */
    } else if ( strcmp(commande, "SCALE2") == 0 ) { /* --> coordonnées fractionnaires */
    } else if ( strcmp(commande, "SCALE3") == 0 ) { /* Sans intérêt ici */
    } else if ( strcmp(commande, "TVECT ") == 0 ) { /* Translation pour les polymères linéaires */
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Cristallo : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section « Coordonnées » -------- */

char *PDB_lit_Coordonnees(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  PDB_Chaine *chaine;
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};
  int type;
  int num_modele = 0;

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp(commande, "ANISOU") == 0 ) { /* Agitation thermique */
      /* Si lu ici, c'est qu'il est dans une chaîne à ignorer */
    } else if ( ( strcmp(commande, "ATOM  ") == 0 ) ||
		( strcmp(commande, "HETATM") == 0 ) ) { /* Atome, premier d'une chaîne */
      type = ligne[ 21 ];

      if ( ( 0 == infos_pdb->ref_type ) || ( infos_pdb->ref_type == type ) ) {
	int num_residu;

	chaine = trouver_chaine( &(infos_pdb->premiere_chaine), type );
	if ( NULL == chaine ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Impossible de trouver/créer une chaîne de type %c\n",
			 __FILE__, __FUNCTION__, __LINE__, type);
	  free( ligne ); ligne = NULL;
	  infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	  return NULL;
	}
	if ( NULL == chaine->chaine ) { /* Pas de molécule associée */
	  char *titre;
	  
	  titre = PDB_creer_titre(infos_pdb->base_nom, infos_pdb->code, type, num_modele);
	  PDB_MOL_chaine(chaine, infos_pdb->premiere_chaine, ( infos_pdb->drapeaux & PDB_OPTIONS_UNE_MOLECULE ), type, titre);
	  free( titre ); titre = NULL;
	}
	if ( NULL == chaine->chaine ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer une molécule dans la chaîne !\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  free( ligne ); ligne = NULL;
	  infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	  return NULL;
	}

	num_residu = atoi(ligne + 22);
	if (chaine->delta_num == -5000) {
	  if (chaine == infos_pdb->premiere_chaine) /* On charge la 1re chaîne, mais acides aminés ne démarrent pas à 1 ! */
	    chaine->delta_num = 0;
	  else
	    chaine->delta_num = num_residu - 1;
	}

	ligne = PDB_lit_chaine(fichier, ligne, infos_pdb, chaine);
	if (ligne == NULL) return NULL;
	continue;
      }
    } else if ( strcmp(commande, "MODEL ") == 0 ) { /* Un des modèles, parmi plusieurs */
      num_modele = atoi(ligne + 10);
#ifdef _DEVELOPPE_
      printf("PDB_lit_Coordonnees : on attaque le modèle n°%d [modèle de référence : %d]\n", num_modele, infos_pdb->ref_modele);
#endif
      if ( (infos_pdb->ref_modele > 0) && (infos_pdb->ref_modele != num_modele) ) { /* Ignorer le modèle */
#ifdef _DEVELOPPE_
	printf( "%s : modèle %d ignoré (seul le modèle %d est désiré)\n", __FUNCTION__, num_modele, infos_pdb->ref_modele);
#endif
	do {
	  free(ligne);
	  ligne = lit_ligne(fichier, 80);
	  if (ligne == NULL) {
	    (void) BI_ERR_alerte(NULL, &PDB_err_ENDMDL, VRAI);
	    infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	    return NULL;
	  }
	  strncpy(commande, ligne, 6);

	  printf( "%s / MODEL, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );

	} while ( strcmp(commande, "ENDMDL") != 0);
      } else if (num_modele > 1) { /* Ce n'est pas le premier modèle */
	PDB_Chaine *chaine;

	for (chaine = infos_pdb->premiere_chaine ; chaine ; chaine = chaine->suivante) {
	  MOL_cherche_barycentre(chaine->chaine, FAUX);
	  MOL_initialise_vue(chaine->chaine);
	  chaine->chaine = NULL; /* On efface les molécules pour en faire de nouvelles */
	}
      }
    } else if ( strcmp(commande, "ENDMDL") == 0 ) { /* Fin d'un modèle */
    } else if ( strcmp(commande, "SIGATM") == 0 ) { /* Incertitude sur un atome */
    } else if ( strcmp(commande, "SIGUIJ") == 0 ) { /* Incertitude sur les U */
    } else if ( ( strcmp(commande, "TER   ") == 0 ) ||
		( strcmp(commande, "TER") == 0 ) ) { /* Pour les fichiers sales qui font sauter les espaces ! */
    } else if ( 0 == *ligne ) { /* Fin de fichier atteinte ! */
      return ligne;
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Coordonnees : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

static char *PDB_lit_chaine(FILE *fichier, char *ligne, PDB_infos *infos_pdb, PDB_Chaine *chaine)
{
  int type, num_residu;
  char commande[7] = {0, 0, 0, 0, 0, 0, 0}, code_aa[4] = {0, 0, 0, 0};
  char *nom_aa;
  Fragment *residu;

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

    if ( strcmp(commande, "ANISOU") == 0 ) { /* Agitation thermique */
    } else if ( (strcmp(commande, "ATOM  ") == 0) ||
		(strcmp(commande, "HETATM") == 0) ) { /* Atome, premier d'une chaîne */
      unsigned int nbr_atomes = 0;

      type = ligne[21];
      if (type != chaine->code) {
#ifdef _DEVELOPPE_
	printf("lit_chaine : %c -> %c [%s]\n", chaine->code, type, ligne);
#endif
	return ligne;
      }

      num_residu = atoi(ligne + 22);
      code_aa[0] = ligne[17];
      code_aa[1] = ligne[18];
      code_aa[2] = ligne[19];

      if (ligne[0] == 'H') { /* On cherche dans les groupes HET lus */
	PDB_hetero *groupe;

	groupe = trouver_hetero( &(infos_pdb->premier_groupe), code_aa);
	if (groupe != NULL) {
	  nbr_atomes = groupe->nbr_atomes;
	  nom_aa = groupe->nom;
	}
#ifdef _DEVELOPPE_
	printf("%s : Hétéroatome / groupe %s (code %s) [chaine : %s]\n\t{%s}\n",
	       __FUNCTION__, groupe->nom, code_aa, chaine->chaine->nom, ligne);
#endif
      }
      if (nbr_atomes == 0)
	nbr_atomes = quel_aa(code_aa, NULL, NULL, &nom_aa);
      if (nbr_atomes < 1) nbr_atomes = 1;

      if (ligne[0] == 'H') /* hétéro numérotés "sans décalage" */
	residu = PDB_trouver_fragment(chaine, chaine->prem_fragment, num_residu, nom_aa);
      else /* Chaînes normales avec "décalage" */ {
/*	printf("Décalage, chaîne %c : %d\n", chaine->code, chaine->delta_num); */
	residu = PDB_trouver_fragment(chaine, chaine->prem_fragment, num_residu - chaine->delta_num, nom_aa);
      }
      if (residu == NULL) {
	infos_pdb->code_retour = PDB_LIGNE_ERREUR;
	free(ligne);
	return NULL;
      }
      ligne = PDB_lit_residu(fichier, ligne, infos_pdb, chaine, num_residu, residu);
      if (ligne == NULL) return NULL;
      continue;
    } else if ( strcmp(commande, "ENDMDL") == 0 ) { /* Fin d'un modèle */
      free(ligne);
      ligne = lit_ligne(fichier, 80);
      return ligne; /* on est donc sûr d'avoir fini la chaîne ! */
    } else if ( strcmp(commande, "SIGATM") == 0 ) { /* Incertitude sur un atome */
    } else if ( strcmp(commande, "SIGUIJ") == 0 ) { /* Incertitude sur les U */
    } else if ( strcmp(commande, "TER   ") == 0 ) {
#ifdef _DEVELOPPE_
      unsigned int i;

      printf("Séquence principale de %c terminée\n", chaine->code);
      /* On vérifie la cohérence des fragments */
      printf("Résidus connus : %lu - Molécule : %u\n",
	     chaine->nbr_residus, chaine->chaine->nbr_fragments);
      for (i = 0, residu = chaine->chaine->fragments ; i < chaine->chaine->nbr_fragments ;
	   i++, residu = residu->suivant)
/*	printf(" résidu %u : %s\n", i, residu->nom)*/;
      printf("Le chaînage paraît correct\n");
#endif
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Chaine : commande non reconnue [%s]\n", commande);
#endif
      return ligne;
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL;
}

static char *PDB_lit_residu(FILE *fichier, char *ligne, PDB_infos *infos_pdb,
			    PDB_Chaine *chaine, int num_residu, Fragment *residu)
{
  int type, nv_residu;
  int num_atome;
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);
    if ( strcmp(commande, "ANISOU") == 0 ) { /* Agitation thermique */
    } else if ( (strcmp(commande, "ATOM  ") == 0) ||
		(strcmp(commande, "HETATM") == 0) ) { /* Atome, premier d'une chaîne */
      type = ligne[21];
      if (type != chaine->code) { /* On change de chaîne, donc de résidu ! */
#ifdef _DEVELOPPE_
	printf("lit_residu : %c -> %c [%s]\n", chaine->code, type, ligne);
#endif
	return ligne;
      }
      nv_residu = atoi(ligne + 22);
      if (nv_residu != num_residu) { /* On change de résidu */
	/* Voir à créer les liaisons internes ? */
	return ligne;
      }

      /* C'est donc bien un atome du résidu que l'on veut trouver */
      num_atome = atoi(ligne + 6);
      ligne = PDB_lit_atome(fichier, ligne, infos_pdb, chaine, num_residu, residu, num_atome);
      if (ligne == NULL) return NULL;
      continue;
    } else if ( strcmp(commande, "ENDMDL") == 0 ) { /* Fin d'un modèle */
      return ligne; /* on a fini le résidu, on renvoit tel quel pour marquer la fini de la chaîne */
    } else if ( strcmp(commande, "SIGATM") == 0 ) { /* Incertitude sur un atome */
    } else if ( strcmp(commande, "SIGUIJ") == 0 ) { /* Incertitude sur les U */
    } else if ( strcmp(commande, "TER   ") == 0 ) {
      return ligne; /* On a fini la chaîne principale, donc le résidu */
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_residu : commande non reconnue [%s]\n", commande);
#endif
      return ligne;
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  return NULL;
}

static char *PDB_lit_atome(FILE *fichier, char *ligne, PDB_infos *infos_pdb,
			   PDB_Chaine *chaine, int num_residu, Fragment *residu, int num_atome)
{
  Atome *atome;
  boolean desordonne = FAUX; /* VRAI si l'atome est désordonné */
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};
  int Z = 0, charge = 0;
  Vecteur3D position;
  float occupation __attribute__((__unused__)), f_therm __attribute__((__unused__));
  int nv_atome;

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

#ifdef _DEVELOPPE_
  printf("PDB_lit_atome : fichier = %p, ligne = %s, infos_pdb = %p, chaine = %p, num_residu = %d, residu = %p, num_atome = %d\n",
	 (void *) fichier, ligne, (void *) infos_pdb, (void *) chaine, num_residu, (void *) residu, num_atome);
#endif

  /* La fonction ne devrait être appelée que sur une ligne ATOM ou HETATM */
  strncpy(commande, ligne, 6);
  if ( (strcmp(commande, "ATOM  ") != 0) && (strcmp(commande, "HETATM") != 0) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Appel sur commande incorrecte [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, commande);
#endif
    return ligne; /* Mais ne devrait jamais arriver */
  }

  /* Cet atome est forcément dans le résidu indiqué => pas de vérif, on le charge. */

  /* On s'occupe du numéro atomique.
     Délicat car au moins trois formats coexistent :
   - premières versions : pas de codage de Z (!), il faut deviner à partir du nom, peu fiable...
   - versions récentes officielles : symbole en colonnes 76 et 77, à une lettre = 77
   - variante de cette version : symbole en colonnes 77 et 78, à une lettre = 77...
  */
  if ( isalpha( ligne[ 76 ] ) ) {
    ligne[ 77 ] = tolower( ligne[ 77 ] );
    Z = ATM_num_atomique( ligne + 76 ); /* Le numéro atomique associé */
    charge = atoi( ligne + 78 );
  } else if ( isalpha(ligne[ 77 ]) ) {
    ligne[78] = tolower( ligne[ 78 ] );
    Z = ATM_num_atomique( ligne + 77 ); /* Le numéro atomique associé */
    charge = 0;
  } else {
    if ( ligne[ 0 ] == 'H' ) { /* On peut avoir un doute CA = Ca ou Calpha ? */
      if ( ligne[ 12 ] != ' ' )
	ligne[ 13 ] = tolower( ligne[ 13 ] ); /* Suppose décalage de 1 entre Ca et Calpha, pas garanti */
    }
    Z = ATM_num_atomique( ligne + 12 ); /* Colonne 13 utilisée pour autre chose ? */
    if ( (Z == 0) || (Z == HYDROGENE) ) { /* Symbole inconnu => probablement 2 lettres, on met la 2e en minuscules
					     D & T donnent HYDROGENE (Deutérium, tritium), mais existe pas dans PDB */
      int nv_Z;

      if ( ligne[12] == ' ' )
	ligne[14] = tolower(ligne[14]);
      else
	ligne[13] = tolower(ligne[13]);
      nv_Z = ATM_num_atomique(ligne + 12);
      if ( (HYDROGENE != Z) || ( nv_Z >= HYDROGENE ) ) { /* Le nouveau code doit convenir... */
	Z = nv_Z;
      }
    }
    if (Z == 0) {
      fprintf(stderr, "PDB_lit_atome : impossible de trouver le symbole de l'atome à partir du nom (%s)\n", ligne + 13);
    }
  }

  ligne[ 72 ] = 0; f_therm = atof( ligne + 60 );    /* Facteur thermique */
  ligne[ 60 ] = 0; occupation = atof( ligne + 54 ); /* Occupation */
  ligne[ 54 ] = 0; position.z = atof( ligne + 46 ); /* z */
  ligne[ 46 ] = 0; position.y = atof( ligne + 38 ); /* y */
  ligne[ 38 ] = 0; position.x = atof( ligne + 30 ); /* x */
  if ( ligne[16] != ' ' ) { /* Atome "désordonné" */
#ifdef _DEVELOPPE_
    /*	printf("Atome désordonné, aa[0] = %c\n", aa[0]); */
#endif
    desordonne = VRAI; /* Donc il y a au moins un autre ATOM de même numéro après */
    /* Pour l'instant, on ignore les suivants */
  }

  atome = ATM_creer( chaine->chaine, NULL, Z, &(position), charge, -1 );
  if ( NULL == atome ) {
    infos_pdb->code_retour = PDB_LIGNE_ERREUR;
    free( ligne ); ligne = NULL;
    (void) fprintf(stderr, "PDB_lit_atome : impossible de créer l'atome [Z = %d, en (%f, %f, %f)]\n", Z, position.x, position.y, position.z);
    return NULL;
  }
  atome->index = num_atome;
  FRG_ajoute_atome(residu, 1, &atome);

  do {
    free(ligne);
    ligne = lit_ligne(fichier, 80);
    if (ligne == NULL) {
      (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
      return NULL;
    }

    if ( strcmp(commande, "ANISOU") == 0 ) { /* Agitation thermique */
    } else if ( (strcmp(commande, "ATOM  ") == 0) ||
		(strcmp(commande, "HETATM") == 0) ) { /* Atome, premier d'une chaîne */
      if (desordonne == FAUX) return ligne;

      nv_atome = atoi(ligne + 6);
      if (nv_atome != num_atome) /* On passe à l'atome suivant */
	return ligne;
    } else if ( strcmp(commande, "SIGATM") == 0 ) { /* Incertitude sur un atome */
    } else if ( strcmp(commande, "SIGUIJ") == 0 ) { /* Incertitude sur les U */
    } else { /* Inclus ENDMDL et TER qui marquent la fin du modèle */
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_atome : commande non reconnue [%s]\n", commande);
#endif
      return ligne;
    }

  } while (ligne);

  return NULL;
}

/* -------- Lecture de la section « Liaisons » -------- */

char *PDB_lit_TableLiaisons(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[ 7 ] = {0, 0, 0, 0, 0, 0, 0};

  if ( NULL == fichier ) return NULL;
  if ( NULL == ligne ) return NULL;

#ifdef _DEVELOPPE_
  printf("PDB_lit_TableLiaisons : fichier = %p, ligne = %s, infos_pdb = %p\n", (void *) fichier, ligne, (void *) infos_pdb);
#endif

  do {
    strncpy( commande, ligne, 6 );

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( strcmp( commande, "CONECT" ) == 0 ) { /* Décrit une liaison ? */
      unsigned long i, atome_central, atomes_lies[ 4 ]; /* Maximum 4 liaisons par ligne... */
      Atome *central;
      Molecule *chaine_reference = NULL;

      /* On lit les numéros des atomes */
      ligne[ 31 ] = 0; atomes_lies[ 3 ] = TXT_CNV_vNombre_UL( ligne + 26, NULL );
      ligne[ 26 ] = 0; atomes_lies[ 2 ] = TXT_CNV_vNombre_UL( ligne + 21, NULL );
      ligne[ 21 ] = 0; atomes_lies[ 1 ] = TXT_CNV_vNombre_UL( ligne + 16, NULL );
      ligne[ 16 ] = 0; atomes_lies[ 0 ] = TXT_CNV_vNombre_UL( ligne + 11, NULL );
      ligne[ 11 ] = 0; atome_central = TXT_CNV_vNombre_UL( ligne + 6, NULL );

      if ( NULL == infos_pdb->premiere_chaine->suivante ) {
	/* Une seule chane => une seule molcule : on cherche dans elle directement */
	central = ATM_index( infos_pdb->premiere_chaine->chaine, atome_central );
	chaine_reference = infos_pdb->premiere_chaine->chaine;
      } else {
	struct chaine *chaine_PDB = infos_pdb->premiere_chaine;

	(void) fprintf(stderr, "%s,%s l.%u : Plusieurs chanes PDB ! Chane 1 : %lu rsidus ; suivante : %p\n",
		       __FILE__, __FUNCTION__, __LINE__,
		       infos_pdb->premiere_chaine->nbr_residus, (void *) infos_pdb->premiere_chaine->suivante );

	do {
	  central = ATM_index( chaine_PDB->chaine, atome_central );
	  if ( central != NULL ) {
	    chaine_reference = chaine_PDB->chaine; /* On a trouv : on postule que les autres seront dans la mme molcule */
	    break;
	  }
	  chaine_PDB = chaine_PDB->suivante;
	} while ( chaine_PDB != NULL );
      }

      /* On repère l'atome central, dans la molécule chargée */
      central = ATM_index( chaine_reference , atome_central );
      if ( NULL == central ) {
	(void) fprintf( stderr, "%s : Atome de base de CONECT non-trouvé [numéro %lu]", __FUNCTION__, atome_central );
	FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      } else {
	for ( i = 0 ; i < 4 ; i++ ) {
	  if ( atomes_lies[ i ] > atome_central ) {
	    Liaison *liaison __attribute__((__unused__));
	    /* 0 : champ vide ; < central : la liaison a déjà été créée, car les liaisons sont en double et classées par numéro d'atome croissant
	        ==> on ne fait la liaison que si l'atome lié est ensuite (soit d'index supérieur) */
	    
	    liaison = creer_liaison( chaine_reference, central, ATM_index( chaine_reference, atomes_lies[ i ] ), 1, LIAISON_COVALENTE );
	  }
	}
      }
    } else if ( 0 == *ligne ) {
      return ligne;
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_TableLiaisons : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande d'une autre section, a priori */
    }

    free( ligne );
    ligne = lit_ligne( fichier, 80 );
  } while ( ligne != NULL );

  (void) BI_ERR_alerte( NULL, &PDB_err_END, VRAI, commande );
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Lecture de la section « Fin » -------- */

char *PDB_lit_Fin(FILE *fichier, char *ligne, PDB_infos *infos_pdb)
{
  char commande[7] = {0, 0, 0, 0, 0, 0, 0};

  if (fichier == NULL) return NULL;
  if (ligne == NULL) return NULL;

  do {
    strncpy(commande, ligne, 6);

#ifdef _DEVELOPPE_
    printf( "%s : Traitement, ligne %s, commande %s\n", __FUNCTION__, ligne, commande );
#endif

    if ( (commande[0] == 'E') && (commande[1] == 'N') && (commande[2] == 'D') ) { /* Fin de chaîne */
      infos_pdb->code_retour = PDB_LIGNE_FINIR;
      free(ligne);
      return NULL;
    } else if ( strcmp(commande, "MASTER") == 0 ) { /* Contrôles internes */
    } else if ( 0 == *ligne ) {
      printf( "%s : Ligne vide !", __FUNCTION__ );
    } else {
#ifdef _DEVELOPPE_
      fprintf(stderr, "PDB_lit_Fin : commande non reconnue [%s]\n", commande);
#endif
      return ligne; /* Commande inconnue ou mal placée, a priori */
    }

    free(ligne);
    ligne = lit_ligne(fichier, 80);
  } while (ligne);

  (void) BI_ERR_alerte(NULL, &PDB_err_END, VRAI, commande);
  infos_pdb->code_retour = PDB_LIGNE_FINIR; /* On essaye quand même de construire ce que l'on a pu lire */
  return NULL; /* Si on est là, c'est que le fichier a un gros problème ! */
}

/* -------- Création de la molécule associée à une chaîne -------- */

static int PDB_MOL_chaine(PDB_Chaine *chaine, PDB_Chaine *premiere,
			  boolean fondre_chaines, int ref_code, const char *titre)
{
  if (chaine == NULL) return -1;
  if (chaine->chaine != NULL) return 0; /* La molécule existe déjà */

  if ( (ref_code > 0) && (ref_code != chaine->code) )
    return -2; /* On ne veut pas charger cette chaîne */

  if ( fondre_chaines != FAUX ) { /* On cherche si une molécule n'existe pas déjà... */
    PDB_Chaine *pre_chaine;

    for (pre_chaine = premiere ; pre_chaine ; pre_chaine = pre_chaine->suivante)
      if (pre_chaine->chaine) {
	chaine->chaine = pre_chaine->chaine;
	return 0;
      }
  }

  /* Si on est là, on doit créer la molécule */
  if (titre == NULL) titre = "PDB/Sans nom";
  chaine->chaine = MOL_creer(titre);

  if (chaine->nbr_residus > 0) { /* On pourrait préparer une série de fragments... */
  }

  return 0;
}

/* -------- Renvoie une chaîne à partir de la chaîne PDB -------- */

static char *PDB_propose_chaine(char *chaine)
{
  size_t lg_chaine;
  char *resultat = NULL;

  if (chaine == NULL) return NULL;
  lg_chaine = strlen(chaine);
  while ( isspace(chaine[0]) ) {
    chaine++;
    lg_chaine--;
    if (lg_chaine == 0) return NULL; /* La chaîne ne comportait que des espaces */
  }

  resultat = strdup(chaine);
  if (resultat == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/lit_pdb.c", "PDB_propose_chaine", "resultat/strdup");
    return NULL;
  }
  lg_chaine = strlen(resultat);
  while ( isspace(resultat[lg_chaine - 1]) ) { /* On ôte les espaces finaux */
    resultat[lg_chaine - 1] = 0;
    lg_chaine--;
  }

  return resultat;
}

/* -------- Création du nom d'une molécule à partir des infos passées --------\
 |  Renvoie le nom créé (à libérer avec free()) - NULL en cas d'erreur        |
 |                                                                            |
 | base   = la base du nom - "Sans nom" si NULL ou de longueur nulle          |
 | code   = le code à 4 lettres de la PDB - si commence par 0, pas utilisé    |
 | chaine = le code de la chaîne PDB - si 0, pas utilisé                      |
 | modele = le numéro du modèle - si 0, pas utilisé                           |
 \--------------------------------------------------------------------------- */

static char *PDB_creer_titre(char *base, char *code, int chaine, unsigned long modele)
{
  char *titre = NULL;
  size_t lg_texte;

  if ( (base == NULL) || (*base == 0) )
    base = "Sans nom";

  lg_texte = strlen(base) + 1; /* 0 final */
  if ( (code != NULL) && (*code != 0) )
    lg_texte += 7; /* + " (xxx)" */
  if (chaine > 0)
    lg_texte += 10; /* + ", chaîne X" */
  if (modele > 0) {
    lg_texte += 8; /* + " modèle " */
    lg_texte += 1 + (int) log10( (double) modele); /* Nombre de chiffres */
  }

#ifdef _DEVELOPPE_
  printf("PDB_creer_titre : Taille demandée = %lu caractères", (unsigned long) lg_texte);
#endif
  titre = (char *) malloc( lg_texte );
  if (titre == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/lit_pdb.c", "PDB_creer_titre", "titre");
    return NULL;
  }

  strcpy(titre, base);
  if ( (code != NULL) && (*code != 0) ) {
    unsigned int i;

    strcat(titre, " (");
    lg_texte = strlen(titre);
    for (i = 0 ; (i < 4) && (code[i] != 0) ; i++) {
      titre[lg_texte++] = code[i];
    }
    titre[lg_texte] = 0;
    strcat(titre, ")");
  }
  if ( chaine > 0 ) {
    strcat(titre, ", cha"i_CIRC"ne ");
    lg_texte = strlen(titre);
    titre[lg_texte] = (char) chaine;
    titre[lg_texte + 1] = 0;
  }
  if ( modele > 0 ) {
    strcat(titre, " mod"e_GRAVE"le ");
    lg_texte = strlen(titre);
    sprintf(titre + lg_texte, "%lu", modele);
  }
#ifdef _DEVELOPPE_
  printf(", taille finale : %u ==> %s\n", (unsigned int)strlen(titre), titre);
#endif

  return titre; /* Fini, il faudra libérer avec free() */
}

/* -------- Reconnaissance de l'acide aminé traité ---------------\
 |   Renvoie le nombre d'atomes (hors H) de l'acide aminé         |
 |      Ne tient pas compte du -OH de l'acide carboxylique        |
 |                       et du -H de l'amine                      |
 | (disparus en H2O lors de la création de la liaison peptidique) |
 |      Ne compte pas les liaisons avec un hydrogène              |
 |                                                                |
 | symbole : le symbole (3 lettres ou 1 lettre) de l'acide aminé  |
 | nbr_H   : le nombre d'atomes d'hydrogène dans l'acide aminé    |
 | nbr_l   : le nombre de liaisons dans l'acide aminé             |
 | nom     : le nom de l'acide aminé reconnu                      |
 \--------------------------------------------------------------- */

static int quel_aa(char *symbole, int *nbr_H, int *nbr_l, char **nom)
{
  if ( symbole == NULL ) return 0;

  /* On fait sauter les éventuels espaces initiaux */
  while ( *symbole == ' ' ) symbole++;

  /* On essaye d'abord le code à trois lettre */
  if ( strlen( symbole ) >= 3 ) {
    char c_1 = toupper( *symbole ), 
      c_2 = toupper( *( symbole + 1 ) ), 
      c_3 = toupper( *( symbole + 2 ) ); /* Les trois premiers codes */

    switch ( c_1 ) /* Première lettre */
    {
    case 'A':
      switch ( c_2 )
      {
      case 'L':
	if ( c_3 == 'A') { /* ALA = alanine, R = -CH3 */
	  if ( nom )   *nom = "Alanine";
	  if ( nbr_H ) *nbr_H = 5;
	  if ( nbr_l ) *nbr_l = 4;
	  return 5;
	}
	break;

      case 'R':
	if ( c_3 == 'G') { /* ARG = arginine, R = -CH2-CH2-CH2-NH-CHNH2=NH */
	  if ( nom )   *nom = "Arginine";
	  if ( nbr_H ) *nbr_H =  5;
	  if ( nbr_l ) *nbr_l = 10;
	  return 11;
	}
	break;

      case 'S':
	if ( c_3 == 'N') { /* ASN = asparagine, R = -CH2-CO-NH2 */
	  if ( nom )   *nom = "Asparagine";
	  if ( nbr_H ) *nbr_H = 7;
	  if ( nbr_l ) *nbr_l = 7;
	  return 8;
	} else if ( c_3 == 'P' ) { /* ASP = acide aspartique, R = -CH2-COOH */
	  if ( nom )   *nom = "Acide aspartique";
	  if ( nbr_H ) *nbr_H = 5; /* Forme acide, aspartate = 1 H de moins */
	  if ( nbr_l ) *nbr_l = 7;
	  return 8;
	}
	break;
      }
      break;

    case 'C':
      switch ( c_2 )
      {
      case 'Y':
	if ( c_3 == 'S' ) { /* CYS = cystéine, R = -CH2-SH */
	  if ( nom )   *nom = "Cyst"e_AIGU"ine";
	  if ( nbr_H ) *nbr_H = 5;
	  if ( nbr_l ) *nbr_l = 5;
	  return 6;
	}
	break;
      }
      break;

    case 'D':
      switch ( c_2 )
      {
      case 'M':
	if ( c_3 == 'S' ) { /* DMS = diméthylsulfoxyde,  CH3-SO-CH3 */
	  if ( nom )   *nom = "Dim"e_AIGU"thylsulfoxyde";
	  if ( nbr_H ) *nbr_H = 6;
	  if ( nbr_l ) *nbr_l = 3;
	  return 4;
	}
	break;
      }
      break;

    case 'G':
      switch ( c_2 )
      {
      case 'L':
	if ( c_3 == 'N' ) { /* GLN = glutamine, R = -CH2-CH2-CO-NH2 */
	  if ( nom )   *nom = "Glutamine";
	  if ( nbr_H ) *nbr_H = 8;
	  if ( nbr_l ) *nbr_l = 8;
	  return 9;
	} else if ( c_3 == 'U' ) { /* GLU = acide glutamique, R = -CH2-CH2-COOH */
	  if ( nom )   *nom = "Acide glutamique";
	  if ( nbr_H ) *nbr_H = 7; /* Y compris le H acide; glutamate = 6 H */
	  if ( nbr_l ) *nbr_l = 8;
	  return 9;
	} else if ( c_3 == 'Y') { /* GLY = glycine (glycocolle), R = -H */
	  if ( nom )   *nom = "Glycine (glycocolle)";
	  if ( nbr_H ) *nbr_H = 3;
	  if ( nbr_l ) *nbr_l = 3;
	  return 4;
	}
	break;

      }
      break;

    case 'H':
      switch ( c_2 )
      {
      case 'E':
	if ( c_3 == 'M' ) { /* HEM = hème, à fer */
	  if (nom) *nom = "H"e_GRAVE"me";
	  if (nbr_H) *nbr_H = 32;
	  if (nbr_l) *nbr_l = 0;
	  return 43;
	}
      case 'I':
	if ( c_3 == 'S' ) /* HIS = histidine, R = -CH2-imidazole */
	{
	  if (nom)   *nom = "Histidine";
	  if (nbr_H) *nbr_H = 8;
	  if (nbr_l) *nbr_l = 10;
	  return 10;
	}
	break;

      case 'O':
	if ( c_3 == 'H' ) /* HOH = molécule d'eau, H2O */
	{
	  if (nom)   *nom = "Eau";
	  if (nbr_H) *nbr_H = 2;
	  if (nbr_l) *nbr_l = 0;
	  return 1;
	}
	break;
      }
      break;

    case 'I':
      switch ( c_2 )
      {
      case 'L':
	if ( c_3 == 'E' ) /* ILE = isoleucine, R = -CH(CH3)-CH2-CH3 */
	{
	  if (nom)   *nom = "Isoleucine";
	  if (nbr_H) *nbr_H = 11;
	  if (nbr_l) *nbr_l = 7;
	  return 8;
	}
	break;
      }
      break;

    case 'L':
      switch ( c_2 )
      {
      case 'E':
	if ( c_3 == 'U' ) /* LEU = leucine, R = -CH(CH3)-CH2-CH3 */
	{
	  if (nom)   *nom = "Leucine";
	  if (nbr_H) *nbr_H = 11;
	  if (nbr_l) *nbr_l = 7;
	  return 8;
	}
	break;

      case 'Y':
	if ( c_3 == 'S' ) /* LYS = lysine, R = -CH2-CH2-CH2-CH2-NH2 */
	{
	  if (nom)   *nom = "Lysine";
	  if (nbr_H) *nbr_H = 12;
	  if (nbr_l) *nbr_l = 8;
	  return 9;
	}
	break;
      }
      break;

    case 'M':
      switch ( c_2 )
      {
      case 'E':
	if ( c_3 == 'T' ) /* MET = méthionine, R = -CH2-CH2-S-CH3 */
	{
	  if (nom)   *nom = "M"e_AIGU"thionine";
	  if (nbr_H) *nbr_H = 9;
	  if (nbr_l) *nbr_l = 7;
	  return 8;
	}
	break;
      }
      break;

    case 'P':
      switch ( c_2 )
      {
      case 'H':
	if ( c_3 == 'E') /* PHE = phénylalanine, R = -CH2-C6H5 */
	{
	  if (nom)   *nom = "Ph"e_AIGU"nylalanine";
	  if (nbr_H) *nbr_H = 9;
	  if (nbr_l) *nbr_l = 11;
	  return 11;
	}
	break;

      case 'R':
	if ( c_3 == 'O') /* PRO = proline,  */
	{
	  if (nom)   *nom = "Proline";
	  if (nbr_H) *nbr_H = 7;
	  if (nbr_l) *nbr_l = 7;
	  return 7;
	}
	break;

      }
      break;

    case 'S':
      switch ( c_2 )
      {
      case 'E':
	if ( c_3 == 'R') /* SER = sérine, R = -CH2-OH */
	{
	  if (nom)   *nom = "S"e_AIGU"rine";
	  if (nbr_H) *nbr_H = 5;
	  if (nbr_l) *nbr_l = 5;
	  return 6;
	}
	break;

      }
      break;

    case 'T':
      switch ( c_2 )
      {
      case 'H':
	if ( c_3 == 'R' ) { /* THR = thréonine, R = -CH2-CH2-OH */
	  if (nom)   *nom = "Thr"e_AIGU"onine";
	  if (nbr_H) *nbr_H = 7;
	  if (nbr_l) *nbr_l = 6;
	  return 7;
	}
	break;

      case 'R':
	if ( c_3 == 'P' ) { /* TRP = tryptophane, R = -CH2-benzimidazole */
	  if (nom)   *nom = "Tryptophane";
	  if (nbr_H) *nbr_H = 7;
	  if (nbr_l) *nbr_l = 15;
	  return 14;
	}
	break;

      case 'Y':
	if ( c_3 == 'R' ) { /* TYR = tyrosine, R = -CH2-C6H4-OH */
	  if (nom)   *nom = "Tyrosine";
	  if (nbr_H) *nbr_H = 7;
	  if (nbr_l) *nbr_l = 12;
	  return 12;
	}
	break;
      }
      break;

    case 'V':
      switch ( c_2 )
      {
      case 'A':
	if ( c_3 == 'L' ) { /* VAL = valine, R = -CH-(CH3)2 */
	  if (nom)   *nom = "Valine";
	  if (nbr_H) *nbr_H = 9;
	  if (nbr_l) *nbr_l = 6;
	  return 7;
	}
	break;
      }
      break;
    }
  }

  if ( strlen( symbole ) == 2 ) { /* Deux codes = on suppose que c'est un symbole chimique */
    if ( nom )   *nom = "Atome seul";
    if ( nbr_H ) *nbr_H = 0;
    if ( nbr_l ) *nbr_l = 0;

    return 1;
  }

  if ( strlen( symbole ) == 1 ) { /* Un code => nucléotide ? */
    if ( *symbole == 'A' ) {
      if ( nom ) *nom = "Ad"e_AIGU"nine";
      return 1;
    } else if ( *symbole == 'C' ) {
      if ( nom ) *nom = "Cytosine";
      return 1;
    } else if ( *symbole == 'G' ) {
      if ( nom ) *nom = "Guanine";
      return 1;
    } else if ( *symbole == 'I' ) {
      if ( nom ) *nom = "Inosine";
      return 1;
    } else if ( *symbole == 'T' ) {
      if ( nom ) *nom = "Thymine";
      return 1;
    } else if ( *symbole == 'U' ) {
      if ( nom ) *nom = "Uracile";
      return 1;
    }
  }

  /* Si l'on est ici, c'est que l'on n'a pas reconnu le composé */
#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "%s,%s l.%u : Commande ATOM avec nom de résidu inconnu [%s]\n",
		 __FILE__, __FUNCTION__, __LINE__, symbole);
#endif

  if ( nom )   *nom = "Compos"e_AIGU" inconnu";
  if ( nbr_H ) *nbr_H = 0;
  if ( nbr_l ) *nbr_l = 0;

  return 0;
}

/* -------- Donne un fragment correspondant au numéro indiqué pour la chaîne précisée ---------\
 |  Renvoie le fragment trouvé ou, en cas d'erreur, NULL                                       |
 |  Si aucun trouvé, un nouveau fragment à la suite de celui de numéro immédiatement inférieur |
 |    ==> les fragments sont toujours organisés par numéro croissant, lors du chargement       |
 |                                                                                             |
 | chaine : la chaîne pour laquelle la recherche a lieu                                        |
 | avant  : le fragment que l'on suppose immédiatement avant, idéalement                       |
 | numero : le numéro du fragment cherché                                                      |
 | nom    : le nom à utiliser, si le fragment n'a pas été trouvé                               |
 \-------------------------------------------------------------------------------------------- */

Fragment *PDB_trouver_fragment(PDB_Chaine *chaine, Fragment *avant, int numero, char *nom)
{
  char *titre;
  Fragment *retour = NULL;

  if (chaine == NULL) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "PDB_trouver_fragment : pas de chaîne !\n");
#endif
    return NULL; /* Pas de chaîne indiquée ! */
  }
  if (chaine->chaine == NULL) { /* Pas de molécule dans la chaîne, donc pas de fragment ! */
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "PDB_trouver_fragment : chaîne appelée sans molécule !\n");
#endif
    return NULL;
  }
  if (avant == NULL) {
    if (chaine->prem_fragment == NULL) { /* Aucun fragment n'existe */
#ifdef _DEVELOPPE_
      printf("PDB_trouver_fragment : aucun fragment dans la chaîne...\n");
#endif
      retour = FRG_creer(chaine->chaine, NULL, nom, 1, NULL, NULL);
      if ( NULL == retour ) {
	(void) fprintf(stderr, "%s,%s l.%u : impossible de créer le premier fragment !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return NULL;
      }
      chaine->prem_fragment = retour;
      retour->numero = numero;

      return retour;
    }
    avant = chaine->prem_fragment;
  }

  if ( avant->numero == (unsigned int) numero ) {
#ifdef _DEVELOPPE_
    printf("  * trouvé n°%u = %s\n", numero, nom);
#endif
    return avant; /* Rien à faire ! */
  }
  if ( avant->numero > (unsigned int) numero ) {
    avant = chaine->prem_fragment;
    if ( NULL == avant ) { /* Devrait jamais arriver, mais bon... */
      avant = chaine->chaine->fragments;
      chaine->prem_fragment = avant;
    }
  }

  /* On cherche le fragment de numéro juste inférieur à celui indiqué */
  while ( avant->suivant != NULL ) {
    if (avant->suivant->numero == (unsigned int) numero) {
#ifdef _DEVELOPPE_
      printf("  * trouvé n°%u = %s\n", numero, nom);
#endif
      return avant->suivant; /* Trouvé ! */
    }

    avant = avant->suivant;
  }

  /* Fragment non trouvé => premier fragment non-numéroté (0), s'il y en a un */
  retour = chaine->prem_fragment;
  while (retour) {
    if (retour->numero == 0) {
      retour->numero = numero;
#ifdef _DEVELOPPE_
      printf("  * trouvé n°%u = %s\n", numero, nom);
#endif
      return retour;
    }
    retour = retour->suivant;
  }

  /* Si on est là, c'est qu'on n'a pas trouvé le fragment cherché : on doit le créer */
/*  printf("Fragment non trouvé, n°%u : on en ajoute un [%s]\n", numero, nom); */

  if ( NULL == nom ) nom = "Nom inconnu";
  titre = TXT_genere_chaine("%s (%u)", nom, numero);
  retour = FRG_creer( chaine->chaine, avant, titre, 1, NULL, NULL );
  free( titre ); titre = NULL;
  if ( NULL == retour ) return NULL;

  retour->numero = numero;

  return retour;
}

/* -------- Donne la chaîne PDB correspondant au code indiqué, dans la famille de chaînes indiquée
   Si la chaîne n'existe pas, ou si premiere == NULL, on crée la chaîne.

   Renvoie NULL en cas d'erreur
*/

PDB_Chaine *trouver_chaine(PDB_Chaine **premiere, char code_chaine)
{
  PDB_Chaine *chaine = NULL;

  if (*premiere != NULL) { /* Existe-t-elle déjà ? */
    for (chaine = *premiere ; chaine ; chaine = chaine->suivante) {
      if (chaine->code == code_chaine) {
#ifdef _DEVELOPPE_
/*	printf("PDB, trouver_chaine : chaîne %c trouvée\n", code_chaine); */
#endif
	return chaine;
      }
    }
    chaine = NULL;
  }

  chaine = (PDB_Chaine *) malloc( sizeof(PDB_Chaine) );
  if (chaine == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/lit_pdb.c", "trouver_chaine", "chaine");
    return NULL;
  }
  chaine->suivante = NULL; /* On l'ajoute en fin de liste chaînée */
  if ( *premiere == NULL ) {
    *premiere = chaine;
  } else {
    PDB_Chaine *avant = *premiere;

    while (avant->suivante)
      avant = avant->suivante;
    avant->suivante = chaine;
  }

  chaine->code = code_chaine;
  chaine->nbr_residus =  0;    /* Nombre de résidus inconnu */
  chaine->delta_num   = -5000; /* Décalage encore inconnu - mais comme certaines chaînes commencent à -1 (sacré format !) */
  chaine->chaine = NULL;
  chaine->prem_fragment = NULL;

  return chaine;
}

/* -------- Donne le groupe PDB correspondant au code indiqué, dans la famille de groupes indiquée
   Si le groupe n'existe pas, ou si premier == NULL, on crée le groupe.

   Renvoie NULL en cas d'erreur
*/

PDB_hetero *trouver_hetero(PDB_hetero **premier, char *code_hetero)
{
  PDB_hetero *groupe = NULL;

  if (*premier != NULL) {
    for (groupe = *premier ; groupe ; groupe = groupe->suivant) {
      char *code = groupe->code;

      if ( (code[0] == code_hetero[0]) && (code[1] == code_hetero[1]) &&
	   (code[2] == code_hetero[2]) )
	return groupe;
    }
  }

  /* Il faut ici créer le groupe */
  groupe = (PDB_hetero *) malloc( sizeof(PDB_hetero) );
  if (groupe == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/lit_pdb.c", "trouver_hetero", "groupe");
    return NULL;
  }

  if ( *premier ) { /* On le place en début de liste */
    groupe->suivant = *premier;
  } else {
    groupe->suivant = NULL;
  }
  *premier = groupe;

  groupe->code[0] = code_hetero[0]; /* Code du groupe */
  groupe->code[1] = code_hetero[1];
  groupe->code[2] = code_hetero[2];
  groupe->code[3] = 0;

  groupe->nom = NULL;
  groupe->nbr_atomes = 0;
  return groupe;
}

/* -------- On libère tout quand on annule un chargement -------- */

static void PDB_annule(PDB_Chaine *premiere_chaine, PDB_hetero *premier_groupe)
{
  if (premiere_chaine != NULL) {
    PDB_Chaine *apres;

    apres = premiere_chaine->suivante;
    if (premiere_chaine->chaine != NULL) {
      PDB_Chaine *chaine;
      Molecule *molecule = premiere_chaine->chaine;

      /* Comme la molécule peut correspondre à plusieurs chaînes, on vérifie... */
      for (chaine = apres ; chaine != NULL ; chaine = chaine->suivante) {
	if (chaine->chaine == molecule)
	  chaine->chaine = NULL;
      }

      /* On n'a plus besoin de la molécule */
      MOL_effacer(molecule, VRAI, VRAI);
    }
    free(premiere_chaine);
    premiere_chaine = apres;
  }

  if ( premier_groupe != NULL ) {
    OPTION_A_FAIRE( "libérer les groupes d'hétéroatomes" );
  }
}

/* -------- On libère tout en confirmant le chargement -------- */

static void PDB_termine(PDB_Chaine *premiere_chaine, PDB_hetero *premier_groupe)
{
  PDB_Chaine *apres;

  /* On libère les différentes chaînes */
  while (premiere_chaine != NULL) {
    apres = premiere_chaine->suivante;
    if (premiere_chaine->chaine) {
      PDB_Chaine *chaine;
      Fragment *fragment;

      /* On élimine les fragments vides */
      fragment = premiere_chaine->chaine->fragments;
      if (fragment == NULL) {
	MOL_effacer(premiere_chaine->chaine, VRAI, VRAI);
      } else {
	do {
	  Fragment *apres = fragment->suivant;

	  if (fragment->nbr_atomes == 0) {
	    FRG_effacer(premiere_chaine->chaine, fragment, VRAI, FAUX);
	  }
	  fragment = apres;
	} while (fragment);
	
	/* On prépare la molécule */
	MOL_cherche_barycentre(premiere_chaine->chaine, FAUX);
	MOL_initialise_vue(premiere_chaine->chaine);
      }

      for (chaine = apres ; chaine ; chaine = chaine->suivante)
	if (chaine->chaine == premiere_chaine->chaine)
	  chaine->chaine = NULL; /* Inutile de le refaire ! */
    }
    free(premiere_chaine);
    premiere_chaine = apres;
  }

  /* Effacer les groupes hétérogènes, maintenant... */
  if ( premier_groupe != NULL ) {
    OPTION_A_FAIRE( "libérer les groupes d'hétéroatomes" );
  }
}
