/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2002, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Gestion des objets éditables
*/

#include <stdio.h>
#include <stdlib.h>

#ifdef _UNIX_
# define _LIBRAIRIE_X_ /* Pour avoir accès aux fonctions internes */
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "bib_interface.h"
#include "txt_utils.h"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int BI_ControlePositionCurseur_TexteInfini(Texte_Infini *texte, unsigned int *position);
static int BI_ControlePositionCurseur_Tedinfo    (TEDINFO *texte, unsigned int *position);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Trouver les objets éditables dans un arbre --------
 * Renvoie 0 si tout s'est bien passé, != 0 sinon
 *
 * arbre   : l'arbre à étudier
 * courant : VRAI = on essaye de conserver l'éditable courant
 */

int ARB_TrouverEditables(unsigned int arbre, boolean courant)
{
  OBJECT *adresse;
  unsigned int i, ex_courant;
  boolean courant_editable = FAUX;

  /* Vérifications */
  if ( arbre >= BI_nbr_arbres ) {
    (void) fprintf( stderr, "ARB_TrouverEditables : numéro d'arbre incorrect [%u]\n", arbre );
    return -1;
  }
  adresse = arbres[ arbre ].adresse;
  if ( NULL == adresse ) {
    (void) fprintf( stderr, "ARB_TrouverEditables : aucun objet dans l'arbre [%u]\n", arbre );
    return -2;
  }

  /* On remet à zéro */
  ex_courant = arbres[ arbre ].edit_courant;
  arbres[ arbre ].premier_edit = ROOT;
  arbres[ arbre ].dernier_edit = ROOT;

  /* On cherche */
  for ( i = 1 ; i < arbres[ arbre ].nbr_objets ; i++ ) {
    if ( ( OB_FLAGS( adresse, i ) & EDITABLE ) &&
	 ( ( OB_STATE( adresse, i ) & DISABLED ) == FAUX ) ) {
      if ( ROOT == arbres[ arbre ].premier_edit )
	arbres[ arbre ].premier_edit = i;
      arbres[ arbre ].dernier_edit = i;
      if ( i == ex_courant )
	courant_editable = VRAI;
    }
  }

  /* On s'occupe de l'éditable courant... */
  if ( ROOT == arbres[ arbre ].premier_edit ) { /* Aucun objet éditable */
    arbres[ arbre ].edit_courant = ROOT;
    arbres[ arbre ].pos_curseur = 0;

    /* Rien d'autre à faire... */
    return 0;
  }
  if ( ROOT == ex_courant ) /* Pas d'éditable avant : impossible de le conserver ! */
    courant = FAUX;

#ifdef _DEVELOPPE_
/*   (void) printf("%s : Éditable trouvé, courant = %d\n", __FUNCTION__, courant); */
#endif

  if ( VRAI == courant ) {  /* On essaye de conserver l'éditable courant */
    unsigned int nv_editable;

    if ( VRAI == courant_editable ) { /* Il est bien encore éditable */
      return 0; /* donc rien de spécial à faire */
    }

    /* On cherche le nouveau, d'abord juste avant... */
    nv_editable = ARB_EditablePrecedent( arbre, ex_courant, FAUX );
    if ( ROOT == nv_editable ) { /* Échec : on cherche juste après */
      nv_editable = ARB_EditableSuivant( arbre, ex_courant, FAUX );
    }

    /* On met le nouvel éditable */
    BI_ChangerEditable( arbres + arbre, arbres[ arbre ].premier_edit );
  } else { /* On se place sur le premier */
    BI_ChangerEditable( arbres + arbre, arbres[ arbre ].premier_edit );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Trouver l'objet éditable après celui indiqué -------- */

unsigned int ARB_EditablePrecedent(unsigned int arbre, unsigned int actuel, boolean boucle)
{
  int precedent;

  if ( arbre >= BI_nbr_arbres ) {
    (void) fprintf( stderr, "%s,%s l.%u : arbre incorrect [%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre );
    return ROOT;
  }
  if ( ROOT == actuel ) { /* Impossible ! */
    (void) fprintf( stderr, "%s,%s l.%u : actuel = ROOT, absurde\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return ROOT;
  }

  if ( ROOT == arbres[ arbre ].premier_edit ) {
    (void) fprintf( stderr, "%s,%s l.%u : aucun objet éditable dans l'arbre %u\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre );
    return ROOT;
  }

  if ( arbres[ arbre ].premier_edit == arbres[ arbre ].dernier_edit ) { /* Un seul objet éditable */
    return arbres[ arbre ].premier_edit;
  }

  if ( actuel == arbres[ arbre ].premier_edit ) { /* Pas d'éditable après */
    if ( FAUX == boucle ) {
      return ROOT;
    } else { /* On passe sur le premier éditable */
      return arbres[ arbre ].dernier_edit;
    }
  }

  /* Là, on cherche effectivement */
  precedent = actuel - 1;
  while ( precedent > ROOT ) {
    if ( OB_FLAGS( arbres[ arbre ].adresse, precedent ) & EDITABLE ) {
      if ( ( OB_STATE( arbres[ arbre ].adresse, precedent ) & DISABLED ) == FAUX ) {
	/* Objet éditable trouvé */
	return precedent;
      }
    }

    precedent--;
  }

  /* Si l'on est là, c'est qu'il y a un problème ! */
  (void) fprintf( stderr, "%s,%s l.%u : impossible de trouver un éditable [arbre %u, actuel = %u]\n",
		  __FILE__, __FUNCTION__, __LINE__, arbre, actuel );
  if ( FAUX == boucle )
    return ROOT;
  else
    return arbres[ arbre ].premier_edit;;
}

/* -------- Trouver l'objet éditable avant celui indiqué -------- */

unsigned int ARB_EditableSuivant(unsigned int arbre, unsigned int actuel, boolean boucle)
{
  unsigned int suivant;

  if ( arbre >= BI_nbr_arbres ) {
    (void) fprintf( stderr, "%s,%s l.%u : arbre incorrect [%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre );
    return ROOT;
  }

  if ( ROOT == arbres[ arbre ].premier_edit ) {
    (void) fprintf( stderr, "%s,%s l.%u : aucun objet éditable dans l'arbre %u\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre );
    return ROOT;
  }

  if ( arbres[ arbre ].premier_edit == arbres[ arbre ].dernier_edit ) { /* Un seul objet éditable */
    return arbres[ arbre ].premier_edit;
  }

  if ( actuel == arbres[ arbre ].dernier_edit ) { /* Pas d'éditable après */
    if ( FAUX == boucle ) {
      return ROOT;
    } else { /* On passe sur le premier éditable */
      return arbres[ arbre ].premier_edit;
    }
  }

  /* Là, on cherche effectivement */
  suivant = actuel + 1;
  while ( suivant < arbres[ arbre ].nbr_objets ) {
    if ( OB_FLAGS( arbres[ arbre ].adresse, suivant ) & EDITABLE ) {
      if ( ( OB_STATE( arbres[ arbre ].adresse, suivant ) & DISABLED ) == FAUX ) {
	/* Objet éditable trouvé */
	return suivant;
      }
    }

    suivant++;
  }

  /* Si l'on est là, c'est qu'il y a un problème ! */
  (void) fprintf( stderr, "%s,%s l.%u : impossible de trouver un éditable [arbre %u, actuel = %u]\n",
		  __FILE__, __FUNCTION__, __LINE__, arbre, actuel );
  if ( FAUX == boucle )
    return ROOT;
  else
    return arbres[ arbre ].premier_edit;;
}

/* -------- Définir un nouvel objet éditable comme contenant le curseur -------- */

int BI_ChangerEditable(Arbre *arbre, unsigned int nv_editable)
{
  if ( nv_editable == arbre->edit_courant ) /* Rien à faire ! */
    return 0;

  /* Si l'arbre est ouvert : on redessine (si nécessaire)... */
  if ( ( arbre->objets != NULL ) && ( arbre->edit_courant != ROOT ) ) {
    /* On remet le curseur au début du champ : on est sûr que l'on affiche le début du texte */
    BI_PlacerCurseur( arbre, 0 );

    /* On redessine l'ancien : ôte le curseur */
    OBJ_dessine( arbre->objets[ arbre->edit_courant ], arbre->adresse + arbre->edit_courant );
  }

  /* On mémorise le nouvel éditables */
  arbre->edit_courant = nv_editable;

  /* On y place le curseur */
  BI_PlacerCurseur( arbre, BI_InitialiserCurseur( arbre->adresse + nv_editable ) );

  /* Tout s'est bien passé */
  return 0;
}

/* -------- On positionne le curseur quand on arrive dans un objet --------------- */

int BI_InitialiserCurseur(OBJECT *objet)
{
  char *texte = NULL;
  unsigned int pos_curseur = 0;

  if ( NULL == objet ) {
    (void) fprintf(stderr, "BI_InitialiserCurseur : pas d'objet\n");
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return 0;
  }

  if ( objet->ob_type == G_TXT_INFINI ) {
    Texte_Infini *txt_infini = objet->ob_spec.txt_infini;

    if ( NULL == txt_infini ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : pas d'infos complémentaires pour un texte infini\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      return 0;
    }
    texte = txt_infini->texte;
  } else if ( G_NOMBRE == objet->ob_type ) {
    struct infos_editable *nombre;

    nombre = objet->ob_spec.nombre;
    if ( NULL == nombre ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas d'infos complémentaires / nombre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      return 0;
    }
    texte = nombre->texte.texte;
  } else if ( (objet->ob_type == G_FTEXT) || (objet->ob_type == G_FBOXTEXT) ) {
    TEDINFO *tedinfo = objet->ob_spec.tedinfo;

    if ( NULL == tedinfo ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : pas d'infos complémentaires pour un texte éditable\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      return 0;
    }
    texte = tedinfo->te_ptext.editable;
  }

  if ( NULL == texte ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : pas de texte\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return 0;
  }

  pos_curseur = strlen( texte ); /* Curseur à la fin */
  return pos_curseur;
}

/* -------- Avancer le curseur d'un cran dans l'arbre -------- */

int BI_AvancerCurseur(Arbre *arbre, unsigned int nombre_caracteres)
{
  unsigned int obj_editable, i, nv_position;
  int retour = -10; /* Comme ça, si éditable de type inconnu, signalé ! */
  char *texte;

  /* Contrôles sur la position */
  obj_editable = arbre->edit_courant;
  if ( ROOT == obj_editable ) { /* Pas d'objet éditable ! */
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : appel pour un arbre sans objet éditable ! [%p, %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, (void *) arbre, nombre_caracteres );
#endif
    return -1;
  }

  /* On récupère la chaîne contenant le texte */
  switch( arbre->adresse[ obj_editable ].ob_type ) {
  case G_TXT_INFINI:
    {
      struct texte_infini *txt_infini;
    
      txt_infini = arbre->adresse[ obj_editable ].ob_spec.txt_infini;
      if ( NULL == txt_infini ) {
	retour = -1;
	texte = NULL;
      } else {
	texte = txt_infini->texte;
	retour = 0;
      }
    }
    break;

  case G_NOMBRE:
    {
      struct infos_editable *infos;

      infos = arbre->adresse[ obj_editable ].ob_spec.nombre;
      if ( NULL == infos ) {
	retour = -2;
      } else {
	texte = infos->texte.texte;
	retour = 0;
      }
    }
    break;

  case G_FTEXT:
  case G_FBOXTEXT:
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : Type d'objet éditable inconnu [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre->adresse[ obj_editable ].ob_type );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -1;
  }
  if ( retour < 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans la détermination du texte [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
  }


  /* Si on avance de 0 caractères : rien à faire */
  if ( 0 == nombre_caracteres ) return 0;

  /* On avance, caractère par caractère */
  for ( i = 0, nv_position = arbre->pos_curseur ; i < nombre_caracteres ; i++ ) {
    int decalage;

    decalage = TXT_LongueurCaractere( texte + nv_position );
    if ( decalage < 0 ) decalage = -decalage;
    nv_position += decalage;
  }

  /* Et on place le curseur à cette position _en octets_ */
  retour = BI_PlacerCurseur( arbre, nv_position );
  if ( retour < 0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : erreur en voulant placer le curseur [objet %u, arbre %u], erreur code %d\n",
		    __FILE__, __FUNCTION__, __LINE__, obj_editable, (unsigned int) (arbre - arbres), retour );
    OBJ_AfficherContenu( arbre->adresse + obj_editable );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  /* Tout s'est bien passé */
  return retour;
}

/* -------- Avancer le curseur d'un cran dans l'arbre -------- */

int BI_ReculerCurseur(Arbre *arbre, unsigned int nombre_caracteres)
{
  unsigned int obj_editable, i, nv_position, obj_type;
  int retour = -10; /* Comme ça, si éditable de type inconnu, signalé ! */
  char *texte = NULL;

  /* Contrôles sur la position */
  obj_editable = arbre->edit_courant;
  if ( ROOT == obj_editable ) { /* Pas d'objet éditable ! */
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : appel pour un arbre sans objet éditable ! [%p, %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, (void *) arbre, nombre_caracteres );
#endif
    return -1;
  }

  /* De quel type est l'objet ? */
  obj_type = arbre->adresse[ obj_editable ].ob_type;

  /* On récupère la chaîne contenant le texte */
  switch( obj_type ) {
  case G_TXT_INFINI:
    {
      struct texte_infini *txt_infini;
    
      txt_infini = arbre->adresse[ obj_editable ].ob_spec.txt_infini;
      if ( NULL == txt_infini ) {
	retour = -1;
	texte = NULL;
      } else {
	texte = txt_infini->texte;
	retour = 0;
      }
    }
    break;

  case G_NOMBRE:
    {
      struct infos_editable *infos;

      infos = arbre->adresse[ obj_editable ].ob_spec.nombre;
      if ( NULL == infos ) {
	AVERTISSEMENT( "Objet NOMBRE sans information associée [%u]",
		       obj_editable );
	retour = -2;
      } else {
	texte = infos->texte.texte;
	retour = 0;
      }
    }
    break;

  case G_FTEXT:
  case G_FBOXTEXT:
    {
      TEDINFO *tedinfo;

      tedinfo = arbre->adresse[ obj_editable ].ob_spec.tedinfo;
      if ( NULL == tedinfo ) {
	AVERTISSEMENT( "Objet FTEXT/FBOXTEXT sans TEDINFO [%u]",
		       obj_editable );
	retour = -3;
      } else {
	texte = tedinfo->te_ptext.editable;
	retour = 0;
      }
    }
    OPTION_A_FAIRE( "Récupérer la chaîne d'un texte G_FTEXT ou G_FBOXTEXT" );
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : Type d'objet éditable inconnu [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, obj_type );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -1;
  }
  if ( retour < 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans la détermination du texte [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
  }

  /* Si on recule de 0 caractères : rien à faire */
  if ( 0 == nombre_caracteres ) return 0;

  /* On recule, caractère par caractère */
  for ( i = 0, nv_position = arbre->pos_curseur ; i < nombre_caracteres ; i++ ) {
    int decalage = 0;
    unsigned long delta = 1;

    do {
      decalage = TXT_LongueurCaractere( texte + nv_position - delta );
      if ( decalage > 0 ) break;
      delta++;
      if ( delta > nv_position ) {
	(void) fprintf( stderr, "%s,%s l.%u : Problème dans la chaîne multi-octets\n",
			__FILE__, __FUNCTION__, __LINE__ );
	decalage = 1;
      }
    } while ( decalage < 0 );

    nv_position -= decalage;
  }

  /* Et on place le curseur à cette position _en octets_ */
  retour = BI_PlacerCurseur( arbre, nv_position );
  if ( retour < 0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : erreur en voulant placer le curseur [objet %u, arbre %u], erreur code %d\n",
		    __FILE__, __FUNCTION__, __LINE__, obj_editable, (unsigned int) (arbre - arbres), retour );
    OBJ_AfficherContenu( arbre->adresse + obj_editable );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  /* Tout s'est bien passé */
  return retour;
}

/* -------- Positionner et dessiner le curseur dans l'arbre --------
   position = position du curseur dans la chaîne de texte
              exprimée en OCTETS depuis le début de la chaîne
	      (attention, si UTF-8 ou autre codage multi-octets, pas forcément équivalent à des caractères)
 */

int BI_PlacerCurseur(Arbre *arbre, unsigned int position)
{
  unsigned int obj_editable;
  int retour = -10; /* Comme ça, si éditable de type inconnu, signalé ! */

  /* Contrôles sur la position */
  obj_editable = arbre->edit_courant;
  if ( ROOT == obj_editable ) { /* Pas d'objet éditable ! */
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : appel pour un arbre sans objet éditable ! [%p, %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, (void *) arbre, position );
#endif
    return -1;
  }

  /* On vérifie que la position est correcte */
  if ( G_TXT_INFINI == arbre->adresse[ obj_editable ].ob_type ) {
    retour = BI_ControlePositionCurseur_TexteInfini( arbre->adresse[ obj_editable ].ob_spec.txt_infini, &position );
  } else if ( G_NOMBRE == arbre->adresse[ obj_editable ].ob_type ) {
    struct infos_editable *infos;

    infos = arbre->adresse[ obj_editable ].ob_spec.nombre;
    if ( NULL == infos ) {
      retour = -1;
      position = 0;
    } else {
      if ( infos->proprietes & BI_EDITABLE_PROBLEME ) { /* Si message d'alerte affiché, curseur au début */
	retour = 0;
	position = 0;
      } else {
	retour = BI_ControlePositionCurseur_TexteInfini( &(infos->texte), &position );
      }
    }
  } else if ( ( G_FTEXT    == arbre->adresse[ obj_editable ].ob_type ) ||
	      ( G_FBOXTEXT == arbre->adresse[ obj_editable ].ob_type ) ) {
    retour = BI_ControlePositionCurseur_Tedinfo( arbre->adresse[ obj_editable ].ob_spec.tedinfo, &position );
  }
  if ( retour < 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Informations manquantes [objet %u, arbre %u], erreur code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, obj_editable, (unsigned int) (arbre - arbres), retour);
    OBJ_AfficherContenu( arbre->adresse + obj_editable );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  /* On fixe la nouvelle position du curseur */
  arbre->pos_curseur = position;

  /* Si l'arbre est ouvert, on dessine le curseur */
  if ( VRAI == arbre->drapeaux.ouvert ) {
    /* Redessin du texte : effacement d'un éventuel ancien curseur */
    OBJ_dessine( arbre->objets[ arbre->edit_courant ], arbre->adresse + arbre->edit_courant );

    /* Affichage du curseur texte */
    BI_DessinerCurseur( arbre );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Contrôler que le curseur est bien placé [texte infini] -------- */

static int BI_ControlePositionCurseur_TexteInfini(Texte_Infini *texte, unsigned int *position)
{
  size_t lg_texte;

  /* Contrôles de sécurité */
  if ( NULL == texte ) {
    *position = 0; /* Pas de texte -> curseur au début, forcément ! */
    return -1;
  }
  if ( NULL == texte->texte ) {
    *position = 0; /* Pas de texte -> curseur au début, forcément ! */
    return +1;
  }

  /* Le curseur ne doit pas être placé après la fin du texte ! */
  lg_texte = strlen( texte->texte );
  if ( *position > lg_texte ) {
    (void) fprintf( stderr, "%s, %s l.%u : CURSEUR TROP LOIN\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    BI_Bip();
    *position = lg_texte;
  }

  /* On doit voir le curseur */
  if ( *position < texte->prem_car )
    texte->prem_car = *position;
  else if ( *position > texte->prem_car + texte->max_visibles ) {
    texte->prem_car = *position;
    if ( ( texte->prem_car + texte->max_visibles ) > lg_texte  ) {
      texte->prem_car = (lg_texte - texte->max_visibles) + 1;
    }
  }

  return 0;
}

/* -------- Contrôler que le curseur est bien placé [Tedinfo] -------- */

static int BI_ControlePositionCurseur_Tedinfo(TEDINFO *tedinfo, unsigned int *position)
{
  size_t lg_texte;

  /* Contrôles */
  if ( NULL == tedinfo ) {
    *position = 0;
    return -1;
  }
  if ( NULL == tedinfo->te_ptext.editable ) {
    *position = 0;
    return -2;
  }

  lg_texte = strlen( tedinfo->te_ptext.editable ); /* Place prévue pour toujours contenir un 0 final ! */
  if ( lg_texte < tedinfo->te_txtlen - 1 ) { /* Il y a des éditables après... */
    if ( *position > lg_texte )
      *position = lg_texte;
  } else { /* On est sur le dernier caractère éditable */
    if ( ( *position >= lg_texte ) && ( lg_texte > 0 ) )
      *position = lg_texte - 1;
  }

  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Gestion des objets contenant un texte infini ~-~-~-~-~-~-~-~-~ */

/* -------- Gestion du texte infini lui-même -------- */

/* ---- Ajouter un caractère dans un texte infini ----
   texte_infini = adresse de la structure à mettre à jour (DOIT ÊTRE NON-NULL)
   position     = où insérer / remplacer le caractère [indiquée en _octets_]
                  si insérer = insère le caractère à cette position en déplaçant celui existant après
		  0 = début du texte
		  peut indiquer le 0 final => forcément insertion et on ajoute à la fin du texte...
   c            = le caractère à ajouter [a priori, UTF-8 ==> peut occuper plusieurs octets]
   inserer      = VRAI si on veut insérer le caractère
                  FAUX si on veut remplacer le caractère
*/

int BI_AjouterCaractere_TexteInfini(Texte_Infini *texte_infini, unsigned int position, const char *c, boolean inserer)
{
  char *texte;
  size_t lg_texte;
  int n_octets;
  long i;

  /* Réservation éventuelle du texte */
  if ( NULL == texte_infini->texte ) { /* On réserve le texte [objet utilisé pour la 1re fois] */
    texte_infini->texte = (char *) calloc( 21, sizeof( char ) );
    texte_infini->max_cars = 20;

    /* La position est forcément 0 */
    position = 0;
  }
  texte = texte_infini->texte;
  if ( NULL == texte ) { /* Réservation a échoué */
    (void) fprintf( stderr, "%s,%s l.%u : Texte NULL, impossible de réserver...\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return -1;
  }

#ifdef _DEVELOPPE_
  /* (void) fprintf( stderr, "%s,%s l.%u : Texte à modifier = [%s] ; texte à insérer = [%s] en position %u [%s] ; mode insertion = %u\n", */
  /* 		  __FILE__, __FUNCTION__, __LINE__, */
  /* 		  texte_infini->texte, c, position, texte_infini->texte + position, inserer ); */
#endif

  /* Sa longueur */
  lg_texte = strlen( texte );
  if ( position > lg_texte ) {
    (void) fprintf( stderr, "%s,%s l.%u : Position absurde, %u, max = %lu\n",
		    __FILE__, __FUNCTION__, __LINE__, position, lg_texte );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -2;
  }
  if ( position == lg_texte ) inserer = VRAI; /* À la fnin du texte, on ne peut qu'insérer devant le 0 final... */

  if ( ( lg_texte > 0 ) && ( FAUX == inserer ) ) {
    int retour;

    /* On commence par supprimer le caractère à la position indiquée, si l'on est en mode remplacement */
    retour = BI_OterCaractere_TexteInfini( texte_infini, position );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : Erreur en voulant ôter un caractère [code %d]\n",
		      __FILE__, __FUNCTION__, __LINE__, retour );
      return -10;
    }
  }

  /* --- Ajout du caractère, en insertion --- */

  /* Nombre d'octets du caractère à ajouter */
  n_octets = TXT_LongueurCaractere( c );
  if ( n_octets == -1 ) n_octets = 1; /* Codes invalides ou trop courts : supposés Latin-1, conservés */

  /* On contrôle que tout tient, y compris le 0 final */
  if ( lg_texte + n_octets >= texte_infini->max_cars ) {
    char *nv_texte;

    nv_texte = (char *) realloc( texte, texte_infini->max_cars + n_octets + 11 ); /* le caractère + 10 octets  + le 0 final */
    if ( NULL == nv_texte ) {
      (void) fprintf( stderr, "%s,%s l.%u : mémoire insuffisante -- nouveau texte infini\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return -20;
    }
    texte_infini->texte = nv_texte;
    texte_infini->max_cars += n_octets + 10; /* Le 0 final n'est pas compté dans max_cars */
    texte = nv_texte;
  }

  /* On prépare pour l'insertion :
     il faut décaler tous les octets d'autant d'octets que nécessaire pour le caractère
     On décale aussi le 0 final : évite de s'interroger sur la position au bord
  */
  for ( i = lg_texte ; i >= position ; i-- ) {
    texte[ i + n_octets ] = texte[ i ];
  }

  /* On place la chaîne à sa position */
  for ( i = 0 ; i < n_octets ; i++ ) {
    texte[ position + i ] = c[ i ];
  }
  texte[ lg_texte + n_octets ] = 0; /* On s'assure de finir correctement la chaîne */

  /* Ajout réussi */
  return 0;
}

/* -------- Ôter un _caractère_ d'un texte infini, à la position i indiquée */

int BI_OterCaractere_TexteInfini(Texte_Infini *texte_infini, unsigned int position)
{
  size_t lg_texte, i;
  int n_octets;
  char *texte;

  if ( NULL == texte_infini ) return 0; /* Rien à effacer */

  texte = texte_infini->texte;
  if ( NULL == texte ) return 0; /* Rien à effacer */

#ifdef _DEVELOPPE_
  /* (void) fprintf( stderr, "%s,%s l.%u : Texte à modifier = [%s] ; ôter en position %u [%s]\n", */
  /* 		  __FILE__, __FUNCTION__, __LINE__, */
  /* 		  texte_infini->texte, position, texte_infini->texte + position ); */
#endif

  lg_texte = strlen( texte );
  if ( position >= lg_texte ) {
    (void) fprintf( stderr, "%s,%s l.%u : Position absurde, %u, max = %lu\n",
		    __FILE__, __FUNCTION__, __LINE__, position, lg_texte );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -2;
  }

  /* Nombre d'octets du caractère à ajouter */
  n_octets = TXT_LongueurCaractere( texte + position );
#ifdef _DEVELOPPE_
  /* (void) fprintf( stderr, "%s,%s l.%u : Caractère à ôter : occupe %d octet(s)\n", */
  /* 		  __FILE__, __FUNCTION__, __LINE__, n_octets ); */
#endif

  if ( -1 == n_octets ) n_octets = 1; /* Codes invalides ou trop courts : supposés Latin-1, conservés */
  if (  0 == n_octets ) return 0;     /* Caractère fantôme... */

  /* On efface le caractère */
  for ( i = position ; i < lg_texte - n_octets ; i++ ) {
    texte[ i ] = texte[ i + n_octets ];
  }

  /* On met des 0 finaux pour être tranquille */
  for ( i = lg_texte - n_octets ; i <= lg_texte ; i++ )
    texte[ i ] = 0;

#ifdef _DEVELOPPE_
  /* (void) fprintf( stderr, "%s,%s l.%u : Texte obtenu = [%s]\n", */
  /* 		  __FILE__, __FUNCTION__, __LINE__, texte_infini->texte ); */
#endif

  /* Terminé */
  return 0;
}
