/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2003, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Lire une valeur (numérique) dans un objet [champ éditable en général]
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h> /* Pour isdigit() et équivalent */
#include <math.h>  /* Pour isinf() ou équivalent */
#include <limits.h> /* Pour les valeurs extrêmes dans les entiers */

#ifdef _UNIX_
# define _LIBRAIRIE_X_
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "txt_utils.h"
#include "bib_interface.h"

#include "objets.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double BI_AnalyserUnite(double valeur, double *erreur, const char *unite_entree, const char *unite_reference);

static double BI_LireNombre_Nombre (struct infos_editable *infos, double *erreur);
static double BI_LireNombre_Tedinfo(TEDINFO *tedinfo);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Lecture d'un nombre entier signé -------- */

long OBJ_LireEntier_S(unsigned int arbre, unsigned int objet)
{
  double valeur;

  valeur = OBJ_LireNombre( arbre, objet, NULL );
  if ( valeur > (double) LONG_MAX ) {
    AVERTISSEMENT( "Valeur entrée trop grande [%g, max=%ld=%g]\n", valeur, LONG_MAX, (double) LONG_MAX );
    return LONG_MAX;
  }
  if ( valeur < (double) LONG_MIN ) {
    AVERTISSEMENT( "Valeur entrée trop petite [%g, min=%ld=%g]\n", valeur, LONG_MIN, (double) LONG_MIN );
    return LONG_MIN;
  }

  return (long) valeur;
}

/* -------- Lecture d'un nombre entier positif -------- */

unsigned long OBJ_LireEntier_U(unsigned int arbre, unsigned int objet)
{
  double valeur;

  valeur = OBJ_LireNombre( arbre, objet, NULL );
  if ( valeur > (double) ULONG_MAX ) {
    AVERTISSEMENT( "Valeur entrée trop grande [%g, max=%lu=%g]\n", valeur, ULONG_MAX, (double) ULONG_MAX );
    return LONG_MAX;
  }
  if ( valeur < 0.0 ) {
    AVERTISSEMENT( "Valeur entrée négative [%g]\n", valeur );
    return 0;
  }

  return (unsigned long) valeur;
}

/* -------- Lecture d'un nombre réel, avec son erreur éventuellement -------- */

double OBJ_LireNombre(unsigned int arbre, unsigned int num_objet, double *erreur)
{
  double val_erreur = 0.0, valeur = 0.0;
  int retour;
  OBJECT *objet;

  /* Contrôles */
  if ( ROOT == num_objet ) return 0.0; /* Objet "manquant", renvoie 0 sans message */
  retour = OBJ_ControlerValeurs(arbre, num_objet);
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramètres absurdes [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return 0.0;
  }

  /* On récupère l'objet */
  objet = arbres[ arbre ].adresse + num_objet;

  switch ( objet->ob_type )
  {
  case G_TXT_INFINI:
    valeur = TXT_CNV_vNombre_D( objet->ob_spec.txt_infini->texte, NULL, &val_erreur );
    break;

  case G_NOMBRE:
    valeur = BI_LireNombre_Nombre( objet->ob_spec.nombre, &val_erreur );
    break;

  case G_SSMENU:		/* On récupère en pratique l'option sélectionnée */
    valeur = (double) ( objet->ob_spec.menu->option_choisie );
    break;

  case G_BOXCHAR:
    {
      int caractere;

      caractere = OBJ_LireCaractere( arbre, num_objet );
      if ( ( caractere >= '0') && ( caractere <= '9' ) ) {
	valeur = (caractere - '0');
      } else {
	valeur = 0.0;
      }
    }
    break;

  case G_BUTTON:
  case G_STRING:
  case G_TITLE:
  case G_SEPARATEUR:
    /* Le texte est directement accessible */
    valeur = TXT_CNV_vNombre_D( objet->ob_spec.texte, NULL, &val_erreur );
    break;

  case G_TEXT:
  case G_BOXTEXT: /* Texte, mais non éditable -> la valeur est dans le texte */
    valeur = TXT_CNV_vNombre_D( objet->ob_spec.tedinfo->te_ptext.constant, NULL, &val_erreur );
    break;

  case G_FTEXT:
  case G_FBOXTEXT:
    valeur = BI_LireNombre_Tedinfo( objet->ob_spec.tedinfo );
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type d'objet non traité [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    return 0.0;
  }

  if ( erreur != NULL ) {
    *erreur = val_erreur;
  }
  return valeur;
}

/* -------- Analyse d'une unité : conversion vers l'unité de référence si nécessaire -------- */

static double BI_AnalyserUnite(double valeur, double *erreur, const char *unite_entree, const char *unite_reference)
{
  int type_reference, type_donne;
  double facteur_reference, facteur_donne;

  if ( NULL == unite_reference ) {
#if _DEVELOPPE_
    (void) fprintf( stderr, "%s, %s l.%u : Pas d'unité de référence !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
#endif
    return valeur;
  }
  if ( 0 == *unite_entree ) return valeur; /* Pas d'unité indiquée = celle de référence par défaut ! */

  /* A-t-on la même unité ? */
  if ( 0 == strcmp( unite_entree, unite_reference ) ) {
    return valeur; /* Oui -> pas de conversion */
  }

  /* A-t-on la même unité, avec juste un caractère avant ? */
  if ( 0 == strcmp( unite_entree + 1, unite_reference ) ) {
    double facteur = 0.0;

    /* On essaye un préfixe multiplicateur SI
       Table des préfixes : d'après le site du ministère de l'industrie */
    switch ( *unite_entree )
    {
    case 'Y': facteur = 1.0e+24; break; /* Y = yotta = 10^24 */
    case 'Z': facteur = 1.0e+21; break; /* Z = zetta = 10^21 */
    case 'E': facteur = 1.0e+18; break; /* E = exa   = 10^18 */
    case 'P': facteur = 1.0e+15; break; /* P = peta  = 10^15 */
    case 'T': facteur = 1.0e+12; break; /* T = téra  = 10^12 */
    case 'G': facteur = 1.0e+9;  break; /* G = giga  = 10^9 */
    case 'M': facteur = 1.0e+6;  break; /* M = méga  = 10^6 */
    case 'k': facteur = 1.0e+3;  break; /* k = kilo  = 10^3 */
    case 'h': facteur = 1.0e+2;  break; /* h = hecto = 10^2 */
      /* Remarque : le déca n'est pas détecté ainsi, car préfixe da = deux caractères... */
    case 'd': facteur = 1.0e-1;  break; /* d = déci  = 10^-1 */
    case 'c': facteur = 1.0e-2;  break; /* c = centi = 10^-2 */
    case 'm': facteur = 1.0e-3;  break; /* m = milli = 10^-3 */
    case '\xB5': facteur = 1.0e-6;  break; /* µ = micro = 10^-6 [codage Latin-1, windows-1252, Latin-9] */
    case '\xE6': facteur = 1.0e-6;  break; /* µ = micro = 10^-6 [codage TOS ; latin-1 : æ, conflit peut-probable...] */
    case 'n': facteur = 1.0e-9;  break; /* n = nano  = 10^-9 */
    case 'p': facteur = 1.0e-12; break; /* p = pico  = 10^-12 */
    case 'f': facteur = 1.0e-15; break; /* f = femto = 10^-15 */
    case 'a': facteur = 1.0e-18; break; /* a = atto  = 10^-18 */
    case 'z': facteur = 1.0e-21; break; /* n = zepto = 10^-21 */
    case 'y': facteur = 1.0e-24; break; /* y = yocto = 10^-24 */
    }

    if ( facteur > 0.0 ) {
      /* Détection faite => on adapte */
      valeur *= facteur;
      *erreur *= facteur; /* L'erreur aussi change d'échelle ! */

      return valeur;
    }
  }

  /* Idem, mais cas de l'UTF-8 : on pourrait avoir un µ
     [séquence C2 B5 ; latin-1 : Âµ donc sera µ aussi, pas gênant même si très peu probable
                       TOS     : אŒ, conflit très improbable...]
   */
  if ( ( '\xC2' == *unite_entree ) && ( '\xB5' == *(unite_entree + 1) ) &&
       ( 0 == strcmp( unite_entree + 2, unite_reference ) ) ) {
    /* Code µ trouvé */
    valeur *= 1.0e-6;
    *erreur *= 1.0e-6;

    return valeur;
  }

  /* On essaye de simplifier les deux pour qu'elles aient même échelle */
  facteur_reference = TXT_NormaliserUnite( unite_reference, FAUX, &type_reference, NULL );
  facteur_donne     = TXT_NormaliserUnite( unite_entree   , FAUX, &type_donne    , NULL );
  if ( type_reference == type_donne ) {
    /* Unités de même type : conversion possible */
    if ( ( facteur_reference != 0.0 ) && ( facteur_donne != 0.0 ) ) {
      valeur *= facteur_donne / facteur_reference;
      return valeur;
    }

    printf( "Entré : %s [%g] --- Voulu : %s [%g]\n", unite_entree, facteur_donne, unite_reference, facteur_reference );
    OPTION_A_FAIRE( "Unités de même type mais conversion non-multiplicatives" );
  }

  AVERTISSEMENT( "Unité non reconnue, ignorée : %s [au lieu de %s]", unite_entree, unite_reference );
  return valeur;
}

/* -------- Conversion à partir d'un éditable spécialisé pour les nombres -------- */

static double BI_LireNombre_Nombre(struct infos_editable *infos, double *erreur)
{
  char *pos_unite = NULL;
  double valeur = 0.0;

  /* Contrôles préliminaires */
  if ( NULL == infos ) {
    (void) fprintf( stderr, "%s, %s l.%u : Pas d'infos pour l'éditable\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return 0.0;
  }
  if ( NULL == infos->texte.texte ) { /* Peut arriver si texte jamais édité... Vide, donc 0.0 */
    *erreur = 0.0;
    return 0.0;
  }

  /* Y a-t-il une unité ? */
  pos_unite = strchr( infos->texte.texte, ' ' );
  if ( pos_unite != NULL ) { /* L'unité commence juste après l'espace */
    *pos_unite = 0; /* On finit la chaîne pour éviter tout problème dans l'analyse.
		       Comme il s'agit d'un texte infini, obligatoirement réservé avec malloc() -> éditable ! */
  }

  /* Quel type de valeur : simple, formule de base ou formule générique ? */
  if ( infos->proprietes & BI_EDITABLE_FONCTION ) {
    *erreur = 0.0; /* Les erreurs ne sont pas gérées */
    OPTION_A_FAIRE( "Formule complexe, quelconque" );
  } else if ( infos->proprietes & BI_EDITABLE_FORMULE ) {
    *erreur = 0.0; /* Les erreurs ne sont pas gérées */
    OPTION_A_FAIRE( "Formule simple" );
  } else { /* Simple valeur : analyse directe */
    valeur = TXT_CNV_vNombre_D( infos->texte.texte, NULL, erreur );
  }

  /* Analyse de l'unité, si elle existe */
  if ( pos_unite != NULL ) {
    *pos_unite = ' '; /* On remet l'espace pour ne pas perturber des réaffichages */

    valeur = BI_AnalyserUnite( valeur, erreur, pos_unite + 1, infos->unite );
  }

  /* Conversion complémentaire si besoin est */
  if ( infos->proprietes & BI_EDITABLE_ENTIER ) {
    valeur = rint( valeur );
  }

  /* Analyse terminée */
  return valeur;
}

/* -------- Conversion à partir d'un TEDINFO -------- */

static double BI_LireNombre_Tedinfo(TEDINFO *tedinfo)
{
  const char *chaine, *masque, *pos_debut, *pos_virgule;
  double valeur;

  /* Contrôles préliminaires */
  if ( NULL == tedinfo ) {
    (void) fprintf( stderr, "%s, %s l.%u : Pas d'infos pour l'éditable\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return 0.0;
  }

  chaine = tedinfo->te_ptext.editable;
  if ( NULL == chaine ) {
    (void) fprintf( stderr, "%s, %s l.%u : Pas de chaîne utilisateur !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return 0.0;
  }

  masque = tedinfo->te_ptmplt;
  if ( NULL == masque ) {
    (void) fprintf( stderr, "%s, %s l.%u : Pas de masque !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return 0.0;
  }

  /* Conversion proprement dite */
  if ( 0 == *chaine ) return 0.0; /* Chaîne vide */

  /* Position du premier caractère éditable */
  pos_debut = strchr( masque, '_' );
  if ( NULL == pos_debut ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucun caractère éditable !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  /* Position de la virgule */
  pos_virgule = strchr( pos_debut, ',' );
  if ( NULL == pos_virgule ) {
    pos_virgule = strchr( pos_debut, '.' );
  }
  if ( NULL == pos_virgule ) {
    /* Seule une valeur entière peut être donnée */
    valeur = TXT_CNV_vNombre_D( chaine, NULL, NULL );
  } else {
    char *resultat;
    unsigned int i = 0;

    /* On réserve pour la chaîne avec la virgule */
    resultat = (char *) calloc( (size_t) (tedinfo->te_txtlen + 2), sizeof(char) );
    /* virgule + 0 final = +2 */
    if (resultat == NULL) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de créer la chaîne convertie\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return 0.0;
    }

    /* On s'occupe de la partie avant le séparateur décimal. Partie entière, sauf si commence par . */
    while ( '_' == *pos_debut ) {
      if ( 0 == *chaine ) break; /* Fin de la chaîne entrée */

      if ( ( ( *chaine >= '0' ) && ( *chaine <= '9' ) ) || /* Chiffre : seuls valables   */
	   ( *chaine == '-' ) || ( *chaine == '.' ) ) { /* Signe moins, point décimal */
        resultat[ i++ ] = *chaine;
      } else if ( *chaine == ',' ) { /* On n'a indiqué que la partie décimale */
        resultat[ i++ ] = '.';
      } else if ( (*chaine != '+') && (*chaine != ' ') ) { /* Les + et espaces sont bêtement ignorés */
#ifdef _DEVELOPPE_
        printf("Caractère incorrect : %c - Ignoré\n", *chaine);
#endif
      }

      pos_debut++;
      chaine++;
    }

    if ( 0 == *chaine ) { /* La chaîne se finit au séparateur, ou avant (sortie de boucle anticipée) */
      valeur = TXT_CNV_vNombre_D( resultat, NULL, NULL );
      free( resultat );
      return valeur;
    }

    /* Séparateur [point ou virgule] décimal, s'il est bien le premier caractère non-éditable du masque... */
    if ( ( *pos_debut == '.' ) || ( *pos_debut == ',' ) ) {
      if ( strchr( resultat, '.' ) == NULL) { /* Il n'y a pas déjà un point décimal */
        resultat[ i++ ] = '.';
      }
      pos_debut++;
    }

    /* Partie au-delà : partie décimale */
    while ( '_' == *pos_debut ) {
      if ( 0 == *chaine ) /* Fin de la chaîne entrée */
	break;

      if ( ( *chaine >= '0' ) && ( *chaine <= '9' ) ) { /* Chiffre : seuls valables */
        resultat[i++] = *chaine;
      } else if ( *chaine != ' ' ) { /* Les espaces sont ignorés silencieusement */
#ifdef _DEVELOPPE_
        printf("Caractère incorrect : %c - Ignoré\n", *chaine);
#endif
      }
      pos_debut++;
      chaine++;
    }

    valeur = TXT_CNV_vNombre_D( resultat, NULL, NULL );
    free(resultat);
  }

  return valeur;
}

/* -------- Lire un seul caractère dans un objet (le premier de la chaîne) -------- */

int OBJ_LireCaractere(unsigned int arbre, unsigned int num_objet)
{
  int retour;
  OBJECT *objet;
  const char *chaine = NULL;

  /* Contrôles */
  retour = OBJ_ControlerValeurs(arbre, num_objet);
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramètres absurdes [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return 0;
  }

  /* On récupère l'objet */
  objet = arbres[ arbre ].adresse + num_objet;

  switch ( objet->ob_type )
  {
  case G_TXT_INFINI:
    chaine = objet->ob_spec.txt_infini->texte;
    break;

  case G_NOMBRE:
    AVERTISSEMENT( "Utilisation d'un champ nombre pour lire un caractère ! [objet %u]", num_objet );
    chaine = objet->ob_spec.nombre->texte.texte;
    break;

  case G_TEXT:
  case G_BOXTEXT: /* Texte, mais non éditable -> la valeur est dans le texte */
    AVERTISSEMENT( "Utilisation d'un texte non-éditable pour lire un caractère ! [objet %u]", num_objet);
    chaine = objet->ob_spec.tedinfo->te_ptext.constant;
    break;

  case G_BUTTON:
  case G_STRING:
  case G_TITLE:
  case G_SEPARATEUR:
    AVERTISSEMENT( "Utilisation d'un bouton, d'une chaîne (ou apparenté) pour lire un caractère ! [objet %u]", num_objet);
    chaine = objet->ob_spec.texte;
    break;

  case G_BOXCHAR:
    return (int) (objet->ob_spec.gbox.caractere);

  case G_FTEXT:
  case G_FBOXTEXT:
    chaine = objet->ob_spec.tedinfo->te_ptext.editable;
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type d'objet non traité [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    return 0;
  }

  if ( NULL == chaine ) {
    (void) fprintf(stderr, "%s,%s l.%u : Chaîne NULLE dans un objet\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  return *chaine;
}
