/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2003, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Valider les caractères pour un objet éditable
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h> /* Pour isdigit() et équivalent */
#include <math.h>  /* Pour isinf() ou équivalent */

#ifdef _UNIX_
# define _LIBRAIRIE_X_
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "txt_utils.h"
#include "bib_interface.h"

#include "objets.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean CAR_ValiderExposant  (char *texte, unsigned int position);
static boolean CAR_ValiderFermeture (int ouvrant, int fermant, char *texte, unsigned int position);
static boolean CAR_ValiderSeparateur(char *texte, unsigned int position, boolean mode_insertion);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Valider un caractère pour un éditable numérique -------- */

boolean BI_ValiderCaractere_Nombre(struct infos_editable *infos, unsigned int position, int code, boolean mode_insertion)
{
  char *texte;
  size_t lg_texte __attribute__((__unused__));

  if ( isdigit( code ) ) return VRAI; /* Les chiffres sont valables en n'importe-quelle position */
  if ( ( code >= 0 ) && ( code < ' ' ) ) return FAUX; /* Les codes de contrôle sont interdits partout */
  /* Attention, on doit autoriser > 128 pour ±, ·, ... */

  texte = infos->texte.texte;
  if ( NULL == texte ) {
    position = 0; /* Forcément ! */
    lg_texte = 0;
  } else {
    lg_texte = strlen( texte );
  }

  /* S'il y a une unité : tout est autorisé, mais seulement _après_ l'espace séparant l'unité de la valeur */
  if ( infos->proprietes & BI_EDITABLE_UNITE ) {
    if ( NULL != texte ) {
      char *pos_espace;

      pos_espace = strchr( texte, ' ' );
      if ( pos_espace != NULL ) { /* Il y a déjà une espace : une seule autorisée */
	if ( position > (unsigned int) ( pos_espace - texte ) ) { /* On est après : tout est autorisé */
	  return VRAI;
	}
      } else { /* On introduit l'espace, sinon traitement classique */
	if ( ' ' == code ) return VRAI;
      }
    }
  }
  /* L'espace est interdite, pour simplifier, sauf pour séparer l'unité éventuelle */
  if ( ' ' == code ) return FAUX;

  /* Notation scientifique toujours autorisée, le e doit suivre un nombre */
  if ( 'e' == code ) {
    if ( VRAI == CAR_ValiderExposant( texte, position ) ) {
      return VRAI;
    }

    /* Non acceptable comme exposant, mais le sera peut-être dans une formule ensuite : on reste... */
  }

  /* Formule simple autorisée : détection des erreurs lors du décodage. Autorisé pour les entiers & les décimaux */
  if ( infos->proprietes & BI_EDITABLE_FORMULE ) {
    if ( ')' == code ) { /* Dans une formule simple, seulement si ouvrante correspondante ! */
      return CAR_ValiderFermeture( '(', ')', texte, position );
    }
    if ( ']' == code ) { /* Dans une formule simple, seulement si ouvrante correspondante ! */
      return CAR_ValiderFermeture( '[', ']', texte, position );
    }

    if ( VRAI == CAR_EstOperateur( code ) ) return VRAI; /* +,-,/,*,^,(),[] */
  }

  /* Indication d'une erreur associée à la valeur autorisée */
  if ( infos->proprietes & BI_EDITABLE_ERREUR ) {
    if ( '±' == code ) {
      return CAR_ValiderExposant( texte, position ); /* Indication de la forme 8,50±0,02, ± doit suivre un chiffre */
    }

    if ( '(' == code ) return VRAI; /* Indication 8,50(2) ou notation (8,50±0,02)e-4 */
    if ( ')' == code ) return CAR_ValiderFermeture( '(', ')', texte, position );

    if ( 'e' == code ) {
      if ( 0 == position ) return FAUX; /* Serait acceptable juste pour une fonction type exp, interdit avec erreurs ! */
      if ( ')' == texte[ position - 1 ] ) return VRAI; /* Cas 8,50(2)e-4 par exemple */

      return FAUX;
    }
  }

  if ( infos->proprietes & BI_EDITABLE_ENTIER ) { /* Seules les valeurs entières sont autorisées */
    if ( '.' == code ) return FAUX; /* Séparateur décimal interdit */
    if ( ',' == code ) return FAUX;

    /* Seul autre caractère possible : - en première position */
    if ( '-' == code ) {
      if ( infos->proprietes & BI_EDITABLE_MIN ) {
	if ( infos->min >= 0.0 ) { /* Seules des valeurs positives sont possibles */
	  return FAUX; /* Le - initial est donc impossible ! */
	}
      }
      if ( 0 == position ) { /* - autorisé en début de chaîne pour indiquer un nombre négatif */
	return VRAI;
      }
    }

    /* Pour un champ entier, tout le reste est interdit ! */
    return FAUX;
  }

  /* Si on est là : valeur décimale autorisée... */
  if ( infos->proprietes & BI_EDITABLE_FONCTION ) {
#ifdef _DEVELOPPE_
    if ( infos->proprietes & BI_EDITABLE_ERREUR ) {
      (void) fprintf(stderr, "%s, %s l.%u : Champ éditable avec à la fois fonction et erreur !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_AVERTISSEMENT );
    }
#endif
    /* On contrôle, pour limiter les erreurs, les parenthésages fermants */
    if ( ')' == code ) return CAR_ValiderFermeture( '(', ')', texte, position );
    if ( ']' == code ) return CAR_ValiderFermeture( '[', ']', texte, position );

    /* Fonctions "complexes" -> a priori, n'importe quoi est valable, erreurs détectées au décodage...
       Il n'y a que ± interdit : introduit une erreur, non géré pour les fonctions */
    if ( '±' == code ) return FAUX;

    return VRAI;
  }

  /* Séparateur décimal */
  if ( ( '.' == code ) || ( ',' == code ) ) {
    if ( infos->proprietes & BI_EDITABLE_FORMULE ) return VRAI; /* Détection d'erreur lors du décodage */

    /* Sinon, il ne peut y en avoir qu'un seul... */
    if ( NULL == texte ) return VRAI; /* .125 = 0,125, pas très propre mais fréquent */

    return CAR_ValiderSeparateur( texte, position, mode_insertion );

    /* Déjà un séparateur décimal ! */
    return FAUX;
  }

  /* Signe */
  if ( '-' == code ) {
    if ( 0 == position ) { /* Signe initial = nombre négatif, seulement si le minimum le permet */
      if ( infos->proprietes & BI_EDITABLE_MIN ) {
	if ( infos->min >= 0.0 ) { /* Seules des valeurs positives sont possibles */
	  return FAUX; /* Le - initial est donc impossible ! */
	}
      }
      return VRAI;
    }

    /* Ici : utilisable seulement après un e d'exposant */
    if ( 'e' == texte[ position - 1] ) {
      if ( ( VRAI == mode_insertion ) && ( '-' == texte[ position ] ) ) return FAUX; /* Insertion devant un - déjà présent */
      return VRAI;
    }
  }

  return FAUX; /* Autres caractères interdits */
}

/* -------- Vérifier si l'on peut ajouter un e = exposant / notation scientifique -------- */

static boolean CAR_ValiderExposant(char *texte, unsigned int position)
{
  if ( ( NULL == texte ) || ( 0 == position ) ) return FAUX; /* Interdit en début de chaîne ! */

  if ( isdigit( texte[ position - 1 ] ) ) return VRAI; /* On est après un chiffre */

  /* Partout ailleurs : interdit (sf formule complexe : exp(),... */
  return FAUX;
}

/* -------- Vérifier si l'on peut ajouter un caractère de fermeture -------- */

static boolean CAR_ValiderFermeture(int ouvrant, int fermant, char *texte, unsigned int position)
{
  int niveau = 0;
  char *car;

  if ( ( NULL == texte ) || ( 0 == position ) ) return FAUX; /* Interdit en début de chaîne ! */

  car = texte;
  while ( *car != 0 ) {
    if ( *car == ouvrant ) niveau++;
    if ( *car == fermant ) niveau--;

    car++;
  }

  if ( niveau > 0 ) /* Plus d'ouvrants que de fermants */
    return VRAI;

  /* Moins d'ouvrant que de fermant, ou autant : entrée impossible */
  return FAUX;
}

/* -------- Vérifier que l'on peut utiliser un séparateur décimal -------- */

static boolean CAR_ValiderSeparateur(char *texte, unsigned int position, boolean mode_insertion)
{
  char *pos_decimal, *pos_erreur;

  /* Y a-t-il déjà un séparateur décimal ? */
  pos_decimal = strchr( texte, '.' );
  if ( NULL == pos_decimal ) {
    pos_decimal = strchr( texte, ',' );
  }
  if ( NULL == pos_decimal ) /* Aucun séparateur décimal */
    return VRAI; /* On l'introduit donc */

  /* Il y a, ici, au moins un séparateur décimal */
  if ( position == (unsigned int) ( pos_decimal - texte ) ) {
    if ( VRAI == mode_insertion ) {
      return FAUX;
    } else {
      return VRAI;
    }
  }
  pos_erreur = strchr( texte, '±' );
  if ( NULL == pos_erreur ) {
    return FAUX; /* Pas d'erreur -> un seul séparateur autorisé */
  }

  if ( position > (unsigned int) ( pos_erreur - texte ) ) {
    OPTION_A_FAIRE( "Contrôle pour la virgule avec erreur ±, côté erreur" );
  } else {
    OPTION_A_FAIRE( "Contrôle pour la virgule avec erreur ±, côté valeur" );
  }
  return FAUX;
}
