/* Bibliothèque X-Window pour les interfaces
            Routines générales
	    Compatible avec gem_gene.c (version GEM : référence)
   (C) 1998-2015, Emmanuel CURIS - Création le 30 juillet 2015 (à partir de X_gene.c)

   Initialisation du programme : connexion à X, initialisation, chargement du fichier RSC
*/

#define _LIBRAIRIE_X_ 

#ifdef _HP_UNIX_
# define _XOPEN_SOURCE_EXTENDED /* Pour que ça passe sous HP-Unix
				  -- mais ne pas le mettre tout le temps : bloque sous Irix 6.2 (...) */
#endif
#define _BSD_COMPAT /* Pour que ça passe à peu-près sous Irix avec le mode POSIX
		       Il reste cependant des warnings pour setitimer -> à voir ! */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>

#ifdef AVEC_OPEN_GL_
# include <X11/Xmu/Xmu.h> /* Pour le choix de la Colormap */
# include <X11/Xmu/StdCmap.h>
#endif

#include "bib_interface.h"

#include "X_motifs.h"

extern Pixmap PIX_inactif; /* Dans X_objets.c */

/* ****** Variables globales limitées à ce module ************** */

#ifdef AVEC_OPEN_GL_
typedef struct {
    int color_mode;        /* WIN_DONTCARE, WIN_RGBA ou WIN_INDEX */
    int stereo;            /* WIN_DONTCARE, WIN_STEREO ou WIN_MONO */
    int ancillary_buffers; /* OR in desired ancillary buffers */
    int aux_buffers;       /* nombre de tampons auxilliaires souhaités */
    int level;             /* underlay < 0, overlay > 0, main planes 0 */
    int transparent_type;  /* WIN_DONTCARE, WIN_OPAQUE ou WIN_TRANSPARENT */
} WinAttributes;
#endif

/**************** Fonctions internes à ce module *****************/

extern int BI_InitRedessins   (void); /* Dans X_redessins.c */

static int BIX_PreanalyseOptions(int *nbr_options, char **liste_parametres[]);
static int BIX_InitialiserFreedesktop(void);

static int      BIX_ChoisirFonte        (const char *nom_ressource);
static XIMStyle BIX_ChoisirMethodeEntree(XIMStyle methode_1, XIMStyle methode_2);

static boolean XAllouerCouleur(const char *nom, XColor *couleur);

#ifdef AVEC_OPEN_GL_
static boolean XGL_initialisation(void);
static int score_visual_class(XVisualInfo *visual);
static int score_visual(WinAttributes *attr, XVisualInfo *vis, WinAttributes *vis_attr);
#endif

/**************** Fonctions proprement dites *********************************** */

/* ---------- Initialisation du programme ---------------------------------------\
 |   Renvoie VRAI si l'initialisation a réussi                                   |
 |           FAUX sinon                                                          |
 |                                                                               |
 | nom_ressource    : le fichier *.RSC à charger                                 |
 | nbr_arbres       : le nombre d'arbres dans le ressource                       |
 | nbr_parametres   : le nombre de paramètres passés en ligne de commande [argc] |
 | liste_parametres : les paramètres passés en ligne de commande          [argv] |
 \------------------------------------------------------------------------------ */

boolean BI_Initialiser(const char *nom_ressource, unsigned int nbr_arbres, int *nbr_parametres, char **liste_parametres[])
{
  XGCValues GC_valeurs;
  unsigned long GC_masque;
  char *nom_fichier_rsc = NULL;
  int retour = 0;

  /* On analyse la ligne de commande pour traiter [et éliminer] les options X */
  retour = BIX_PreanalyseOptions( nbr_parametres, liste_parametres );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Options en ligne de commande incorrectes [code %d] ! Démarrage annulé\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    return FAUX;
  }

  /* On démarre la connexion avec X-Window */
  BI_Display = XOpenDisplay( NULL ); /* Ouverture du « display » */
  if ( NULL == BI_Display ) {
    (void) fprintf( stderr, "**** LASE / Interface X : initialise ****\n" );
    (void) fprintf( stderr, "Ouverture de l'affichage (« Display ») impossible\n" );
    (void) fprintf( stderr, "V"e_AIGU"rifiez que vous utilisez bien X-Window\n" );
    (void) fprintf( stderr, "V"e_AIGU"rifiez que la variable DISPLAY est correctement configur"e_AIGU"e\n" );
    (void) fprintf( stderr, " -------- Fin du programme --------\n" );
    return FAUX;
  }

  /* Veut-on une communication synchronisée avec le serveur X ? */
  if ( VRAI == BI_synchronisation ) {
    XSynchronize( BI_Display, True ); /* On synchronise les événements X et le programme : débogage
					 Ralentit énormément... */
    (void) printf( "BI_Initialiser : synchronisation X activée\n" );
  }

  /* Test : la bibliothèque supporte-t-elle les locales ? */
  if ( True == XSupportsLocale() ) {
    char *modification_locales;
    XIMStyle methode_choisie = 0;

#ifdef _DEVELOPPE_
    (void) printf( "%s,%s l.%u : Gestion des localisations possibles, support UTF-8 possible.\n",
		   __FILE__, __FUNCTION__, __LINE__ );
#endif

    /* On installe les modificateurs locaux définis
       La valeur de retour ne doit être ni modifiée, ni libérée.
       Elle sera libérée par la Xlib quand nécessaire (fin du programme ?)
     */
    modification_locales = XSetLocaleModifiers( "@im=none" );
    if ( NULL == modification_locales ) {
      (void) fprintf( stderr, "%s,%s l.%u : Codes de modificateurs locaux non-reconnus. Gestion UTF-8 impossible\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    }
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : Modificateurs locaux possibles = %s\n",
		    __FILE__, __FUNCTION__, __LINE__, modification_locales );
#endif

    /* Ouverture d'une méthode d'entrée */
    BI_Methode_entree = XOpenIM( BI_Display, NULL, NULL, NULL );
    if ( NULL == BI_Methode_entree ) {
      (void) fprintf( stderr, "%s,%s l.%u : Impossible d'ouvrir une méthode d'entrée des données...\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      BI_Methode_choisie = None;
    } else {
      XIMStyles *methodes_disponibles;
      char *retour;
      
      /* On récupère les méthodes disponibles pour entrer du texte */
      retour = XGetIMValues( BI_Methode_entree, XNQueryInputStyle, &methodes_disponibles, NULL);
      if ( retour != NULL ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible d'obtenir les méthodes d'entrée disponibles [%s]\n",
			__FILE__, __FUNCTION__, __LINE__, retour );
	methode_choisie = XIMPreeditNone | XIMStatusNone; /* Par sécurité... */
      } else {
	XIMStyle methode_supportee = XIMPreeditNothing | XIMPreeditNone | XIMStatusNothing | XIMStatusNone;
	XIMStyle methode;
	unsigned short i;

#ifdef _DEVELOPPE_
	(void) printf( "%s,%s l.%u : %d méthodes disponibles...\n",
		       __FILE__, __FUNCTION__, __LINE__, methodes_disponibles->count_styles );
#endif
	/* On parcourt la liste pour choisir le meilleur style disponible
	   [Code inspiré de l'exemple du livre X, chapitre 11]
	*/
	for ( i = 0 ; i < methodes_disponibles->count_styles ; i++ ) {
	  methode = methodes_disponibles->supported_styles[ i ];
#ifdef _DEVELOPPE_
	  printf( "** Méthode %d [%lu]", i, (unsigned long) methode );
#endif
	  if ( ( methode_supportee & methode ) == methode ) {
	    methode_choisie = BIX_ChoisirMethodeEntree( methode, methode_choisie );
#ifdef _DEVELOPPE_
	    printf( " UTILISABLE **\n" );
	  } else {
	    printf( " TROP COMPLEXE **\n" );
#endif
	  }
	}

	/* On n'a plus besoin de la liste... */
	XFree( methodes_disponibles ); methodes_disponibles = NULL;
      }

#ifdef _DEVELOPPE_
      printf( "** Méthode retenue : %lu\n", (unsigned long) methode_choisie );
#endif
    }

    BI_Methode_choisie = methode_choisie;
  } else {
    (void) fprintf( stderr, "%s,%s l.%u : La version de X installée ne gère pas les localisations. Gestion de l'UTF-8 incomplète\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    BI_Methode_entree  = NULL;
    BI_Methode_choisie = None;
  }

  /* Lecture des paramètres essentiels à l'utilisation de X */
  BI_Ecran        = DefaultScreenOfDisplay( BI_Display );
  BI_Num_ecran    = DefaultScreen(          BI_Display ); /* Écran utilisé pour l'affichage */

  BI_Visual       = DefaultVisual(   BI_Display, BI_Num_ecran ); /* ????                           */
  BI_Fen_racine   = RootWindow(      BI_Display, BI_Num_ecran ); /* Fenêtre racine X-Window        */
  BI_Palette      = DefaultColormap( BI_Display, BI_Num_ecran ); /* Palette de couleurs par défaut */

  /* Taille de l'écran totales, seront éventuellement corrigées ensuites par l'extension FreeDesktop */
  BI_larg_ecran = WidthOfScreen(  BI_Ecran );
  BI_haut_ecran = HeightOfScreen( BI_Ecran );

  /* Définition des couleurs de base [ancienne table de couleurs VDI/TOS] */
  XAllouerCouleur( "black"           , &BI_Couleurs[ NOIR ]        );
  XAllouerCouleur( "white"           , &BI_Couleurs[ BLANC ]       );
  XAllouerCouleur( "red"             , &BI_Couleurs[ ROUGE ]       );
  XAllouerCouleur( "green"           , &BI_Couleurs[ VERT ]        );
  XAllouerCouleur( "blue"            , &BI_Couleurs[ BLEU ]        );
  XAllouerCouleur( "cyan"            , &BI_Couleurs[ CYAN ]        );
  XAllouerCouleur( "yellow"          , &BI_Couleurs[ JAUNE ]       );
  XAllouerCouleur( "magenta"         , &BI_Couleurs[ VIOLET ]      );
  XAllouerCouleur( "light grey"      , &BI_Couleurs[ GRIS ]        );
  XAllouerCouleur( "dim grey"        , &BI_Couleurs[ GRIS_FONCE ]  );
  XAllouerCouleur( "indian red"      , &BI_Couleurs[ ROUGE_FONCE ] );
  XAllouerCouleur( "dark green"      , &BI_Couleurs[ VERT_FONCE ]  );
  XAllouerCouleur( "midnight blue"   , &BI_Couleurs[ BLEU_FONCE ]  );
  XAllouerCouleur( "light steel blue", &BI_Couleurs[ CYAN_FONCE ]  );
  XAllouerCouleur( "orange"          , &BI_Couleurs[ JAUNE_FONCE ] );
  XAllouerCouleur( "violet red"      , &BI_Couleurs[ VIOLET_FONCE ] );
  XAllouerCouleur( "black"           , &BI_Couleurs[ COULEUR_SPECIALE ] );

  /* Sélection de la fonte */
  retour = BIX_ChoisirFonte( nom_ressource );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Erreur en voulant choisir des fontes [code %d]\n*** Exécution annulée ***\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    BI_Terminer();
    return FAUX;
  }

  /* Initialisation graphique */
  /*   BI_GC_programme = DefaultGC( BI_Display, BI_Num_ecran ); */ /* GC du programme */
  GC_masque = GCFunction | GCForeground | GCBackground | GCLineWidth |
    GCLineStyle | GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule | GCFont;

  GC_valeurs.function   = GXcopy;    /* Équivalent du mode remplacement */
  GC_valeurs.foreground = BI_Couleurs[ NOIR  ].pixel;
  GC_valeurs.background = BI_Couleurs[ BLANC ].pixel;
  GC_valeurs.line_width = 1;
  GC_valeurs.line_style = LineSolid; /* Ligne continue */
  GC_valeurs.cap_style  = CapButt;   /* On trace _tous_ les points */
  GC_valeurs.join_style = JoinMiter; /* On relie les points */
  GC_valeurs.fill_style = FillSolid;
  GC_valeurs.fill_rule  = WindingRule;
  GC_valeurs.font       = BI_Fonte->fid;
  BI_GC_programme = XCreateGC( BI_Display, BI_Fen_racine, GC_masque, &GC_valeurs );

  /* chargement d'une fonte X... */
  /* XSetFont( BI_Display, BI_GC_programme, BI_Fonte->fid ); */

  GC_valeurs.function   = 0x3; /* Mode: remplacement */
  GC_valeurs.background = BI_Couleurs[ GRIS ].pixel;
  GC_valeurs.foreground = BI_Couleurs[ NOIR ].pixel;
  GC_valeurs.font       = BI_Fonte->fid;
  GC_masque = GCForeground | GCBackground | GCFunction | GCFont;

  BI_GC_objets = XCreateGC( BI_Display, BI_Fen_racine, GC_masque, &GC_valeurs );

  /* on s'occupe de l'init. OpenGL, si nécessaire */
#ifdef AVEC_OPEN_GL_
  BI_OpenGL_actif = VRAI; /* Il y a Open-GL supposé actif, puisque compilé avec */

  if ( glXQueryExtension( BI_Display, NULL, NULL ) == False ) { /* Pas d'Open-GL */
    (void) fprintf( stderr, "Open GL non installé ! Compilez une version sans Open GL\n" );
    (void) fprintf( stderr, "Aucun affichage 3d ne sera possible\n" );
    BI_OpenGL_actif = FAUX;
  } else {
    if ( XGL_initialisation() == FAUX ) {
      (void) fprintf( stderr, "Erreur lors de l'init Open GL - Affichages 3d indisponibles\n" );
      (void) fprintf( stderr, "Essayez une version sans Open GL (recompilez le programme !)\n" );
      BI_OpenGL_actif = FAUX;
    }
  }
#endif

  /* Protocoles de communication avec le gestionnaire de fenêtres */
  BI_ProtocoleFermerFenetre = XInternAtom( BI_Display, "WM_DELETE_WINDOW", True );
  if ( (Atom) None == BI_ProtocoleFermerFenetre ) {
    (void) fprintf( stderr, "%s,%s l.%u : protocole de fermeture des fenêtres inconnu !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
  }

  /* Extensions freedesktop (adapte au besoin la taille de l'écran) */
  BIX_InitialiserFreedesktop();

  /* Init. pour les redessins */
  retour = BI_InitRedessins();	/* Dans X_redessins.c */
  if ( 0 != retour ) {
    (void) fprintf( stderr, "%s,%s l.%u : erreur dans BI_InitRedessins() [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    BI_Terminer();
    return FAUX;
  }

  /* On s'occupe de la VDI */
  _PRG_station = VDI_ouvre( 1 ); /* On ouvre une station sur l'écran */

  gdos = -2; /* Pas de GDOS sous cette version... On verra plus tard */

  /* Combien d'arbres au total ? */
  BI_nbr_arbres  = nbr_arbres; /* Arbres créés au vol par le programme */
  BI_nbr_arbres += 3; /* Arbres internes : sélecteur, alerte, sous-menus,... */

  /* On cherche où se trouve le fichier RSC */
  nom_fichier_rsc = BI_TrouverRSC( nom_ressource );
  if ( ( NULL == nom_fichier_rsc ) && ( nom_ressource != NULL ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de trouver le fichier RSC [%s]\n",
		    __FILE__, __FUNCTION__, __LINE__, nom_ressource );
    (void) fprintf( stderr, " ---- Chargement du programme annulé ----\n" );
    BI_Terminer();
    return FAUX;
  }

  if ( NULL != nom_fichier_rsc ) { /* On veut charger un RSC */
    struct rsc_infos infos;

    /* Combien d'arbres dans le RSC ? */
    retour = BI_RSC_LireEnTeteFichier( nom_fichier_rsc, &infos );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR %d en lisant l'en-tête du fichier RSC [%s]\n",
		      __FILE__, __FUNCTION__, __LINE__, retour, nom_fichier_rsc );
      (void) fprintf( stderr, " ---- Chargement du programme annulé ----\n" );
      free( nom_fichier_rsc ); nom_fichier_rsc = NULL;
      BI_Terminer();
      return FAUX;
    }

    BI_nbr_arbres += infos.nbr_arbres;
    (void) fclose( infos.fichier ); infos.fichier = NULL;
  }

  /* On crée les arbres */
  arbres = BI_ARB_CreerArbres( BI_nbr_arbres );
  if ( NULL == arbres ) {
    (void) fprintf( stderr, "%s,%s l.%u : Mémoire insuffisante pour réserver les %u arbres\n",
		    __FILE__, __FUNCTION__, __LINE__, BI_nbr_arbres);
    free( nom_fichier_rsc ); nom_fichier_rsc = NULL;
    BI_Terminer();
    return FAUX;
  }

  /* On charge éventuellement le RSC */
  if ( nom_fichier_rsc != NULL ) {
    retour = BI_ChargerRSC( nom_fichier_rsc, 0, NULL );
    free( nom_fichier_rsc ); nom_fichier_rsc = NULL;
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR %d en lisant le fichier RSC [%s]\n",
		      __FILE__, __FUNCTION__, __LINE__, retour, nom_ressource );
      (void) fprintf( stderr, " ---- Chargement du programme annulé ----\n" );
      BI_Terminer();
      return FAUX;
    }
  }

  /* Préparation du sélecteur de fichiers */
  retour = BIX_FSEL_Initialiser();
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : ERREUR à l'initialisation du sélecteur de fichiers [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    BI_Terminer();
    return FAUX;
  }

  /* Init. des différents motifs de remplissage */
  PIX_inactif = XCreateBitmapFromData( BI_Display, BI_Fen_racine, (const char *) RMP_inactif, 32, 32 );

  return VRAI; /* Tout s'est bien passé, on peut continuer */
}

/* -------- Choix de la police à utiliser -------- */

static int BIX_ChoisirFonte(const char *nom_ressource)
{
#ifdef _utf8_
  char **liste_manquants;
  char *code_manquant = NULL;
  int nombre_manquants = 0;
#endif

  /* Y a-t-il déjà un nom de police proposé ? 
     [passé comme option en ligne de commande, traitée par BIX_PreanalyseOptions() : option -fn
      NE PAS CHANGER LE NULL de la défition dans X_var.c [car doit être une chaîne libérable par free()
     ]
  */
  if ( BI_nom_fonte != NULL ) {
    BI_Fonte = XLoadQueryFont( BI_Display, BI_nom_fonte );
    if ( NULL == BI_Fonte  ) {
      (void) fprintf( stderr, "%s,%s l.%u : Nom de fonte indiqué incorrect [%s]. La fonte par défaut sera utilisée\n",
		      __FILE__, __FUNCTION__, __LINE__, BI_nom_fonte );
      /* On libère, car sera modifié ensuite donc perdu... */
      free( BI_nom_fonte ); BI_nom_fonte = NULL;
    }
  }

  if ( NULL == BI_Fonte ) { /* Pas de fonte passée en paramètre, ou nom de fonte incorrect */
    /* On recherche la fonte dans le .Xdefaults, d'après le nom du ressource utilisé [s'il existe] */
    if ( nom_ressource != NULL ) { /* Il y a bien un nom de ressource passé */
      char *nom_programme;

      nom_programme = strdup( nom_ressource );
      if ( NULL == nom_programme ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de déterminer le nom de ressource X à utiliser [%s]\n",
			__FILE__, __FUNCTION__, __LINE__, nom_ressource );
	BI_nom_fonte = NULL;
      } else {
	char *position_point;
	boolean decale = FAUX;

	/* Si le nom du ressource contient un suffixe, on le supprime 
	   [Exemple : pour LASE, exafs.rsc => on fait exafs]
	 */
	position_point = strchr( nom_programme, '.' );
	if ( position_point != NULL ) {
	  if ( position_point == nom_programme ) { /* Nom de la forme .xyz : on  garde xyz */
	    nom_programme++;
	    decale = VRAI;
	  } else { /* Nom de la forme A.xyz */
	    nom_programme[ position_point - nom_programme ] = 0;
	  }
	}

	/* Remarque : la chaîne doit être libérée par la Xlib, d'après la doc.
	   ==> rien à faire en quittant... mais message valgrind sur certains postes !*/
	BI_nom_fonte = XGetDefault( BI_Display, nom_programme, "fonte" );

	/* On essaye de trouver la fonte indiquée, s'il y en avait une */
	if ( BI_nom_fonte != NULL ) {
	  BI_Fonte = XLoadQueryFont( BI_Display, BI_nom_fonte );
	  if ( NULL == BI_Fonte ) { /* Nom de fonte incorrect */
	    (void) fprintf( stderr, "%s,%s l.%u : Fonte du .Xdefaults incorrecte pour la ressource %s : %s\n",
			    __FILE__, __FUNCTION__, __LINE__, nom_programme, BI_nom_fonte );
	  }
#ifdef _DEVELOPPE_
	} else {
	  (void) printf( "%s,%s l.%u : Pas de fonte spécifiée dans la ressource %s\n",
			 __FILE__, __FUNCTION__, __LINE__, nom_programme );
#endif
	}

	/* Plus besoin du nom du programme : on libère */
	if ( VRAI == decale ) nom_programme--; /* On annule le décalage auparavant, si nécessaire */
	free( nom_programme ); nom_programme = NULL;
      }
    }
  }

  if ( NULL == BI_Fonte ) {
    /* Pas de fonte trouvée dans la ressource associée au programme
       ==> on cherche une fonte dans la ressource associée à la bibliothèque X/GEM
     */
    /* Remarque : la chaîne doit être libérée par la Xlib, d'après la doc.
       ==> rien à faire en quittant... mais message valgrind sur certains postes !*/
    BI_nom_fonte = XGetDefault( BI_Display, "bib_X", "fonte" );
    if ( BI_nom_fonte != NULL ) {
      BI_Fonte = XLoadQueryFont( BI_Display, BI_nom_fonte );
      if ( NULL == BI_Fonte ) { /* Nom de fonte incorrect */
	(void) fprintf( stderr, "%s,%s l.%u : Fonte du .Xdefaults incorrecte pour la ressource bib_X : %s\n",
			    __FILE__, __FUNCTION__, __LINE__, BI_nom_fonte );
      }
#ifdef _DEVELOPPE_
    } else {
      (void) printf( "%s,%s l.%u : Pas de fonte spécifiée dans la ressource bib_X\n",
		      __FILE__, __FUNCTION__, __LINE__ );
#endif
    }
  }

  if ( NULL == BI_Fonte ) {
    /* Tout a raté, on utilise la fonte par défaut */
    BI_nom_fonte = (char *) BI_fonte_defaut; /* Cast possible car on ne libèrera pas... */
    BI_Fonte = XLoadQueryFont( BI_Display, BI_fonte_defaut );
  }

  if ( NULL == BI_Fonte ) { /* Nom de fonte incorrect */
    (void) fprintf( stderr, "%s,%s l.%u : Fonte par défaut incorrecte : %s\n",
		    __FILE__, __FUNCTION__, __LINE__, BI_fonte_defaut );
    (void) fprintf( stderr, "Chargement de la fonte impossible...\n Chargement du programme annul"e_AIGU"\n" );
    (void) fprintf( stderr, "Essayez de modifier la fonte par défaut (BI_fonte_defaut) dans X_var.c\n" );
    (void) fprintf( stderr, "   (Fonte actuelle : %s)\n", BI_fonte_defaut );
    return -1;
  }

  /* On définit les dimensions des caractères pour cette fonte */
  BI_larg_car = BI_Fonte->max_bounds.width;
  BI_haut_car = BI_Fonte->max_bounds.ascent + BI_Fonte->max_bounds.descent + 2;
  /* +2 = +1 car ligne de base oubliée sinon ; +1 pour avoir un peu de place en-dessous */

  /* On déduit de cela le jeu de polices pour l'UTF-8
     Le code manquant ne doit pas être libéré, il le sera par XFreeFontSet
   */
#ifdef _utf8_
  BI_FonteUTF8 = XCreateFontSet( BI_Display, BI_nom_fonte, 
				 &liste_manquants, &nombre_manquants, &code_manquant );
  if ( NULL == BI_FonteUTF8 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de créer la fonte UTF-8 %s - support UTF-8 partiel.\n",
		    __FILE__, __FUNCTION__, __LINE__, BI_nom_fonte );
  } else {
    int i;

    (void) printf( "%s,%s l.%u : Fonte UTF-8 avec %d jeux de caractères manquants, le code %s sera utilisé à la place\n",
      __FILE__, __FUNCTION__, __LINE__, nombre_manquants, code_manquant );
    for ( i = 0 ; i < nombre_manquants ; i++ ) {
      (void) printf( "Jeu manquant : %s\n", liste_manquants[ i ] );
    }

    /* Plus besoin de la liste des fontes manquantes */
    if ( liste_manquants != NULL ) XFreeStringList( liste_manquants );
  }
#endif

  /* Chargement de la police réussi */
  return 0;
}

/* -------- Recherche, initialisation des protocoles avancés :
            convention freedesktop [http://standards.freedesktop.org/wm-spec/1.4]
 */

static int BIX_InitialiserFreedesktop(void)
{
  Atom _NET_SUPPORTING_WM_CHECK, _NET_WORKAREA, type_reel;
  Window fenetre_fille;
  int format_reel;
  unsigned long nombre_atomes, reste_ensuite;
  void *proprietes_X = NULL;

  /* La propriété _NET_SUPPORTING_WM_CHECK existe-t-elle ? */
  _NET_SUPPORTING_WM_CHECK = XInternAtom( BI_Display, "_NET_SUPPORTING_WM_CHECK", True );
  if ( None == _NET_SUPPORTING_WM_CHECK ) {
    /* Non : mauvais signe... inutile de chercher plus loin... */
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s : protocole freedesktop inconnu !\n", __FUNCTION__ );
#endif
      BI_Freedesktop_OK = FAUX;
      return 0;			/* Pas de problème particulier */
  }

  /* Est-elle connue de la fenêtre racine ? [Doc dit type WINDOW, mais en fait type ATOM : ??? ==> "any" pour être tranquille] */
  if ( XGetWindowProperty( BI_Display, BI_Fen_racine, _NET_SUPPORTING_WM_CHECK, 0, 1, False, AnyPropertyType,
			   &type_reel, &format_reel, &nombre_atomes, &reste_ensuite, (unsigned char **) &proprietes_X ) != Success ) {
    (void) fprintf( stderr, "%s : ERREUR dans XGetWindowProperty !\n", __FUNCTION__ );
    BI_Freedesktop_OK = FAUX;
    return -1;
  }
  if ( ( None == type_reel ) || ( 0 == nombre_atomes ) || ( NULL == proprietes_X ) ) {
    (void) fprintf( stderr, "\n%s : _NET_SUPPORTING_WM_CHECK défini comme atome, mais pas associé à la fenêtre racine !\n\n"
		    "type_reel = %d ; nombre_atomes = %lu ; proprietes_X = %p\n",
		    __FUNCTION__, (int) type_reel, nombre_atomes, proprietes_X );
    BI_Freedesktop_OK = FAUX;
    XFree( proprietes_X );
    return 0;
  }
  if ( XA_WINDOW != type_reel ) {
    (void) fprintf( stderr, "%s : _NET_SUPPORTING_WM_CHECK non-défini comme XA_WINDOW [type : %d] !\n",
		    __FUNCTION__, (int) type_reel );
    BI_Freedesktop_OK = FAUX;
    XFree( proprietes_X );
    return 0;
  }

  fenetre_fille = ( (Window *) proprietes_X)[ 0 ];
  XFree( proprietes_X ); proprietes_X = NULL;

  /* Est-elle connue de la fenêtre fille ? */
  if ( XGetWindowProperty( BI_Display, fenetre_fille, _NET_SUPPORTING_WM_CHECK, 0, 1, False, AnyPropertyType,
			   &type_reel, &format_reel, &nombre_atomes, &reste_ensuite, (unsigned char **) &proprietes_X ) != Success ) {
    (void) fprintf( stderr, "%s : ERREUR dans XGetWindowProperty !\n", __FUNCTION__ );
    BI_Freedesktop_OK = FAUX;
    return -1;
  }
  if ( ( None == type_reel ) || ( 0 == nombre_atomes ) || ( NULL == proprietes_X ) ) {
    (void) fprintf( stderr, "%s : _NET_SUPPORTING_WM_CHECK pas associée à la fenêtre fille indiquée !\n"
		    "type_reel = %d ; nombre_atomes = %lu ; proprietes_X = %p --- fille = %lu\n",
		    __FUNCTION__, (int) type_reel, nombre_atomes, (void *) proprietes_X, (unsigned long) fenetre_fille );
    BI_Freedesktop_OK = FAUX;
    return 0;
  }
  if ( XA_WINDOW != type_reel ) {
    (void) fprintf( stderr, "%s : _NET_SUPPORTING_WM_CHECK non-défini comme XA_WINDOW [type : %d] !\n",
		    __FUNCTION__, (int) type_reel );
    BI_Freedesktop_OK = FAUX;
    XFree( proprietes_X );
    return 0;
  }
  if ( fenetre_fille != *((Window *) proprietes_X) ) {
    (void) fprintf( stderr, "%s : Les fenêtres indiquées dans _NET_SUPPORTING_WM_CHECK ne coïncident pas !\n",
		    __FUNCTION__ );
    BI_Freedesktop_OK = FAUX;
    XFree( proprietes_X );
    return 0;
  }

  /* Plus besoin de cela */
  XFree( proprietes_X ); proprietes_X = NULL;

#ifdef _DEVELOPPE_
  (void) fprintf( stderr, "initialise : protocole freedesktop SUPPORTÉ !\n" );
#endif
  BI_Freedesktop_OK = VRAI;

  /* On essaye de récupérer la taille utilisable du bureau */
  _NET_WORKAREA = XInternAtom( BI_Display, "_NET_WORKAREA", True );
  if ( None != _NET_WORKAREA ) {
    unsigned long *tailles = NULL;

    /* Existe bien => on essaye de le récupérer */
    if ( XGetWindowProperty( BI_Display, BI_Fen_racine, _NET_WORKAREA, 0, 32 * 4 * 6, False, XA_CARDINAL,
			     &type_reel, &format_reel, &nombre_atomes, &reste_ensuite, (unsigned char **) &tailles ) != Success ) {
      (void) fprintf( stderr, "%s : ERREUR dans XGetWindowProperty !\n", __FUNCTION__ );
      BI_Freedesktop_OK = FAUX;
      return -1;
    }
    if ( ( None == type_reel ) || ( 0 == nombre_atomes ) || ( NULL == tailles ) ) {
      (void) fprintf( stderr, "%s : _NET_WORKAREA pas associée à la fenêtre racine, ou avec le mauvais type !\n"
		      "type_reel = %d ; nombre_atomes = %lu ; tailles = %p\n",
		      __FUNCTION__, (int) type_reel, nombre_atomes, (void *) tailles );
    } else {
      unsigned int nombre_ecrans, i;
      unsigned long x_ecran __attribute__((__unused__)), y_ecran __attribute__((__unused__)), larg_ecran, haut_ecran;

      if ( reste_ensuite > 0 ) {
	AVERTISSEMENT( "Plus de 6 écrans virtuels [%lu]: pas géré...", reste_ensuite >> 2 );
      }
      /* Nombre d'écrans connus : il y a 4 atomes par écran */
      nombre_ecrans = nombre_atomes >> 2;
      for ( i = 0 ; i < nombre_ecrans ; i++ ) {
	x_ecran = tailles[ 4 * i  ];
	y_ecran = tailles[ 4 * i + 1];
	larg_ecran = tailles[ 4 * i + 2 ];
	haut_ecran = tailles[ 4 * i + 3 ];

	if ( ( 0 == larg_ecran ) || ( 0 == haut_ecran ) ) {
	  AVERTISSEMENT( "Dimensions d'écran absurdes dans _NET_WORKAREA [%lu * %lu] : on ignore ce flag...", larg_ecran, haut_ecran );
	  break;
	}

	if ( larg_ecran < BI_larg_ecran ) BI_larg_ecran = larg_ecran;
	if ( haut_ecran < BI_haut_ecran ) BI_haut_ecran = haut_ecran;
      }

#ifdef _DEVELOPPE_
      fprintf( stderr, "%s : Dimensions de travail : %u * %u\n", __FUNCTION__, BI_larg_ecran, BI_haut_ecran );
#endif
    }
  }

  /* On récupère les ATOM pour ce que l'on utilise */
  _NET_ACTIVE_WINDOW = XInternAtom( BI_Display, "_NET_ACTIVE_WINDOW", True );

  /* Terminé */
  return 0;
}

/* -------- Sélection de la meilleure méthode d'entrée --------
      [code repris & francisé de l'exemple 11.5 du chapitre 11 du manuel X11R5]
 */

static XIMStyle BIX_ChoisirMethodeEntree(XIMStyle methode_1, XIMStyle methode_2)
{
  XIMStyle s,t;
  XIMStyle preedit = XIMPreeditArea | XIMPreeditCallbacks | XIMPreeditPosition | XIMPreeditNothing | XIMPreeditNone;
  XIMStyle status  = XIMStatusArea  | XIMStatusCallbacks                       | XIMStatusNothing  | XIMStatusNone ;

  /* Les cas triviaux */
  if ( methode_1 == methode_2 ) return methode_1;
  if ( methode_1 == 0 ) return methode_2;
  if ( methode_2 == 0 ) return methode_1;

  /* Mêmes bits pour la pré-édition et le statut, mais les autres peuvent différer ==> arbitrairement, la 1re */
  if ( (methode_1 & (preedit | status) ) == ( methode_2 & (preedit | status) ) )
    return methode_1;

  /* On contrôle d'abord la préédition */
  s = methode_1 & preedit;
  t = methode_2 & preedit;
  if ( s != t ) {
    /* Elles diffèrent : on choisit la plus performante */
    if ( s | t | XIMPreeditCallbacks ) /* Au moins une des deux supporte les « callbacks » */
      return ( XIMPreeditCallbacks == s ) ? methode_1 : methode_2;

    if ( s | t | XIMPreeditPosition ) /* Au moins une des deux supporte la gestion d'un curseur */
      return ( XIMPreeditPosition  == s ) ? methode_1 : methode_2;

    if ( s | t | XIMPreeditArea ) /* Au moins une des deux supporte une région spécifique */
      return ( XIMPreeditArea      == s ) ? methode_1 : methode_2;

    if ( s | t | XIMPreeditNothing ) /* Au moins une deux supporte la prise en charge par le système */
      return ( XIMPreeditNothing   == s ) ? methode_1 : methode_2;

    /* Si on est là :
       - soit s et t n'ont toutes les deux que XIMPreeditNone de mis ==> elles sont égales : impossible !
       - soit s a plus d'un bit mis, théoriquement impossible !
    */
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans les méthodes d'entrée, plusieurs bits « Preedit » mis ? [s = %lu, t = %lu]\n",
		    __FILE__, __FUNCTION__, __LINE__, (unsigned long) s, (unsigned long) t );

    return 0;			/* Problème dans les méthodes... */
  }

  /* Si les méthodes de préédition sont identiques, on compare l'affichage du statut */
  s = methode_1 & status;
  t = methode_2 & status;
  if ( s | t | XIMStatusCallbacks ) /* Au moins une des deux supporte les « callbacks » */
    return ( XIMStatusCallbacks == s ) ? methode_1 : methode_2;

  if ( s | t | XIMStatusArea ) /* Au moins une des deux supporte une région spécifique */
    return ( XIMStatusArea      == s ) ? methode_1 : methode_2;

  if ( s | t | XIMStatusNothing ) /* Au moins une deux supporte la prise en charge par le système */
    return ( XIMStatusNothing   == s ) ? methode_1 : methode_2;

    /* Si on est là :
       - soit s et t n'ont toutes les deux que XIMStatusNone de mis ==> elles sont égales : impossible !
       - soit s a plus d'un bit mis, théoriquement impossible !
    */
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans les méthodes d'entrée, plusieurs bits « Status » mis ? [s = %lu, t = %lu]\n",
		    __FILE__, __FUNCTION__, __LINE__, (unsigned long) s, (unsigned long) t );

    return 0;			/* Problème dans les méthodes... */
}


static boolean XAllouerCouleur(const char *nom, XColor *couleur)
{
  XColor reelle;
  Status allouee; /* (0 = non, reste = oui) */

  if ( (couleur == NULL) || (nom == NULL) )
    return FAUX;

  allouee = XAllocNamedColor(BI_Display, BI_Palette, nom, couleur, &reelle);
  if (allouee == FAUX) {
#ifdef _ENGLISH_
    printf("Can't allocate color %s\n", nom);
#else
    printf("XAllouerCouleur : impossible de d"e_AIGU"clarer la couleur « %s »\n", nom);
#endif
    return FAUX;
  }

  return VRAI;
}

/* -------- Analyse des options de la ligne de commande : depend du systeme... -------- */

static int BIX_PreanalyseOptions(int *nbr_options, char **liste_parametres[])
{
  boolean option_supprimee = FAUX;
  unsigned int i, nombre_options;
  char *option;

  if ( ( NULL == nbr_options) || ( NULL == liste_parametres ) ) return 0;	/* Pas d'option passee */

  nombre_options = *nbr_options;
  if ( ( nombre_options < 2 ) || ( NULL == *liste_parametres ) ) { /* La premiere option est le nom du programme... */
    return 0;
  }

  for ( i = 0 ; i < nombre_options ; i++ ) {
    /* Quelle option est concernee ? */
    option = (*liste_parametres)[ i ];
    if ( NULL == option ) {
      continue;
    }

    /* Est-ce reellement une option ? */
    if ( '-' != *option ) {
      continue;
    }
    if ( '-' == *(option + 1) ) { /* -- : la suite correspond a des noms de fichier uniquement */
      break;
    }

    /* On fait une comparaison exacte */
    if ( strcmp( option + 1, "b2=3" ) == 0 ) {
      /* Option -b3=2 : bouton 3 (droit) identique au bouton 2 (milieu) */
      BI_2_egale_3 = VRAI;
      (*liste_parametres)[ i ] = NULL; /* On supprime cette option de la liste */
    } else if ( strcmp( option + 1, "direct" ) == 0 ) {
      BI_GL_direct = VRAI; /* Affichages Open-gL directs à l'écran */
      (*liste_parametres)[ i ] = NULL; /* On supprime cette option de la liste */
    } else if ( strcmp( option + 1, "fn" ) == 0 ) {
      /* Option -fn : choix de la fonte */
      (*liste_parametres)[ i ] = NULL; /* On supprime cette option de la liste */

      /* On lit le nom de la fonte dans le parametre suivant */
      i++;
      if ( i == nombre_options ) { /* Il n'y a rien après ! */
	fprintf(stderr, "Erreur : -fn doit être suivi d'un nom de fonte !\n");
	return -1;
      }
      BI_nom_fonte = strdup( (*liste_parametres)[ i ] );
      (*liste_parametres)[ i ] = NULL;
    } else if ( strcmp( option + 1, "geom" ) == 0 ) {
      /* Option -geom XXXxYYY = taille par défaut des fenêtres, en pixels */
      (*liste_parametres)[ i ] = NULL; /* On supprime cette option de la liste */
      i++;
      if ( i == nombre_options ) { /* Il n'y a rien après ! */
	fprintf(stderr, "Erreur : -geom doit être suivi des dimensions en pixels !\n");
	return -2;
      }
      sscanf( (*liste_parametres)[ i ], "%ux%u", &BI_larg_defaut, &BI_haut_defaut);
      (*liste_parametres)[ i ] = NULL;
    } else if ( strcmp( option + 1, "indirect" ) == 0 ) {
#ifdef _DEVELOPPE_
      BI_GL_direct = FAUX; /* Affichages Open-gL directs à l'écran */
#endif
      (*liste_parametres)[ i ] = NULL;
    } else if ( strcmp( option + 1, "sync" ) == 0 ) {
#ifdef _DEVELOPPE_
      printf( "On force la synchronisation X\n" );
      BI_synchronisation = VRAI;
#else
      (void) fprintf( stderr, "Option -sync : version de développement uniquement\n" );
#endif
      (*liste_parametres)[ i ] = NULL;
    }
  }

  /* On supprime de la liste les options mises a NULL */
  do {
    option_supprimee = FAUX;
    for ( i = 0 ; i < nombre_options ; i++ ) {
      if ( NULL == (*liste_parametres)[ i ] ) {
	unsigned int j;

	/* On decale les suivants */
	for ( j = i + 1 ; j < nombre_options ; j++ ) {
	  (*liste_parametres)[ j - 1 ] = (*liste_parametres)[ j ];
	}

	/* Une option de moins */
	option_supprimee = VRAI;
	nombre_options--;
      }
    }
    /* En cas de plusieurs NULL successifs, on ne fait sauter ainsi que le premier ! */
  } while ( VRAI == option_supprimee );

#ifdef _DEVELOPPE_
  printf( "%s : %u option(s) restante(s) : ", __FUNCTION__, nombre_options );
  for ( i = 0 ; i < nombre_options ; i++ ) {
    printf( "%s ;", (*liste_parametres)[ i ] );
  }
  printf( "\n" );
#endif

  /* On memorise le nouveau nombre d'options */
  *nbr_options = nombre_options;

  /* Preanalyse terminee sans erreur */
  return 0;
}

#ifdef AVEC_OPEN_GL_

/* ~~~~~~~~ Gestion Open GL ~~~~~~~~
   Le code ici est inspiré du code dans glx_samples fourni avec Digital Unix
   (Code de John Denis) pour la recherche du meilleur Visual et de la meilleure Colormap
 */

#define WIN_COLOR_MODE_ATTR		1
#define WIN_STEREO_ATTR			2
#define WIN_ANCILLARY_BUFFERS_ATTR	4
#define WIN_AUX_BUFFERS_ATTR		8
#define WIN_LEVEL_ATTR			16
#define WIN_TRANSPARENT_TYPE_ATTR	32

#define WIN_DONTCARE		2

#define WIN_INDEX      	 	0  /* Matches value returned by glXGetConfig */
#define WIN_RGBA       	 	1  

#define WIN_MONO      	 	0  /* Matches value returned by glXGetConfig */
#define WIN_STEREO     	 	1  

#define WIN_OPAQUE         	0		
#define WIN_TRANSPARENT 	1

#define WIN_DOUBLEBUFFER       	1
#define WIN_SINGLEBUFFER       	2
#define WIN_ACCUMBUFFER         4
#define WIN_ALPHABUFFER       	8
#define WIN_DEPTHBUFFER        	16
#define WIN_STENCILBUFFER      	32

/* -------- Initialisation = recherche du meilleur Visual et de la meilleur Colormap
   on ne cherche pas à utiliser l'alpha
   on utilise un seul tampon écran (Buffer)
   -------- */

static boolean XGL_profondeur_ok(int profondeur, int *liste_profondeurs, int nbr_profondeurs)
{
  int i;

  for (i = 0 ; i < nbr_profondeurs ; i++) {
    if ( profondeur == liste_profondeurs[i] ) { /* Profondeur gérée */
      if ( profondeur <= DefaultDepth(BI_Display, DefaultScreen(BI_Display)) ) { /* Ok avec celle par défaut */
	return VRAI;
      }
    }
  }

  return FAUX; /* Profondeur non supportée par l'écran indiqué */
}

static boolean XGL_initialisation(void)
{
  int i, nvisuals, val, best_score;
  int retour;
  int attributs[] = {GLX_RGBA,  /* On recherche un truc en RGB alpha pour commencer */
		     GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, GLX_DEPTH_SIZE, 1,
		     None, None, None};
  int *profondeurs, nbr_profondeurs;
  XVisualInfo *visual;

  /* On récupère la liste des profondeurs supportées par l'écran */
  profondeurs = XListDepths(BI_Display, DefaultScreen(BI_Display), &nbr_profondeurs);
  if (profondeurs == NULL) {
    fprintf(stderr, "Impossible de connaître la liste des profondeurs d'écran gérées\n");
    return FAUX;
  }

  BI_OpenGL_dBuffer = FAUX;
  visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
  if (visual == NULL) {
    fprintf(stderr, "glXChooseVisual, RGBA/Depth = 1/Single : pas de Visual trouvé\n");
    attributs[7] = GLX_DOUBLEBUFFER; /* On essaye en DoubleBuffer */
    attributs[8] = None;
    visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
    BI_OpenGL_dBuffer = VRAI;
  }
  if (visual == NULL) {
    fprintf(stderr, "glXChooseVisual, RGBA/Depth = 1/Double : pas de Visual trouvé\n");
    attributs[7] = None; /* On essaye sans DepthBuffer, en simple */
    visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
    BI_OpenGL_dBuffer = FAUX;
  }
  if (visual == NULL) {
    fprintf(stderr, "glXChooseVisual, RGBA/Depth = 0 : pas de Visual trouvé\n");
    attributs[1] = None; /* On essaye un truc brutal : Visual RGB sans contrainte */
    visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
    BI_OpenGL_dBuffer = FAUX;
  }
  if (visual == NULL) {
    fprintf(stderr, "glXChooseVisual, RGBA : pas de Visual trouvé\n");
    attributs[1] = GLX_DOUBLEBUFFER; /* On essaye Visual RGB avec double buffer */
    attributs[2] = None;
    visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
    BI_OpenGL_dBuffer = VRAI;
  }
  if (visual == NULL) {
    fprintf(stderr, "glXChooseVisual, RGBA/Double buffer : pas de Visual trouvé\n");
    attributs[0] = None; /* On essaye un truc brutal */
    visual = glXChooseVisual(BI_Display, DefaultScreen(BI_Display), attributs);
    BI_OpenGL_dBuffer = FAUX;
  }
  if (visual == NULL) {
    fprintf(stderr, "Impossible de trouver un Visual RGB convenable pour Open-GL\n");
    XFree(profondeurs);
    return FAUX;
  }
  /*  if ( visual->visualid != XVisualIDFromVisual( DefaultVisual(BI_Display, DefaultScreen(BI_Display) ) ) )
      {
      printf("ID des Visual différents - problèmes possibles\n");
      } */

  if ( XGL_profondeur_ok(visual->depth, profondeurs, nbr_profondeurs) == VRAI ) {
    XGL_Visual = *visual;
  } else { /* Ça se produit, apparemment, parfois sur Silicon Graphics - arrgh.
	      En effet, donne un visual de profondeur > à celle de l'écran (!) d'où
	      BadMatch. On doit donc ruser pour que ça fonctionne... */
    XVisualInfo *vis_list, *this_vis, *best_vis  __attribute__((__unused__)), sample_vis;
    WinAttributes this_vis_attr;
    WinAttributes all_attr;

# ifdef _DEVELOPPE_ /* Affichage de la raison pour laquelle ça n'a pas marché */
    if ( visual->depth > DefaultDepth(BI_Display, DefaultScreen(BI_Display)) )
      fprintf(stderr, "Profondeur de glXChooseVisual > celle de l'écran : risque de problèmes...\n");
    else
      fprintf(stderr, "Profondeur de glXChooseVisual pas dans la liste de celles supportées\n");
# endif
    
    all_attr.color_mode = WIN_DONTCARE;
    all_attr.stereo = WIN_DONTCARE;
    all_attr.ancillary_buffers = 0;
    all_attr.aux_buffers = 0;
    all_attr.level = 0;
    all_attr.transparent_type = WIN_DONTCARE;

    /* Liste des Visual disponibles pour l'écran par défaut */
    sample_vis.screen = DefaultScreen(BI_Display);
    /*  sample_vis.visualid = XVisualIDFromVisual( DefaultVisual(BI_Display, DefaultScreen(BI_Display) ) ); */
    vis_list = XGetVisualInfo(BI_Display, VisualScreenMask /*|VisualIDMask*/, &sample_vis, &nvisuals);
    if (nvisuals <= 0) {
      fprintf(stderr, "Aucun Visual pour cet écran sur ce Display\n");
      XFree(profondeurs);
      return FAUX;
    }

    /* On parcourt la liste pour trouver le meilleur */
    best_score = -1; best_vis = NULL;
    for ( i = 0 ; i < nvisuals ; i++ ) {
      int score = 0;
      this_vis = &vis_list[i];

# ifdef _DEVELOPPE_
      printf("Visual numéro %d\n", i);
# endif

      /* Le Visual doit être reconnu par GLX */
      retour = glXGetConfig( BI_Display, this_vis, GLX_USE_GL, &val);
      if ( retour != 0 ) {
# ifdef _DEVELOPPE_
	printf("  il ne supporte pas Open GL - erreur %d\n", retour);
# endif
	continue;
      }
      if ( val == False ) {
# ifdef _DEVELOPPE_
	printf("  il ne supporte pas Open GL\n");
# endif
	continue;
      }

      /* Visual must be in main planes which is level 0 */ 
      retour = glXGetConfig( BI_Display, this_vis, GLX_LEVEL, &val);
      if ( retour != 0 ) {
# ifdef _DEVELOPPE_
	printf("  erreur glXGetConfig : %d\n", retour);
# endif
	continue;
      }
      if ( val != 0 ) {
# ifdef _DEVELOPPE_
	printf("  inadapté au plan principal\n");
# endif
	continue;
      }

      if ( XGL_profondeur_ok(this_vis->depth, profondeurs, nbr_profondeurs) == FAUX ) {
# ifdef _DEVELOPPE_
	fprintf(stderr, "  visual avec une profondeur de couleur non gérée par l'écran\n");
# endif
	continue;
      }
   
      /* 
       * Score the visual. If any of the requested buffers are not available 
       * the returned score will be negative.
       */
      score = score_visual(&all_attr, this_vis, &this_vis_attr);
      if ( score < 0 ) {
# ifdef _DEVELOPPE_
	printf("  score négatif\n");
# endif
	continue;
      }

      if ( score > best_score ) {
	best_score = score;
	XGL_Visual = *this_vis;
# ifdef _DEVELOPPE_
	printf("  sélectionné, score = %d\n", score);
	printf("   fenêtre racine associée : %ld\n", RootWindow(BI_Display, XGL_Visual.screen));
	printf("   visual associé : %p - id = %ld\n", (void *) XGL_Visual.visual, XGL_Visual.visualid);
	printf("   profondeur : %d\n", XGL_Visual.depth);
	printf("   type : %d\n", XGL_Visual.class);
	/* break; */
      } else {
	printf("  rejeté, score = %d\n", score);
	printf("   fenêtre racine associée : %ld\n", RootWindow(BI_Display, this_vis->screen));
	printf("   visual associé : %p - id = %ld\n", (void *) (this_vis->visual), this_vis->visualid);
	printf("   profondeur : %d\n", this_vis->depth);
	printf("   type : %d\n", this_vis->class);
# endif
      }
    }

    if ( XGL_Visual.visual == NULL ) {
      fprintf(stderr, "Impossible de trouver un Visual convenable\n");
      XFree(profondeurs);
      return FAUX;
    }
    
    XFree(vis_list);
    
  }

# ifdef _DEVELOPPE_
  printf("Visual sélectionné pour Open-GL :\n");
  printf("   fenêtre racine associée : %ld\n", RootWindow(BI_Display, XGL_Visual.screen));
  printf("   visual associé : %p - id = %ld\n", (void *) XGL_Visual.visual, XGL_Visual.visualid);
  printf("   profondeur : %d\n", XGL_Visual.depth);
  printf("   type : %d [3 = Pseudo, 4 = True, 5 = Direct]\n", XGL_Visual.class);

  printf("Visual par défaut de l'écran :\n");
  printf("   fenêtre racine associée : %ld\n", RootWindow(BI_Display, DefaultScreen(BI_Display)));
  printf("   visual associé : %p - id = %ld\n", (void *) BI_Visual,
	 XVisualIDFromVisual( DefaultVisual(BI_Display, DefaultScreen(BI_Display) ) ));
  printf("   profondeur : %d\n", DefaultDepth(BI_Display, DefaultScreen(BI_Display)));
# endif

  /* Si on est ici, c'est qu'on a réussi à trouver un Visual convenable
     on passe donc à la ColorMap
  */

  switch (XGL_Visual.class)
  {
    int etat;

  case PseudoColor:  /* Sur Digital Unix au moins, il semblerait que XCreateColormap sur un Visual
			de type PseudoColor conduise à une BadValue...
			On utilise alors la palette par défaut */
    XGL_Palette = DefaultColormap(BI_Display, DefaultScreen(BI_Display));
    break;

  case TrueColor:
  case DirectColor:
    XGL_Palette = None;
    etat = XmuLookupStandardColormap(BI_Display, XGL_Visual.screen, XGL_Visual.visualid,
				     (unsigned) XGL_Visual.depth, XA_RGB_DEFAULT_MAP, False, True);
    if (etat == 1) { /* Ça a marché */
      XStandardColormap *std_cmaps;
      int nbr_cmaps;

      etat = XGetRGBColormaps(BI_Display, RootWindow( BI_Display, XGL_Visual.screen),
			      &std_cmaps, &nbr_cmaps, XA_RGB_DEFAULT_MAP);
      if (etat == 1) { /* Idem */
	if ( nbr_cmaps > 0) {
	  for (i = 0; i < nbr_cmaps; i++) {
	    if ( std_cmaps[i].visualid == XGL_Visual.visualid ) {
	      XGL_Palette = std_cmaps[i].colormap;
	      XFree(std_cmaps);

# ifdef _DEVELOPPE_
	      printf("On a trouvé une Colormap standard : %lu\n", XGL_Palette);
# endif
	      XFree(profondeurs);
	      return VRAI;
	    }
	  }
# ifdef _DEVELOPPE_
	  printf("Aucune Colormap standard avec le bon visualid\n");
# endif
	  XFree( std_cmaps);
	}
# ifdef _DEVELOPPE_
	else
	  printf("Aucune Colormap standard\n");
# endif
      }
# ifdef _DEVELOPPE_
      else {
	printf("Erreur dans XGetRGBColormaps\n");
      }
# endif
    }
# ifdef _DEVELOPPE_
    else {
      printf("Impossible de regarder les Colormaps standards\n");
    }
# endif

    XGL_Palette = XCreateColormap(BI_Display, BI_Fen_racine, XGL_Visual.visual, AllocNone);

    break;

  case StaticColor: /* Supportés pas MESA uniquement (?) */
  case StaticGray:
  case GrayScale:
    fprintf(stderr, "Type de Visual pas encore supporté : %d\n", XGL_Visual.class);
    break;
  }

  XFree(profondeurs);
  return VRAI; /* On a pu initialiser correctement pour Open GL */
}

/*  
 * Preference is given to the visual that supports the most colors
 * and also to pseudocolor over gray scale and true color over pseudocolor. 
 * Since there is a large number of colors available on 24 bit true/direct 
 * color devices, this algorithm heavily favors these visuals. (That is, 
 * the number of available colors is weighted much heavier than other factors 
 * such as depth buffer size.)
 */

static int score_visual_class(XVisualInfo *visual)
{
  unsigned int mask;
  int score = 0;

  switch ( visual->class )
  {
  case StaticGray:
  case GrayScale:
    /* favor pseudo color over gray scale */
    score = visual->colormap_size - 2;
    break;
  case StaticColor:
  case PseudoColor:
    /* favor true/direct over pseudo color */
    score = visual->colormap_size - 1;
    break;
  case TrueColor:
  case DirectColor:
    mask = visual->red_mask | 
      visual->green_mask | 
      visual->blue_mask;
    if ( mask )
      for ( score = 1; mask > 0; mask >>= 1 ) if ( mask & 1 ) score <<= 1;
    break;
  }

  return score;
}

/* 
 * Checks the visual attributes. If the specified attributes are not
 * supported then -1 is returned. If all the specified attributes are
 * supported then the visual is scored according to the depth of the
 * specified ancillary buffers, the depth of the color buffers, and
 * the type of visual. The score for the visual is returned.
 * (Note that alpha is treated differently than other visual attributes: 
 * if alpha is not specified then a visual with an alpha buffer will not 
 * be selected. See the comment in get_visual_and_colormap() for more details.)
 */

static int score_visual(WinAttributes *attr, XVisualInfo *vis, WinAttributes *vis_attr)
{
  int val; 
  int score = 0;

  vis_attr->ancillary_buffers = 0;
   
  /* Color Index or RGBA? It must match the requested value */
  glXGetConfig( BI_Display, vis, GLX_RGBA, &vis_attr->color_mode);
  if ( attr->color_mode != WIN_DONTCARE &&
       attr->color_mode != vis_attr->color_mode ) return( -1);
   
  /* Stereo or mono visual? It must match the requested value */
  glXGetConfig( BI_Display, vis, GLX_STEREO, &vis_attr->stereo);
  if ( attr->stereo != WIN_DONTCARE &&
       attr->stereo != vis_attr->stereo ) return( -1);

  /* 
   * If double buffering is requested then visual must have a back buffer 
   * If single buffering is requested then it cannot have a back buffer   
   */
  glXGetConfig( BI_Display, vis, GLX_DOUBLEBUFFER, &val);
  vis_attr->ancillary_buffers |= (val) ? (WIN_DOUBLEBUFFER) : (WIN_SINGLEBUFFER);
  if ( (attr->ancillary_buffers & WIN_DOUBLEBUFFER) &&
       !(vis_attr->ancillary_buffers & WIN_DOUBLEBUFFER) ) return( -1);
  if ( (attr->ancillary_buffers & WIN_SINGLEBUFFER)  &&
       !(vis_attr->ancillary_buffers & WIN_SINGLEBUFFER) ) return( -1);

  /* 
   * If alpha requested then alpha size must be > 0. If alpha is not requested
   * then the size must be zero. 
   * NOTE: You may wish to check for a minimum alpha size. We assume
   * that size > 0 is sufficient if alpha is requested.
   */ 
  if ( vis_attr->color_mode == WIN_RGBA ) {
    glXGetConfig( BI_Display, vis, GLX_ALPHA_SIZE, &val);
    if ( val > 0 ) vis_attr->ancillary_buffers |= WIN_ALPHABUFFER;
    if ( (attr->ancillary_buffers & WIN_ALPHABUFFER) != 
	 (vis_attr->ancillary_buffers & WIN_ALPHABUFFER) ) return( -1);
  }

  /* 
   * The number of requested aux buffers must be available 
   * NOTE: You may wish to adjust the score based on the number 
   * of available aux buffers.
   */
  glXGetConfig( BI_Display, vis, GLX_AUX_BUFFERS, &vis_attr->aux_buffers);
  if ( attr->aux_buffers &&
       vis_attr->aux_buffers < attr->aux_buffers ) return( -1);

  /* 
   * We just check that the specified ancillary buffers are supported;
   * that is the size just has to be > 0. If an ancillary buffer is 
   * supported then it's depth adds to the score. 
   * NOTE: You may want to make it possible to specify a minimum 
   * size for the ancillary buffers.
   */

  /* 
   * Accumulation buffers, if requested, are scored according the the 
   * size of the red, green, blue and alpha components.               
   */
  if ( vis_attr->color_mode == WIN_RGBA ) {
    glXGetConfig( BI_Display, vis, GLX_ACCUM_RED_SIZE, &val);
    if ( val > 0 )
      vis_attr->ancillary_buffers |= WIN_ACCUMBUFFER;
    if ( attr->ancillary_buffers & WIN_ACCUMBUFFER ) {
      if ( val <= 0 ) return( -1);
      score += val;
      glXGetConfig( BI_Display, vis, GLX_ACCUM_GREEN_SIZE, &val); score += val;
      glXGetConfig( BI_Display, vis, GLX_ACCUM_BLUE_SIZE, &val); score += val;
      glXGetConfig( BI_Display, vis, GLX_ACCUM_ALPHA_SIZE, &val); score += val;
    }
  }

  /* The depth buffer, if requested, is scored according to it's size */
  /* The visual is rejected if the size is zero                       */
  glXGetConfig( BI_Display, vis, GLX_DEPTH_SIZE, &val);
  if ( val > 0 ) 
    vis_attr->ancillary_buffers |= WIN_DEPTHBUFFER;
  if ( attr->ancillary_buffers & WIN_DEPTHBUFFER ) {
    if ( val <= 0 ) return( -1);
    score += val;
  }

  /* The stencil buffer is also scored according to it's size */
  /* The visual is rejected if the size is zero               */
  glXGetConfig( BI_Display, vis, GLX_STENCIL_SIZE, &val);
  if ( val > 0 ) 
    vis_attr->ancillary_buffers |= WIN_STENCILBUFFER;
  if ( attr->ancillary_buffers & WIN_STENCILBUFFER ) {
    if ( val <= 0 ) return( -1);
    score += val;
  }

  /* Score the visual class based on number of colors supported */
  score += score_visual_class( vis);

  return score;
}

#endif /* OPEN_GL */
