/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2003 Emmanuel CURIS
   Création : le 8 juillet 2003 (Budapest)

   Récupération des données expérimentales pour obtenir µ total

   Fichiers enregistrés à Bessy2 (Berlin) : format HDF
*/

#include <stdio.h>
#include <stdlib.h>

#include <mfhdf.h> /* Bibliothèque HDF v.4, NCSA */
# ifndef H4_MAX_VAR_DIMS
#  define H4_MAX_VAR_DIMS MAX_VAR_DIMS /* Ancienne version de la bibliothèque HDF */
# endif

#include "systeme.h"
#include "serie.h"

#include "import_xas.h"

#include "ext_bessy2.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static char *HDF_Attribut_LireChaine(int32 id, int32 attribut, size_t lg_chaine);
static int   HDF_LireTableDoubles   (int32 id_table, int32 num_colonne, int32 nbr_valeurs, double *table);
static int32 HDF_TrouverTable_Name  (int32 id_fichier, int32 nbr_tables, const char *nom_table);

static double HDF_LireDouble(int32 id_table, int32 num_colonne, int32 position);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

int BESSY2_LireSpectre(const char *nom, Infos_Extraction *infos, EXT_ListeGraphes *lst_graphes, struct serie *serie, unsigned int num_spectre)
{
  boolean err_traitee = FAUX;
  int32 dimension, tailles[ H4_MAX_VAR_DIMS ], type, nbr_attributs;
  int32 id_fichier, id_i0, id_ifluo, id_energies;
  int32 nbr_tables, nbr_attributs_globaux;
  int32 idx_energie, idx_i0, idx_ifluo;
  intn retour, table_vide;
  int type_reel = GR_AUTRETYPE;
  unsigned long i, nombre_points;

  id_fichier = SDstart( nom, DFACC_READ );
  if ( FAIL == id_fichier ) {
    /* Format incorrect, car fichier existe et est accessible [sinon fopen() aurait échoué] */
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    return -1;
  }

  /* Infos sur le fichier */
  retour = SDfileinfo( id_fichier, &nbr_tables, &nbr_attributs_globaux );
  if ( FAIL == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDfileinfo() [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDend( id_fichier );
    return -2;
  }
  if ( nbr_tables < 3 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas assez de tables dans le fichier [3 minimum, 8 en général ?][%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    (void) SDend( id_fichier );
    return -4;
  }

  /* Les attributs globaux ne servent à rien, a priori ? */

  /* Table pour les énergies : numéro 2 ? */
  idx_energie = HDF_TrouverTable_Name( id_fichier, nbr_tables, "Energy" );
  if ( FAIL == idx_energie ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de trouver une table d'énergies [fichier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
#endif
    (void) SDend( id_fichier );
    (void) BI_ERR_alerte( NULL, &ERR_SansEnergies, VRAI, nom );
    return -105; /* On n'affiche pas d'autre message ! */
  }

  /* On choisit cette table pour connaître le nombre de points et remplir la table des X */
  id_energies = SDselect( id_fichier, idx_energie );
  if ( FAIL == id_energies ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDselect(), table des énergies [index %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, (unsigned int) idx_energie);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDend( id_fichier );
    return -6;
  }

  /* Taille des données ?
     Le nom ne nous intéresse pas : Motor, non informatif !
  */
  retour = SDgetinfo( id_energies, NULL, &dimension, tailles, &type, &nbr_attributs);
  if ( FAIL == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDgetinfo()\n",
		   __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDendaccess( id_energies );
    (void) SDend( id_fichier );
    return -12;
  }
  if ( dimension > 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Plus d'une colonne dans la table des énergies. La première seule sera utilisée.\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }
  if ( type != DFNT_FLOAT64 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Les données ne sont pas des doubles, mais de type %d. Résultat imprédictible\n",
		   __FILE__, __FUNCTION__, __LINE__, (int) type);
  }
  nombre_points = tailles[ 0 ];

  /* On vérifie que la table n'est pas vide */
  retour = SDcheckempty( id_energies, &table_vide );
  if ( table_vide ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Spectre vide / fichier %s / %lu points\n",
		   __FILE__, __FUNCTION__, __LINE__, nom, nombre_points);
#endif
    (void) SDendaccess( id_energies );
    (void) SDend( id_fichier );
    (void) BI_ERR_alerte( NULL, &ERR_TableVide, VRAI, nom );
    return -106;
  }

  /* On peut donc préparer les graphes [Pour l'instant, un graphe par fichier...] */
  retour = EXTRAIT_PreparerGraphes( infos, 1, nombre_points, nom, lst_graphes, serie, 1, num_spectre + 1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans EXTRAIT_PreparerGraphes() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, (int) retour);
    (void) SDendaccess( id_energies );
    (void) SDend( id_fichier );
    return -2;
  }

  /* On peut lire les X */
  retour = HDF_LireTableDoubles( id_energies, 0, nombre_points, lst_graphes->x );
  (void) SDendaccess( id_energies ); /* Plus besoin des X, quoi qu'il arrive */
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans HDF_LireTableDoubles() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, (int) retour);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDend( id_fichier );
    return -2;
  }

  /* On cherche pour I0 : courrant dans l'anneau ? */
  idx_i0 = HDF_TrouverTable_Name( id_fichier, nbr_tables, "Ring_1" );
  if ( FAIL == idx_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de trouver une table des \"I0\"\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) SDend( id_fichier );
    return -5;
#ifdef _DEVELOPPE_
  } else {
/*     printf( "%s : colonne du courrant anneau = %u", __FUNCTION__, (unsigned int) idx_i0 ); */
#endif
  }

  /* On cherche pour If : courant / MCA ? */
  idx_ifluo = HDF_TrouverTable_Name( id_fichier, nbr_tables, "Roentec-MCA" );
  if ( FAIL == idx_ifluo ) {
    /* Changement de nom : mca_roentec maintenant -> on essaye */
    idx_ifluo = HDF_TrouverTable_Name( id_fichier, nbr_tables, "mca_roentec" );
  }
  if ( FAIL == idx_ifluo ) {
    /* Pas de fluo : scan de I0 seul ? soit PGM_K617_1, soit l'autre diode... */
    idx_ifluo = HDF_TrouverTable_Name( id_fichier, nbr_tables, "PGM_K617_1" );
    type_reel = GR_I0; /* On charge I0, normalisé par le courrant dans l'anneau */
  } else {
    type_reel = GR_I1; /* On charge If, normalisé par le courrant dans l'anneau */
  }
  if ( FAIL == idx_ifluo ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de trouver une table des \"Ifluo\"\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) SDend( id_fichier );
    return -5;
  }

  /* On ouvre les tables correspondantes */
  id_i0 = SDselect( id_fichier, idx_i0 );
  if ( FAIL == id_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDselect(), table des I0 [index %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, (unsigned int) idx_i0);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDend( id_fichier );
    return -8;
  }
  id_ifluo = SDselect( id_fichier, idx_ifluo );
  if ( FAIL == id_ifluo ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDselect(), table des If [index %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, (unsigned int) idx_ifluo);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    (void) SDendaccess( id_i0 );
    (void) SDend( id_fichier );
    return -8;
  }

  /* Lecture point par point */
  for ( i = 0 ; i < nombre_points ; i++ ) {
    double i0, i_fluo;

    /*  Lecture des valeurs */
    i0 = HDF_LireDouble( id_i0, 0, i );
    i_fluo = HDF_LireDouble( id_ifluo, 0, i );

    /* On stocke I0, si nécessaire */
    if ( lst_graphes->i_0 != NULL ) {
      lst_graphes->i_0->y[ i ] = i0;
    }

    /* On stocke I1, si nécessaire */
    if ( lst_graphes->i_1 != NULL ) {
      lst_graphes->i_1[ 0 ]->y[ i ] = i_fluo;
    }

    /* On fait le calcul si nécessaire */
    if ( lst_graphes->spectres != NULL ) {
      if ( ( i0 <= 0.0 ) || ( i_fluo < 0.0 ) ) {
	lst_graphes->spectres[ 0 ]->y[ i ] = 0.0; /* Arbitrairement */
	if ( FAUX == err_traitee ) {
	  int reponse;

	  reponse = BI_ERR_alerte( NULL, &ERR_CoupureFaisceau, VRAI, nom );
	  if ( 3 == reponse ) { /* On annule */
	    return -150; /* Pas besoin d'un nouveau message ! */
	  }
	  if ( 2 == reponse ) { /* On arrête ici */
	    return i;
	  }
	  err_traitee = VRAI;
	}
      } else {
	lst_graphes->spectres[ 0 ]->y[ i ] = i_fluo / i0;
      }
    }
  }

  /* Plus besoin des tables de fluo. et I0 */
  (void) SDendaccess( id_i0 );
  (void) SDendaccess( id_ifluo );

  /* Fermeture du fichier */
  retour = SDend( id_fichier );
  if ( FAIL == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDend()\n",
		   __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    /* Pas très grave */
    return 0;
  }

  /* Modifications post-traitement pour coller aux résultats */
  GRF_ChangerType( lst_graphes->i_0, GR_AUTRETYPE ); /* Courant dans l'anneau */
  if ( lst_graphes->spectres != NULL ) {
    GRF_ChangerType( lst_graphes->spectres[ 0 ], type_reel );
  }

  /* Le chargement a réussi */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Routines concernant le format HDF en général ~-~-~-~-~-~-~-~- */

/* -------- Lire un attribut contenant une chaîne de caractères -------- */

static char *HDF_Attribut_LireChaine(int32 id, int32 attribut, size_t lg_chaine)
{
  char *texte;
  intn retour;

  if ( 0 == lg_chaine ) return NULL;

  /* L'attribut ne contient pas de 0 final, mais on en veut un */
  texte = ( char * ) malloc( sizeof( char ) * (lg_chaine + 1) );
  if ( NULL == texte ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Lecture effective */
  retour = SDreadattr( id, attribut, (void *) texte );
  if ( FAIL == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDreadattr()\n",
		   __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    free( texte );
    return NULL;
  }

  /* On ajoute un 0 final */
  texte[ lg_chaine ] = 0;

  /* Le texte est prêt */
  return texte;
}

/* -------- Lire une seule valeur dans une table -------- */

static double HDF_LireDouble(int32 id_table, int32 num_colonne, int32 position)
{
  double valeur;
  int32 positions[ H4_MAX_VAR_DIMS ], nombre[ H4_MAX_VAR_DIMS ];
  int32 i;
  intn retour;

  /* Init des tableaux */
  for ( i = 0 ; i < H4_MAX_VAR_DIMS ; i++ ) {
    positions[ i ] = 0;
    nombre[ i ] = 0;
  }

  /* Pour la colonne précisée, on donne la position initiale et le nombre de valeurs */
  positions[ num_colonne ] = position;
  nombre[ num_colonne ] = 1;

  /* Et on lit */
  retour = SDreaddata( id_table, positions, NULL, nombre, &valeur );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDreaddata()\n",
		   __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    return 0.0;
  }

  /* Tout s'est bien passé */
  return valeur;
}

/* -------- Lire d'un coup une table de doubles à partir d'une colonne de tables -------- */

static int HDF_LireTableDoubles(int32 id_table, int32 num_colonne, int32 nbr_valeurs, double *table)
{
  int32 positions[ H4_MAX_VAR_DIMS ], nombre[ H4_MAX_VAR_DIMS ];
  int32 i;
  intn retour;

  /* Init des tableaux */
  for ( i = 0 ; i < H4_MAX_VAR_DIMS ; i++ ) {
    positions[ i ] = 0;
    nombre[ i ] = 0;
  }

  /* Pour la colonne précisée, on donne le nombre de valeurs */
  nombre[ num_colonne ] = nbr_valeurs;

  /* Et on lit */
  retour = SDreaddata( id_table, positions, NULL, nombre, table );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDreaddata()\n",
		   __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
    HEprint( stderr, 0 );
#endif
    return -1;
  }

  /* Tout s'est bien passé */
  return 0;
}


/* -------- Trouver une table avec un attribut "Name" contenant le nom indiqué -------- */

static int32 HDF_TrouverTable_Name(int32 id_fichier, int32 nbr_tables, const char *nom_table)
{
  int32 i, i_attribut;
  int32 id_table;
  int32 dimension, tailles[ H4_MAX_VAR_DIMS ], type, nbr_attributs;
  intn retour;
  size_t lg_nom;

  lg_nom = strlen( nom_table ); /* Nombre de caractères, 0 exclu */

  /* On essaye chacune des tables */
  for ( i = 0 ; i < nbr_tables ; i++ ) {
    /* On choisit la table */
    id_table = SDselect( id_fichier, i );
    if ( FAIL == id_table ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDselect\n",
		     __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
      HEprint( stderr, 0 );
#endif
      return FAIL;
    }

    /* Combient d'attributs associés ? */
    retour = SDgetinfo( id_table, NULL/* Le nom ne nous intéresse pas ici */, &dimension, tailles, &type, &nbr_attributs);
    if ( FAIL == retour ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDgetinfo()\n",
		     __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
      HEprint( stderr, 0 );
#endif
      (void) SDendaccess( id_table );
      return FAIL;
    }

    /* Pour chacun des attributs */
    for ( i_attribut = 0 ; i_attribut < nbr_attributs ; i_attribut++ ) {
      char nom_attribut[ 70 ]; /* Probablement 64 max. comme les noms de tables, mais la doc ne l'indique pas */
      char *texte_attribut;
      int32 type_donnees, nombre_donnees;

      retour = SDattrinfo( id_table, i_attribut, nom_attribut, &type_donnees, &nombre_donnees );
      if ( FAIL == retour ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDfileinfo()\n",
		       __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
	HEprint( stderr, 0 );
#endif
	(void) SDendaccess( id_table );
	return FAIL;
      }
      if ( (size_t) nombre_donnees != lg_nom ) /* On est sûr que l'on n'est pas bon ! */
	continue;

      if ( strcmp( nom_attribut, "Name" ) != 0 ) /* Pas le bon attribut */
	continue;

      if ( ( type_donnees != DFNT_UCHAR8 ) && ( type_donnees != DFNT_CHAR8 ) ) {
	(void) fprintf(stderr, "%s,%s l.%u : Attribut name, ne contient pas un texte !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	continue; /* Ce n'est pas un texte ! */
      }

      /* Ici, contient-on le bon nom ? Il faut lire un texte ! */
      texte_attribut = HDF_Attribut_LireChaine( id_table, i_attribut, nombre_donnees );
      if ( NULL == texte_attribut ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : Attention, problème dans HDF_Attribut_LireChaine()\n",
		       __FILE__, __FUNCTION__, __LINE__);
#endif
	continue;
      }

      if ( 0 == strcmp( texte_attribut, nom_table ) ) {
	/* Attribut trouvé : inutile de continuer */
	free( texte_attribut ); /* On libère donc... */
	retour = SDendaccess( id_table );
	if ( FAIL == retour ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDendaccess()\n",
			 __FILE__, __FUNCTION__, __LINE__);
	}

	/* Et on indique que l'on a trouvé ce que l'on cherchait */
	return i;
      }

      free( texte_attribut );
    }

    /* Plus besoin de ces données */
    retour = SDendaccess( id_table );
    if ( FAIL == retour ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SDendaccess()\n",
		     __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
      HEprint( stderr, 0 );
#endif
      return FAIL;
    }
  }

  /* Si on est là, c'est que l'on n'a rien trouvé */
  return FAIL;
}
