/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1996 - 2010, Emmanuel CURIS
   Cration : 14 juin 2010 (repris de N_couches/feff_modele.c)

   Cration d'un modle  partir des fichiers de sortie de FEFF

   1) Utilise le fichiers PATHS.DAT pour connatre les paramtres des diffrents chemins
   2) Cre les chemins correspondants
   3) Lit les diffrents rsultats  partir des fichiers FEFFnnnn.DAT
   4) Recalcule les paramtres importants (phase et amplitude)  partir de ces donnes
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"

#include "atomes.h"
#include "chemins.h"
#include "couches.h"
#include "feff.h"
#include "fichiers.h"
#include "modeles.h"
#include "theorie.h" /* Pour SEUIL_K */

#include "feff_paths_dat.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FEFF_PathsDat_ControlerAbsorbeur(Absorbeur *absorbeur, FEFF_PATHS_DAT *infos, Vecteur3D *pos_absorbeur, char symbole[ 3 ]);
static int FEFF_PathsDat_LireAmplitudePhase(FILE *fichier, Graphe *phase, Graphe *amplitude, double longueur, FEFF_PATHS_DAT *infos);
static int FEFF_PathsDat_LireCouche        (FILE *fichier, Serie_Graphes *serie_amp_pha, FEFF_PATHS_DAT *infos, struct modele_ajuste *modele_feff);
static int FEFF_PathsDat_LireFEFFxxxDAT    (FEFF_PATHS_DAT *infos, Serie_Graphes *serie_amp_pha, Couche *couche, unsigned int numero_chemin, double longueur );

static Chemin *FEFF_PathsDat_LireChemin(FILE *fichier, FEFF_PATHS_DAT *infos, unsigned int numero_chemin, unsigned int n_reflexions );
static struct modele_ajuste *FEFF_PathsDat_TrouverModele(Absorbeur *absorbeur);

/* ~~~~~~~~ Ralisation pratique des fonctions ~~~~~~~~ */

struct modele_ajuste *FEFF_LirePathsDat(FEFF_PATHS_DAT *infos, struct absorbeur *absorbeur, const char *nom_fichier)
{
  FILE *fichier;
  Serie_Graphes *grf_amp_pha;
  Vecteur3D pos_absorbeur;
  char *ligne = NULL,  symbole[3] = {0, 0, 0};
  int retour;
  long debut_chemins;
  unsigned long nbr_non_lus = 0;
  struct modele_ajuste *modele_feff = NULL;

  /* Contrles */
  if ( ( NULL == infos ) || ( NULL == nom_fichier ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramtres incorrects {NULL}\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On essaye d'ouvrir le fichier */
  fichier = fopen( nom_fichier, "rb" );
  if ( NULL == fichier ) {
    infos->code_erreur = PATHSDAT_ERREUR_FICHIER_INACCESSIBLE;
    return NULL;
  }

  /* On saute l'en-tte */
  do {
    free( ligne );
    ligne = lit_ligne( fichier, 0 );
    if ( NULL == ligne ) {
      infos->code_erreur = PATHSDAT_ERREUR_FICHIER_ILLISIBLE;
      return NULL;
    }
  } while ( strncmp( ligne, " ----------", 10 ) != 0 );
  /* On est sur une ligne de tirets : a priori, sparateur... */
  free( ligne ); ligne = NULL;

  debut_chemins = ftell( fichier ); /* On mmorise cette rgion */
  /* On saute les 3 lignes qui suivent : (==> on est sur la ligne de l'absorbeur)
      - description du 1er chemin
      - titres des colonnes qui suivent
      - premier rflecteur */
  sauter_lignes( fichier, 3 );

  /* On lit les informations de l'absorbeur
     Suppose implicitement que le nom de l'atome est son symbole, cas des fichiers crs avec LASE */
  (void) fscanf( fichier, "%lf %lf %lf %*d '%c%c", &(pos_absorbeur.x), &(pos_absorbeur.y), &(pos_absorbeur.z), symbole, symbole + 1 );

  /* On vrifie que l'absorbeur convient, sinon on fait le ncessaire */
  retour = FEFF_PathsDat_ControlerAbsorbeur( absorbeur, infos, &pos_absorbeur, symbole );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans le contrle de l'absorbeur [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    infos->code_erreur = retour;
    return NULL;
  }

  /* On cherche le modle o faire le chargement */
  modele_feff = FEFF_PathsDat_TrouverModele( infos->absorbeur );
  if ( NULL == modele_feff ) {
    (void) fprintf(stderr, "%s,%s l.%u : Problme dans la cration du modle\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos->code_erreur = PATHSDAT_ERREUR_PAS_DE_MODELE;
    return NULL;
  }

  /* On revient au dbut pour lire les chemins dans l'ordre */
  fseek( fichier, debut_chemins, SEEK_SET );

  /* On prpare la srie des phases & amplitudes */
  grf_amp_pha = SERIE_GRF_CreerSerie(NULL, nom_serie_amplitudes_phases);
  if ( NULL == grf_amp_pha ) {
    (void) fprintf(stderr, "%s,%s l.%u : impossible de crer la srie des phases & amplitudes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  /*  partir d'ici, on peut attaquer la lecture & la construction proprement dites... */
  do {
    retour = FEFF_PathsDat_LireCouche( fichier, grf_amp_pha, infos, modele_feff );
    if ( retour > 0 ) { /* Erreur mineure = graphes d'amplitude et phase non lus */
      nbr_non_lus++;
    }
  } while ( retour != EOF );

  (void) fclose( fichier );

  /* On renvoie le modle d'ajustement cr */
  return modele_feff;
}

/* -------- Contrle de la cohrence avec l'absorbeur -------- */

static int FEFF_PathsDat_ControlerAbsorbeur(Absorbeur *absorbeur, FEFF_PATHS_DAT *infos, Vecteur3D *pos_absorbeur, char symbole[ 3 ])
{
  int Z_abs;
  Atome *atome;

  /* On rcupre le numro atomique de l'absorbeur du fichier */
  Z_abs = ATM_num_atomique( symbole );
  if ( SYMBOLE_CHIMIQUE_INCONNU == Z_abs ) {
    (void) fprintf(stderr, "%s,%s l.%u : symbole inconnu [%s : x=%f, y=%f, z=%f], chargement annul\n",
		   __FILE__, __FUNCTION__, __LINE__, symbole, pos_absorbeur->x, pos_absorbeur->y, pos_absorbeur->z );
    return PATHSDAT_ERREUR_Z_INCONNU;
  }

  /* Contrle premier : a-t-on un candidat absorbeur ? */
  if ( ( absorbeur != NULL ) && ( infos->options & PATHSDAT_FORCER_ABSORBEUR_INDIQUE ) ) {
    if ( Z_abs != absorbeur->absorbeur->num_atomique ) {
      (void) fprintf(stderr, "%s,%s l.%u : Numros atomiques diffrents !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      infos->Z_absorbeur_feff = Z_abs;
      return PATHSDAT_ERREUR_Z_INCOMPATIBLE;
    }

    /* Si mme Z, on postule que convient, mais si (x, y, z diffrent : translation) */
    infos->translation.x = absorbeur->absorbeur->position.x - pos_absorbeur->x;
    infos->translation.y = absorbeur->absorbeur->position.y - pos_absorbeur->y;
    infos->translation.z = absorbeur->absorbeur->position.z - pos_absorbeur->z;

    /* Par dfinition, OK */
    infos->absorbeur = absorbeur;
    return 0;
  }

  /* Pas de candidat indiqu : on cherche... */
  /* On cherche l'absorbeur, si possible dans la molcule courante */
  atome = ATM_chercher( pos_absorbeur, Z_abs, mol_courante );
  if ( atome != NULL ) {
    /* Atome trouv : OK */
    infos->molecule = quelle_molecule( atome, &absorbeur ); /* Molcule  utiliser */
  } else { /* On doit crer la molcule qui contiendra le modle */
    infos->molecule = MOL_creer( "Mod"e_GRAVE"le FEFF" );
    if ( NULL == infos->molecule ) {
      (void) fprintf( stderr, "%s,%s l.%u : impossible de crer la molcule du modle !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return PATHSDAT_ERREUR_PAS_DE_MOLECULE;
    }

    atome = ATM_creer( infos->molecule, NULL, Z_abs, pos_absorbeur, 0.0, -1.0 );
    if ( NULL == atome ) {
      (void) fprintf( stderr, "%s,%s l.%u : impossible de crer l'atome absorbeur du modle !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      (void) MOL_effacer( infos->molecule, VRAI, VRAI);
      infos->molecule = NULL;
      return PATHSDAT_ERREUR_PAS_DE_MOLECULE;
    }
  }

  /* On vrifie que l'atome trouv est bien un absorbeur */
  if ( FAUX == atome->marques.absorbeur ) {
    infos->absorbeur = creer_absorbeur( atome, "Absorbeur FEFF", 1.0 );
    if ( NULL == infos->absorbeur ) {
      return PATHSDAT_ERREUR_PAS_DE_MOLECULE;
    }
  } else {
    /* On retrouve cet absorbeur : a t fait dans ATM_chercher */
    infos->absorbeur = absorbeur;
  }

  /* L'absorbeur a t trouv, tout va bien... */
  return 0;
}

/* -------- Trouver le modle dans lequel crer les couches,... -------- */

static struct modele_ajuste *FEFF_PathsDat_TrouverModele(Absorbeur *absorbeur)
{
  struct modele_ajuste *modele;

  /* 1er essai : le modle actif */
  modele = absorbeur->modele_actif;
  if ( modele != NULL ) {
    if ( MODELE_EXAFS_CHEMINS == modele->type ) {
      if ( NULL == modele->infos_modele.feff->couches ) {
	/* Aucune couche : on postule que l'on peut le mettre dedans */
	return modele;
      }
    }

    /* Ce n'est pas le modle actif : on vrifie les autres */
    for ( modele = absorbeur->modeles ; 
	  modele != NULL ;
	  modele = modele->suivant ) {
      if ( MODELE_EXAFS_CHEMINS == modele->type ) {
	if ( NULL == modele->infos_modele.feff->couches ) {
	  /* Aucune couche : on postule que l'on peut le mettre dedans */
	  return modele;
	}
      }
    }
  }

  /* Toujours pas : on cre un nouveau modle */
  modele = ABS_CreerModele( absorbeur, MODELE_EXAFS_CHEMINS, NULL );
  MODELE_Nommer( modele, txt_nom_modele_par_defaut );
  return modele;
}

/* -------- Lire tout un chemin de diffusion -------- */

static Chemin *FEFF_PathsDat_LireChemin(FILE *fichier, FEFF_PATHS_DAT *infos, unsigned int numero_chemin, unsigned int N_atomes )
{
  Vecteur3D position;
  unsigned int potentiel, i;
  double r, beta, eta;
  Chemin *chemin = NULL;
  int n_lu;
  char *nom_chemin;

  /* Pour le nom : Zn -> O -> ... -> Zn + 0 final + 4 chiffres/n du chemin + ", ",
     avec 3 cars pour le symbole chimique */
  nom_chemin = (char *) calloc( 7 * (N_atomes + 1) + 10, sizeof(char));
  if ( nom_chemin != NULL ) { /* On a pu crer le nom du chemin */
    sprintf( nom_chemin, "%4d, %s -> ", numero_chemin, ATM_Z_vers_symbole( infos->absorbeur->absorbeur->num_atomique ) );
  }

  chemin = creer_chemin( N_atomes );
  if ( NULL == chemin ) {
    (void) fprintf(stderr, "%s,%s l.%u : impossible de crer le chemin %u\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_chemin );
    return NULL;
  }

  for ( i = 0 ; i < N_atomes ; i++ ) {
    Atome *atome;
    char symbole[ 7 ] = {0, 0, 0, 0, 0, 0, 0};
    int Z;

    n_lu = fscanf( fichier, "%lf %lf %lf %u '%6c' %lf %lf %lf",
		   &position.x, &position.y, &position.z, &potentiel, symbole, &r, &beta, &eta);
#ifdef _DEVELOPPE_
    /*    printf("Atome lu : Z = %s en (%f, %f, %f)\n", symbole, x, y, z); */
#endif
    if ( n_lu == EOF ) /* Fin de fichier atteinte */
      return NULL;
    if ( n_lu != 8 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ligne du chemin %u incomplte, n_lu=%d [atome %u : %s (%f, %f, %f), pot. %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, numero_chemin, n_lu, i, symbole, position.x, position.y, position.z, potentiel );
      (void) SYS_ERREUR(ERREUR_LECTURE, "paths.dat", fichier, "Ligne d'atome du chemin incomplte");
      return NULL;
    }

    /* On corrige la position pour tenir compte de la translation ventuelle */
    position.x += infos->translation.x;
    position.y += infos->translation.y;
    position.z += infos->translation.z;

    Z = ATM_num_atomique(symbole);
    atome = ATM_chercher( &position, Z, infos->molecule );
    if ( NULL == atome ) { /* Pas trouv : il faut le crer */
      if ( Z <= 0 ) Z = OXYGENE;
      atome = ATM_creer( infos->molecule, NULL, Z, &position, 0.0, 0.0 );
      if ( NULL == atome ) {
	(void) fprintf(stderr, "%s,%s l.%u : impossible de crer l'atome du chemin %u\n",
		       __FILE__, __FUNCTION__, __LINE__, numero_chemin );
	return NULL;
      }
    }

    chemin->atomes[ i ] = atome; /* i-me atome du chemin */
    if ( nom_chemin != NULL ) { /* On a pu crer le nom du chemin */
      strcat( nom_chemin, ATM_donnees[Z - 1].symbole );
      if ( i < N_atomes - 1 ) /* On n'est pas sur le dernier... */
	strcat( nom_chemin, " -> " );
    }
  }
  nommer_chemin( chemin, nom_chemin );
  free( nom_chemin ); nom_chemin = NULL;

  return chemin;
}

static int FEFF_PathsDat_LireCouche(FILE *fichier, Serie_Graphes *serie_amp_pha, FEFF_PATHS_DAT *infos, struct modele_ajuste *modele_feff)
{
  int n_lu, N;
  unsigned int num_chemin;
  unsigned int n_reflexion;
  double longueur;
  Couche *couche;

  if ( NULL == fichier ) return EOF;

  /* Premire ligne : index de chemin, dgnrescence, nombre d'atomes impliqus, longueur */
  n_lu = fscanf( fichier, "%u %u %d %*[^=] = %lf", &num_chemin, &n_reflexion, &N, &longueur );
  if ( n_lu == EOF ) /* Fin de fichier atteinte */
    return EOF;
  if ( n_lu != 4 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ligne du chemin incomplte, n_lu=%d [num=%u n_ref=%u N=%d lg=%f]\n",
		   __FILE__, __FUNCTION__, __LINE__, n_lu, num_chemin, n_reflexion, N, longueur );
    (void) SYS_ERREUR( ERREUR_LECTURE, "paths.dat", fichier, "Ligne de chemin incompl"e_GRAVE"te" );
    return EOF;
  }

  /* Premiers contrles, simples, sur le chargement de cette couche */
  if ( ( infos->options & PATHSDAT_LIMITER_LONGUEUR ) && ( longueur > infos->longueur_maximale ) ) { /* Chemins classs par longueur croissante :
						    s'il ne convient pas, les suivants non plus */
    return EOF;
  }
  if ( ( infos->options & PATHSDAT_LIMITER_NOMBRE_CHEMINS ) && ( num_chemin > infos->i_maximal ) ) { /* Mme chose pour les index de chemins */
    return EOF;
  }

  (void) ignore_ligne( fichier ); /* Fin de la ligne avec les valeurs du chemin */
  if ( EOF == ignore_ligne( fichier ) ) { /* Ligne des lgendes */
    (void) fprintf(stderr, "%s,%s l.%u : impossible de trouver les premires coordonnes, format de fichier incorrect\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return EOF;
  }

  /* On cre la couche concerne */
  couche = COUCHE_Creer( modele_feff->infos_modele.feff, NULL );
  if ( NULL == couche ) {
    (void) fprintf(stderr, "%s,%s l.%u : impossible de crer la couche !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return EOF;
  }

  /* Et on l'initialise */
  couche->R.valeur_initiale = longueur; /* Longueur apparente; variation de 10% autour autorise */
  couche->R.min = longueur * 0.9;
  couche->R.max = longueur * 1.1;
  couche->R.contenu.valeur = longueur;
  if ( infos->options & PATHSDAT_R_CONSTANT ) {
    couche->R.type = VAR_CONSTANTE;
  }

  couche->N.valeur_initiale = N; /* Dgnrescence, variation de 0,5 autour autorise */
  couche->N.min = N - 0.5;
  couche->N.max = N + 0.5;
  couche->N.contenu.valeur = N;
  if ( infos->options & PATHSDAT_N_CONSTANT ) {
    couche->N.type = VAR_CONSTANTE;
  }

/*   couche->Debye_Waller.valeur_initiale = 1e-3; */
  couche->Debye_Waller.min = 1e-5;
  couche->Debye_Waller.max = 0.5; /* Intervalle large car grandes variations dans modles multi-couches... */

  /* Chargement du chemin proprement dit */
  couche->chemin = FEFF_PathsDat_LireChemin( fichier, infos, num_chemin, n_reflexion);
  if ( NULL == couche->chemin ) {
    char *nom_couche;

    nom_couche = TXT_genere_chaine( "Chemin %d (sans chemin)", num_chemin );
    COUCHE_Nommer( couche, nom_couche );
    free( nom_couche ); nom_couche = NULL;
    return EOF;
  } else {
    char *nom_couche;

    nom_couche = TXT_genere_chaine( "Chemin %d : %s", num_chemin, couche->chemin->nom);
    COUCHE_Nommer( couche, nom_couche );
    free( nom_couche );  nom_couche = NULL;
  }
  couche->chemin->amplitude = couche->amplitude;
  couche->chemin->phase = couche->phase;
  couche->chemin->multiplicite = N;
  couche->chemin->pseudo_longueur = longueur;
  couche->chemin->longueur_reelle = longueur;

  /* Chargement des fichiers phases & amplitudes [si on les veut] */
  if ( serie_amp_pha != NULL ) {
    int retour;

    retour = FEFF_PathsDat_LireFEFFxxxDAT( infos, serie_amp_pha, couche, num_chemin, longueur);
    if (retour != 0) {
      return +1;
    }
  }

  return 0; /* Tout s'est bien pass */
}

static int FEFF_PathsDat_LireFEFFxxxDAT(FEFF_PATHS_DAT *infos, Serie_Graphes *serie_amp_pha, Couche *couche, unsigned int numero_chemin, double longueur)
{
  FILE *feff;
  Graphe *phase = NULL, *amplitude = NULL;
  char *nom_feff;
  int c_lu[2];
  unsigned long nbr_lignes;

  if ( NULL == couche ) /* Cration de la couche impossible ! */
    return -1;

  /* Maintenant, lecture des phases et amplitudes */
  nom_feff = (char *) calloc( strlen( infos->chemin_travail ) + 14, sizeof(char));
  if (nom_feff == NULL) {
    ERR_Memoire( "Nom du fichier feffxxx.dat", strlen( infos->chemin_travail ) + 14 );
    return -2;
  }

  sprintf( nom_feff, "%s/feff%.4d.dat", infos->chemin_travail, numero_chemin );
  /*  printf("Fichier feff contenant phases et amplitudes : %s\n", nom_feff); */

  feff = fopen( nom_feff, "rb" );
  if ( NULL == feff ) {
    (void) fprintf( stderr, "%s,%s l.%u : erreur en voulant lire %s [Chemin %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, nom_feff, numero_chemin );
    free( nom_feff );
    couche->utilisee = FAUX; /* On ne l'utilise pas : pas d'amplitude et phase associes ! */
    return -3;
  }

  /* On passe l'en-tte */
  c_lu[0] = 0; c_lu[1] = 0;
  while ( ! ( (c_lu[0] == '@') && (c_lu[1] == '#') ) ) {
    c_lu[0] = c_lu[1];
    c_lu[1] = fgetc(feff);

    if (c_lu[1] == EOF) {
      fprintf(stderr, "Erreur : fin de fichier atteinte, marque @# non trouve !\n");
      fprintf(stderr, "tes-vous sr qu'il s'agisse d'un fichier FEFF ?\n");
      (void) fclose(feff);
      free(nom_feff);
      couche->utilisee = FAUX; /* On ne l'utilise pas : pas d'amplitude et phase associes ! */
      return -4;
    }
  }
  ignore_ligne(feff); /* On se place au dbut de la ligne suivante... */

  /* On compte le nombre de lignes */
  nbr_lignes = compte_lignes(feff, 0);

#ifdef _ENGLISH_
  sprintf(nom_feff, "Phase-shift for path %d", numero_chemin);
#else
  sprintf(nom_feff, "Phase du chemin %d", numero_chemin );
#endif
  phase = Serie_nv_graphe( serie_amp_pha, NULL, nbr_lignes, GR_EXAFSPHA, GR_NORMAL, nom_feff);
  if (phase == NULL) {
    (void) fclose(feff);
    free(nom_feff);
    couche->utilisee = FAUX; /* On ne l'utilise pas : pas de phase associe ! */
    return -5;
  }

#ifdef _ENGLISH_
  sprintf( nom_feff, "Amplitude for path %d", numero_chemin );
#else
  sprintf( nom_feff, "Amplitude du chemin %d", numero_chemin );
#endif
  amplitude = Serie_nv_graphe( serie_amp_pha, phase, nbr_lignes, GR_EXAFSAMP, GR_NORMAL, nom_feff );
  if (amplitude == NULL) {
    GRF_efface(phase, VRAI, VRAI);
    (void) fclose(feff);
    free(nom_feff);
    couche->utilisee = FAUX; /* On ne l'utilise pas : pas d'amplitude associe ! */
    return -6;
  }

  couche->phase     = phase;
  couche->amplitude = amplitude;

  /* On lit les phases et amplitudes _avec corrections ncessaires_ */
  (void) FEFF_PathsDat_LireAmplitudePhase( feff, phase, amplitude, longueur, infos );

  GRF_couleur( phase, (numero_chemin % 15) + 1, -1, -1 );
  GRF_couleur( amplitude, (numero_chemin % 15) + 1, -1, -1 );

  (void) fclose(feff);
  free(nom_feff);
  return 0;
}

/* -------- Lecture & correction au vol des phases et amplitudes ---------\
 |  Renvoie <0 en cas d'erreur, le nombre de points lus sinon             |
 |                                                                        |
 | amplitude : graphe contenant l'amplitude corrige                      |
 | phase     : graphe contenant le dphasage corrig                      |
 | longueur  : longueur apparente du chemin (pour correction d'amplitude) |
 \----------------------------------------------------------------------- */

static int FEFF_PathsDat_LireAmplitudePhase(FILE *fichier, Graphe *phase, Graphe *amplitude, double longueur, FEFF_PATHS_DAT *infos)
{
  double k, phi_central, amp, pha, red_fac, lpm, re_p;
  register double *x_amp, *x_pha, *y_amp, *y_pha;
  unsigned long i;

  for (i = 0, x_amp = amplitude->x, y_amp = amplitude->y, x_pha = phase->x, y_pha = phase->y ; 
       i < phase->nombre_point ;
       i++, x_amp++, y_amp++, x_pha++, y_pha++)
  {
    fscanf(fichier, "%lf %lf %lf %lf %lf %lf %lf", &k, &phi_central, &amp, &pha, &red_fac, &lpm, &re_p);

    *x_amp = k;
    *x_pha = k;

    *y_amp = amp * red_fac;
    *y_pha = phi_central + pha;

    if ( infos->options & PATHSDAT_INCLURE_LPM ) {
      *y_amp *= exp(-2.0 * longueur / lpm); /* On inclus le libre-parcours moyen */
    }
  }

  GRF_bornes(amplitude, ORDONNE, DESORDONNE); /* Calcul des bornes, pour l'affichage */
  GRF_bornes(phase    , ORDONNE, DESORDONNE);

  return (int) i; /* Tout s'est bien pass, joie */
}
