/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Création de ce fichier : 23 juin 1998

   Module de filtrage des spectres (essentiellement par transformée de Fourier)

   Paramétrage de la transformée de Fourier inverse
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "exafs.h"
#include "filtrage.h"
#include "interface.h"
#include "operations.h" /* Pour les infos de troncature d'un graphe */
#include "tracer.h"

#include "tf_inverse.txt"

/* ******** Variables locales ************************ */

static int num_tfi_fenetre = FEN_HANNING; /* Fenêtre à l'ouverture du module */

/* ******** Fonctions locales (gestion de l'interface) ******** */

static void RDS_tfi(Fenetre *, int x, int y, unsigned int larg , unsigned int haut);
static int  MSG_tfi(Fenetre *fen, Message *msg);
static int  GER_tfi(unsigned int num_objet, int x, int y);
static int  CLV_tfi(Fenetre *fen, unsigned int touche, int etat);
static int  CLC_tfi(Fenetre *fen, int bouton, int n_clic, int etat, int x, int y);
static int  TFI_PrepareArbreFenetre(unsigned int numero_arbre);

static int  GER_paramtfi(unsigned int num_objet, int x, int y);
static void prepare_arbre_tfi_inverse(void);

static int  TFI_DemanderLesR   (Fenetre *fenetre, TFI_infos *infos);
static int  TFI_InstallerFiltre(Fenetre *fenetre, TFI_infos *infos, int nv_filtre);
static void TFI_InstallerLesR  (Fenetre *fenetre, TFI_infos *infos);
static void TFI_MiseAJour      (Fenetre *fenetre, TFI_infos *infos);

/* ******** Réalisation pratique des fonctions **************** */

Graphe *TFI_CalculInverse(TFI_infos *infos, double k_min, double k_max, double pas,
			  int methode, boolean erreurs, boolean covariance)
{
  Infos_Nyquist *nyquist;
  Graphe *resultat, *original, *filtre;
  char *titre, *unite;
  const char *unite_X;

  if ( NULL == infos) return NULL;
  if ( NULL == infos->graphe ) return NULL;

  /* Contrôle du pas */
  if ( k_min > k_max ) {
    double k = k_min; k_min = k_max; k_max = k;
  }
  if (pas < 0.0) pas = 0.0;
  if ( pas > (k_max - k_min) ) {
    return NULL;
  }

  /* Contrôle pour le calcul des erreurs */
  if ( NULL == infos->graphe->erreur_y )
    erreurs = FAUX;
  if ( FAUX == erreurs )
    covariance = FAUX;

  /* ---- On calcule le graphe original à transformer ---- */

  /* 0. Extraction de la région intéressante */
  original = TF_IsoleSignal( infos->graphe, &(infos->filtre) );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'extraire la région intéressante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* 1. Le filtre [graphe isolé] */
  filtre = GRF_nv_TableX( NULL, GR_FENETRE, GRF_SERIE, NULL, VRAI,
			  original->x, original->nombre_point, FAUX );
  if ( NULL == filtre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de précalculer le filtre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    if ( original != infos->graphe ) {
      /* C'est donc un graphe isolé, à libérer tout de suite */
      GRF_eff_isole( original );
      original = NULL;
    }
    return NULL;
  }
  TF_fenetre( filtre, &(infos->filtre) );

  /* 2. Produit graphe * filtre, dans filtre */
  if ( NULL == GRF_ProduitGraphes( filtre, original, VRAI, VRAI, NULL ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le produit des graphes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    if ( original != infos->graphe ) {
      GRF_eff_isole( original ); /* Plus besoin de lui */
      original = NULL;
    }
    GRF_eff_isole( filtre );
    filtre = NULL;
    return NULL;
  }

  /* Log dans l'historique */
  HIST_Titre(1, "Transformation de Fourier inverse", "TFI");
  HIST_Texte(1, 0, "Graphe original : %s\n", infos->graphe->nom);
  HIST_Texte(3, 0, "Calcul pour k compris entre %.10f et %.10f avec un pas de %.10e\n", k_min, k_max, pas);
  if ( infos->filtre.type != PAS_DE_FENETRE ) {
    HIST_Texte(2, 0, "Filtre : type = %s\n", Types_fenetres[infos->filtre.type]);
    HIST_Texte(3, 0, "         [Rmin, Rmax] = [%.10f, %.10f]\n", infos->filtre.x_min, infos->filtre.x_max);
    if ( Autorise_Symetrique(infos->filtre.type) ) { /* Fenêtres définies par 4 points */
      HIST_Texte(3, 0, "Fin de la mont"e_AIGU"e     : %.10g\n", infos->filtre.x_1);
      HIST_Texte(3, 0, "D"e_AIGU"but de la descente : %.10g\n", infos->filtre.x_2);
    }
    if ( Avec_Parametre(infos->filtre.type) ) {
      HIST_Texte(3, 0, "Param"e_GRAVE"tre : %.10f\n", infos->filtre.parametre);
    }
  } else {
    HIST_Texte(2, 0, "Pas de filtre\n");
  }

  /* ---- Calcul de la TFI ---- */

  switch ( methode )
  {
  case TF_CALC_RAPIDE:
    HIST_Texte(4, 0, "M"e_AIGU"thode : algorithme rapide pour 2^n points\n");
    {
      Graphe *tf_inv;
      double *nv_x, *nv_y;
      unsigned long nbr_pts;

      tf_inv = TF_prep_rapide( filtre, k_min, k_max, pas, TF_SENS_INVERSE);
      if ( NULL == tf_inv ) {
	HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de pr"e_AIGU"parer le graphe r"e_AIGU"sultat");
	GRF_eff_isole( filtre );
	return NULL;
      }
      TFI_rapide(filtre, tf_inv); /* Calcul effectif */

      /* On récupère la seule partie demandée */
      nbr_pts = GRF_extrait_table(tf_inv, k_min, k_max, FAUX, &nv_x, &nv_y);
      if (0 != nbr_pts ) { /* On réalise la troncature */
	free(tf_inv->x);
	free(tf_inv->y);
	tf_inv->x = nv_x;
	tf_inv->y = nv_y;
	tf_inv->nombre_point = nbr_pts;

	/* Remarque : TF rapide, pas d'erreur ni de matrice de covariance sur le résultat */
      }
      resultat = tf_inv;
    }
    break;

  case TF_CALC_NORMAL:
    HIST_Texte(4, 0, "M"e_AIGU"thode : int"e_AIGU"gration par la m"e_AIGU"thode des trap"e_GRAVE"zes\n");
    resultat = TF_prep_normal(k_min, k_max, pas, TF_SENS_INVERSE, erreurs, covariance);
    if ( NULL == resultat ) {
      HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de pr"e_AIGU"parer le graphe r"e_AIGU"sultat");
      GRF_eff_isole( filtre );
      return NULL;
    }
    TFI_Trapezes(filtre, resultat);
    resultat->drapeaux.protege = FAUX;
    break;

  default:
    (void) fprintf(stderr, "TFI_CalculInverse : méthode de calcul inconnue [%d]\n", methode);
    return NULL;
  }

  /* --- Plus besoin du filtre --- */
  GRF_eff_isole( filtre ); filtre = NULL;
  if ( original != infos->graphe ) {
    /* On ne peut effacer l'original copié (graphe isolé) que maintenant,
       car ses X sont aussi ceux de filtre ! */
    GRF_eff_isole( original );
    original = NULL;
  }

  /* ---- On élimine la fonction d'apodisation ---- */

  if ( infos->graphe->pointeur.infos_tf != NULL ) {
    Graphe *fenetre;

    fenetre = GRF_nv_isole( resultat->nombre_point, GR_FENETRE, GR_NORMAL);
    if (NULL == fenetre ) {
      (void) BI_ERR_alerte(NULL, &ERR_TFI_SansFenetre, VRAI);
    } else {
      int code_erreur;

      HIST_Texte(4, 0, "Division par la fonction d'apodisation utilis"e_AIGU"e pour la TF\n");
      memcpy( fenetre->x, resultat->x, sizeof(double) * resultat->nombre_point );

      /* Calcul de la fonction d'apodisation */
      TF_fenetre(fenetre, infos->graphe->pointeur.infos_tf );

      /* Division */
      GRF_QuotientGraphes(resultat, fenetre, VRAI, VRAI, NULL, &code_erreur);
      GRF_eff_isole( fenetre );
      if ( ERR_GRF_DIV_ZERO == code_erreur ) {
	(void) BI_ERR_alerte(NULL, &ERR_TFI_FenetreNulle, VRAI);
      }
    }
  } else {
    (void) BI_ERR_alerte(NULL, &ERR_TFI_SansFenetre, VRAI);
  }

  /* On cherche les bornes en Y */
  GRF_bornes(resultat, NE_PAS_FAIRE, DESORDONNE);

  /* ---- Calcul terminé ---- */

  /* Titre du résultat */
  titre = TXT_genere_chaine(titre_grf_tfi, infos->graphe->nom);
  GRF_titre( resultat, titre);
  free( titre ); titre = NULL;
  HIST_Texte(1, 0, "R"e_AIGU"sultat : %s\n", resultat->nom);

  /* Unité en X : l'inverse de l'unité en X du graphe avant TFI */
  unite_X = GRF_QuelleUnite( infos->graphe, AXE_X );
  unite = TXT_ProduitUnites( NULL, unite_X, -1 );
  GRF_ChangerUnite( resultat, AXE_X, unite );
  free( unite ); unite = NULL;

  /* Unité en Y : unité en Y avant TFI * unité en X avant TFI */
  unite = TXT_ProduitUnites( GRF_QuelleUnite( infos->graphe, AXE_Y ), unite_X, 1 );
  GRF_ChangerUnite( resultat, AXE_Y, unite );
  free( unite ); unite = NULL;

  /* On conserve couleur & style */
  GRF_CopierAffichage( resultat, infos->graphe, GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* On mémorise l'intervalle pris en compte */
  nyquist = (Infos_Nyquist *) malloc(sizeof(Infos_Nyquist));
  if (nyquist == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Filtrage/TF_inverse.c", "MSG_tfi", "nyquist");
  } else {
    nyquist->r_min = infos->filtre.x_min;
    nyquist->r_max = infos->filtre.x_max;
  }
  resultat->pointeur.nyquist = nyquist;

  return resultat;
}

/* -------------- Ouverture de la fenêtre ----------------------------------------------- */

void TFI_ouvre(Graphe *tf)
{
  Fenetre *GEM_fenetre;
  TFI_infos *infos;

  if (tf == NULL) return; /* Rien à faire */
  if (tf->drapeaux.protege == VRAI) {
    EXF_ERREUR(GRF_PROTEGE); /* Graphe protégé -> appel impossible */
    return;
  }

  GEM_fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fen_tfi, NULL,
				    PARAM_FOUR_INVER, BI_FEN_ARBRE_HAUT, TFI_PrepareArbreFenetre,
				    RDS_tfi, CLV_tfi, CLC_tfi, MSG_tfi, RCT_pointeur,
				    ( int (*)(Fenetre *) ) NULL);
  if (GEM_fenetre == NULL) return; /* Rien à faire... */

  FEN_Ouvrir( GEM_fenetre );

  GEM_fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_tf */
  GEM_fenetre->drapeaux.redessin    = VRAI;

  infos = (TFI_infos *) FEN_LireComplement( GEM_fenetre, FEN_COMPLEMENT_2 );
  if ( infos != NULL ) { /* Déjà des infos sur une TF -> on les perd */
    if ( infos->graphe == tf ) { /* Même graphe -> pas à se fouler */
      return;
    } else {
      /* On efface ces infos pour en créer de nouvelles */
      /* On les délie donc de la fenêtre */
      FEN_GRF_LierGraphe( GEM_fenetre, NULL );
      FEN_StockerComplement( GEM_fenetre, FEN_COMPLEMENT_2, NULL );

      /* On peut donc effacer les graphes */
      GRF_efface( infos->fenetre, VRAI, VRAI ); infos->fenetre = NULL;

      num_tfi_fenetre = infos->filtre.type;
      infos->graphe->drapeaux.protege = VRAI; /* Pas utilisé -> on peut l'effacer */
      free( infos ); infos = NULL;
    }
  }

  infos = (TFI_infos *) malloc( sizeof(TFI_infos) ); /* On réserve les infos de TF inverse */
  if (infos == NULL) {
    FEN_Detruire( GEM_fenetre ); /* On ferme la fenêtre : sert à rien... */
    SYS_ERREUR(MANQUE_MEMOIRE, "f_inv.c", "TFI_ouvre", "infos");
    return;
  }

  OPTION_A_FAIRE( "Créer la fenêtre de la TFI en graphe isolé" ); /* ,"Longueur du chemin", "Pond"e_AIGU"ration", ANGSTROM, NULL */
  infos->fenetre = GRF_nouveau( tf, tf->nombre_point, GR_INTERNE, GRF_FCT_COMPLEXE, titre_grf_filtre );
  if (infos->fenetre == NULL) { /* Mémoire insuffisante */
    free( infos ); infos = NULL;
    FEN_Detruire( GEM_fenetre );
    return;
  }
  memcpy(infos->fenetre->x, tf->x, sizeof(double) * tf->nombre_point );
  infos->fenetre->x_min = tf->x_min;
  infos->fenetre->x_max = tf->x_max;

  GRF_couleur( infos->fenetre, BLEU, -1, -1 );
  infos->fenetre->drapeaux.protege = VRAI; /* Protégé -> ineffaçable */

  infos->graphe = tf; /* Le graphe sur lequel on travaille */
  tf->drapeaux.protege = VRAI;  /* Il est donc ineffaçable... */
  
  /* Paramètres de la TFI (fenêtre en particulier) */
  infos->filtre.x_min = tf->x_min + 1.0; /* Intervalle pour lequel on calculera */
  infos->filtre.x_max = tf->x_max - 1.0;
  infos->filtre.x_1 = infos->filtre.x_min + 1.0;
  infos->filtre.x_2 = infos->filtre.x_max - 1.0;
  OBJ_MettreIntervalle( PARAM_FOUR_INVER, TF_INV_RMIN, TF_INV_RMAX, 0, 0, infos->filtre.x_min, infos->filtre.x_max, 0.0, 0 );
  OBJ_MettreIntervalle( PARAM_FOUR_INVER, PRPTFI_R1, PRPTFI_R_2, 0, 0, infos->filtre.x_1, infos->filtre.x_2, 0.0, 0 );
  infos->filtre.symetrique = OBJ_choisit(PARAM_FOUR_INVER, PRPTFI_SYMETRIK, QUESTION);

  /* Calcul de la TF-inverse : fenêtre */
  infos->filtre.type = num_tfi_fenetre;
  infos->num_depart   = FEN_RECTANGLE; /* Affichage à partir de la première */

  ARB_PreparerDefileur_T(PARAM_FOUR_INVER, TF_INV_FEN1, TF_INV_FEN_4, 0,
			 TF_INV_FEN_AVANT, TF_INV_FEN_FOND, TF_INV_FEN_BOUT, TF_INV_FEN_APRES, 
			 NOMBRE_FENETRES, Types_fenetres, NULL, num_tfi_fenetre, FEN_RECTANGLE);

  infos->filtre.parametre = OBJ_LireNombre(PARAM_FOUR_INVER, TF_INV_PARAM, NULL );
  TF_fenetre( infos->fenetre, &(infos->filtre) ); /* On calcule la fenêtre */

  /* On les lie à la fenêtre, maintenant que tout est prêt */
  FEN_GRF_LierGraphe( GEM_fenetre, tf );
  FEN_StockerComplement( GEM_fenetre, FEN_COMPLEMENT_2, infos );
}

/* ----------- Redessin de cette fenêtre ------------------------------------------------ */

static void RDS_tfi(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Graphe *graphe, *produit;
  InfosRegion2D region;
  TFI_infos *infos;
  double echelle_X, echelle_Y;
  int x_barre, retour;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  infos = (TFI_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return;
  graphe = infos->graphe;
  if ( ( NULL == graphe ) || ( graphe->x_max == graphe->x_min ) || ( graphe->y_max == graphe->y_min ) ) {
    return;
  }

  /* On prépare la région 2D : aide pour les dessins après en fait */
  retour = TRC_Z2D_Initialiser( &region, fenetre, graphe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* Le graphe dont on fait la TF  */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &echelle_X, &echelle_Y );

  /* Dessin du graphe multiplié par le filtre - en violet */
  /* a. On copie, en isolé, la fenêtre */
  produit = GRF_Copier( infos->fenetre, NULL, VRAI );
  if ( produit != NULL ) {
    /* On calcule le produit */
    GRF_ProduitGraphes( produit, graphe, VRAI, FAUX, NULL );

    /* On trace, avec les mêmes échelles */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, produit, VIOLET, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

    /* On n'a plus besoin de cet isolé */
    GRF_eff_isole( produit ); produit = NULL;

    OPTION_A_FAIRE( "Tracé du produit graphe * filtre par un graphe isolé à optimiser !" );
  }

  /* B. Dessin du filtre : forcément compris entre 0 et 1, occupe toute la fenêtre */
  region.axe_Y.min = 0.0;
  region.axe_Y.max = 1.0;
  TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->fenetre, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

  /* Mise en place des barres pour k_min (rouge) et k_max (vert) */
  x_barre = TRC_ConvertitX_PvE( infos->filtre.x_min, graphe->x_min, 0, echelle_X );
  TRC_Couleur( ROUGE );
  TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

  x_barre = TRC_ConvertitX_PvE( infos->filtre.x_max, graphe->x_min, 0, echelle_X );
  TRC_Couleur( ROUGE );
  TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

  /* Mise en place des barres pour k_1 et k_2, si la fenêtre s'y prête */
  if ( Autorise_Symetrique(infos->filtre.type) ) {
    x_barre = TRC_ConvertitX_PvE( infos->filtre.x_1, graphe->x_min, 0, echelle_X );
    TRC_Couleur( ROUGE_FONCE );
    TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

    if (infos->filtre.symetrique) {
      x_barre = TRC_ConvertitX_PvE( infos->filtre.x_max - (infos->filtre.x_1 - infos->filtre.x_min), graphe->x_min, 0, echelle_X );
    } else {
      x_barre = TRC_ConvertitX_PvE( infos->filtre.x_2, graphe->x_min, 0, echelle_X );
    }
    TRC_Couleur( VERT_FONCE );
    TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );
  }
}

/* ------------ Traitement des messages reçus par la fenêtre --------------------------------- */

static int MSG_tfi(Fenetre *fenetre, Message *message)
{
  Graphe *graphe = NULL, *tf_fenetre = NULL, *fourier;
  TFI_infos *infos; /* Les infos pour la fenêtre, pour la TF */

  infos = (TFI_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL != infos ) {
    graphe     = infos->graphe;
    tf_fenetre = infos->fenetre;
  }

  if ( message->type == WM_CLOSED ) { /* On demande la fermeture de la fenêtre */
    int reponse;

    if ( message->fenetre.haut == -1 ) /* On force la fermeture */
      reponse = 2;
    else {
      reponse = EXF_ERREUR(FIN_MODULE);
      if (reponse == 3) return BI_EVENEMENT_TRAITE; /* On annule */
    }

    num_tfi_fenetre = infos->filtre.type;
    if (reponse == 2) {
      /* On quitte sans calculer la TF inverse
	 ==> on délie donc les informations de la fenêtre */
      FEN_GRF_LierGraphe( fenetre, NULL );
      FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

      /* On peut maintenant effacer tranquillement */
      GRF_efface( tf_fenetre, VRAI, VRAI ); /* On efface ce graphe */
      if ( graphe != NULL )
	graphe->drapeaux.protege = FAUX;
      free( infos ); infos = NULL; /* On libère la mémoire */
      FEN_Detruire( fenetre ); /* On ferme la fenêtre */
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* On ferme la fenêtre : les choses sérieuses commencent... */
    FEN_Fermer( fenetre, FAUX ); /* Évite un sale message de redessin */

    if ( TFI_init(infos) == FAUX) { /* Calcul annulé (utilisateur) */
      FEN_Reouvrir( fenetre );
      return BI_EVENEMENT_FEN_FERMEE; /* Les index de fenêtre ont pu changer en réouvrant ! */
    }
    FEN_Detruire( fenetre ); /* On ferme définitivement la fenêtre */

    GRF_efface( tf_fenetre, VRAI, VRAI );

    { /* On stocke les paramètres dans l'arbre de reproduction */
      extern struct infos_reproduit_filtre infos_reproduire;

      OBJ_MettreIntervalle( TF_GENERAL, TFG_TFI_KMIN, TFG_TFI_KMAX, TFG_TFI_PAS, 0,
			    OBJ_LireNombre(PARAM_TFI, TFI_KMIN, NULL ), OBJ_LireNombre(PARAM_TFI, TFI_KMAX, NULL ), OBJ_LireNombre(PARAM_TFI, TFI_PAS, NULL ), 0 );

      infos_reproduire.type_fen_tfi = infos->filtre.type;
      infos_reproduire.tfi_symetrique = infos->filtre.symetrique;
      OBJ_MettreIntervalle( TF_GENERAL, RF_VAL_R1, RF_VAL_R2, 0, 0, infos->filtre.x_1, infos->filtre.x_2, 0.0, 0 );
      OBJ_MettreIntervalle( TF_GENERAL, RF_VAL_RMIN, RF_VAL_RMAX, 0, 0, infos->filtre.x_min, infos->filtre.x_max, 0.0, 0 );
      OBJ_MettreNombre( TF_GENERAL, TFG_TFI_PARAM, infos->filtre.parametre, 0.0 );
    }

    /* Calcul de la transformée de Fourier */
    if ( OBJ_choisit(PARAM_TFI, TFI_RAPIDE, QUESTION) ) { /* On veut la méthode rapide */
      reponse = TF_CALC_RAPIDE;
    } else if ( OBJ_choisit(PARAM_TFI, TFI_GENERALE, QUESTION) == VRAI ) {
      reponse = TF_CALC_NORMAL;
    }
    
    fourier = TFI_CalculInverse(infos, OBJ_LireNombre(PARAM_TFI, TFI_KMIN, NULL ), OBJ_LireNombre(PARAM_TFI, TFI_KMAX, NULL ), OBJ_LireNombre(PARAM_TFI, TFI_PAS, NULL ),
				reponse, OBJ_choisit(PARAM_TFI, TFI_ESTIME_ERR, QUESTION), OBJ_choisit(PARAM_TFI, TFI_MAT_COVAR, QUESTION));

    /* On récupère le spectre EXAFS proprement dit == la partie réelle du spectre transformé... */
    if ( OBJ_choisit(PARAM_TFI, TFI_SPECTRE, QUESTION) ) { /* On garde le signal EXAFS */
      char *titre;
      const char *unite_Y;
      int ponderation;
      Graphe *exafs;

      exafs = CMP_reelle(fourier); /* S'occupe des erreurs et tout */

      /* On met un titre plus adapté */
      titre = TXT_genere_chaine(titre_grf_exafs, infos->graphe->nom);
      GRF_titre(exafs, titre);
      free(titre);

      /* On installe une légende en Y plus adapté [en utilisant l'unité comme indice pour la pondération] */
      unite_Y = GRF_QuelleUnite( exafs, AXE_Y );
      if ( VRAI == TXT_UniteEstSimple( unite_Y, &ponderation ) ) {
	if ( 0 == ponderation ) {
	  GRF_ChangerLegende( exafs, AXE_Y, "khi(k)", VRAI );
	} else if ( -1 == ponderation ) {
	  GRF_ChangerLegende( exafs, AXE_Y, "k * khi(k)", VRAI );
	} else if ( -2 == ponderation ) {
	  GRF_ChangerLegende( exafs, AXE_Y, "k"DEUX_EXPOSANT" * khi(k)", VRAI );
	} else if ( -3 == ponderation ) {
	  GRF_ChangerLegende( exafs, AXE_Y, "k"TROIS_EXPOSANT" * khi(k)", VRAI );
	} else {
	  char *legende;

	  legende = TXT_genere_chaine( "k^%d * khi(k)", -ponderation );
	  GRF_ChangerLegende( exafs, AXE_Y, legende, FAUX );
	  /* la légende sera libérée avec le graphe */
	}
      }
    }

    if ( OBJ_choisit(PARAM_TFI,TFI_AMPLITUDE, QUESTION) ) { /* On veut garder l'amplitude = module */
      Graphe *module;

      module = CMP_module(fourier);
      if ( NULL != module ) {
	char *titre;

	/* On change le type */
	GRF_ChangerType( module, GR_EXAFSAMP );

	/* On change le titre */
	titre = TXT_genere_chaine(titre_grf_amplitude, infos->graphe->nom);
	GRF_titre(module, titre);
	free(titre);
	module = corr_amplitude(module, 0.0, 0.0, 0.0); /* Demande des corrections d'amplitude */
      }
    }

    if ( OBJ_choisit(PARAM_TFI, TFI_PHASE, QUESTION) ) { /* On veut le déphasage = argument */
      Graphe *phase;
      
      phase = CMP_argument( fourier );
      if ( phase != NULL ) {
	char *titre;

	/* On change le type */
	GRF_ChangerType( phase, GR_EXAFSPHA );

	/* On change le titre */
	titre = TXT_genere_chaine( titre_grf_phase, infos->graphe->nom );
	GRF_titre( phase, titre );
	free(titre); titre = NULL;

	/* On fait la conversion éventuelle [en demandant les valeurs à l'utilisateur] */
	phase = TFI_CorrigerPhase( phase, 0, 0.0 );
      }
    }

    /* On peut supposer que l'on a fini ici... */
    fourier->drapeaux.protege = FAUX; /* On a fini de calculer sur ces graphes */
    graphe->drapeaux.protege  = FAUX;
    GRF_aller(fourier); /* On met ce graphe en graphe courant */
    free(infos); infos = NULL; /* Ne sert plus à rien */

    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non traité [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Gestion du clavier pour la fenêtre de TF inverse -------- */

static int CLV_tfi(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour;
  TFI_infos *infos;
  Message message;
  unsigned long nv_filtre, prem_visible;

  /* Ctrl-touche : raccourci pour le menu -> non pris en compte ici */
  if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE;

  infos = (TFI_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Infos = NULL\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  switch (touche)
  {
  case 'r':
  case 'R':
    (void) TFI_DemanderLesR( fenetre, infos );
    return BI_EVENEMENT_TRAITE;

  case 'p':
  case 'P':
    if ( Avec_Parametre(infos->filtre.type) ) {
      retour = ARB_SimulerClic( PARAM_FOUR_INVER, TF_INV_PARAM, 1 );
      if ( retour != BI_EVENEMENT_NON_TRAITE )
	return retour;
    }
    return BI_EVENEMENT_TRAITE; /* Traité, mais sans effet */

  case 's':
  case 'S':
    if ( Autorise_Symetrique(infos->filtre.type) ) {
      retour = ARB_SimulerClic( PARAM_FOUR_INVER, PRPTFI_SYMETRIK, 1 );
      if ( retour != BI_EVENEMENT_NON_TRAITE )
	return retour;
    }
    return BI_EVENEMENT_TRAITE; /* Traité, mais sans effet */

  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    retour = (fenetre->message)(fenetre, &message);
    if ( retour != BI_EVENEMENT_NON_TRAITE )
      return retour;
    return BI_EVENEMENT_TRAITE; /* Traité, mais sans effet */

  case Fleche_Haut:
  case Fleche_Bas:
    /* Préparation de l'appel */
    nv_filtre    = (unsigned long) infos->filtre.type;
    prem_visible = (unsigned long) infos->num_depart;

    /* Appel « clavier » du défileur */
    if ( touche == Fleche_Haut ) {
      ARB_GererDefileur_T( PARAM_FOUR_INVER, TF_INV_FEN1, TF_INV_FEN_4, 0,
			   TF_INV_FEN_AVANT, TF_INV_FEN_FOND, TF_INV_FEN_BOUT, TF_INV_FEN_APRES,
			   NOMBRE_FENETRES, Types_fenetres, NULL, &nv_filtre, &prem_visible,
			   ROOT, +1 );
    } else {
      ARB_GererDefileur_T( PARAM_FOUR_INVER, TF_INV_FEN1, TF_INV_FEN_4, 0,
			   TF_INV_FEN_AVANT, TF_INV_FEN_FOND, TF_INV_FEN_BOUT, TF_INV_FEN_APRES,
			   NOMBRE_FENETRES, Types_fenetres, NULL, &nv_filtre, &prem_visible,
			   ROOT, -1 );
    }

    /* Mémorisation des résultats de l'appel */
    infos->num_depart = prem_visible;
    (void) TFI_InstallerFiltre( fenetre, infos, nv_filtre );

    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* --------- Gestion d'un clic sur la fenêtre de TF inverse --------------------------------- */

static int CLC_tfi(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)), int etat, int x, int y __attribute__((__unused__)))
{
  TFI_infos *infos;
  double x_echelle, x_clic;

  infos = (TFI_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) {
    return BI_EVENEMENT_TRAITE;
  }
  if ( ( NULL == infos->graphe ) || ( infos->graphe->x_max == infos->graphe->x_min ) ) {
    return BI_EVENEMENT_TRAITE;
  }
  x_echelle = ( (double) fenetre->w_larg ) / (infos->graphe->x_max - infos->graphe->x_min);
  x_clic  = x - fenetre->w_x;
  x_clic /= x_echelle;
  x_clic += infos->graphe->x_min;

  if (bouton == 1) { /* Bouton gauche */
    if ( etat & ( K_LSHIFT | K_RSHIFT ) ) { /* On appuie sur shift en même temps */
      if ( Autorise_Symetrique( infos->filtre.type ) )
	infos->filtre.x_1 = x_clic;
      else
	infos->filtre.x_min = x_clic;
    } else
      infos->filtre.x_min = x_clic;
  } else { /* Bouton droit */
    if ( etat & ( K_LSHIFT | K_RSHIFT ) ) { /* On appuie sur shift en même temps */
      if ( Autorise_Symetrique( infos->filtre.type ) && (infos->filtre.symetrique == FAUX) )
	infos->filtre.x_2 = x_clic;
      else
	infos->filtre.x_max = x_clic;
    } else
      infos->filtre.x_max = x_clic;
  }

  /* On installe les valeurs modifiées */
  TFI_InstallerLesR( fenetre, infos );

  return BI_EVENEMENT_TRAITE;
}

/* -------- Préparation de la TFI (choix de la fenêtre principalement) ----------------------------- */

static int GER_tfi(unsigned int num_objet, int x_souris __attribute__((__unused__)), int y_souris)
{
  Fenetre *fen_cour;
  Graphe *fenetre __attribute__((__unused__));
  TFI_infos *infos;
  int reponse;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, {0}, {0}, {0} };

  fen_cour = ARB_QuelleFenetre(arbres + PARAM_FOUR_INVER, FAUX); /* Fenêtre actuelle */
  infos = (TFI_infos *) FEN_LireComplement( fen_cour, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return BI_EVENEMENT_TRAITE;
  fenetre = infos->fenetre;

  /* 1. S'agit-il du sélecteur de filtre ? */
  if ( ( ( num_objet >= TF_INV_FEN1 ) && ( num_objet <= TF_INV_FEN_4 ) ) ||
       ( TF_INV_FEN_AVANT == num_objet ) || ( TF_INV_FEN_FOND == num_objet ) || ( TF_INV_FEN_BOUT == num_objet ) || ( TF_INV_FEN_APRES == num_objet ) ) {
    unsigned long nv_filtre;
    unsigned long prem_visible;

    /* Préparation de l'appel */
    nv_filtre    = (unsigned long) infos->filtre.type;
    prem_visible = (unsigned long) infos->num_depart;

    /* On gère le défileur */
    ARB_GererDefileur_T( PARAM_FOUR_INVER, TF_INV_FEN1, TF_INV_FEN_4, 0,
			 TF_INV_FEN_AVANT, TF_INV_FEN_FOND, TF_INV_FEN_BOUT, TF_INV_FEN_APRES,
			 NOMBRE_FENETRES, Types_fenetres, NULL, &nv_filtre, &prem_visible, num_objet, y_souris );

    /* On mémorise quel est le nouveau premier visible */
    infos->num_depart = (int) prem_visible;

    /* Si nécessaire, on installe le filtre */
    (void) TFI_InstallerFiltre( fen_cour, infos, nv_filtre );

    /* Terminé */
    return BI_EVENEMENT_TRAITE;
  }

  switch (num_objet)
  {
  case TF_INV_RMIN: /* Demande de R min à l'utilisateur */
  case PRPTFI_R1:
  case PRPTFI_R_2:
  case TF_INV_RMAX: /* Demande de k max à l'utilisateur */
    (void) TFI_DemanderLesR( fen_cour, infos );
    return BI_EVENEMENT_TRAITE;

  case TF_INV_PARAM:
    nv_valeur.proprietes = 0;
    nv_valeur.texte = "p =";
    OPTION_A_FAIRE( "Mettre les limites en fonction du type de fenêtre" );
    nv_valeur.valeur_initiale.reel = infos->filtre.parametre;

    reponse = BI_DemandeValeur( fen_cour, &Q_ValeurParametre, &nv_valeur );
    if ( Q_ValeurParametre.bouton_annuler == (unsigned int) reponse ) return VRAI; /* On annule */

    infos->filtre.parametre = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( PARAM_FOUR_INVER, TF_INV_PARAM, infos->filtre.parametre, 0.0 );

    /* On met à jour */
    TFI_MiseAJour( fen_cour, infos );
    return BI_EVENEMENT_TRAITE;

  case PRPTFI_SYMETRIK:
    infos->filtre.symetrique = OBJ_choisit(PARAM_FOUR_INVER, PRPTFI_SYMETRIK, ECHANGE);
    OBJ_active(PARAM_FOUR_INVER, PRPTFI_R_2, !(infos->filtre.symetrique) );

    /* On met à jour */
    TFI_MiseAJour( fen_cour, infos );
    return BI_EVENEMENT_TRAITE;
  }
  
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Demander les valeurs de R1, ... -------- */

static int TFI_DemanderLesR(Fenetre *fenetre, TFI_infos *infos)
{
  const char *unite_X = NULL;
  int reponse;
  struct infos_demande nv_valeurs[ 4 ] = {
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } } };
  unsigned int i, nbr_valeurs_demandees;

  if ( NULL == infos->graphe ) return -1;

  /* Initialisation des valeurs */
  for ( i = 0 ; i < 4 ; i++ ) {
    nv_valeurs[ i ].proprietes = BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeurs[ i ].min = 0.0;
    nv_valeurs[ i ].max = infos->graphe->x_max;
  }

  /* Installation de l'unité, éventuellement */
  unite_X = GRF_QuelleUnite( infos->graphe, AXE_X );
  if ( unite_X != NULL ) {
    for ( i = 0 ; i < 4 ; i++ ) {
      nv_valeurs[ i ].proprietes |= BI_EDITABLE_UNITE;
      nv_valeurs[ i ].unite = unite_X;
    }
#ifdef _DEVELOPPE_
  } else {
    (void) fprintf(stderr, "%s,%s l.%u : Graphe sans unité [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->graphe->nom);
#endif
  }

  /* Borne inférieure */
  nv_valeurs[ 0 ].texte = "Rmin =";
  nv_valeurs[ 0 ].valeur_initiale.reel = infos->filtre.x_min;

  /* Borne inférieure du plateau, s'il existe */
  if ( VRAI == Autorise_Symetrique( infos->filtre.type ) ) {
    nv_valeurs[ 1 ].texte = "R1 =";
    nv_valeurs[ 1 ].valeur_initiale.reel = infos->filtre.x_1;

    /* Faut-il une borne supérieure du plateau ? */
    if ( VRAI == infos->filtre.symetrique ) {
      /* Non : automatique par symétrie */
      nbr_valeurs_demandees = 3;
    } else {
      /* Oui */
      nv_valeurs[ 2 ].texte = "R2 =";
      nv_valeurs[ 2 ].valeur_initiale.reel = infos->filtre.x_2;

      nbr_valeurs_demandees = 4;
    }
  } else {
    nbr_valeurs_demandees = 2;
  }

  /* Borne supérieure */
  nv_valeurs[ nbr_valeurs_demandees - 1 ].texte = "Rmax =";
  nv_valeurs[ nbr_valeurs_demandees - 1 ].valeur_initiale.reel = infos->filtre.x_max;

  /* On demande les valeurs à l'utilisateur */
  reponse = BI_DemandeMultiple( fenetre, &Q_ValeursR, nbr_valeurs_demandees, nv_valeurs );
  if ( reponse != 1 ) return +1; /* On annule ou il y a eu une erreur */

  /* On mémorise les valeurs */
  infos->filtre.x_min = nv_valeurs[ 0 ].valeur_donnee.reel;
  if ( VRAI == Autorise_Symetrique( infos->filtre.type ) ) {
    infos->filtre.x_1 = nv_valeurs[ 1 ].valeur_donnee.reel;
    if ( VRAI == infos->filtre.symetrique ) {
      infos->filtre.x_max = nv_valeurs[ 2 ].valeur_donnee.reel;
    } else {
      infos->filtre.x_2   = nv_valeurs[ 2 ].valeur_donnee.reel;
      infos->filtre.x_max = nv_valeurs[ 3 ].valeur_donnee.reel;
    }
  } else {
    infos->filtre.x_max = nv_valeurs[ 1 ].valeur_donnee.reel;
  }

  /* On prend en compte les nouvelles valeurs */
  TFI_InstallerLesR( fenetre, infos );

  /* Terminé sans problème */
  return 0;
}

/* -------- Installer un nouveau type de filtre -------- */

static int TFI_InstallerFiltre(Fenetre *fenetre, TFI_infos *infos, int nv_filtre)
{
  boolean lim_internes, av_parametre;

  if ( nv_filtre == infos->filtre.type ) return 0; /* Rien à faire */

  /* On mémorise le type de filtre */
  infos->filtre.type = nv_filtre;

  /* En fonction des possibilités, on active ou non les objets */
  if ( PAS_DE_FENETRE == nv_filtre ) {
    OBJ_active( PARAM_FOUR_INVER, TF_INV_RMIN, FORCE_NON );
    OBJ_active( PARAM_FOUR_INVER, TF_INV_RMAX, FORCE_NON );
  } else {
    OBJ_active( PARAM_FOUR_INVER, TF_INV_RMIN, FORCE_OUI );
    OBJ_active( PARAM_FOUR_INVER, TF_INV_RMAX, FORCE_OUI );
  }

  /* La fenetre a-t-elle des limites internes ? */
  lim_internes = Autorise_Symetrique( infos->filtre.type ); /* VRAI -> asymétrie possible */
  OBJ_active( PARAM_FOUR_INVER, PRPTFI_SYMETRIK, lim_internes );
  OBJ_active( PARAM_FOUR_INVER, PRPTFI_R1      , lim_internes );

  /* Si oui, force-t-on une fenetre symétrique ? */
  if ( VRAI == lim_internes ) {
    if ( OBJ_choisit( PARAM_FOUR_INVER, PRPTFI_SYMETRIK, QUESTION ) == VRAI ) {
      OBJ_active( PARAM_FOUR_INVER, PRPTFI_R_2, FORCE_NON );
    } else {
      OBJ_active( PARAM_FOUR_INVER, PRPTFI_R_2, FORCE_OUI );
    }
  } else {
    OBJ_active( PARAM_FOUR_INVER, PRPTFI_R_2, FORCE_NON );
  }

  /* La fenetre demande-t-elle un paramètre ? */
  av_parametre = Avec_Parametre( infos->filtre.type );
  OBJ_active( PARAM_FOUR_INVER, TF_INV_PARAM, av_parametre ); /* Seule fenêtre avec un paramètre... */

  /* On met à jour */
  TFI_MiseAJour( fenetre, infos );

  /* Terminé sans souci */
  return 0;
}

/* -------- Met les valeurs des R après tri (pour la cohérence) -------- */

static void TFI_InstallerLesR(Fenetre *fenetre, TFI_infos *infos)
{
  /* On controle l'ordre des valeurs */
  if (infos->filtre.x_min > infos->filtre.x_max) {
    double r = infos->filtre.x_min;
    infos->filtre.x_min = infos->filtre.x_max;
    infos->filtre.x_max = r;
  }
  if ( Autorise_Symetrique( infos->filtre.type ) ) {
    if (infos->filtre.symetrique) {
      if ( infos->filtre.x_1 < infos->filtre.x_min ) {
	double r = infos->filtre.x_min;
	infos->filtre.x_min = infos->filtre.x_1;
	infos->filtre.x_1 = r;
      }
      if ( infos->filtre.x_1 > infos->filtre.x_max ) {
	double r = infos->filtre.x_max;
	infos->filtre.x_max = infos->filtre.x_1;
	infos->filtre.x_1 = r;
      }
    } else {
      double r[4];
      int i, j;

      r[0] = infos->filtre.x_min;
      r[1] = infos->filtre.x_1;
      r[2] = infos->filtre.x_2;
      r[3] = infos->filtre.x_max;

      for (i = 0 ; i < 4 ; i++)
	for (j = i ; j < 4 ; j++)
	  if ( r[i] > r[j] ) {
	    double l = r[i];
	    r[i] = r[j]; r[j] = l;
	  }

      infos->filtre.x_min = r[0];
      infos->filtre.x_1   = r[1];
      infos->filtre.x_2   = r[2];
      infos->filtre.x_max = r[3];
    }
  }

  /* On affiche les valeurs dans l'arbre */
  OBJ_MettreIntervalle( PARAM_FOUR_INVER, TF_INV_RMIN, TF_INV_RMAX, 0, 0, infos->filtre.x_min, infos->filtre.x_max, 0.0, 0 );
  if ( Autorise_Symetrique( infos->filtre.type ) ) {
    OBJ_MettreNombre( PARAM_FOUR_INVER, PRPTFI_R1, infos->filtre.x_1, 0.0 );
    if (infos->filtre.symetrique == FAUX)
      OBJ_MettreNombre( PARAM_FOUR_INVER, PRPTFI_R_2, infos->filtre.x_2, 0.0 );
  }

  /* Si on a changé les valeurs, il faut recalculer et mettre à jour */
  TFI_MiseAJour( fenetre, infos );
}

/* -------- Mettre à jour les affichages -------- */

static void TFI_MiseAJour(Fenetre *fenetre, TFI_infos *infos)
{
  /* On recalcule le filtre */
  TF_fenetre( infos->fenetre, &(infos->filtre) );

  /* On redessine */
  FEN_redessin( fenetre, BI_RDS_UTIL );
}

/* ~~~~~~~~ Paramètres de calcul pour la TFI, autres que la fenêtre : type de méthode,... ~~~~~~~~ */

/* ----------------- Initialisation du calcul de la TF inverse ----------------------------\
 |     Renvoie le graphe qui contiendra la TF inverse                                      |
 |                                                                                         |
 | infos        : les informations de calcul de la TF inverse                              |
 \---------------------------------------------------------------------------------------- */

boolean TFI_init(TFI_infos *infos)
{
  Infos_FonctionApodisation *param_tf;
  boolean etat;

  if ( NULL == infos ) return FAUX;
  if ( NULL == infos->graphe ) return FAUX;

  /* On installe les infos de la TF, si on les a */
  infos->num_depart = FEN_RECTANGLE;
  if ( infos->graphe->pointeur.infos_tf ) { /* Infos de TF présentes */
    param_tf = infos->graphe->pointeur.infos_tf;

    OBJ_MettreNombre( PARAM_TFI, TFI_PARAMETRE, param_tf->parametre, 0.0 );
    OBJ_MettreIntervalle( PARAM_TFI, TFI_KMIN, TFI_KMAX, 0, 0, param_tf->x_min, param_tf->x_max, 0.0, 0 );
    OBJ_MettreIntervalle( PARAM_TFI, TFI_KMIN_FEN, TFI_KMAX_FEN, 0, 0, param_tf->x_min, param_tf->x_max, 0.0, 0 );
    OBJ_MettreIntervalle( PARAM_TFI, TFI_K_1, TFI_K_2, 0, 0, param_tf->x_1, param_tf->x_2, 0.0, 0 );
    OBJ_choisit(PARAM_TFI, TF_FEN_SYMETRIK, param_tf->symetrique);

    infos->num_choisi = param_tf->type;
  } else {
    param_tf = (Infos_FonctionApodisation *) malloc( sizeof(Infos_FonctionApodisation) );
    if (param_tf == NULL) {
      SYS_ERREUR(MANQUE_MEMOIRE, "f_inv.c", "TFI_init", "param_tf");
      return FAUX;
    }

    infos->graphe->pointeur.infos_tf = param_tf; /* On installe les paramètres de fenêtre
						    associé à un graphe -> sera libéré quand le graphe sera effacé... */
    infos->num_choisi = FEN_RECTANGLE;
    OBJ_choisit(PARAM_TFI, TF_FEN_SYMETRIK, FAUX);
    param_tf->symetrique = FAUX;
  }

  if (arbres[PARAM_TFI].drapeaux.prepare == FAUX)
    prepare_arbre_tfi_inverse();

  OBJ_active(PARAM_TFI, TFI_ESTIME_ERR, infos->graphe->covariance != NULL);
  OBJ_active(PARAM_TFI, TFI_MAT_COVAR, FORCE_NON);
  OBJ_choisit(PARAM_TFI, TFI_ESTIME_ERR, FORCE_NON);

  ARB_PreparerDefileur_T(PARAM_TFI, TFI_FENETRE_1, TFI_FENETRE_5, 0,
			 TFI_FEN_AVANT, TFI_FOND_ASC, TFI_BOUTON_ASC, TFI_FEN_APRES,
			 NOMBRE_FENETRES, Types_fenetres, NULL, infos->num_choisi, FEN_RECTANGLE);
  
  etat = Autorise_Symetrique( infos->num_choisi );
  OBJ_active(PARAM_TFI, TF_FEN_SYMETRIK, etat); /* Seules fenêtres pour lesquelles ça a un sens */
  OBJ_active(PARAM_TFI, TFI_K_1, etat);
  OBJ_active(PARAM_TFI, TFI_K_2, etat & (param_tf->symetrique));

  ARB_MettreComplement( PARAM_TFI, &infos );
  if ( ARB_Bloquant( PARAM_TFI, titre_arbre_tfi, principale, 0 ) == TFI_ANNULE) { /* On annule le calcul */
    ARB_MettreComplement( PARAM_TFI, NULL );
    return FAUX;
  }
  ARB_MettreComplement( PARAM_TFI, NULL );

  /* Paramètres de la fenêtre ayant servi lors du calcul de la TF directe */
  param_tf->x_min     = OBJ_LireNombre( PARAM_TFI, TFI_KMIN_FEN, NULL );
  param_tf->x_max     = OBJ_LireNombre( PARAM_TFI, TFI_KMAX_FEN, NULL );
  param_tf->parametre = OBJ_LireNombre( PARAM_TFI, TFI_PARAMETRE, NULL );
  param_tf->type      = infos->num_choisi;

  return VRAI; /* On a fini l'initialisation */
}

/* -------- Gestion de l'arbre de paramétrage de calcul de la TF inverse -------- */

static int GER_paramtfi(unsigned int num_objet, int x_souris __attribute__((__unused__)), int y_souris)
{
  TFI_infos *infos;
  boolean etat;

  infos = (TFI_infos *) ARB_LireComplement( PARAM_TFI );
  if ( (num_objet >= TFI_FENETRE_1) && (num_objet <= TFI_FENETRE_5) ) {
    unsigned long nv_fenetre   = (unsigned long) (infos->num_choisi);
    unsigned long prem_visible = (unsigned long) (infos->num_depart);

    ARB_GererDefileur_T(PARAM_TFI, TFI_FENETRE_1, TFI_FENETRE_5, 0,
			TFI_FEN_AVANT, TFI_FOND_ASC, TFI_BOUTON_ASC, TFI_FEN_APRES,
			NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible, num_objet, y_souris);

    if (infos->num_choisi == (int) nv_fenetre)
      return BI_EVENEMENT_TRAITE; /* Rien à faire */

    infos->num_choisi = (int) nv_fenetre;
    etat = (nv_fenetre == FEN_HANNING) || (nv_fenetre == FEN_HAMMING)
      || (nv_fenetre == FEN_BARTLETT); /* VRAI -> asymétrie possible */
    OBJ_active(PARAM_TFI, TF_FEN_SYMETRIK, etat);
    OBJ_active(PARAM_TFI, TFI_K_1, etat);
    
    etat &= ! (OBJ_choisit(PARAM_TFI, TF_FEN_SYMETRIK, QUESTION));
    OBJ_active(PARAM_TFI, TFI_K_2, etat);

    etat = (nv_fenetre == FEN_HAMMING) || (nv_fenetre == FEN_NORTON_F3)
      || (nv_fenetre == FEN_BESSEL)  || (nv_fenetre == FEN_GAUSSIENNE);
    OBJ_active(PARAM_TFI, TFI_PARAMETRE, etat); /* Seule fenêtre avec un paramètre... */

    return BI_EVENEMENT_TRAITE;
  }

  switch (num_objet)
  {
  case TFI_FEN_AVANT:
  case TFI_FEN_APRES:  /* Gestion de la liste de fenêtres */
  case TFI_FOND_ASC:
  case TFI_BOUTON_ASC:
    {
      unsigned long nv_fenetre   = (unsigned long) (infos->num_choisi);
      unsigned long prem_visible = (unsigned long) (infos->num_depart);
	
      ARB_GererDefileur_T(PARAM_TFI, TFI_FENETRE_1, TFI_FENETRE_5, 0,
			  TFI_FEN_AVANT, TFI_FOND_ASC, TFI_BOUTON_ASC, TFI_FEN_APRES,
			  NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible, num_objet, y_souris);
      infos->num_depart = (int) prem_visible;
    }
    return BI_EVENEMENT_TRAITE;

  case TF_FEN_SYMETRIK:
    etat = OBJ_choisit(PARAM_TFI, TF_FEN_SYMETRIK, ECHANGE);
    OBJ_active(PARAM_TFI, TFI_K_2, !etat); /* Inutile si symétrique... */
    return BI_EVENEMENT_TRAITE;

  case TFI_ESTIME_ERR:
    etat = OBJ_choisit(PARAM_TFI, TFI_ESTIME_ERR, ECHANGE);
    OBJ_active(PARAM_TFI, TFI_MAT_COVAR, etat);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Préparation de l'arbre indiquant comment calculer la TFI -------- */

static void prepare_arbre_tfi_inverse(void)
{
  OBJ_deplace(PARAM_TFI, TFI_AMPLITUDE, 0, 1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_PHASE    , 0, 2, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_COMPLEXE , 0, 3, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_MIT      , 0, 1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_GENERALE , 0, 1, DEPLACE_PIXELS);

  ARB_EnjoliverDefileur( PARAM_TFI, TFI_LST_FENETRE, TFI_FENETRE_1, TFI_FENETRE_5,
			 TFI_FEN_AVANT, TFI_FOND_ASC, TFI_BOUTON_ASC, TFI_FEN_APRES, ARB_ASCENSEUR_DROITE );

  OBJ_deplace (PARAM_TFI, TFI_LST_FENETRE, 0, -2, DEPLACE_PIXELS);
  OBJ_deplace (PARAM_TFI, TFI_FEN_AVANT, 0, -2, DEPLACE_PIXELS);
  OBJ_deplace (PARAM_TFI, TFI_FOND_ASC , 0, -1, DEPLACE_PIXELS);
  OBJ_deplace (PARAM_TFI, TFI_FEN_APRES, 0,  2, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_TFI, TFI_KMAX, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_PAS , 0, +2, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_TFI, TFI_PARAMETRE, 0, -2, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_KMIN_FEN , 0, -1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_K_2      , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_TFI, TFI_KMAX_FEN , 0, +2, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_TFI, TFI_MAT_COVAR, 0, +1, DEPLACE_PIXELS);

#ifndef _DEVELOPPE_
  OBJ_active(PARAM_TFI, TFI_MIT, FORCE_NON);
  OBJ_active(PARAM_TFI, TFI_COMPLEXE, FORCE_NON);
#endif

  OBJ_MettreNombre( PARAM_TFI, TFI_PAS, 0.02, 0.0 );

  arbres[PARAM_TFI].fonction.arbre = GER_paramtfi;
  arbres[PARAM_TFI].drapeaux.prepare = VRAI;
}

/* -------- Préparation de l'arbre dans la fenêtre de la TF inverse -------- */

static int TFI_PrepareArbreFenetre(unsigned int numero_arbre)
{
  if ( FAUX == arbres[ numero_arbre ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( numero_arbre, "menu_filtre.html#tfi", GER_tfi, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return -1;
    }
  }

#ifdef AVEC_TRADUCTION_
#endif

  arbres[ PARAM_FOUR_INVER ].drapeaux.prepare = VRAI;

  return 0;
}
