/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Création : 4 février 1999

   Filtrage du spectre : calcul par TF et TF inverse
   Calcul de la TF et de la TFI par la méthode directe (méthode des trapèzes)
   Permet en particulier le calcul des barres d'erreur, voire de la matrice de covariance.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* Pour memcpy() */

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"
#include "science.h"

#include "filtrage.h"
#include "matrices.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Complexe TF_avec_erreur(Graphe *graphe, double x, Complexe *erreur);
static Complexe TF_complexe   (Graphe *graphe, double x);
static void     TF_covar      (Graphe *source, Graphe *fourier);
static Complexe TF_reelle     (Graphe *graphe, double x);

static Complexe TFI_complexe  (Graphe *graphe, double x);
static void     TFI_covar (Graphe *source, Graphe *fourier);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Calcul dans le sens direct ~-~-~-~-~-~-~-~- */

/* -------- TF d'un graphe, méthode brutale --------------------------------------\
 |    Ne renvoie rien                                                             |
 |    Suppose que les X ont déjà été calculés et sont dans fourier-x...           |
 |                                                                                |
 | infos_tf : le graphe dont on veut la TF                                        |
 | fourier  : le graphe recevant la TF calculée (il doit être de type complexe)   |
 | pas      : le pas du calcul                                                    |
 \------------------------------------------------------------------------------- */

void TF_simple(Graphe *source, Graphe *fourier)
{
  register double   *x_tf;
  register Complexe *y_tf;

  if ( (NULL == source) || (NULL == fourier) ) return;

  x_tf = fourier->x;
  y_tf = (Complexe *) fourier->y;

  if ( VRAI == GRF_EstFonctionReelle( source ) ) { /* Graphe en entrée à valeurs réelles */
    if ( fourier->erreur_y != NULL ) {
      if ( fourier->covariance != NULL ) { /* Calcul de la matrice de covariance */
	TF_covar( source, fourier );
      } else { /* Juste les erreurs */
	register Complexe *err_y = (Complexe *) (fourier->erreur_y->y);
	unsigned long i;

	for (i = 0 ; i < fourier->nombre_point ; i++, x_tf++, y_tf++, err_y++) {
	  *y_tf = TF_avec_erreur(source, *x_tf, err_y);
	}

	GRF_bornes(fourier->erreur_y, NE_PAS_FAIRE, DESORDONNE);
	fourier->erreur_y->y_min = 0.0; /* Forcément positif, ce sont des écarts-types ! */
      }
    } else { /* On ne se préoccupe pas des erreurs */
      unsigned long i;

      for (i = 0 ; i < fourier->nombre_point ; i++, x_tf++, y_tf++) {
	*y_tf = TF_reelle(source, *x_tf);   /* On calcule en ce point */
      }
    }
  } else if ( GRF_FCT_COMPLEXE == source->genre_donnees ) {
    if ( fourier->erreur_y != NULL ) {
      if ( fourier->covariance != NULL ) { /* Calcul de la matrice de covariance */
	fprintf(stderr, "TF_simple : graphe source complexe, avec matrice de covariance\n");
	fprintf(stderr, "CAS À PROGRAMMER\n");
      } else { /* Juste les erreurs */
	register Complexe *err_y = (Complexe *) (fourier->erreur_y->y);
	unsigned long i;

	for (i = 0 ; i < fourier->nombre_point ; i++, x_tf++, y_tf++, err_y++) {
	  *y_tf = TF_avec_erreur(source, *x_tf, err_y);
	}

	GRF_bornes(fourier->erreur_y, NE_PAS_FAIRE, DESORDONNE);
	fourier->erreur_y->y_min = 0.0; /* Forcément positif, ce sont des écarts-types ! */
      }
    } else { /* On ne se préoccupe pas des erreurs */
      unsigned long i;

      for (i = 0 ; i < fourier->nombre_point ; i++, x_tf++, y_tf++) {
	*y_tf = TF_complexe(source, *x_tf); /* On calcule en ce point */
      }
    }    
  } else {
    fprintf(stderr, "TF_simple : Graphe à valeurs ni réelles, ni complexes\n");
  }

  GRF_bornes( fourier, NE_PAS_FAIRE, DESORDONNE );
}

/* -------- TF d'une fonction complexe (méthode "brutale") -----------------------\
 |   Renvoie la valeur calculée au point considéré                                |
 |                                                                                |
 | graphe : le graphe dont on veut calculer la TF - doit être de type complexe... |
 | x      : le point pour lequel on calcule la TF                                 |
 \------------------------------------------------------------------------------- */

static Complexe TF_complexe(Graphe *graphe, double x)
{
  Complexe calcul = {0.0 , 0.0}; /* Le résultat de l'intégration */
  register double *zone_x;
  register Complexe *zone_y;
  double dx_x_r,reelle,imag,delta_x;
  double inter;
  double cosinus, sinus;
  unsigned long i;

  x *= 2.0; /* Toujours par la suite... */

  zone_x = graphe->x;
  zone_y = (Complexe *) graphe->y; /* On se place au début du graphe */

  dx_x_r = x * *(zone_x);
  cosinus = cos( dx_x_r );
  sinus   = sin( dx_x_r );

  reelle =  cosinus * zone_y->re + sinus   * zone_y->im;
  imag   =  -sinus  * zone_y->re + cosinus * zone_y->im;

  for (i = 0 ; i < graphe->nombre_point - 1 ; i++) {
    delta_x = *(zone_x + 1) - *zone_x;
    zone_x++; zone_y++;

    dx_x_r = x * *zone_x;
    cosinus = cos( dx_x_r );
    sinus   = sin( dx_x_r );
			
    inter = cosinus * zone_y->re + sinus * zone_y->im;
    reelle    += inter;
    calcul.re += reelle * delta_x;
    reelle     = inter;

    inter = -sinus * zone_y->re + cosinus * zone_y->im;
    imag      += inter;
    calcul.im += imag * delta_x;
    imag       = inter;
  }

  calcul.re *= 0.3989422804; /* racine (2/pi) / 2 [ /2 == méthode des trapèzes pour l'intégration... ] */
  calcul.im *= 0.3989422804;

  return calcul;
}

/* -------- TF d'une fonction réelle pure (méthode "brutale") --------------------\
 |   Renvoie la valeur calculée au point considéré                                |
 |   Calcule l'intégrale en utilisant la méthode des rectangles                   |
 |                                                                                |
 | graphe : le graphe dont on veut calculer la TF                                 |
 | x      : le point pour lequel on calcule la TF                                 |
 \------------------------------------------------------------------------------- */

static Complexe TF_reelle(Graphe *graphe, double x)
{
  Complexe calcul = {0.0 , 0.0}; /* Le résultat de l'intégration */
  const double coefficient = 1.0 / sqrt(2.0 * M_PI);
  double dx_x_r,reelle,imag,delta_x;
  double inter_reel, inter_imag, y;
  register double *zone_x;
  register double *zone_y;
  unsigned long i;

  x *= 2.0; /* Toujours par la suite... */

  zone_x = graphe->x;
  zone_y = graphe->y;

  y = *zone_y;

  dx_x_r = x * *(zone_x);

  reelle =  cos( dx_x_r ) * y;
  imag   =  -sin( dx_x_r ) * y;

  for (i = 0 ; i < graphe->nombre_point - 1 ; i++) {
    delta_x = /* *(zone_x + 1) */ - *zone_x;
    zone_x++;
    delta_x += *zone_x; /* Ici, delta_x = x(i+1) - x(i) */
    dx_x_r = x * *zone_x;

    zone_y++;
    y = *zone_y;

    inter_reel = y * cos( dx_x_r );
    reelle    += inter_reel;
    calcul.re += reelle * delta_x;
    reelle     = inter_reel;

    inter_imag = - y * sin( dx_x_r );
    imag      += inter_imag;
    calcul.im += imag * delta_x;
    imag       = inter_imag;
  }

  calcul.re *= coefficient;
  calcul.im *= coefficient;

  return calcul;
}

/* -------- TF d'une fonction avec calcul des barres d'erreur sur la TF ----------\
 |   Renvoie un nombre complexe contenant la valeur calculée au point considéré   |
 |        et un autre, passé en paramètre, contient l'erreur sur Re() et Im()     |
 |   Repose sur la linéarité de la méthode des rectangles, qui se propage donc    |
 |        aux variances (si on suppose l'indépendance des X)                      |
 |        suppose que converge bien vers la variance de l'intégrale...            |
 |                                                                                |
 | Note : on ne peut pas utiliser l'algorithme utilisé pour le calcul sans erreur |
 |   parce que les (xi+1-xi)^2 ne sont pas simplement réarrangeables...           |
 |                                                                                |
 | Formules utilisées :
 |
 |          1 (                       N-1                                                       )
 | Re(Tf) = - (f(x )cos(2x r)(x -x ) + S  f(x )cos(2x r)(x   - x   ) + f(x )cos(2x r)(x - x   ) )
 |          2 (   0       0    1  0   i=1    i       i    i+1   i-1       N       N    N   N-1  )
 |
 |          1 (                       N-1                                                       )
 | Im(Tf) = - (f(x )sin(2x r)(x -x ) + S  f(x )sin(2x r)(x   - x   ) + f(x )sin(2x r)(x - x   ) )
 |          2 (   0       0    1  0   i=1    i       i    i+1   i-1       N       N    N   N-1  )
 |
 |     2    1 ( 2    2             2 N-1  2    2                  2    2    2                2 )
 | Re(s ) = - (s  cos (2x r)(x -x ) + S  s  cos (2x r)(x   - x   )  + s  cos (2x r)(x - x   )  )
 |          4 ( 0        0    1  0   i=1  i        i    i+1   i-1      N        N    N   N-1   )
 |
 |     2    1 ( 2    2             2 N-1  2    2                  2    2    2                2 )
 | Im(s ) = - (s  sin (2x r)(x -x ) + S  s  sin (2x r)(x   - x   )  + s  sin (2x r)(x - x   )  )
 |          4 ( 0        0    1  0   i=1  i        i    i+1   i-1      N        N    N   N-1   )
 |
 |                                                                                |
 | graphe : le graphe dont on veut calculer la TF, supposé de type réel           |
 |                         ou complexe, mais seule la partie réelle est non-nulle |
 | x      : le point pour lequel on calcule la TF                                 |
 | erreur : contiendra les barres d'erreur sur la partie réelle et imaginaire     |
 \------------------------------------------------------------------------------- */

static Complexe TF_avec_erreur(Graphe *graphe, double x, Complexe *erreur)
{
  Complexe calcul = {0.0 , 0.0}; /* Le résultat de l'intégration */
  const double coefficient = 1.0 / sqrt(2.0 * M_PI); /* Coeff. de la TF
							Attention, il contient déjà le terme 1/2 
							dû à la méthode des trapèzes */
  double dx_x_r, delta_x;
  double erreur_y; /* Pour l'erreur */
  double cosinus, sinus;
  register double *k, *err_y;
  unsigned long i;

  x *= 2.0; /* Toujours par la suite... */

  erreur->re = 0.0; /* On initialise l'erreur */
  erreur->im = 0.0;

  /* Premier terme de la somme */
  k = graphe->x;
  dx_x_r = x * *k;
  cosinus = cos( dx_x_r ); sinus = -sin( dx_x_r );

  err_y = graphe->erreur_y->y;
  erreur_y = *err_y * *err_y; /* La relation utilise la variance,
				 mais l'erreur est enregistrée sous forme d'écart-type */

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) { /* Calcul sur un graphe complexe, supposé réel pur */
    register Complexe *zone_y = (Complexe *) (graphe->y);

    /* Premiers termes des sommes */
    delta_x = *(k+1) - *k;
    calcul.re = zone_y->re * cosinus * delta_x;
    calcul.im = zone_y->re * sinus * delta_x;
    delta_x *= delta_x;

    erreur->re = erreur_y * cosinus * cosinus * delta_x;
    erreur->im = erreur_y * sinus * sinus * delta_x;

    /* La somme proprement dite */
    k++; zone_y++; err_y++;

    for (i = 1 ; i < graphe->nombre_point - 1 ; i++, k++, zone_y++, err_y++) {
      dx_x_r = x * *k;
      delta_x = *(k+1) - *(k-1);

      cosinus = cos(dx_x_r) * delta_x; sinus = -sin(dx_x_r) * delta_x;

      calcul.re += zone_y->re * cosinus;
      calcul.im += zone_y->re * sinus;

      erreur_y = *err_y * *err_y; /* On stocke l'écart-type, mais travail sur les variances ! */
      erreur->re += erreur_y * cosinus * cosinus;
      erreur->im += erreur_y * sinus * sinus;
    }

    /* Le dernier terme de la somme - sauf erreur, tous les pointeurs pointent sur la dernière valeur */
    dx_x_r = x * *k;
    delta_x = *k - *(k - 1);

    cosinus = cos(dx_x_r) * delta_x; sinus = -sin(dx_x_r) * delta_x;

    calcul.re += zone_y->re * cosinus;
    calcul.im += zone_y->re * sinus;

    erreur_y = *err_y * *err_y; /* On stocke l'écart-type, mais travail sur les variances ! */
    erreur->re += erreur_y * cosinus * cosinus;
    erreur->im += erreur_y * sinus * sinus;
  } else { /* Travail sur un graphe à valeurs réelles */
    register double *zone_y = graphe->y;

    /* Premiers termes des sommes */

    delta_x = *(k+1) - *k;
    calcul.re = *zone_y * cosinus * delta_x;
    calcul.im = *zone_y * sinus * delta_x;
    delta_x *= delta_x;

    erreur->re = erreur_y * cosinus * cosinus * delta_x;
    erreur->im = erreur_y * sinus * sinus * delta_x;

    /* La somme proprement dite */
    k++; zone_y++; err_y++;

    for (i = 1 ; i < graphe->nombre_point - 1 ; i++, k++, zone_y++, err_y++) {
      dx_x_r = x * *k;
      delta_x = *(k+1) - *(k-1);

      cosinus = cos(dx_x_r) * delta_x; sinus = -sin(dx_x_r) * delta_x;

      calcul.re += *zone_y * cosinus;
      calcul.im += *zone_y * sinus;

      erreur_y = *err_y * *err_y; /* On stocke l'écart-type, mais travail sur les variances ! */
      erreur->re += erreur_y * cosinus * cosinus;
      erreur->im += erreur_y * sinus * sinus;
    }

    /* Le dernier terme de la somme - sauf erreur, tous les pointeurs pointent sur la dernière valeur */
    dx_x_r = x * *k;
    delta_x = *k - *(k - 1);

    cosinus = cos(dx_x_r) * delta_x; sinus = -sin(dx_x_r) * delta_x;

    calcul.re += *zone_y * cosinus;
    calcul.im += *zone_y * sinus;

    erreur_y = *err_y * *err_y; /* On stocke l'écart-type, mais travail sur les variances ! */
    erreur->re += erreur_y * cosinus * cosinus;
    erreur->im += erreur_y * sinus * sinus;
  }

  erreur->re = sqrt(erreur->re); /* On repasse en écart-type */
  erreur->im = sqrt(erreur->im);

  calcul.re *= coefficient; erreur->re *= coefficient;
  calcul.im *= coefficient; erreur->im *= coefficient;

  return calcul;
}

/* -------- Calcul de la TF avec calcul simultané de la matrice de covariance --------\
 |   On fait une routine à part entière parce que la technique devient délicate...    |
 |                                                                                    |
 | La matrice de covariance est symétrique -> on ne mémorise que la moitié inférieure |
 | Les valeurs sont complexes, donc il faut Re() et Im() : colonnes / lignes mélangées|
 | Principe : la méthode des trapèzes est linéaire -> il y a une matrice A associée   |
 |    le calcul de la TF revient donc à faire la transformation TF = A X              |
 |    où X est le vecteur des v.a. originales et T.F. le vecteur d'arrivée dans C^n   |
 |    On suppose de plus les v.a. de X indépendantes : SIGMA(X) est diagonale.        |
 |    On a finalement SIGMA(TF) = A SIGMA(X) tA                                       |
 |                                                                                    |
 | Si X est un vecteur gaussien, TF est aussi un vecteur gaussien                     |
 |                                                                                    |
 | Algorithme ultra-basique pour l'instant - certainement optimisable !               |
 \----------------------------------------------------------------------------------- */

static void TF_covar(Graphe *source, Graphe *fourier)
{
  size_t nbr_elements __attribute__((__unused__)), i, j;
  double *covariance;
  Complexe erreur;
  register double *x_tf;
  register Complexe *y_tf;
  double *x = source->x, *z = fourier->x, *sigma = source->erreur_y->y;
  const double coefficient = 1.0 / (2.0 * M_PI); /* Coeff. de la TF, au carré !
						    Attention, il contient déjà le 1/4 venant de
						    l'utilisation de la méthode des trapèzes */
  Graphe *erreur_y;

  if ( NULL != source->covariance ) {
    fprintf(stderr, "TF_covar : matrice de covariance pour le graphe original, NON TRAITÉ\n");
  }

  covariance = fourier->covariance->valeurs;

  /* TF a n points complexes -> matrice de covariance est de taille (2n) * (2n)
     On garde les éléments de la moitié inférieure (avec la diagonale) -> (2n)(2n + 1) / 2
     soit un total de n(2 * n + 1) éléments */
  nbr_elements = fourier->nombre_point * (2 * fourier->nombre_point + 1);

  /* Calcul des termes diagonaux : la méthode de la routine d'avant convient parfaitement...
     Calcule en même temps la TF elle-même, d'ailleurs */
  for (i = 0, y_tf = (Complexe *) (fourier->y), x_tf = fourier->x ;
       i < fourier->nombre_point ; i++, y_tf++, x_tf++) {
    *y_tf = TF_avec_erreur(source, *x_tf, &erreur);
    covariance[ TF_element(2 * i    , 2 * i    ) ] = erreur.re * erreur.re;
    covariance[ TF_element(2 * i + 1, 2 * i + 1) ] = erreur.im * erreur.im;
  }

  /* Calcul des termes hors-diagonale, pour l'instant sans recherche d'optimisation
     i = indice de ligne, j = indice de colonne, travail sur la moitié inférieure */

  /* Premier cas : i impair (lignes des parties imaginaires) */
  for (i = 1 ; i < 2 * fourier->nombre_point ; i += 2) {
    double R = 2.0 * z[ (i - 1) >> 1 ];
    /* On calcule simultanément pour j pair et impair : seul le a_jk change !
       Remarque : ainsi, recalcule les termes diagonaux impairs... Voir à améliorer !
    */
    /*    printf("Calcul pour R = %f\n", R/2.0); */

    for (j = 0 ; j < i ; j += 2) {
      double a_ik, a_jk, s_kk;
      unsigned long k;
      double R_j = 2 * z[j >> 1];
      double dx;
      unsigned long n_1, n_2;

      n_1 = TF_element(i, j);
      n_2 = n_1 + 1;

      /* Premier terme de la somme : a_i0 * s_00 * a_j0 */
      dx = x[1] - x[0];
      a_ik = dx * -sin(x[0] * R); /* Terme a_i0 */
      s_kk = sigma[0] * sigma[0];
      a_jk = dx *  cos(x[0] * R_j); /* Terme a_j0 */
      covariance[ n_1 ] = a_ik * s_kk * a_jk;

      a_jk = dx * -sin(x[0] * R_j); /* Terme a_j0 */
      covariance[ n_2 ] = a_ik * s_kk * a_jk;

      /* Termes intermédiaires */
      for (k = 1 ; k < source->nombre_point - 1 ; k++)
      {
	dx = x[k + 1] - x[k - 1];
	a_ik = dx * -sin(x[k] * R);
	s_kk = sigma[k] * sigma[k];
	a_jk = dx *  cos(x[k] * R_j);
	covariance[ n_1 ] += a_ik * s_kk * a_jk;

	a_jk = dx * -sin(x[k] * R_j);
	covariance[ n_2 ] += a_ik * s_kk * a_jk;
      }

      /* Dernier terme */
      /*      k = fourier->nombre_point - 1;*/
      a_ik = (x[k] - x[k - 1]) * -sin(x[k] * R);
      s_kk = sigma[k] * sigma[k];
      a_jk = (x[k] - x[k - 1]) *  cos(x[k] * R_j);
      covariance[ n_1 ] += a_ik * s_kk * a_jk;
      covariance[ n_1 ] *= coefficient;

      a_jk = (x[k] - x[k - 1]) * -sin(x[k] * R_j);
      covariance[ n_2 ] += a_ik * s_kk * a_jk;
      covariance[ n_2 ] *= coefficient;
      /*      printf("Éléments (%ld, %ld) [%d] et (%ld, %ld) [%d] faits\n", i, j, TF_element(i, j),
	      i, j+1, TF_element(i, j + 1)); */
    }
  }

  /* Deuxième cas : i pair (lignes des parties réelles) */
  for (i = 2 ; i < 2 * fourier->nombre_point ; i += 2) {
    double a_ik, a_jk, s_kk;
    double d_x;
    double R = 2 * z[i >> 1]; /* Celui étudié = de la ligne */
    unsigned long k;

    /* Calcul simultané pair / impair
       _avant_ la diagonale sinon on écrase le premier élément de la ligne suivante */
    for (j = 0 ; j < i ; j += 2) {
      double R_j = 2 * z[j >> 1]; /* Celui de la colonne */
      unsigned long n_1, n_2;

      n_1 = TF_element(i, j);
      n_2 = n_1 + 1;

      /* Premier terme de la somme : a_i0 * s_00 * a_j0 */
      d_x = x[1] - x[0];
      a_ik = d_x *  cos(x[0] * R); /* Terme a_i0 */
      s_kk = sigma[0] * sigma[0];
      a_jk = d_x *  cos(x[0] * R_j); /* Terme a_j0 */
      covariance[ n_1 ] = a_ik * s_kk * a_jk;

      a_jk = d_x * -sin(x[0] * R_j); /* Terme a_j0 */
      covariance[ n_2 ] = a_ik * s_kk * a_jk;

      /* Termes intermédiaires */
      for (k = 1 ; k < source->nombre_point - 1 ; k++) {
	d_x = x[k + 1] - x[k - 1];
	a_ik = d_x *  cos(x[k] * R);
	s_kk = sigma[k] * sigma[k];
	a_jk = d_x *  cos(x[k] * R_j);
	covariance[ n_1 ] += a_ik * s_kk * a_jk;

	a_jk = d_x * -sin(x[k] * R_j);
	covariance[ n_2 ] += a_ik * s_kk * a_jk;
      }

      /* Dernier terme */
      /*      k = fourier->nombre_point - 1;*/
      d_x = x[k] - x[k - 1];
      a_ik = d_x * cos(x[k] * R);
      s_kk = sigma[k] * sigma[k];
      a_jk = d_x * cos(x[k] * R_j);
      covariance[ n_1 ] += a_ik * s_kk * a_jk;
      covariance[ n_1 ] *= coefficient;

      a_jk = d_x * -sin(x[k] * R_j);
      covariance[ n_2 ] += a_ik * s_kk * a_jk;
      covariance[ n_2 ] *= coefficient;

      /*      covariance[ TF_element(i, i) ] += a_ik * a_ik * s_kk; */
      /*      printf("Éléments (%ld, %ld) [%d] et (%ld, %ld) [%d] faits\n", i, j, TF_element(i, j),
	      i, j+1, TF_element(i, j + 1)); */
    }
  }

  /* Calcul normalement terminé -> on réalise un graphe de barres d'erreur à partir de ça */
  erreur_y = fourier->erreur_y;
  fourier->drapeaux.erreur_y = VRAI;
  erreur_y->x_min = fourier->x_min;
  erreur_y->x_max = fourier->x_max;

  for (i = 0, y_tf = (Complexe *) (erreur_y->y) ; i < fourier->nombre_point ; i++, y_tf++) {
    y_tf->re = sqrt(covariance[ TF_element(2 * i    , 2 * i    ) ]);
    y_tf->im = sqrt(covariance[ TF_element(2 * i + 1, 2 * i + 1) ]);
  }
  GRF_bornes(erreur_y, NE_PAS_FAIRE, DESORDONNE);
  erreur_y->y_min = 0.0;

#ifdef _DEVELOPPE_
  verifie_covariance(covariance, fourier->nombre_point);
#endif
}

/* ~-~-~-~-~-~-~-~- Calcul de la réciproque ~-~-~-~-~-~-~-~- */

/* -------- Interface avec l'extérieur -------- */

int TFI_Trapezes(Graphe *source, Graphe *resultat)
{
  int retour = FAUX; /* Code de retour */

  if ( (NULL == source) || (NULL == resultat) ) return ERR_GRF_NULL;

  if ( GRF_FCT_REELLE == source->genre_donnees ) { /* Graphe en entrée à valeurs réelles */
    fprintf(stderr, "TFI_Trapezes : Graphe à valeurs réelles : À FAIRE\n");
  } else if ( GRF_FCT_COMPLEXE == source->genre_donnees ) {
    if ( resultat->erreur_y != NULL ) {
      if ( resultat->covariance != NULL ) {
	TFI_covar( source, resultat );
      } else { /* On ne veut pas la matrice de covariance */
	fprintf(stderr, "TFI_Trapezes : Graphe à valeurs complexes, avec erreurs, sans covariance : À FAIRE\n");
      }
    } else { /* On néglige les barres d'erreur */
      unsigned long i;
      register double *x_tfi;
      register Complexe *y_tfi;

      for (i = 0, x_tfi = resultat->x, y_tfi = (Complexe *) resultat->y ;
	   i < resultat->nombre_point; i++, x_tfi++, y_tfi++) {
	*y_tfi = TFI_complexe( source, *x_tfi );
      }
    }
  } else {
    fprintf(stderr, "TFI_Trapezes : Graphe à valeurs ni réelles, ni complexes\n");
  }

  return retour;
}

/* -------- TFI d'une fonction complexe (méthode "brutale") ----------------------\
 |   Renvoie la valeur calculée au point considéré                                |
 |                                                                                |
 | graphe : le graphe dont on veut calculer la TF - doit être de type complexe... |
 | x      : le point pour lequel on calcule la TF                                 |
 \------------------------------------------------------------------------------- */

static Complexe TFI_complexe(Graphe *graphe, double x)
{
  Complexe calcul = {0.0 , 0.0}; /* Le résultat de l'intégration */
  register Complexe *zone_y;
  double dx_x_r, reelle, imag, delta_x;
  double inter;
  double cosinus, sinus;
  register double *zone_x;
  unsigned long i;

  x *= 2.0; /* Toujours par la suite... */

  zone_x = graphe->x;
  zone_y = (Complexe *) graphe->y; /* On se place au début du graphe */

  dx_x_r = x * *(zone_x);
  cosinus = cos( dx_x_r );
  sinus   = sin( dx_x_r );

  reelle =  cosinus * zone_y->re -   sinus * zone_y->im;
  imag   =    sinus * zone_y->re + cosinus * zone_y->im;

  for (i = 0 ; i < graphe->nombre_point - 1 ; i++) {
    delta_x = *(zone_x + 1) - *zone_x;
    zone_x++;
    zone_y++;
    dx_x_r = x * *zone_x;
    cosinus = cos( dx_x_r );
    sinus   = sin( dx_x_r );
			
    inter = cosinus * zone_y->re -   sinus * zone_y->im;
    reelle    += inter;
    calcul.re += reelle * delta_x;
    reelle     = inter;

    inter =   sinus * zone_y->re + cosinus * zone_y->im;
    imag      += inter;
    calcul.im += imag * delta_x;
    imag       = inter;
  }

  calcul.re *= 0.3989422804; /* racine (2/pi) / 2 [ /2 == méthode des trapèzes pour l'intégration... ] */
  calcul.im *= 0.3989422804;

  return calcul;
}

/* -------- Calcul de la TFI, méthode avec calcul de la matrice de covariance et tout et tout --------\
 |   Principe : on a une TF avec sa matrice de covariance S, la TF est un vecteur gaussien            |
 |   On intègre par les trapèzes : transformation linéaire, de matrice A                              |
 |   On calcule la matrice de covariance de la TFI, S' = A S tA                                       |
 |                                                                                                    |
 | Matrices S et S' forcément symétriques : on en tient compte                                        |
 | indices impairs = parties imaginaires, indices pairs = parties réelles                             |
 | Note : assimile le tableau de Complexe (re, im) à un tableau de doubles par commodité              |
 |     -> pas forcément extrêmement propre : à retravailler si besoin est !                           |
 \--------------------------------------------------------------------------------------------------- */

static void TFI_covar(Graphe *source, Graphe *tf_inv)
{
  Complexe *y_tf, *y_err;
  double *covar_tfi, *covar_tf, *x, *z;
  unsigned long i, j, k, l;
  unsigned long n_tf, n_tfi, nbr_elements_tfi __attribute__((__unused__));
  double *inter;
  const double coefficient = sqrt(2.0 / M_PI) / 2.0; /* Coeff. de la TFI - le 2 : méthode des trapèzes */

  if ( (source == NULL) || (tf_inv == NULL) ) return;
  if ( tf_inv->covariance == NULL ) return;

  covar_tf = source->covariance->valeurs;
  if (covar_tf == NULL) {
#ifdef _DEVELOPPE_
    fprintf(stderr, "Bizarre : appel à TFI_covar avec un source sans matrice de covariance\n"); 
#endif
    return;
  }

  /* On effectue le calcul de la matrice de covariance */
  n_tf = source->nombre_point;

  n_tfi = tf_inv->nombre_point;

  nbr_elements_tfi = n_tfi * (2 * n_tfi + 1);
  covar_tfi = tf_inv->covariance->valeurs;

  z = source->x;
  x = tf_inv->x;

#ifdef _DEVELOPPE_
  printf("Calcul de la TFI : calcul de la matrice de covariance finale\n");
#endif

  /* Formule :

            2 * n_tf [      (l - 1                  2 * n_tf            ) ]
     s'   =    S     [ a    (  S    a    * s     +     S     a    * s   ) ]
      i,j    l = 0   [  j,l (k = 0   i,k    l,k      k = l    i,k    k,l) ]

  */

#define Matrice(i, j) ( 2 * n_tf * (i) + (j) ) /* Indiçage horizontal : 1 2 3 // 4 5 6 ... */

  /* inter stocke de façon temporaire la matrice A de calcul de la TF
     A = matrice rectangulaire (2 * n_tfi, 2 * n_tf) - pair : p. réelle, impair : p. imaginaire */

  inter = (double *) malloc( sizeof(double) * (size_t) 4 * n_tf * n_tfi );
  if (inter == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Filtrage/brutal.c", "TFI_covar", "inter");
    free(covar_tfi);
    return;
  }

  /* Remplissage : les cosinus (pp & ii) et sinus (pi & ip) */
#ifdef _DEVELOPPE_
  printf("Calcul de la matrice A\n");
#endif

  for (i = 0 ; i < 2 * n_tfi ; i += 2) { /* Indices de lignes */
    double cosinus, sinus, d_x, xf2;
    unsigned long i_2 = i >> 1;
    unsigned long j_2;

    /* Premières colonnes : (R[1] - R[0]) * cos/sin(2 * R[0] * k[i/2]) */
    d_x = (z[1] - z[0]) * coefficient;
    xf2 = 2 * x[i_2];
    cosinus = cos(xf2 * z[0]);
    sinus   = sin(xf2 * z[0]);

    inter[ Matrice(i    , 0) ] = d_x * cosinus;
    inter[ Matrice(i    , 1) ] = -d_x * sinus;
    inter[ Matrice(i + 1, 0) ] = d_x * sinus;
    inter[ Matrice(i + 1, 1) ] = d_x * cosinus;

    /* Colonnes suivantes */
    for (j = 2 ; j < 2 * (n_tf - 1) ; j += 2) { /* Indices de colonnes */
      j_2 = j >> 1;
      d_x = (z[j_2 + 1] - z[j_2 - 1]) * coefficient;
      cosinus = cos(xf2 * z[j_2]);
      sinus   = sin(xf2 * z[j_2]);

      inter[ Matrice(i, j)     ] = d_x * cosinus;
      inter[ Matrice(i, j + 1) ] = -d_x * sinus;
      inter[ Matrice(i + 1, j) ] = d_x * sinus;
      inter[ Matrice(i+1, j+1) ] = d_x * cosinus;
    }

    /* Dernières colonnes */
    j_2 = j >> 1;

    d_x = (z[j_2] - z[j_2 - 1]) * coefficient;
    cosinus = cos(xf2 * z[j_2]);
    sinus   = sin(xf2 * z[j_2]);

    inter[ Matrice(i, j)     ] = d_x * cosinus;
    inter[ Matrice(i, j + 1) ] = -d_x * sinus;
    inter[ Matrice(i + 1, j) ] = d_x * sinus;
    inter[ Matrice(i+1, j+1) ] = d_x * cosinus;

#ifdef _DEVELOPPE_
    printf("Ligne %ld terminée\r", i);
#endif
  }

  /* La matrice A est prête, maintenant on calcule S' = A S tA (sachant que S est symétrique) */
#ifdef _DEVELOPPE_
  printf("\nCalcul de la matrice S' = A S tA\n");
#endif

  for (i = 0 ; i < 2 * n_tfi ; i++) { /* Indice de ligne */
    for (j = 0 ; j <= i ; j++) { /* Indice de colonnes */
      unsigned long n_ij = TF_element(i, j);
      double s = 0.0;

      /* Premier terme, l = 0 (pour éviter boucle infinie avec k < l - 1, k et l étant non signés */
      for (k = 0 ; k < 2 * n_tf ; k++) {
	s += inter[ Matrice(j, k) ] * covar_tf[ TF_element(k, 0) ]; 
      }
      covar_tfi[n_ij] = s * inter[ Matrice(i, 0) ];

      for (l = 1 ; l < 2 * n_tf ; l++) {
	s = 0.0;
	for (k = 0 ; k < l ; k++) {
	  s += inter[ Matrice(j, k) ] * covar_tf[ TF_element(l, k) ];
	}
	for (k = l ; k < 2 * n_tf ; k++) {
	  s += inter[ Matrice(j, k) ] * covar_tf[ TF_element(k, l) ];
	}
	covar_tfi[n_ij] += s * inter[ Matrice(i, l) ];
      }
    }

#ifdef _DEVELOPPE_
    printf("Ligne %ld terminée\r", i);
#endif
  }

  /* Maintenant, on calcule la TFI proprement dite */

#ifdef _DEVELOPPE_
  printf("\nCalcul de la TF inverse proprement dite\n");
#endif

  for (i = 0, y_tf = (Complexe *) tf_inv->y ; i < 2 * n_tfi ; i += 2, y_tf++) {
    Complexe *y_lu = (Complexe *) source->y;

    for (j = 0 ; j < 2 * n_tf ; j += 2, y_lu++) {
      y_tf->re += inter[ Matrice(i    , j) ] * y_lu->re + inter[ Matrice(i    , j + 1) ] * y_lu->im;
      y_tf->im += inter[ Matrice(i + 1, j) ] * y_lu->re + inter[ Matrice(i + 1, j + 1) ] * y_lu->im;
    }
  }

#ifdef _DEVELOPPE_
  verifie_covariance(covar_tfi, n_tfi);
  printf("Calcul de la TFI : calcul des barres d'erreur et de la TFI proprement dite\n");
#endif

  /* Calcul final : de la TFI et des barres d'erreur */
  tf_inv->drapeaux.erreur_y = VRAI;

  for (i = 0, y_err = (Complexe *) (tf_inv->erreur_y->y) ;
       i < tf_inv->nombre_point ; i++, y_err++) {

    y_err->re = sqrt(covar_tfi[ TF_element(2 * i    , 2 * i    ) ]);
    y_err->im = sqrt(covar_tfi[ TF_element(2 * i + 1, 2 * i + 1) ]);
  }

  GRF_bornes(tf_inv->erreur_y, NE_PAS_FAIRE, DESORDONNE);
  tf_inv->erreur_y->y_min = 0.0;
}

#ifdef _DEVELOPPE_
/* -------- Essais de vérification de la cohérence des matrices de corrélation -------- */

void verifie_covariance(double *covariance, unsigned long nombre_point)
{
  unsigned long total = 2 * nombre_point;
  unsigned long i, j;

  /* Première vérification : on doit avoir | Cov(i, j)/sqrt(V(i), V(j)) | < 1 */

  printf("Vérification de la matrice de covariance\n");

  for (i = 0 ; i < total ; i++) {
    double V_i = covariance[ TF_element(i, i) ];

    if (V_i < 0.0) {
      fprintf(stderr, "Problème : élément diagonal %ld négatif !\n", i);
      continue;
    }
    V_i = sqrt(V_i);

    for (j = 0 ; j < i ; j++) {
      double Cov_ij; 
      double V_j = covariance[ TF_element(j, j) ];
      if (V_j < 0.0) {
	fprintf(stderr, "Problème : élément diagonal %ld négatif !\n", j);
	continue;
      }
      V_j = sqrt(V_j);

      Cov_ij = covariance[ TF_element(i, j) ];
      if (Cov_ij == 0.0) {
	printf("Remarque : éléments %ld (point %ld) et %ld (point %ld) non corrélés\n",
	       i, i >> 1, j, j >> 1);
      }

      if ( (V_i > 0.0) && (V_j > 0.0) ) {
	Cov_ij /= V_i * V_j;
	if ( fabs(Cov_ij) > 1.0 )
	  fprintf(stderr, "Problème : coefficient de corrélation supérieur à 1 - Cov(%ld, %ld) = %f\n",
		  i, j, Cov_ij);
#ifdef verification_totale
	if ( fabs(Cov_ij) > 0.1 ) /* NB : essais montrent que pour des points indépendants, on trouve
				      des coefficients de l'ordre de 10^-2... */
	  printf("Points %ld et %ld corrélés à plus de 0,1 - rho = %f\n", i, j, Cov_ij);
#endif
	if ( fabs(Cov_ij) > 0.9 ) /* Très forte corrélation */
	  printf("Points %ld et %ld corrélés à plus de 0,9 - rho = %f\n", i, j, Cov_ij);
      } else
	printf("Remarque : un des points a une variance nulle !\n");
    }

  }

  printf("Vérification terminée\n");
}
#endif
