/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier : 15 septembre 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Calcul d'une transforme de Fourier (directe ou inverse) rapide
   Source original rcupr de ../FOURIER.c
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "filtrage.h"
#include "science.h"

/* ******** Variables locales ************************ */

#define SENS_DIRECT  1 /* TF directe */
#define SENS_INVERSE 2 /* TF inverse */

#define ancienne_methode
/* #define num_recipes */

/* ******** Fonctions locales ************************ */

static void      FFT_rapide(Complexe *source, Complexe *fourier, Complexe *vecteur, unsigned long nbr_pts, boolean sens);
static Complexe *FFT_omega (unsigned long nbr_points, int sens);

/* ******** Ralisation pratique des fonctions ******* */

/* ----------------- TF d'un graphe, mthode de la TF rapide ------------------------------\
 |    Ne renvoie rien                                                                      |
 |    Suppose que le nombre de points est une puissance de 2 et que Dx = constante         |
 |    Les x sont dj dans fourier->x                                                      |
 |                                                                                         |
 | source   : le graphe dont on veut la TF inverse                                         |
 | fourier  : le graphe recevant la TF inverse calcule                                    |
 \---------------------------------------------------------------------------------------- */

void TF_rapide(Graphe *source, Graphe *fourier)
{
  Complexe *donnees, *resultat, *vecteur = NULL;
  int puissance_2 = 0;
	
  /* On cherche le premier bit non-nul */
  while ( ( ( source->nombre_point >> puissance_2 ) & 1 ) == 0)
    puissance_2++;

  if ( (1U << puissance_2) != source->nombre_point ) {
    (void) fprintf(stderr, "TF_rapide : le nombre de points n'est pas une puissance de 2 [%lu]\n", source->nombre_point);
    (void) EXF_ERREUR(PUISSANCE_2);
    return;
  }

  donnees  = (Complexe *) source->y;
  resultat = (Complexe *) fourier->y;

  if ( puissance_2 == 0 ) { /* Un seul point dans le graphe dont on veut la TF */
    resultat->re = donnees->re; /* Rien  faire... */
    resultat->im = donnees->im;
    return;
  }

  if ( puissance_2 == 1 ) { /* Deux points dans le graphe dont on veut la TF */
    const double coefficient = sqrt(2.0 / M_PI); /* Coefficient de la TF */
    resultat[0].re = (donnees[0].re + donnees[1].im) * coefficient;
    resultat[0].im = (donnees[0].im + donnees[1].im) * coefficient; 
    resultat[1].re = (donnees[0].re - donnees[1].re) * coefficient;
    resultat[1].im = (donnees[0].im - donnees[1].im) * coefficient;
  }

#ifdef ancienne_methode
  vecteur = FFT_omega( source->nombre_point >> 1, SENS_DIRECT );
  if (vecteur == NULL) return;
#endif

  GRF_bornes(source, ORDONNE, DESORDONNE);

  FFT_rapide(donnees, resultat, vecteur, source->nombre_point, SENS_DIRECT);

  fourier->nombre_point >>= 1;      /* On divise par 2 : seule la premire moiti nous intresse */
  GRF_bornes(fourier, ORDONNE, DESORDONNE);  /* Les nouvelles bornes qui vont avec */
  
  free(vecteur); /* On a fini avec lui... */
}

/* ----------------- TF inverse d'un graphe, mthode de la TF rapide ----------------------\
 |    Ne renvoie rien                                                                      |
 |    Suppose que le nombre de points est une puissance de 2 et que Dx = constante         |
 |    Les x sont dj dans fourier->x                                                      |
 |                                                                                         |
 | source   : le graphe dont on veut la TF inverse                                         |
 | fourier  : le graphe recevant la TF inverse calcule                                    |
 \---------------------------------------------------------------------------------------- */

void TFI_rapide(Graphe *source, Graphe *fourier)
{
  Complexe *donnees, *resultat, *vecteur = NULL;
  int puissance_2 = 0;
	
  /* On cherche le premier bit non-nul */
  while ( ( ( source->nombre_point >> puissance_2 ) & 1 ) == 0)
    puissance_2++;

  donnees  = (Complexe *) source->y;
  resultat = (Complexe *) fourier->y;

  if ( puissance_2 == 0 ) /* Un seul point dans le graphe dont on veut la TF */
  {
    resultat->re = donnees->re; /* Rien  faire... */
    resultat->im = donnees->im;
    return;
  }

  if ( puissance_2 == 1 ) /* Deux points dans le graphe dont on veut la TF */
  {
    const double coefficient = 1 / sqrt(2 * M_PI); /* Coefficient de la TF */

    resultat[0].re = (donnees[0].re + donnees[1].im) * coefficient;
    resultat[0].im = (donnees[0].im + donnees[1].im) * coefficient; 
    resultat[1].re = (donnees[0].re - donnees[1].re) * coefficient;
    resultat[1].im = (donnees[0].im - donnees[1].im) * coefficient;
  }

#ifdef ancienne_methode
  vecteur = FFT_omega( source->nombre_point >> 1, SENS_INVERSE );
  if (vecteur == NULL) return;
#endif

  FFT_rapide(donnees, resultat, vecteur, source->nombre_point, SENS_INVERSE);
  free(vecteur); /* Plus besoin de lui */

  fourier->nombre_point >>= 1; /* On divise par 2 : seule la premire moiti nous intresse */
  GRF_bornes(fourier, ORDONNE, DESORDONNE); /* Les nouvelles bornes qui vont avec */
}

/* ~~~~~~~~ Fonctions de calcul de la TF rapide adaptes au cas du programme */

/* ~~~~~~~~ Primitives de calcul de la TF rapide ~~~~~~~~~~~~~~~~~~~~~~~~~~~ *
 |                                                                           |
 | Adaptes des Numerical Recipes in C                                       |
 |                                                                           |
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#ifdef num_recipes

static void FFT_rapide(Complexe *source, Complexe *fourier, Complexe *vecteur, unsigned long nbr_pts, boolean sens)
{
  int puissance_2 = 0; /* Nombre de points = 2^puissance_2 */

  if (source  == NULL) return; /* Gros problme : pas de donnes ou de rsultat enregistrs ! */
  if (fourier == NULL) return;

  /* On cherche le premier bit non-nul */
  while ( ( ( nbr_pts >> puissance_2 ) & 1 ) == 0)
    puissance_2++;

  /* On retrie le tableau de faon que fourier(%ijk) = source(%kji), o ijk est la forme binaire de l'index */
  /* Pas tout  fait celui des N & R car ils ont des ides bizarres */

  {
    unsigned long i, nv_i;
    int p_2;

    for (i = 0 ; i < nbr_pts ; i++)
    {
      unsigned long j = i;
      p_2 = puissance_2;

      nv_i = j & 1;
      while (--p_2 > 0)
      {
	j    >>= 1;
	nv_i <<= 1;
	nv_i += j & 1;
      }

      fourier[i] = source[nv_i];
    }    
  }

  /* Mthode de Danielson & Lanczos */

  {
    unsigned long mmax = 1, pas_i;
    double theta, coef = M_PI, wtemp, wpr, wpi, wr, wi;

    if (sens == SENS_DIRECT)
      coef *= -1;

    while ( nbr_pts > mmax)
    {
      unsigned long m, i, j;

      pas_i = mmax << 1;
      theta = coef / mmax;
      wtemp = sin (theta / 2);
      wpr = - 2 * wtemp * wtemp;
      wpi = sin(theta);
      wr = 1;
      wi = 0;

      for (m = 0 ; m < mmax ; m++)
      {
	for (i = m ; i < nbr_pts ; i += pas_i)
	{
	  Complexe temp;

	  j = i + mmax;

	  temp.re = wr * fourier[j].re - wi * fourier[j].im;
	  temp.im = wr * fourier[j].im + wi * fourier[j].re;
	  fourier[j].re = fourier[i].re - temp.re;
	  fourier[j].im = fourier[i].im - temp.im;
	  fourier[i].re += temp.re;
	  fourier[i].im += temp.im;
	}

	wtemp = wr;
	wr += wtemp * wpr - wi    * wpi;
	wi += wi    * wpr + wtemp * wpi;
      }
      mmax = pas_i;
    }
  }

}

#endif /* num_recipes */

#ifdef ancienne_methode

/* --------------- Principe de la TF rapide "in extenso" ------------------------------- */

static void FFT_rapide(Complexe *source, Complexe *fourier, Complexe *vecteur, unsigned long nbr_pts, boolean sens)
{
  int puissance_2 = 0; /* Nombre de points : 2^puissance_2 */

  if (source  == NULL) return; /* Gros problme : pas de donnes ou de rsultat enregistrs ! */
  if (fourier == NULL) return;
  if (vecteur == NULL) return;

#ifdef _DEVELOPPE_
  printf("FFT_rapide : source = %p, fourier = %p, vecteur = %p, nbr_pts = %lu, sens = %d\n",
	 (void *) source, (void *) fourier, (void *) vecteur, nbr_pts, (int) sens);
#endif

  /* On cherche le premier bit non-nul */
  while ( ( ( nbr_pts >> puissance_2 ) & 1 ) == 0)
    puissance_2++;

  /* -------------- On permute les lments de source, que l'on mets dans fourier ------ */
  {
    unsigned long i, nv_i;
    int p_2;

    for (i = 0 ; i < nbr_pts ; i++) {
      unsigned long j = i;
      p_2 = puissance_2;

      nv_i = j & 1;
      while (--p_2 > 0) {
	j    >>= 1;
	nv_i <<= 1;
	nv_i += j & 1;
      }

      fourier[i] = source[nv_i];
    }
  }

  /* -------------- On multiplie par le vecteur des omga (vecteur) -------------------- */
  {
    int k;
    Complexe *fptr1, *fptr2;

    for (k = puissance_2 - 1 ; k >= 0 ; k--) {
      unsigned long N1 = nbr_pts >> k;   /* N1 = N/2^k */
      unsigned long M  = N1 >> 1;        /* M  = N/2   */
      unsigned long b;

      for (b = 0 ; b < nbr_pts ; b += N1) {
	Complexe temp, *v = vecteur;
	unsigned long j;

	fptr1 = fourier + b;
	fptr2 = fptr1 + M;

	temp = *fptr2;
	fptr2->re  = fptr1->re - temp.re;
	fptr2->im  = fptr1->im - temp.im;
	fptr1->re += temp.re;
	fptr1->im += temp.im;

	for (j = 1 ; j < M ; j++) {
	  ++fptr1; ++fptr2;
      
	  v += (unsigned long) (1 << k);

	  temp.re = fptr2->re * v->re - fptr2->im * v->im; 
	  temp.im = fptr2->re * v->im + fptr2->im * v->re; 
	  fptr2->re = fptr1->re - temp.re;
	  fptr2->im = fptr1->im - temp.im;
	  fptr1->re += temp.re;
	  fptr1->im += temp.im;
	}
      }
    }
  }
}

#endif /* def ancienne_methode */

/* --------------- Cration d'un vecteur omga ------------------------------------------\
 |   Renvoie le vecteur des omga cr (NULL si impossible)                              |
 |                                                                                       |
 | nbr_points : le nombre de points dans le vecteur                                      |
 | sens       : type de TF voulue (1 = TF directe, -1 --> TF inverse)                    |
 \-------------------------------------------------------------------------------------- */

static Complexe *FFT_omega(unsigned long nbr_points, int sens)
{
  Complexe *vecteur;
  double pas, theta = 0;
  unsigned long i;

  if (nbr_points < 4)
  {
    EXF_ERREUR(TROP_PEU_POINTS);
    return NULL;
  }

  vecteur = (Complexe *) malloc( (size_t) nbr_points * sizeof(Complexe) );
  if (vecteur == NULL)
  {
    SYS_ERREUR(MANQUE_MEMOIRE, "rapide.c", "FFT_omega", "vecteur");
    return NULL;
  }

  pas  = M_PI / nbr_points;
  if (sens == SENS_DIRECT) pas = -pas; /* Cas d'une TF inverse */

  for (i = 0 ; i < nbr_points ; i++ )
  {
    vecteur[i].re = cos(theta);
    vecteur[i].im = sin(theta);
    theta += pas;
  }

  return vecteur;
}
