/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2003, Emmanuel CURIS

   Chargement des graphes, fichiers au format interne
   Voir Graphes/sauver.c pour la description du format
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef VERSION_SIMPLIFIEE_
# define _LIMITE_ 1
#endif

#include "systeme.h"
#include "bin_utils.h"

#include "serie.h"

#include "filtrage.h" /* Pour chargement de Infos_FonctionApodisation */
#include "matrices.h" /* Pour chargement de la matrice de covariance */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define VERIFIE(x, r, ch) if (n_lu < (x)) \
                          { (void) SYS_ERREUR(ERREUR_LECTURE, NULL, fichier, NULL); \
                            (void) fprintf(stderr, "VERIFIE, chargement : %s\n", ch); \
                            return (r); }

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Graphe *Charger_v3   ( FILE *fichier, int version, Serie_Graphes *serie );
static Graphe *ChargerAncien( FILE *fichier, int version, Serie_Graphes *serie );

static int   ch_affichage (FILE *fichier, Graphe *graphe, int version);
static int   ch_covariance(FILE *fichier, Graphe *graphe);
static int   ch_table_C   (FILE *fichier, unsigned long max_points, Complexe *table, double *min, double *max);
static int   ch_table_I   (FILE *fichier, unsigned long max_points, int *table);
static int   ch_table_R   (FILE *fichier, unsigned long max_points, double *table, double *min, double *max);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~ Chargement des séries ~-~-~-~-~-~-~-~ */

Serie_Graphes *Serie_Charger(FILE *fichier, int version, const char *nom_fichier)
{
  Graphe *graphe;
  Serie_Graphes *serie_lue = NULL;
  char *nom_serie = NULL;
  int n_lu, retour;
  short int drapeau;
  unsigned long i, nombre_graphes;

  if ( NULL == fichier ) return NULL;

  /* On lit le nom de la série */
  nom_serie = BIN_CH_chaine( fichier );

  /* On crée la série, après la courante */
  serie_lue = SERIE_GRF_CreerSerie( serie_courante, nom_serie );
  if ( NULL == serie_lue ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de créer la série %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_serie);
    free( nom_serie ); nom_serie = NULL;
    return NULL;
  }
  free( nom_serie ); nom_serie = NULL;

  /* La série contient-elle une table des X ?*/
  n_lu = fread( &drapeau, sizeof( short int ), 1, fichier );
  VERIFIE( 1, NULL, "drapeau table des X" );
  if ( drapeau ) {
    unsigned long nombre_points;
    double *x;
    char *legende __attribute__((__unused__)), *unite __attribute__((__unused__));

    /* Oui : on la lit */
    /* 1. Combien de points ? */
    n_lu  = fread( &nombre_points, sizeof(unsigned long), 1, fichier );
    VERIFIE( 1, NULL, "drapeau table des X" );

    /* 2. Création de la table */
    x = (double *) malloc( sizeof( double ) * nombre_points );
    if ( NULL == x ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver les X [%lu valeurs]!\n",
		     __FILE__, __FUNCTION__, __LINE__, nombre_points);
      return NULL;
    }

    /* 3. Lecture des valeurs */
    retour = ch_table_R( fichier, nombre_points, x, NULL, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en lisant les X, série %s\n",
		     __FILE__, __FUNCTION__, __LINE__, serie_lue->nom);
      free( x ); x = NULL;
      return NULL;
    }

    /* 4. Assignation des valeurs [les X seront libérés avec la série] */
    retour = Serie_AssocierTable( serie_lue, nombre_points, x, FAUX );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en assignant les X à la série %s [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, serie_lue->nom, retour);
      free( x ); x = NULL;
      return NULL;
    }

    /* 5. Y a-t-il des erreurs en X ? */
    n_lu = fread( &drapeau, sizeof( short int ), 1, fichier );
    VERIFIE( 1, NULL, "drapeau table des erreurs en X" );
    if ( drapeau ) {
      OPTION_A_FAIRE( "Lecture des erreurs en X" );
    }

    /* 6. La légende en X */
    legende = BIN_CH_chaine( fichier );
    OPTION_A_FAIRE( "Installer la légende en X" );

    /* 7. Les unités en X */
    unite = BIN_CH_chaine( fichier );
    OPTION_A_FAIRE( "Installer l'unité en X" );
  }

  /* Combien de graphes faut-il lire ? */
  n_lu = fread( &nombre_graphes, sizeof( unsigned long ), 1, fichier );
  VERIFIE( 1, NULL, "Nombre de graphes" );

  /* Les graphes, un par un */
  for ( i = 0 ; i < nombre_graphes ; i++ ) {
    /* 1. Les X sont-ils ceux de la série ? */
    n_lu = fread( &drapeau, sizeof( short int ), 1, fichier );
    VERIFIE( 1, NULL, "drapeau table des X dans la série" );

    /* 2. Lecture du graphe */
    graphe = GRF_Charger( fichier, version, nom_fichier, serie_lue );

    /* 3. On installe les X de la série, si nécessaire */
    if ( drapeau ) {
      graphe->x = serie_lue->table_X;
      /* Normalement, x_min et x_max ont été sauvés, donc sont corrects ! */
    }
  }

  /* Chargement de la série terminé */
  return serie_lue;
}

/* ~-~-~-~-~-~-~-~ Chargement des graphes ~-~-~-~-~-~-~-~ */

Graphe *GRF_Charger(FILE *fichier, int version, const char *nom_fichier, Serie_Graphes *serie)
{
  Graphe *graphe;

  if ( NULL == fichier ) return NULL;

  if ( version >= 30 ) {
    graphe = Charger_v3( fichier, version, serie );
  } else {
    graphe = ChargerAncien( fichier, version, serie );
  }

  if ( NULL == graphe ) {
    (void) fprintf(stderr, "GRF_Charger : ERREUR de chargement\n");
  } else {
    /* Le graphe a été sauvé... puisqu'il vient d'être chargé ! */
    graphe->drapeaux.sauve = VRAI;

    /* On associe le dossier & le nom du fichier */
    GRF_AssocierFichier( graphe, nom_fichier, VRAI );
  }

  return graphe;
}

static Graphe *Charger_v3( FILE *fichier, int version, Serie_Graphes *serie )
{
  char *titre;
  short int drapeau;
  int type, genre, retour;
  unsigned long nombre_points;
  size_t n_lu;
  Graphe *graphe;

  ENTREE_FONCTION( __FUNCTION__, "fichier=%p version=%d serie=%p", fichier, version, serie );

  /* Informations préalables : nombre de points, nature des données */
  n_lu  = fread(&nombre_points, sizeof(unsigned long), 1, fichier);
  n_lu += fread(&type, sizeof(int), 1, fichier);
  n_lu += fread(&genre, sizeof(int), 1, fichier);
  VERIFIE(3, NULL, "ch_graphe - en_tete (version > 3.0)\n");

  /* On crée le graphe, dans la série indiquée */
  graphe = Serie_nv_graphe(serie, NULL, nombre_points, type, genre, NULL);
  if ( NULL == graphe ) {
    (void) fprintf(stderr, "Charger_v3 : ERREUR dans Serie_nv_graphe [v. %f]\n", version / 10.0 );
    SORTIE_FONCTION( __FUNCTION__, "%p [Impossible d'ajouter à la série]", NULL );
    return NULL;
  }

  /* On lit la table des X */
  retour = ch_table_R( fichier, graphe->nombre_point, graphe->x, &(graphe->x_min), &(graphe->x_max) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "Charger_v3 : ERREUR dans ch_table_R\n");
    SORTIE_FONCTION( __FUNCTION__, "%p [ch_table_R]", NULL );
    return NULL;
  }

  /* On lit la table des Y et, éventuellement, des Z */
  switch ( genre )
  {
  case GRF_FCT_REELLE:
  case GRF_SERIE:
  case GRF_COURBE_PLANE:
    retour = ch_table_R( fichier, graphe->nombre_point, graphe->y, &(graphe->y_min), &(graphe->y_max) );
    break;

  case GRF_FCT_COMPLEXE:
    retour = ch_table_C( fichier, graphe->nombre_point, (Complexe *) graphe->y, &(graphe->y_min), &(graphe->y_max) );
    break;

  case GRF_COURBE_ESPACE:
  case GRF_SURFACE:
    retour = ch_table_R( fichier, graphe->nombre_point, graphe->y, &(graphe->y_min), &(graphe->y_max) );
    retour = ch_table_R( fichier, graphe->nombre_point, graphe->z.z, NULL, NULL);
    break;

  case GRF_HISTOGRAMME:
    (void) fprintf(stderr, "GRF_Charger : À FAIRE --- charger les Y d'un histogramme\n");
    break;

  case GRF_FAISCEAU:
    (void) fprintf(stderr, "GRF_Charger : À FAIRE --- charger les Y d'un faisceau de graphes\n");
    break;

  case GRF_MULTI_SERIE:
    (void) fprintf(stderr, "GRF_Charger : À FAIRE --- charger les Y d'un graphe multi-séries\n");
    break;

  case GRF_ERREUR_R:
    retour = ch_table_R( fichier, graphe->nombre_point, graphe->y, &(graphe->y_min), &(graphe->y_max) );
    retour = ch_table_I( fichier, graphe->nombre_point, graphe->z.ddl);
    break;

  case GRF_ERREUR_C:
    retour = ch_table_C( fichier, graphe->nombre_point, (Complexe *) graphe->y, &(graphe->y_min), &(graphe->y_max) );
    retour = ch_table_I( fichier, graphe->nombre_point, graphe->z.ddl);
    break;

  default:
    (void) fprintf(stderr, "GRF_Charger : type de données inconnu [%d]\n", graphe->genre_donnees );
    retour = -5;
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "GRF_Charger : ERREUR dans le chargement des Y\n");
    SORTIE_FONCTION( __FUNCTION__, "%p [les Y & Z]", NULL );
    return NULL;
  }

  /* On lit d'éventuelles informations complémentaires */
  switch ( type )
  {
  case GR_DEVIATION: /* Table des ddl */
    retour = ch_table_I( fichier, graphe->nombre_point, graphe->z.ddl);
    break;

  case GR_TF: /* Informations sur la fenêtre */
    drapeau = (graphe->pointeur.infos_tf != NULL);
    n_lu = fread(&drapeau, sizeof( short int ), 1, fichier);
    VERIFIE(1, NULL, "GRF_Charger : GR_TF, drapeau");
    if ( drapeau ) {
      Infos_FonctionApodisation *infos;

      /* On réserve pour cela */
      infos = (Infos_FonctionApodisation *) malloc( sizeof(Infos_FonctionApodisation) );
      if ( NULL == infos ) {
	(void) fprintf(stderr, "GRF_Charger : impossible de réserver les infos pour la TF\n");
	SORTIE_FONCTION( __FUNCTION__, "%p [Infos TF]", NULL );
	return NULL;
      }

      n_lu  = fread( &(infos->x_min)     , sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->x_max)     , sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->parametre) , sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->type)      , sizeof(int) , 1, fichier);
      n_lu += fread( &(infos->x_1)       , sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->x_2)       , sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->symetrique), sizeof(boolean), 1, fichier);
      VERIFIE(7, NULL, "GRF_Charger : GR_TF, infos de fenêtre");

      graphe->pointeur.infos_tf = infos;
    }
    break;

  case GR_TFI: /* Informations sur le filtre */
  case GR_EXAFS:
    n_lu = fread(&drapeau, sizeof( short int ), 1, fichier);
    VERIFIE(1, NULL, "GRF_Charger : GR_TFI / GR_EXAFS, drapeau");
    if ( drapeau ) {
      Infos_Nyquist *infos;

      infos = (Infos_Nyquist *) malloc( sizeof(Infos_Nyquist) );
      if ( NULL == infos ) {
	(void) fprintf(stderr, "GRF_Charger : GR_TFI / GR_EXAFS, imposible de réserver les infos\n");
	SORTIE_FONCTION( __FUNCTION__, "%p [Infos TFI]", NULL );
	return NULL; /* Si on échoue ici, on échoue probablement après, alors... */
      }

      n_lu  = fread( &(infos->r_min), sizeof(double) , 1, fichier);
      n_lu += fread( &(infos->r_max), sizeof(double) , 1, fichier);
      VERIFIE(2, NULL, "GRF_Charger : GR_TFI / GR_EXAFS, infos de filtrage");

      graphe->pointeur.nyquist = infos;
    }
    break;
  }

  graphe->drapeaux.sauve = VRAI;

  /* Là, on s'occupe des incertitudes
     1. Erreurs sur les X */
  fread(&drapeau, sizeof(short int), 1, fichier);
  if ( drapeau == VRAI ) { /* Il y a des erreurs en X */
    Graphe *erreur;

    erreur = GRF_creer_erreurs(graphe, AXE_X);
    if (erreur != NULL) { /* Le graphe existe bien */
      retour = ch_table_R( fichier, erreur->nombre_point, erreur->y, &(erreur->y_min), &(erreur->y_max) );
      retour = ch_table_I( fichier, erreur->nombre_point, erreur->z.ddl );
    } else { /* On saute purement et simplement... */
      retour = ch_table_R( fichier, 0, NULL, NULL, NULL );
      retour = ch_table_I( fichier, 0, NULL );
    }

    erreur->drapeaux.sauve = VRAI;
  }

  /* 2. Erreurs sur les Y */
  fread(&drapeau, sizeof(short int), 1, fichier);
  if ( drapeau == VRAI ) { /* Il y a des erreurs en X */
    Graphe *erreur;

    erreur = GRF_creer_erreurs(graphe, AXE_Y);
    if (erreur != NULL) { /* Le graphe existe bien */
      if ( GRF_FCT_COMPLEXE == genre ) {
	retour = ch_table_C( fichier, erreur->nombre_point, ( Complexe * ) erreur->y, &(erreur->y_min), &(erreur->y_max) );
      } else {
	retour = ch_table_R( fichier, erreur->nombre_point, erreur->y, &(erreur->y_min), &(erreur->y_max) );
	retour = ch_table_I( fichier, erreur->nombre_point, erreur->z.ddl );
      }
    } else { /* On saute purement et simplement... */
      if ( GRF_FCT_COMPLEXE == genre ) {
	retour = ch_table_C( fichier, 0, NULL, NULL, NULL );
      } else {
	retour = ch_table_R( fichier, 0, NULL, NULL, NULL );
	retour = ch_table_I( fichier, 0, NULL );
      }
    }

    erreur->drapeaux.sauve = VRAI;
  }

  /* Matrice de covariance, si elle existe */
  n_lu = fread( &drapeau, sizeof(short int), 1, fichier );
  VERIFIE(1, NULL, "Charger_v3, drapeau / matrice de covariance");
  if ( VRAI == drapeau ) { /* Il y a une matrice de covariance à lire */
    retour = ch_covariance(fichier, graphe);
    if ( retour != 0 ) {
      (void) fprintf(stderr, "Charger_v3 : ERREUR dans ch_covariance()\n");
      SORTIE_FONCTION( __FUNCTION__, "%p [Matrice de covariance]", NULL );
      return NULL;
    }
  }

  if ( version >= 36 ) {
    /* Depuis la version 3.6 : drapeaux de propriétés du graphe */
    n_lu = fread( &drapeau, sizeof(short int), 1, fichier );
    graphe->drapeaux.X_tries = drapeau;

    n_lu += fread( &drapeau, sizeof(short int), 1, fichier );
    graphe->drapeaux.X_uniques = drapeau;

    n_lu += fread( &drapeau, sizeof(short int), 1, fichier );
    graphe->drapeaux.theorique = drapeau;

    n_lu += fread( &drapeau, sizeof(short int), 1, fichier );
    graphe->drapeaux.normalise = drapeau;

    VERIFIE( 4, NULL, "Charger_v3, drapeaux de propriétés du graphe" );
  } else {
    /* On essaye de déduire les drapeaux des types de graphes "anciens" */
    switch ( graphe->type )
    {
    case GR_EXAFS:
      graphe->drapeaux.normalise = VRAI;
      break;

    case GR_AJUSTE:
    case GR_FENETRE:
      graphe->drapeaux.theorique = VRAI;
      break;
    }
  }

  /* Titre du graphe */
  titre = BIN_CH_chaine( fichier );
  GRF_titre( graphe, titre );
  free( titre );

  /* Paramètres d'affichage, s'ils existent */
  retour = ch_affichage( fichier, graphe, version );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "Charger_v3 : ERREUR dans ch_affichage()\n");
    SORTIE_FONCTION( __FUNCTION__, "%p [Affichage]", NULL );
    return NULL;
  }

  /* Tout s'est bien passé, chargement terminé */
  SORTIE_FONCTION( __FUNCTION__, "%p", graphe );
  return graphe;
}

static Graphe *ChargerAncien( FILE *fichier, int version, Serie_Graphes *serie )
{
  char *titre;
  short int genre, drapeau;
  int type, couleur, style;
  unsigned long nombre_points, lg_titre;
  size_t n_lu;
  Graphe *graphe;

  if (version < 13) {
    int ex_genre;
    struct {
      unsigned long marque;     /* Est censée valoir EMC\233 */
      int           version;    /* Numéro de version */
      unsigned long nbr_points; /* Dans le graphe */
      int           type;       /* Type du graphe */
      unsigned long lg_titre;   /* Longueur du titre */
    } en_tete; /* uniquement pour l'ancienne version du programme */

    fseek(fichier, 0L, 0);
    n_lu = fread(&en_tete, sizeof(en_tete), 1, fichier);
    VERIFIE(1, NULL, "ch_graphe - en_tete (version < 1.3)\n");
    nombre_points = en_tete.nbr_points;
    type = en_tete.type;
    lg_titre = en_tete.lg_titre;
    if (version > 10) {
      n_lu = fread(&ex_genre, sizeof(int), 1, fichier);
      VERIFIE(1, NULL, "ch_graphe - genre (version < 1.3)\n");
      genre = (short int) ex_genre;
    } else {
      if ( en_tete.type == GR_TF )
	genre = GRF_FCT_COMPLEXE;
      else
	genre = GRF_FCT_REELLE;
    }
  } else {
    n_lu  = fread(&nombre_points, sizeof(unsigned long), 1, fichier); /* Points dans le graphe */
    n_lu += fread(&type, sizeof(int), 1, fichier);
    if ( version < 30 ) {
      n_lu += fread(&genre, sizeof(short int), 1, fichier);
    } else {
      n_lu += fread(&genre, sizeof(int), 1, fichier);
    }
    n_lu += fread(&lg_titre, sizeof(unsigned long), 1, fichier); /* Le titre */
    VERIFIE(4, NULL, "ch_graphe - en_tete (version > 1.2)\n");
  }

  titre = (char *) calloc( lg_titre + 1, sizeof(char) );
  if (titre == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_graphe", "titre");
    return NULL;
  }
  fgets(titre, (int) (lg_titre + 1), fichier);

  graphe = Serie_nv_graphe(serie, NULL, nombre_points, type, genre, titre);
  free( titre ); titre = NULL;
  if ( NULL == graphe ) {
    return NULL;
  }

  if ( version > 10 ) {
    char *legende;
    int unit_x, unit_y;

    fread(&lg_titre, sizeof(unsigned long), 1, fichier); /* La légende en X */
    legende = (char *) calloc( lg_titre + 1, sizeof(char) );
    if (legende == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_graphe", "legende / X");
      return NULL;
    }
    fgets(legende, (int) (lg_titre + 1), fichier);
    fread( &unit_x, sizeof(int), 1, fichier );
    GRF_ChangerLegende( graphe, AXE_X, legende, FAUX );
    OPTION_A_FAIRE( "Convertir les codes d'unité [axe des X]" );

    fread(&lg_titre, sizeof(unsigned long), 1, fichier); /* La légende en Y */
    legende = (char *) calloc( lg_titre + 1, sizeof(char) );
    if (legende == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_interne", "leg_y");
      return NULL;
    }
    fgets(legende, (int) (lg_titre + 1), fichier);
    fread(&unit_y, sizeof(int), 1, fichier);
    GRF_ChangerLegende(graphe, AXE_Y, legende, FAUX);
    OPTION_A_FAIRE( "Convertir les codes d'unité [axe des Y]" );
  }

  graphe->drapeaux.sauve = VRAI; /* Comme ça, ok pour la suite quelle que soit la version */

  fread( &(graphe->x_min), sizeof(double), 1, fichier ); /* Bornes en X */
  fread( &(graphe->x_max), sizeof(double), 1, fichier );
  fread( &(graphe->y_min), sizeof(double), 1, fichier ); /* Bornes en Y */
  fread( &(graphe->y_max), sizeof(double), 1, fichier );

  fread( graphe->x, sizeof(double), nombre_points, fichier ); /* Les valeurs en X */
  if ( graphe->genre_donnees == GRF_FCT_REELLE ) {            /* Les valeurs en Y */
    fread( graphe->y, sizeof(double), graphe->nombre_point, fichier );
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    fread( graphe->y, sizeof(Complexe), graphe->nombre_point, fichier );
  } else if ( GRF_COURBE_ESPACE == graphe->genre_donnees ) { /* Version 1.5 min et graphe non complexe, non réel ! */
    n_lu = fread(graphe->y, sizeof(double), graphe->nombre_point, fichier);
    n_lu = fread(graphe->z.z, sizeof(double), graphe->nombre_point, fichier);
  } else {
    fprintf(stderr, "ch_graphe : cas non traité, genre = %d\n", (int) genre);
  }

  if ( version > 22 ) { /* v. 2.3 : les ddl sont sauvés */
    if ( GR_DEVIATION == graphe->type ) {
      fread( graphe->z.ddl, sizeof(int), graphe->nombre_point, fichier );
    }
  }

  fread( &style, sizeof(int), 1, fichier); /* Le style de dessin, et la couleur */
  fread( &couleur, sizeof(int), 1, fichier);
  if ( style > 6 ) {
    GRF_couleur( graphe, couleur, 0, style - 6 );
  } else {
    GRF_couleur( graphe, couleur, style, 0 );
  }
  
  if ( graphe->type == GR_TF ) { /* Une transformée de Fourier : infos complémentaires ? */
    boolean infos_tf;

    fread(&infos_tf, sizeof(boolean), 1, fichier);

    if (infos_tf == VRAI) { /* Oui : infos de fenêtrage... */
      Infos_FonctionApodisation *infos = (Infos_FonctionApodisation *) malloc( sizeof(Infos_FonctionApodisation) );

      if (infos == NULL) {
	long taille;
	taille = 3 * sizeof(double);
	if (version >= 12) {
	  taille += 2 * sizeof(double) + sizeof(boolean);
	  if (version >= 15)
	    taille += 2 * sizeof(int);
	  else
	    taille += 2 * sizeof(double);
	}
	(void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_interne", "infos (de TF)");
	fseek(fichier, taille, SEEK_CUR); /* On ignore */
      } else {
	fread( &(infos->x_min)    , sizeof(double), 1, fichier);
	fread( &(infos->x_max)    , sizeof(double), 1, fichier);
	fread( &(infos->parametre), sizeof(double), 1, fichier);
	if (version < 15) { /* Bizarrement, sauvé en double dans les versions avant la 1.5 ! */
	  double degre;
	  fread( &(infos->type) , sizeof(double), 1, fichier);
	  fread( &degre, sizeof(double), 1, fichier);
	} else {
	  fread( &(infos->type) , sizeof(int), 1, fichier);
	  if (version < 16) { /* Pas sauvé depuis, car n'existe plus */
	    int degre;
	    fread( &degre, sizeof(int), 1, fichier);
	  }
	}
	if (version >= 12) { /* Sauvé depuis la version 1.2 */
	  fread( &(infos->x_1)          , sizeof(double) , 1, fichier);
	  fread( &(infos->x_2)          , sizeof(double) , 1, fichier);
	  fread( &(infos->symetrique)   , sizeof(boolean), 1, fichier);
	} else {
	  infos->x_1 = infos->x_min + 1.0;
	  infos->x_2 = infos->x_max - 1.0;
	  infos->symetrique = VRAI;
	}

	if ( (version > 16) && (version < 22) ) { /* Versions 1.7 à 2.1 : on sauve ici la matrice de covariance */
	  size_t n_lu __attribute__((__unused__));
	  boolean covariance;

	  n_lu = fread( &covariance, sizeof(boolean), 1, fichier);
	  if ( covariance == VRAI ) {
	    ch_covariance(fichier, graphe);
	  }
	} /* Aucune matrice de covariance sauvée avant la v. 1.7 */

	graphe->pointeur.infos_tf = infos;
      }
    }
  } else if ( (graphe->type == GR_TFI) || (graphe->type == GR_EXAFS) ) {
    if (version >= 17) { /* Ajouté en v. 1.7 */
      boolean infos_nyquist;

      fread(&infos_nyquist, sizeof(boolean), 1, fichier);

      if (infos_nyquist == VRAI) { /* Oui : infos de fenêtrage... */
	boolean covariance;
	Infos_Nyquist *infos = (Infos_Nyquist *) malloc( sizeof(Infos_Nyquist) );

	if (infos == NULL) {
	  long taille;
	  taille  = 2 * sizeof(double);
	  (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_graphe", "infos (Nyquist)");
	  fseek(fichier, taille, SEEK_CUR); /* On ignore */
	} else {
	  fread( &(infos->r_min)        , sizeof(double), 1, fichier);
	  fread( &(infos->r_max)        , sizeof(double), 1, fichier);
	  graphe->pointeur.nyquist = infos;
	}

	if ( version < 22 ) { /* Changé de place dans la version 2.2 */
	  n_lu = fread(&covariance, sizeof(boolean), 1, fichier);
	  if (covariance == VRAI)	{
	    ch_covariance(fichier, graphe);
	  }
	}
      }
    }
  }

  /* On essaye de déduire les drapeaux des types de graphes "anciens" */
  switch ( graphe->type )
  {
  case GR_EXAFS:
    graphe->drapeaux.normalise = VRAI;
    break;

  case GR_AJUSTE:
  case GR_FENETRE:
    graphe->drapeaux.theorique = VRAI;
    break;
  }

  HIST_Texte(4, 0, "Graphe charg"e_AIGU" : %s\n", graphe->nom);
  if (version < 13) {/* Version avant la 1.3 ne sauve rien d'autre */
    return graphe;
  }

  fread(&drapeau, sizeof(short int), 1, fichier);
  if (drapeau == VRAI) { /* Il y a des erreurs en X */
    Graphe *erreur;

    erreur = GRF_creer_erreurs(graphe, AXE_X);
    if (erreur != NULL) { /* Le graphe existe bien */
      fread(erreur->y, sizeof(double), nombre_points, fichier);
      if ( version > 22 ) { /* v. 2.3 : les ddl sont sauvés */
	fread( erreur->z.ddl, sizeof(int), nombre_points, fichier );
      }

      GRF_bornes(erreur, NE_PAS_FAIRE, DESORDONNE);
      erreur->drapeaux.sauve = VRAI;
      graphe->drapeaux.erreur_x = VRAI;
    }
  } else
    graphe->erreur_x = NULL;

  fread(&drapeau, sizeof(short int), 1, fichier);
  if (drapeau == VRAI) { /* Il y a des erreurs en Y */
    short int type = GR_NORMAL;
    Graphe *erreur;

    if (version > 13) /* Erreurs complexes apparues avec la v. 1.4 */
      fread(&type, sizeof(short int), 1, fichier);

    erreur = GRF_creer_erreurs(graphe, AXE_Y);
    if (erreur != NULL) { /* Le graphe existe bien */
      if (type == GR_NORMAL) {
	fread(erreur->y, sizeof(double), nombre_points, fichier);
	if ( version > 22 ) { /* v. 2.3 : les ddl sont sauvés */
	  fread( erreur->z.ddl, sizeof(int), nombre_points, fichier );
	}
      } else
	fread(erreur->y, sizeof(Complexe), nombre_points, fichier);

      GRF_bornes(erreur, NE_PAS_FAIRE, DESORDONNE);
      erreur->drapeaux.sauve = VRAI;
      graphe->drapeaux.erreur_y = VRAI;
    }
  } else
    graphe->erreur_y = NULL;

  if ( version > 21 ) { /* Matrice de covariance toujours présente, depuis la v. 2.2 */
    short int mat_presente;
    size_t n_lu;

    n_lu = fread( &mat_presente, sizeof(short int), 1, fichier );
    VERIFIE(1, graphe, "ch_graphe(), drapeaux / matrice de covariance");
    if ( VRAI == mat_presente ) { /* Il y a une matrice de covariance à lire */
      ch_covariance(fichier, graphe);
    }
  }

  return graphe;
}

/* -------- Chargement des informations d'affichage d'un graphe -------- */

static int ch_affichage(FILE *fichier, Graphe *graphe, int version)
{
  char *legende_X, *legende_Y, *unite_X, *unite_Y;
  short int drapeau, style, marque;
  int entiers[ 4 ];

  fread( &drapeau, sizeof( short int ), 1, fichier );
  if ( version == 30 ) /* Version 3.0 : erreur dans la sauvegarde, 0 si affichage au lieu de 1 ! */
    drapeau = !drapeau;
  if ( 0 == drapeau ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_affichage : Pas d'information d'affichage !\n");
#endif

    return 0;
  }

  /* On lit les légendes des axes */
  legende_X = BIN_CH_chaine( fichier );
  GRF_ChangerLegende( graphe, AXE_X, legende_X, FAUX );

  legende_Y = BIN_CH_chaine( fichier );
  GRF_ChangerLegende( graphe, AXE_Y, legende_Y, FAUX );

  /* Couleur, style, unités */
  if ( version <= 33 ) {
    fread( entiers, sizeof( int ), 4, fichier );

    if ( entiers[ 1 ] < 6 ) {
      style = entiers[ 1 ];
      marque = 0;
    } else {
      style = 0;
      marque = entiers[ 1 ] - 6;
    }
  } else if ( version <= 35 ) {
    fread( entiers, sizeof( int ), 2, fichier );

    if ( entiers[ 1 ] < 6 ) {
      style = entiers[ 1 ];
      marque = 0;
    } else {
      style = 0;
      marque = entiers[ 1 ] - 6;
    }
  } else {
    fread( entiers, sizeof( int ), 1, fichier ); /* La couleur */
    fread( &style, sizeof( short int ), 1, fichier ); /* Le style */
    fread( &marque, sizeof( short int ), 1, fichier ); /* Le style */
  }

  /* On vérifie style et marques [en cas de problème de sauvegarde... ou de style pas encore supporté : version trop ancienne] */
  if ( ( style > BI_NOMBRE_STYLES_LIGNES ) || ( style < 0 ) ) {
    /* On passe en continu [Rq. 0 : aucune ligne tracé ! */
    AVERTISSEMENT( "Style de ligne incorrect ou non supporté : %d", style );
    style = BI_CONTINU;
  }
  if ( ( marque > BI_NOMBRE_MARQUES ) || ( marque < 0 ) ) {
    /* On passe en points : on est sûr de voir quelque chose même s'il n'y a pas de ligne associée ! */
    AVERTISSEMENT( "Type de marque incorrect ou non supporté : %d", marque );
    marque = BI_MARQUE_AUCUNE;
  }

  /* On installe cela dans le graphe */
  GRF_couleur( graphe, entiers[ 0 ], style, marque );

  if ( version < 34 ) {
    OPTION_A_FAIRE( "Convertir les anciens codes d'unités" );
    return 0;
  }

  /* On lit les unités des axes */
  unite_X = BIN_CH_chaine( fichier );
  if ( ( version < 35 ) && ( NULL == unite_X ) ) {
    GRF_ChangerUnite( graphe, AXE_X, GRF_UniteParDefaut( graphe->type, AXE_X ) );
  } else {
    GRF_ChangerUnite( graphe, AXE_X, unite_X );
    free( unite_X ); unite_X = NULL;
  }

  unite_Y = BIN_CH_chaine( fichier );
  if ( ( version < 35 ) && ( NULL == unite_Y ) ) {
    GRF_ChangerUnite( graphe, AXE_Y, GRF_UniteParDefaut( graphe->type, AXE_Y ) );
  } else {
    GRF_ChangerUnite( graphe, AXE_Y, unite_Y );
    free( unite_Y ); unite_Y = NULL;
  }

  return 0;
}

/* -------- chargement d'une matrice de covariance
   Renvoie 0 si tout s'est bien passé, != 0 en cas d'erreur
*/

static int ch_covariance(FILE *fichier, Graphe *graphe)
{
  double *covariance = NULL;
  unsigned long n_reel;
  size_t n_lu;

  if ( fichier == NULL ) return -1;
  if ( NULL == graphe ) return -2;

  n_reel = graphe->nombre_point;
  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees )
    n_reel *= 2;

  /* Nombre de valeurs mémorisées pour la matrice de covariance */
  n_reel = ( n_reel * (n_reel + 1) ) >> 1;

  GRF_CreerCovariance( graphe );
  if ( graphe->covariance != NULL ) {
    covariance = graphe->covariance->valeurs;
  }

  if (covariance == NULL) { /* On se contente de sauter */
    fseek(fichier, (int) (n_reel * sizeof(double)), SEEK_CUR);
    return 0; /* On se contente de sauter... */
  }

  n_lu = fread(covariance, sizeof(double), n_reel, fichier);
  VERIFIE(n_reel, FAUX, "ch_covariance");

  HIST_Texte(5, 0, " .. matrice de covariance\n");

  return 0;
}

/* -------- Chargement d'une table de complexe -------- */

static int ch_table_C(FILE *fichier, unsigned long max_points, Complexe *table, double *min, double *max)
{
  unsigned long nbr_points, i;
  double m, M;
  size_t n_lu;

  n_lu = fread( &nbr_points, sizeof( unsigned long ), 1, fichier );
  if ( 0 == nbr_points ) {
    if ( table != NULL ) { /* On initialise à 0 */
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_table_C : Appel avec une table, mais pas de table dans le fichier !\n");
#endif
      for ( i = 0 ; i < max_points ; i++ ) {
	table[ i ].re = 0.0;
	table[ i ].im = 0.0;
      }
    }
    if ( min != NULL ) *min = 0.0;
    if ( max != NULL ) *max = 0.0;

    return 0;
  }

  if ( max_points < nbr_points ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_table_C : Appel avec une table de %lu points, mais table de %lu points dans le fichier !\n",
		   max_points, nbr_points);
#endif
    n_lu = fread( table, sizeof( Complexe ), max_points, fichier );

    /* On saute les points restants */
    fseek( fichier, sizeof( Complexe ) * ( nbr_points - max_points ), SEEK_CUR );
  } else {
    n_lu = fread( table, sizeof( Complexe ), nbr_points, fichier );
  }
  if ( n_lu != nbr_points ) {
    (void) fprintf(stderr, "ch_table_C : ERREUR, lecture de la table\n");
    return -2;
  }

  /* On initialise à zéro la fin de la table, si nécessaire */
  for ( i = nbr_points ; i < max_points ; i++ ) {
    table[ i ].re = 0.0;
    table[ i ].im = 0.0;
  }

  /* Valeurs extrêmes */
  fread( &m, sizeof(double), 1, fichier );
  fread( &M, sizeof(double), 1, fichier );
  if ( min != NULL ) *min = m;
  if ( max != NULL ) *max = M;

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Chargement d'une table de valeurs entières -------- */

static int ch_table_I(FILE *fichier, unsigned long max_points, int *table)
{
  unsigned long nbr_points, i;
  size_t n_lu;

  n_lu = fread( &nbr_points, sizeof( unsigned long ), 1, fichier );
  if ( 0 == nbr_points ) {
    if ( table != NULL ) { /* On initialise à 0 */
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_table_I : Appel avec une table, mais pas de table dans le fichier !\n");
#endif
      for ( i = 0 ; i < max_points ; i++ )
	table[ i ] = 0;
    }

    return 0;
  }

  if ( max_points < nbr_points ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_table_I : Appel avec une table de %lu points, mais table de %lu points dans le fichier !\n",
		   max_points, nbr_points);
#endif
    n_lu = fread( table, sizeof( int ), max_points, fichier );

    /* On saute les points restants */
    fseek( fichier, sizeof( int ) * ( nbr_points - max_points ), SEEK_CUR );
  } else {
    n_lu = fread( table, sizeof( int ), nbr_points, fichier );
  }
  if ( n_lu != nbr_points ) {
    (void) fprintf(stderr, "ch_table_I : ERREUR, lecture de la table\n");
    return -2;
  }

  /* On initialise à zéro la fin de la table, si nécessaire */
  for ( i = nbr_points ; i < max_points ; i++ ) {
    table[ i ] = 0;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Chargement d'une table de valeurs réelles -------- */

static int ch_table_R(FILE *fichier, unsigned long max_points, double *table, double *min, double *max)
{
  unsigned long nbr_points, i;
  double m, M;
  size_t n_lu;

  n_lu = fread( &nbr_points, sizeof( unsigned long ), 1, fichier );
  if ( 0 == nbr_points ) {
    if ( table != NULL ) { /* On initialise à 0 */
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_table_R : Appel avec une table, mais pas de table dans le fichier !\n");
#endif
      for ( i = 0 ; i < max_points ; i++ )
	table[ i ] = 0.0;

      if ( min != NULL ) *min = 0.0;
      if ( max != NULL ) *max = 0.0;
    } else {
      /* On charge une série de valeurs... */
      if ( min != NULL ) *min = 0.0;
      if ( max != NULL ) *max = (double) (max_points - 1);      
    }

    return 0;
  }

  if ( max_points < nbr_points ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_table_R : Appel avec une table de %lu points, mais table de %lu points dans le fichier !\n",
		   max_points, nbr_points);
#endif
    n_lu = fread( table, sizeof( double ), max_points, fichier );

    /* On saute les points restants */
    fseek( fichier, sizeof( double ) * ( nbr_points - max_points ), SEEK_CUR );
  } else {
    n_lu = fread( table, sizeof( double ), nbr_points, fichier );
  }
  if ( n_lu != nbr_points ) {
    (void) fprintf(stderr, "ch_table_R : ERREUR, lecture de la table\n");
    return -2;
  }

  /* On initialise à zéro la fin de la table, si nécessaire */
  for ( i = nbr_points ; i < max_points ; i++ ) {
    table[ i ] = 0.0;
  }

  /* Valeurs extrêmes */
  fread( &m, sizeof(double), 1, fichier );
  fread( &M, sizeof(double), 1, fichier );
  if ( min != NULL ) *min = m;
  if ( max != NULL ) *max = M;

  /* Tout s'est bien passé */
  return 0;
}
