/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Interpolation d'une image, sur une grille prcise ou impose par une autre image
   Lissage d'une image
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "images.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

typedef struct {
  Image *reference;
  unsigned int methode;
} Infos_Interpole_Image;

#define LGMAX_TXT_NOM_METHODE 20 /* Objets dans l'arbre / interface */
#define LGMAX_TXT_NOM_IMAGE   23

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void recalcule_image(Image *source, Image *resultat);

static int GER_IMG_interpole(unsigned int objet, int x, int y);
static void    prepare_arbre_interpole(void);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Interpolation d'une image pour en avoir une plus fine -------- */

Image *IMG_interpole(Image *source, unsigned int nv_larg, unsigned int nv_haut)
{
  char *nom_image;
  Image *image;

  if (source == NULL) return NULL;
  if ( (nv_larg == source->larg) && (nv_haut == source->haut) ) {
    /* Rien  faire */
    return source;
  }
  if ( source->genre != IMG_GENRE_REELLE ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Image non-relle\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  image = IMG_CreerImage( NULL, source, source->genre, source->type, nv_larg, nv_haut );
  if (image == NULL)
    return NULL;
  nom_image = TXT_genere_chaine("%s interpol"e_AIGU"e", source->nom);
  IMG_Titre( image, nom_image );
  free(nom_image);

  /* Nouvelle dimension d'un pixel */
  image->l_reel = ((source->larg - 1) * source->l_reel) / (float) (image->larg - 1);
  image->h_reel = ((source->haut - 1) * source->h_reel) / (float) (image->haut - 1);

  /* Valeurs minimales inchanges */
  image->x_min = source->x_min;
  image->y_min = source->y_min;

  recalcule_image(source, image);
  IMG_bornes(image); /* Les bornes */

  return image; /* On a fini */
}

/* -------- Plaquage d'une image sur une grille impose / interpolation -------- */

Image *IMG_plaquer(Image *ref, Image *source)
{
  Image *resultat;

  if ( (ref == NULL) || (NULL == source) ) return NULL;

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, ref->larg, ref->haut );
  if (resultat == NULL) return NULL;

  /* Dimension des cases : fixe par la rfrence */
  resultat->l_reel = ref->l_reel;
  resultat->h_reel = ref->h_reel;

  /* Valeurs minimales   : fixes par la rfrence */
  resultat->x_min = ref->x_min;
  resultat->y_min = ref->y_min;

  recalcule_image(source, resultat);
  IMG_bornes(resultat); /* Les bornes */
  return resultat;
}

/* -------- Idem, mais le squelette de l'image est fourni... -------- */

int IMG_CalerSurImage(Image *image, Image *originale)
{
  if ( (image == NULL) || (NULL == originale) ) return -1;

  /* Lissage proprement dit */
  recalcule_image( originale, image );

  IMG_bornes( image ); /* Les bornes */

  /* Calcul termin, tout s'est bien pass */
  return 0;
}

/* -------- Calcul des valeurs d'une image sur une grille prdfinie -------- */

static void recalcule_image(Image *source, Image *resultat)
{
  unsigned int i, j;
  boolean dehors;

  if ( (source == NULL) || (resultat == NULL) ) return;

  /* On calcule */
  for (i = 0 ; i < resultat->larg ; i++) {
    float x;

    x = i * resultat->l_reel + resultat->x_min;
    for (j = 0 ; j < resultat->haut ; j++) {
      float y, valeur;

      y = j * resultat->h_reel + resultat->y_min;
      valeur = IMG_lit_valeur(source, x, y, IMG_INTERPOLE_BILINEAIRE, &dehors);
#ifdef _DEVELOPPE_
      if ( VRAI == dehors ) {
	(void) fprintf(stderr, "%s,%s l.%u : Point (%g, %g) [%u, %u] hors de %s\n",
		       __FILE__, __FUNCTION__, __LINE__, x, y, i, j, source->nom);
      }
#endif
      resultat->valeurs.R[ i + j * resultat->larg ] = valeur;
    }
  }
}

/* ~~~~~~~~ Interface graphique ~~~~~~~~ */

Image *IMG_lissage(Image *source)
{
  int reponse;
  unsigned int nv_larg, nv_haut;
  Infos_Interpole_Image infos = {NULL, IMG_INTERPOLE_BILINEAIRE};

  if ( arbres[LISSE_IMAGE].drapeaux.prepare == FAUX )
    prepare_arbre_interpole();

  /* Pas d'image de rfrence par dfaut */
  if ( OBJ_choisit(LISSE_IMAGE, LI_SUR_GRILLE, QUESTION) == VRAI ) { /* On doit donc le dsactiver : on simule un clic */
    GER_IMG_interpole(LI_SUR_GRILLE, 0, 0);
  }

  ARB_MettreComplement( LISSE_IMAGE, &infos );
  reponse = ARB_Bloquant( LISSE_IMAGE, "Lisser l'image", principale, 0 );
  ARB_MettreComplement( LISSE_IMAGE, NULL );
  if ( LI_ANNULER == reponse )
    return NULL;

  if ( OBJ_choisit(LISSE_IMAGE, LI_SUR_GRILLE, QUESTION) == VRAI ) {
    return IMG_plaquer(infos.reference, source);
  }

  if ( OBJ_choisit(LISSE_IMAGE, LI_X_PIXELS, QUESTION) == VRAI )
    nv_larg = OBJ_LireEntier_U( LISSE_IMAGE, LI_VALEUR_X );
  else {
    float pas;

    pas = OBJ_LireNombre( LISSE_IMAGE, LI_VALEUR_X, NULL );
    if (pas <= 0.0 ) {
      return NULL;
    }
    nv_larg = 1 + (unsigned int) rint( (source->l_reel * (float) (source->larg - 1)) / pas );
  }

  if ( OBJ_choisit(LISSE_IMAGE, LI_Y_PIXELS, QUESTION) == VRAI )
    nv_haut = OBJ_LireEntier_U( LISSE_IMAGE, LI_VALEUR_Y );
  else {
    float pas;

    pas = OBJ_LireNombre( LISSE_IMAGE, LI_VALEUR_Y, NULL );
    if (pas <= 0.0 ) {
      return NULL;
    }
    nv_haut = 1 + (unsigned int) rint( (source->h_reel * (float) (source->haut - 1)) / pas );
  }

  return IMG_interpole(source, nv_larg, nv_haut);
}

static int GER_IMG_interpole(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int etat;
  Infos_Interpole_Image *infos;

  infos = (Infos_Interpole_Image *) ARB_LireComplement( LISSE_IMAGE );

  switch (objet)
  {
  case LI_METHODE:   /* Choix de la mthode d'interpolation */
    OBJ_choisit(LISSE_IMAGE, objet, FORCE_OUI);
    OBJ_choisit(LISSE_IMAGE, objet, FORCE_NON);
    return VRAI;

  case LI_SUR_GRILLE: /* Doit-on recaler sur une image ? */
    etat = OBJ_choisit(LISSE_IMAGE, LI_SUR_GRILLE, ECHANGE);
    OBJ_active(LISSE_IMAGE, LI_CHX_IMAGE, etat); /* N'a de sens que si on recale sur une image... */
    if (etat == FAUX) {
      infos->reference = NULL;
      met_chaine(LISSE_IMAGE, LI_CHX_IMAGE, LGMAX_TXT_NOM_IMAGE, "Aucune image choisie", TXT_COPIER);
    }
    etat = !etat;
    OBJ_active(LISSE_IMAGE, LI_X_PIXELS, etat); /* N'a pas de sens si on recale sur une image */
    OBJ_active(LISSE_IMAGE, LI_X_MICRONS, etat);
    OBJ_active(LISSE_IMAGE, LI_VALEUR_X, etat);
    OBJ_active(LISSE_IMAGE, LI_Y_PIXELS, etat);
    OBJ_active(LISSE_IMAGE, LI_Y_PAS, etat);
    OBJ_active(LISSE_IMAGE, LI_VALEUR_Y, etat);
    return VRAI;

  case LI_CHX_IMAGE: /* Choix de l'image de rfrence pour la grille */
    OBJ_choisit(LISSE_IMAGE, objet, FORCE_OUI);
    OBJ_choisit(LISSE_IMAGE, objet, FORCE_NON);
    return VRAI;
  }

  return FAUX;
}

static void prepare_arbre_interpole(void)
{
  /* Optimisation des positions des objets */
  OBJ_deplace(LISSE_IMAGE, LI_METHODE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(LISSE_IMAGE, LI_CHX_IMAGE, 0, +2, DEPLACE_PIXELS);

  /* Fonction de gestion */
  arbres[LISSE_IMAGE].fonction.arbre = GER_IMG_interpole;
  arbres[LISSE_IMAGE].drapeaux.prepare = VRAI;
}
