/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Cration de profils, coupes,...
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "txt_utils.h"
#include "graphes.h"

#include "exafs.h"
#include "images.h"
#include "science.h" /* Pour M_PI, si ncessaire */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_TXT_NOM_IMAGE 47

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean verifie_angle         (Image *image, double angle, int position);
static boolean Profil_Position_Depart(Image *image, double angle_rad, int position, double *x_depart, double *y_depart);
static boolean Profil_Position_Fin   (Image *image, double angle_rad, int position, double *x_depart, double *y_depart);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Renvoie VRAI si la demi-droite partant de l'origine et avec l'angle indiqu coupe l'image */

static boolean verifie_angle(Image *image, double angle, int position)
{
/*  double angle_min, angle_max;*/ /* Angles limites, en radians */

  if (NULL == image) return FAUX;

  switch ( position )
  {
  case IMG_CENTRE_DANS_IMAGE: /* Tous les angles sont valables */
    return VRAI;

  case IMG_CENTRE_COIN_HG: /* Angles entre 270 et 360 */
    if ( (angle == 0.0) || (angle >= 270.0) ) return VRAI;
    break;

  case IMG_CENTRE_COIN_HD:
    if ( (angle >= 180.0) && (angle <= 270.0) ) return VRAI;
    break;

  case IMG_CENTRE_COIN_BG:
    if ( angle <= 90.0 ) return VRAI;
    break;

  case IMG_CENTRE_COIN_BD:
    if ( (angle >= 90.0) && (angle <= 180.0) ) return VRAI;
    break;

    /* RESTE A FAIRE LES AUTRES POSITIONS DU CENTRE ! */
  default:
    fprintf(stderr, "IMG_Position_Centre : position non trouve !\n");
  }
  return FAUX;
}

/* -------- Position du premier point, pour un profil radial -------- */

static boolean Profil_Position_Depart(Image *image, double angle_rad, int position, double *x_depart, double *y_depart)
{
  if ( NULL == image ) return FAUX;
  if ( (NULL == x_depart) || (NULL == y_depart) ) return FAUX;

#ifdef _DEVELOPPE_
  printf("Profil_Position_Depart : image = %p, angle_rad = %g, position = %d, x_depart = %p, y_depart = %p\n",
	 (void *) image, angle_rad, position, (void *) x_depart, (void *) y_depart);
#endif

  switch ( position )
  {
  case IMG_CENTRE_DANS_IMAGE: /* Centre appartient  l'image -> on part du centre... */
  case IMG_CENTRE_COIN_HG:
  case IMG_CENTRE_COIN_HD:
  case IMG_CENTRE_COIN_BG:
  case IMG_CENTRE_COIN_BD:
  case IMG_CENTRE_BORD_B:
  case IMG_CENTRE_BORD_H:
  case IMG_CENTRE_BORD_G:
  case IMG_CENTRE_BORD_D:
    *x_depart = 0.0;
    *y_depart = 0.0;
    break;

/* A FAIRE : les autres cas (centre hors de l'image) */
  }

  return VRAI; /* Tout s'est bien pass */
}

static boolean Profil_Position_Fin(Image *image, double angle, int position, double *x_fin, double *y_fin)
{
  double angle_rad;
  double x_lim = 0.0, y_lim = 0.0; /* La frontire qu'il ne faut pas franchir... */
  double y_px, x_py;

  if ( NULL == image ) return FAUX;
  if ( (NULL == x_fin) || (NULL == y_fin) ) return FAUX;

  angle_rad = angle * M_PI / 180.0;

  if ( angle == 0.0 ) { /* Coupe horizontale */
    *x_fin = image->x_min + (image->larg - 1) * image->l_reel;
    *y_fin = 0.0;
    return VRAI;
  } else if ( angle == 90.0 ) { /* Coupe verticale */
    *x_fin = 0.0;
    *y_fin = image->y_min + (image->haut - 1) * image->h_reel;
    return VRAI;
  } else if ( angle == 180.0 ) { /* Coupe horizontale */
    *x_fin = image->x_min;
    *y_fin = 0.0;
    return VRAI;
  } else if ( angle == 270.0 ) { /* Coupe verticale */
    *x_fin = 0.0;
    *y_fin = image->y_min;
    return VRAI;
  }

  /* Ici, on sait que l'on coupe en biais */
  /* On cherche les bords de l'image que l'on risque de couper */
  if ( VRAI == IMG_CentreEstDansImage(position) ) {
    if ( angle < 90.0 ) { /* Premier quadrant */
      x_lim = image->x_min + (image->larg - 1) * image->l_reel;
      y_lim = image->y_min + (image->haut - 1) * image->h_reel;
    } else if ( angle < 180.0 ) { /* Deuxime quadrant */
      x_lim = image->x_min;
      y_lim = image->y_min + (image->haut - 1) * image->h_reel;
    } else if ( angle < 270.0 ) { /* Troisime quadrant */
      x_lim = image->x_min;
      y_lim = image->y_min;
    } else { /* Dernier quadrant */
      x_lim = image->x_min + (image->larg - 1) * image->l_reel;
      y_lim = image->y_min;
    }
  } else {
    /* A FAIRE : CAS DU CENTRE HORS DE L'IMAGE */
  }

  /* Position des deux points d'intersection */
  y_px = x_lim * tan( angle_rad ); /* Y de l'intersection de la droite de coupe avec le bord vertical de l'image */
  x_py = y_lim / tan( angle_rad ); /* X de l'intersection de la droite de coupe avec le bord horizontal de l'image */
  if ( fabs(y_px) > fabs(y_lim) ) { /* C'est le bord horizontal qui limite */
    *x_fin = x_py;
    *y_fin = y_lim;
  } else { /* C'est le bord vertical qui limite */
    *x_fin = x_lim;
    *y_fin = y_px;
  }

#ifdef _DEVELOPPE_
  printf("Dimensions de l'image (%f, %f) - (%f, %f)\n", image->x_min, image->y_min,
	 image->x_min + (image->larg - 1) * image->l_reel, image->y_min + (image->haut - 1) * image->h_reel);
  printf("Point d'arrt de la coupe : (%f, %f)\n", *x_fin, *y_fin);
#endif

  return VRAI;
}

/* ~~~~~~~~ Interface graphique ~~~~~~~~ */

Graphe *IMG_profil(Image *image)
{
  int retour, i;
  double pas;
  Graphe *profil = NULL;

  if (image == NULL) return NULL;

  met_chaine(CALC_PROFIL, CP_IMAGE, LGMAX_TXT_NOM_IMAGE, image->nom, TXT_COPIER);

  retour = ARB_Bloquant( CALC_PROFIL, "Estimer un profil", principale, 0 );
  if (retour == CP_ANNULER) return NULL;

  pas = OBJ_LireNombre(CALC_PROFIL, CP_PAS, NULL );
  if (retour == CP_OK) { /* Calcul pour un angle donn */
    double angle;

    /* On travaille ici en degrs -> OBJ_LireAngle pas utilisable */
    angle = OBJ_LireNombre(CALC_PROFIL, CP_ANGLE, NULL );
    /* Il faudrait, si on le veut, effacer le profil prcdent */
    profil = init_profil(image, (int) rint(angle), pas);
    if (image == image_courante) {
      IMG_AllerImage( image ); /* Pour tout remettre  zro */
    }
    return profil;
  }

  /* Ici, on cherche  trouver le profil d'angle optimal */
  for (i = 0 ; i < 360 ; i += 5) { /* Par pas de 5 */
    Graphe *nv_profil;

    nv_profil = init_profil(image, i, pas);
    if (nv_profil == NULL)
      return NULL;

    if ( (profil == NULL) || ( (nv_profil->y_max - nv_profil->y_min) > ( profil->y_max - profil->y_min ) ) ) {
      GRF_efface(profil, VRAI, VRAI); /* On efface l'ancien meilleur */
      profil = nv_profil;
    }
    if (profil != nv_profil) /* On ne le garde pas comme meilleur profil */
      GRF_efface(nv_profil, VRAI, VRAI);
  }
/*   image->profil = profil; */
  return profil;
}

/* -------- Gnrer un profil  partir d'une image --------------------------\
 |   Renvoie NULL en cas d'erreur                                            |
 |                                                                           |
 | image = l'image dont on veut extraire le profil                           |
 | angle = l'angle par rapport  l'horizontale, en degrs                    |
 |   valable entre 0 et 360 degrs                                           |
 | pas = pas  utiliser pour le calcul...                                    |
 |   si absurde (<=0), on essaye de trouver un pas raisonnable               |
 \-------------------------------------------------------------------------- */

Graphe *init_profil(Image *image, int angle, double pas)
{
  int position_centre;
  double angle_rad;
  double x_depart, y_depart, x_fin, y_fin;
  Graphe *profil;
  Profil *infos_profil;
  char *titre;

  if (image == NULL) return NULL;

  /* On se ramne dans 0-360 */
  while (angle >= 360) angle -= 360;
  while (angle < 0) angle += 360;

  /* Contrle de l'angle */
  position_centre = IMG_Position_Centre(image);
  if ( position_centre < 0 ) {
    fprintf(stderr, "Image %s : position du centre non trouve !\n", image->nom);
    return NULL;
  }

  if ( verifie_angle(image, angle, position_centre) == FAUX ) {
    fprintf(stderr, "Image %s : l'angle indiqu (%d) ne permet pas de couper l'image !\n",
	    image->nom, angle);
    return NULL;
  }

  /* Conversion de l'angle en radians */
  angle_rad = M_PI / 180.0 * angle ;

  /* Position des deux points limitant le profil */
  Profil_Position_Depart(image, angle_rad, position_centre, &x_depart, &y_depart);
  Profil_Position_Fin(image, angle, position_centre, &x_fin, &y_fin);

  /* On fait la coupe entre ces deux points */
  profil = IMG_coupe(image, x_depart, y_depart, x_fin, y_fin, pas);
  if (profil == NULL) return NULL;

  titre = TXT_genere_chaine("%s, profil "a_GRAVE" %d"DEGRE, image->nom, angle);
  GRF_titre(profil, titre);
  free(titre);
  GRF_bornes(profil, ORDONNE, DESORDONNE);

  infos_profil = (Profil *) malloc( sizeof(Profil) );
  if (infos_profil == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Images/profils.c", "init_profil", "infos_profil");
  } else {
    profil->pointeur.profil = infos_profil;
    infos_profil->unite = pas;
    infos_profil->angle = angle; /* En degrs */
  }

  /* On essaye d'attribuer la coupe  l'image */
  if ( image->infos_affichage != NULL ) {
    if ( NULL == image->infos_affichage->coupe ) {
      IMG_AssocierCoupe( image, profil, x_depart, y_depart, x_fin, y_fin );
    }
  }

  return profil; /* Tout s'est bien passe */
}

/* -------- Ralise une coupe de l'image considre, avec le bon profil... --------\
 |  Renvoie NULL en cas d'erreur, un graphe de type Profil sinon                   |
 |
 | image  : l'image  couper
 | x1, y1 : point de dpart de la coupe (en coordonnes mtriques)
 | x2, y2 : point de fin de la coupe (en coordonnes mtriques)
 | pas    : le pas  utiliser pour la coupe (en microns)
 \-------------------------------------------------------------------------------- */

Graphe *IMG_coupe(Image *image, double x1, double y1, double x2, double y2, double pas)
{
  unsigned long n_pts, i;
  register double *x, *y;
  double lg_coupe;
  char *titre;
  Graphe *coupe;

  if (image == NULL) return NULL;
  pas = fabs(pas);

  lg_coupe = sqrt( (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1) );
  if ( lg_coupe < 3.0 * pas ) { /* Points trops proches ! */
    (void) fprintf(stderr, "IMG_coupe : longueur de coupe trop courte [%g / %g minimum]\n", lg_coupe, 3.0*pas);
    (void) fprintf(stderr, "            [points : (%g, %g) -- (%g, %g)]\n", x1, y1, x2, y2);
    return NULL;
  }
  if ( 0.0 == pas ) { /* Pas nul -> dtection automatique du pas : arbitraire, 101 points dans la coupe (= pas du 100e ) */
    pas = lg_coupe / 100.0;
  }

  n_pts = (unsigned long) rint(lg_coupe / pas);
  titre = TXT_genere_chaine("%s : coupe de (%f, %f) "a_GRAVE" (%f, %f)", image->nom, x1, y1, x2, y2);
  coupe = GRF_nouveau( NULL, n_pts, GR_PROFIL, GR_NORMAL, titre );
  free(titre);
  if (coupe == NULL) return NULL;
  printf("Longueur de coupe : %f , pas %f  => %lu points\n", lg_coupe, pas, n_pts);

  if (x2 == x1) { /* Coupe verticale */
    if (y2 < y1) pas = -pas;
    for (i = 0, x = coupe->x, y = coupe->y ; i < coupe->nombre_point ; i++, x++, y++) {
      *x = fabs(pas * (double) i);
      *y = IMG_lit_valeur(image, x1, y1 + pas * (double) i, IMG_INTERPOLE_BILINEAIRE, NULL);
    }
  } else if (y2 == y1) { /* Coupe horizontale */
    if (x2 < x1) pas = -pas;
    for (i = 0, x = coupe->x, y = coupe->y ; i < coupe->nombre_point ; i++, x++, y++) {
      *x = fabs(pas * (double) i);
      *y = IMG_lit_valeur(image, x1 + pas * (double) i, y1, IMG_INTERPOLE_BILINEAIRE, NULL);
    }
  } else {
    float dx, dy;

    dx = pas * (x2 - x1) / lg_coupe;
    dy = pas * (y2 - y1) / lg_coupe;
    for (i = 0, x = coupe->x, y = coupe->y ; i < coupe->nombre_point ; i++, x++, y++) {
      *x = pas * (double) i;
      *y = IMG_lit_valeur(image, x1 + dx * (double) i, y1 + dy * (double) i,
			  IMG_INTERPOLE_BILINEAIRE, NULL);
/*      printf("Valeur en d = %f (%f, %f) : %f\n", *x, x1 + dx * (float) i, y1 + dy * (float) i, *y); */
    }
  }

  GRF_bornes(coupe, ORDONNE, DESORDONNE);
  return coupe;
}
