/* LASE - Logiciel d'Analyse des Spectres EXAFS
   Tout ce qui concerne le filtrage (en gros, les transformées de Fourier)
   (c) Emmanuel CURIS, 1998-1999
   Création : 4 mai 1998
*/

#ifndef FOURIER_H_
# define FOURIER_H_ /* Contre inclusion multiple */
# include "graphes.h"   /* Pour les graphes */
# include "interface.h" /* Pour GRF_liste   */

/* Correspondance a(i, j) <-> index du tableau associé pour une matrice triangulaire inférieure */

#define TF_element(i, j) ( ( ( (i) * ((i) + 1) ) >> 1 ) + (j) )

/* ~~~~~~~~~ Transformée de Fourier ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* --------- Infos nécessaires au calcul... --------------------- */

/* ---- Description d'un filtre = fonction d'apodisation ---- */

typedef struct infos_fenetre
{
  int type;      /* Le type de la fenêtre                    */
  double parametre;  /* Paramètre pour le calcul de la fenêtre   */
  double x_min, x_max;   /* Les extrêmes pour la fenêtre             */
  double x_1, x_2;       /* Les internes, pour fenêtre asymétrique   */
  boolean symetrique;    /* VRAI si la fenêtre est symétrique        */
} Infos_FonctionApodisation;

typedef struct tf_infos
{
  Graphe *graphe;        /* Graphe dont on veut la TF                */
  Graphe *fenetre;       /* La fenêtre, calculée                     */
  Graphe *apercu;        /* Le graphe qui contient l'aperçu...       */
  struct infos_fenetre apodisation;  /* Informations sur la fonction d'apodisation */
  unsigned int num_depart; /* Pour affichage du sélecteur              */
  Graphe *amplitude;     /* Correction d'amplitude                   */
  Graphe *phase;         /* Correction des phases                    */
} TF_infos;

#define TF_CALC_RAPIDE 1 /* TF rapide, 2^n points */
#define TF_CALC_NORMAL 2 /* Intégration directe */
#define TF_CALC_FFTW   3 /* TF rapide, bibliothèque FFTW */

#define TF_SENS_DIRECT  -1 /* Calcul de la TF  */
#define TF_SENS_INVERSE -2 /* Calcul de la TFI */

Graphe *TF_CalculDirect(TF_infos *infos, double r_min, double r_max, double pas,
			int methode, boolean erreurs, boolean covariance);

/* --------- Paramétrage de la façon de faire la TF ------------- */

Graphe *TF_ouvre(Graphe *graphe, boolean chaine); /* Le graphe dont on veut la TF */
boolean TF_init (TF_infos *infos_tf);

/* --------- Initialisation de la TF ---------------------------- */

Graphe *TF_IsoleSignal(Graphe *original, struct infos_fenetre *filtre);
Graphe *TF_prep_normal(double r_min, double r_max, double pas, int sens, boolean erreurs, boolean covariance);
Graphe *TF_prep_rapide(Graphe *original, double r_min, double r_max, double pas, int sens);

/* --------- Corrections avant la TF ---------------------------- */

boolean TF_coramp(Graphe *graphe, Graphe *correct); /* d'amplitude */
boolean TF_corpha(Graphe *graphe, Graphe *correct); /* de phase    */

/* --------- Calcul de la TF proprement dit -------------------- */

void TF_simple(Graphe *source, Graphe *fourier); /* Méthode bête et méchante */
void TF_rapide(Graphe *source, Graphe *fourier); /* Méthode de la TF rapide  */

/* ~~~~~~~~~ Transformée de Fourier inverse ~~~~~~~~~~~~~~~~~~~~ */

typedef struct
{
  Graphe *graphe,      /* Le graphe dont on veut une TF inverse   */
         *fenetre;     /* La fenêtre utilisée pour le calcul      */
  struct infos_fenetre filtre; /* Infos sur le filtre */
  int    num_depart,   /* Pour le choix de la fenêtre de la TF    */
         num_choisi;
} TFI_infos;

typedef struct nyquist
{
  double r_min, r_max; /* Bornes de la TFI pour pouvoir appliquer le théorème de Nyquist */
} Infos_Nyquist;

/* --------- Paramétrage de la façon de faire la TFI ----------- */

void    TFI_ouvre(Graphe *graphe);
boolean TFI_init (TFI_infos *infos);

/* --------- Initialisations du calcul ------------------------- */

Graphe *TFI_CalculInverse(TFI_infos *infos, double k_min, double k_max, double pas,
			  int methode, boolean erreurs, boolean covariance);

/* --------- Calcul effectif de la TF inverse ------------------ */

int TFI_Trapezes(Graphe *source, Graphe *resultat);

void TFI_simple(Graphe *source, Graphe *fourier); /* Méthode simple */
void TFI_rapide(Graphe *source, Graphe *fourier); /* Méthode rapide */

void TFI_fenetre(Graphe *fourier, TF_infos *infos); /* On déconvolue la fenêtre */

/* -------- Corrections après la TFI -------- */

Graphe *TFI_CorrigerPhase(Graphe *graphe, int sens, double longueur); /* Correction de la phase */
Graphe *corr_amplitude(Graphe *graphe, double n_voisin, double sigma, double longueur);

/* ~~~~~~~~~ Calcul des fenêtres pour la TF et la TFI ~~~~~~~~~~ */

/* Les fenêtres de TF utilisables - index du tableau Types_fenetres dans fourier.c */

#define PAS_DE_FENETRE 0 /* Basique...    */
#define FEN_RECTANGLE  1 /* même remarque */
#define FEN_BESSEL     2 /* Kaiser-Bessel, symétrique, sans plateau */
#define FEN_GAUSSIENNE 3 /* Gaussienne, symétrique, sans plateau    */
#define FEN_HANNING    4 /* Cas particulier de Hamming, pour alpha = 0,5 */
#define FEN_HAMMING    5 /* Symétrique ou asymétrique */
#define FEN_NORTON_F3  6 /* Norton-Beer, F3, symétrique */
#define FEN_BARTLETT   7 /* Linéaire, symétrique ou asymétrique */
#define FEN_BLACKMAN   8 /* ??? */

#define NOMBRE_FENETRES 8 /* Nombre total de fenêtres disponibles */

extern const char *Types_fenetres[]; /* Noms des fenêtres - dans variables.c */

/* ~~~~~~~~~ Quelques macros pour simplifier les choses ~~~~~~~~~~~~ */

#define Autorise_Symetrique(i) ( ((i) == FEN_HANNING) || ((i) == FEN_HAMMING) \
                              || ((i) == FEN_BARTLETT) )
#define Avec_Parametre(i) ( ((i) == FEN_BESSEL) || ((i) == FEN_GAUSSIENNE) || ((i) == FEN_HAMMING) \
                         || ((i) == FEN_NORTON_F3) )

void   TF_fenetre(Graphe *fenetre, Infos_FonctionApodisation *infos_fenetre); /* Complète */

double TF_calcfen(double x, Infos_FonctionApodisation *infos_fenetre); /* En un point */

	/* Macros de calcul des fenêtres un peu complexes... */

int TFEN_bartlett(Graphe *fenetre, double k_1, double k_2, double k_3, double k_4);
int TFEN_hamming (Graphe *fenetre, double k_1, double k_2, double k_3, double k_4, double param);
int TFEN_hanning (Graphe *fenetre, double k_1, double k_2, double k_3, double k_4);

#ifdef _DEVELOPPE_
void verifie_covariance(double *covariance, unsigned long nombre_point);
#endif

/* ~~~~~~~~ Pour l'automatisation de la reproduction ~~~~~~~~ */

struct infos_reproduit_filtre
{
  Graphe *corr_amplitude;
  Graphe *corr_phase;
  int type_fen_tf;
  int type_fen_tfi;
  boolean tf_symetrique, tfi_symetrique;
};

void TF_reproduit(Graphe *graphe);

#endif /* FOURIER_H_ */
