/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2007, Emmanuel CURIS
   Création : janvier 2007

   Interface graphique avec l'utilisateur

   Options sur les absorbeurs (en particulier l'absorbeur courant et ses modèles)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "couches.h"		/* Pour lire les chemins FEFF ! */
#include "feff.h"
#include "interface.h"
#include "modeles.h"

#include "int_absorbeur.txt"

/* -------- Variables locales -------- */

struct infos_absorbeur {
  Absorbeur *absorbeur;		/* L'absorbeur en cours d'édition */
  unsigned long modele_actif;	/* Pour le défileur de modèles */
  unsigned long premier_modele_affiche;
};

#define LGMAX_NOM_MODELE 20	/* À vérifier dans le RSC */

/* -------- Fonctions locales -------- */

static int  I_MOL_GererEditerAbsorbeur(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)));
static void I_MOL_LireContribution    (Variable *variable);
static void I_MOL_MettreContribution  (Variable *variable);
static void I_MOL_PreparerArbreEditeAbsorbeur(void);
static void I_MOL_PreparerListeModeles(struct infos_absorbeur *infos);

static const char *I_MOL_Modele_NVTexte(unsigned int arbre, unsigned long numero_objet);

/* ~-~-~-~-~-~-~-~-~ Modification des paramètres d'un absorbeur ~-~-~-~-~-~-~-~-~ */

/* -------- Appel depuis l'extérieur -------- */

int I_MOL_EditerAbsorbeur(Absorbeur *absorbeur)
{
  int retour;
  struct infos_absorbeur infos = { NULL, 0UL, 0UL };

  if ( NULL == absorbeur ) return 0;

  /* On prépare l'arbre, au besoin */
  if ( FAUX == arbres[ EDITE_ABSORBEUR ].drapeaux.prepare ) {
    I_MOL_PreparerArbreEditeAbsorbeur( );
  }

  /* On initialise l'arbre */
  infos.absorbeur = absorbeur;

  /* a. Coordonnées de l'absorbeur & autres informations « chimiques » */
  OBJ_MettreCoordonnees( EDITE_ABSORBEUR, CABS_X, CABS_Y, CABS_Z,
			  &(absorbeur->absorbeur->position), absorbeur->absorbeur->err_position );
  OBJ_CopierTexte( EDITE_ABSORBEUR, CABS_NUMERO_ATOM,
		   ATM_Z_vers_symbole( absorbeur->absorbeur->num_atomique ), 3 );
  
  /* b. Nom de l'absorbeur, modifiable (texte infini) */
  OBJ_CopierTexte( EDITE_ABSORBEUR, CABS_NOM, absorbeur->nom, 0 );

  /* c. Contribution au spectre de la molécule */
  I_MOL_MettreContribution( &(absorbeur->contribution) );

  /* d. Modèles associés à l'absorbeur */
  I_MOL_PreparerListeModeles( &infos );

  /* On associe les infos à l'arbre */
  ARB_MettreComplement( EDITE_ABSORBEUR, &infos );
  
  /* On appelle l'arbre  */
  retour = ARB_Bloquant( EDITE_ABSORBEUR, titre_arbre_edite_absorbeur, fen_vue_3d, 0 );

  /* Plus besoin des infos */
  ARB_MettreComplement( EDITE_ABSORBEUR, NULL );

  if ( CABS_ANNULER == retour ) {
    /* Modification annulée... */
    return 0;
  }

  /* On lit les nouvelles valeurs pour l'absorbeur */
  /* a. Position, Z : pour info uniquement => non modifiables */
  /* b. Nom de l'absorbeur */
  lit_chaine( EDITE_ABSORBEUR, CABS_NOM, 0, NULL, TXT_REMPLACE );

  /* c. Contribution */
  I_MOL_LireContribution( &(absorbeur->contribution) );

  /* d. Modèles : modifiés en temps réel, pas d'annulation possible */

  /* On a terminé, sans souci */
  return 0;
}

/* -------- Gestion de l'arbre -------- */

static int I_MOL_GererEditerAbsorbeur(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int retour;
  struct infos_absorbeur *infos;

  infos = (struct infos_absorbeur *) ARB_LireComplement( EDITE_ABSORBEUR );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associées à l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  switch ( objet )
  {
  case CABS_CONSTANTE:
  case CABS_VARIABLE:
    OBJ_choisit( EDITE_ABSORBEUR, objet, FORCE_OUI);
    OBJ_active ( EDITE_ABSORBEUR, CABS_MIN, objet == CABS_VARIABLE );
    OBJ_active ( EDITE_ABSORBEUR, CABS_MAX, objet == CABS_VARIABLE );
    OBJ_active ( EDITE_ABSORBEUR, CABS_PAS, objet == CABS_VARIABLE );
    return BI_EVENEMENT_TRAITE;

  case CABS_RESULTAT: /* On réutilise les résultats obtenus - variable choisie uniquement */
    OBJ_MettreNombre( EDITE_ABSORBEUR, CABS_CONTRIB_INI,
		      VAR_LireValeur( &(infos->absorbeur->contribution), VARIABLE_LIRE_VALEUR_AJUSTEE ) * 100.0, 0.0 );
    return BI_EVENEMENT_TRAITE;

  case CABS_NOUVEAU:		/* Créer un nouveau modèle */
    OBJ_choisit( EDITE_ABSORBEUR, objet, FORCE_OUI );
    retour = I_ModeleAbsorbeur_Nouveau( infos->absorbeur, infos->modele_actif );
    if ( retour != 0 ) {
      AVERTISSEMENT( "Erreur en voulant créer un modèle... [Code %d]", retour );
    }
    /* On refait la liste des modèles */
    I_MOL_PreparerListeModeles( infos );
    OBJ_choisit( EDITE_ABSORBEUR, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Lire les valeurs de la contribution -------- */

static void I_MOL_LireContribution(Variable *variable)
{
  /* La valeur initiale [stockée en fraction, affichée en pourcentage */
  variable->valeur_initiale = OBJ_LireNombre( EDITE_ABSORBEUR, CABS_CONTRIB_INI, NULL ) / 100.0;

  if ( OBJ_choisit( EDITE_ABSORBEUR, CABS_CONSTANTE, QUESTION ) ) {
    variable->type = VAR_CONSTANTE;
  } else {
    variable->type = VAR_LIBRE;

    /* On lit donc min, max et pas */
    (void) OBJ_LireIntervalle( EDITE_ABSORBEUR, CABS_MIN, CABS_MAX, CABS_PAS, ROOT,
			       &(variable->min), &(variable->max), &(variable->delta), NULL );
    /* Conversion en fraction */
    variable->min /= 100.0;
    variable->max /= 100.0;
    variable->delta /= 100.0;
  }
}

/* -------- Mettre dans l'arbre les valeurs de la contribution -------- */

static void I_MOL_MettreContribution(Variable *variable)
{
  /* La valeur initiale */
  OBJ_MettreNombre( EDITE_ABSORBEUR, CABS_CONTRIB_INI, (variable->valeur_initiale) * 100.0, 0.0 );

  /* La contribution est stockée en fraction [0;1], mais affichée en pourcentage */
  if ( VAR_CONSTANTE == variable->type ) {
    OBJ_choisit( EDITE_ABSORBEUR, CABS_CONSTANTE, FORCE_OUI );
    OBJ_choisit( EDITE_ABSORBEUR, CABS_VARIABLE , FORCE_NON );

    OBJ_active( EDITE_ABSORBEUR, CABS_MIN, FORCE_NON );
    OBJ_active( EDITE_ABSORBEUR, CABS_MAX, FORCE_NON );
    OBJ_active( EDITE_ABSORBEUR, CABS_PAS, FORCE_NON );
  } else {
    OBJ_choisit( EDITE_ABSORBEUR, CABS_CONSTANTE, FORCE_NON );
    OBJ_choisit( EDITE_ABSORBEUR, CABS_VARIABLE , FORCE_OUI );

    OBJ_active( EDITE_ABSORBEUR, CABS_MIN, FORCE_OUI );
    OBJ_active( EDITE_ABSORBEUR, CABS_MAX, FORCE_OUI );
    OBJ_active( EDITE_ABSORBEUR, CABS_PAS, FORCE_OUI );

    /* Seulement dans ce cas, min/max/pas est intéressant */
    OBJ_MettreIntervalle( EDITE_ABSORBEUR, CABS_MIN, CABS_MAX, CABS_PAS, ROOT,
			  (variable->min) * 100.0, (variable->max) * 100.0, (variable->delta) * 100.0, 0 );
  }
}

/* -------- Installer la liste des modèles & adapter les boutons en conséquence -------- */

static void I_MOL_PreparerListeModeles(struct infos_absorbeur *infos)
{
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associées à l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  /* Cas 1 : aucun modèle */
  if ( 0 == infos->absorbeur->nbr_modeles ) {
    OBJ_active( EDITE_ABSORBEUR, CABS_COPIER , FORCE_NON );
    OBJ_active( EDITE_ABSORBEUR, CABS_EFFACER, FORCE_NON );

    OBJ_active( EDITE_ABSORBEUR, CABS_TYPE_MODELE, FORCE_NON );

    OBJ_active( EDITE_ABSORBEUR, CABS_CHEMINS, FORCE_NON );

    infos->modele_actif = 0;
    infos->premier_modele_affiche = 0;
  } else {
    OBJ_active( EDITE_ABSORBEUR, CABS_COPIER , FORCE_OUI );
    OBJ_active( EDITE_ABSORBEUR, CABS_EFFACER, FORCE_OUI );
    
    OBJ_active( EDITE_ABSORBEUR, CABS_TYPE_MODELE, FORCE_OUI );
  }

  /* On prépare le défileur en fonction de cela */
  ARB_PreparerDefileur_O( EDITE_ABSORBEUR, CABS_MODELE_1, CABS_MODELE_N, LGMAX_NOM_MODELE,
			  CABS_MODELE_AVAN, CABS_FOND_ASCE, CABS_BOUTON_ABS, CA_MODELE_APRES,
			  infos->absorbeur->nbr_modeles, NULL, infos->modele_actif, infos->premier_modele_affiche,
			  I_MOL_Modele_NVTexte );
}

/* -------- Défileur : renvoie les textes demandés -------- */

static const char *I_MOL_Modele_NVTexte(unsigned int arbre, unsigned long numero_objet)
{
  struct infos_absorbeur *infos;
  struct modele_ajuste *modele;
  unsigned long i;

  infos = (struct infos_absorbeur *) ARB_LireComplement( arbre );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associées à l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  if ( numero_objet >= infos->absorbeur->nbr_modeles ) {
    AVERTISSEMENT( "Appel pour objet %lu, mais %u seulement", numero_objet, infos->absorbeur->nbr_modeles );
    return NULL;
  }

  
  for ( i = 0, modele = infos->absorbeur->modeles ;
	i < numero_objet ;
	i++, modele = modele->suivant );

  return modele->nom_modele;
}

/* -------- Préparation de l'arbre -------- */

static void I_MOL_PreparerArbreEditeAbsorbeur(void)
{
  /* Conversions en sous-menus */
  OPTION_A_FAIRE( "Sous-menu des types de modèles" );

  /* Adaptations automatiques */
  if ( FAUX == arbres[ EDITE_ABSORBEUR ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation, boîte de dialogue - ajustement d'un atome marqué comme absorbeur" );
    retour = ARB_Initialiser( EDITE_ABSORBEUR, NULL, I_MOL_GererEditerAbsorbeur, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  /* Ajustements graphiques */
  OBJ_deplace( EDITE_ABSORBEUR, CABS_TXT_Y, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EDITE_ABSORBEUR, CABS_Y    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EDITE_ABSORBEUR, CABS_TXT_Z, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( EDITE_ABSORBEUR, CABS_Z    , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( EDITE_ABSORBEUR, CABS_VARIABLE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EDITE_ABSORBEUR, CABS_TXT_MIN , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( EDITE_ABSORBEUR, CABS_MIN     , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( EDITE_ABSORBEUR, CABS_TXT_MAX , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( EDITE_ABSORBEUR, CABS_MAX     , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( EDITE_ABSORBEUR, CABS_TXT_PAS , 0, +3, DEPLACE_PIXELS);
  OBJ_deplace( EDITE_ABSORBEUR, CABS_PAS     , 0, +3, DEPLACE_PIXELS);

  ARB_EnjoliverDefileur( EDITE_ABSORBEUR, CABS_FOND_MODEL, CABS_MODELE_1, CABS_MODELE_N,
			 CABS_MODELE_AVAN, CABS_FOND_ASCE, CABS_BOUTON_ABS, CA_MODELE_APRES, ARB_ASCENSEUR_DROITE );

  /* Traduction, au besoin */
#ifdef AVEC_TRADUCTION_
#endif

  /* Terminé */
  arbres[ EDITE_ABSORBEUR ].drapeaux.prepare = VRAI;
}

/* ~-~-~-~-~-~-~-~-~ Opérations de base autour d'un absorbeur ~-~-~-~-~-~-~-~-~ */

/* -------- Effacer un absorbeur -------- */

int I_MOL_EffacerAbsorbeur(Molecule *molecule, unsigned int numero_absorbeur)
{
  Atome *atome_absorbeur;
  int reponse;

  if ( NULL == molecule ) return 0;
  if ( numero_absorbeur >= molecule->nbr_absorbeurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Numéro d'absorbeur incorrect [%u, max %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_absorbeur, molecule->nbr_absorbeurs - 1 );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* On demande confirmation de l'effacement */
  reponse = BI_ERR_alerte( principale, &Q_EffacerAbsorbeur, VRAI );
  if ( I_ABS_EFFACER_ANNULER == reponse ) return +1;  /* Annulation */

  /* On note quel atome est cet absorbeur */
  atome_absorbeur = molecule->absorbeurs[ numero_absorbeur ].absorbeur;

  /* On efface l'absorbeur */
  ABS_oter( molecule, numero_absorbeur, FAUX );

  /* Si demandé, on efface aussi l'atome associé */
  if ( I_ABS_EFFACER_ATOME == reponse ) {
    ATM_efface( molecule, atome_absorbeur, VRAI );
  }

  /* Effacement réussi */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Modèles pour un absorbeur ~-~-~-~-~-~-~-~-~ */

/* -------- Interface générique : demande aussi le type,... -------- */

int I_ModeleAbsorbeur_Nouveau(Absorbeur *absorbeur, unsigned long numero_modele_avant)
{
  int retour;

  if ( NULL == absorbeur ) {
    /* On travaille sur l'absorbeur courant */
    if ( NULL == mol_courante ) {
      (void) fprintf(stderr, "%s, %s l.%u : Aucune molécule en mémoire !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
    if ( NULL == mol_courante->absorbeurs ) {
      OPTION_A_FAIRE( "La molécule ne contient pas d'absorbeur ! Proposer d'en créer un" );
      return -2;
    }

    absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
  }

  OPTION_A_FAIRE( "Proposer d'autres types de modèles que FEFF !" );

  /* On crée un modèle FEFF pour l'absorbeur choisi */
  retour = I_ModeleAbsorbeur_NouveauFEFF( absorbeur, numero_modele_avant );

  return retour;
}

/* -------- Modèle de type FEFF -------- */

int I_ModeleAbsorbeur_NouveauFEFF(Absorbeur *absorbeur, unsigned long numero_modele_avant __attribute__((__unused__)))
{
  boolean coherent = VRAI;
  char *chemin, *nom_complet;
  int reponse;
  struct modele_ajuste *modele;

  /* On demande quel type de modèle FEFF on veut */
  reponse = BI_ERR_alerte( NULL, &Q_OrigineModeleFEFF, VRAI );
  if ( I_ABS_MODELE_FEFF_ANNULER == reponse ) {
    return +1;			/* Annulation : mineur */
  }

  switch( reponse )
  {
  case I_ABS_MODELE_FEFF_CONSTRUIRE:
    /* On construit le modèle */
    modele = ABS_CreerModele( absorbeur, MODELE_EXAFS_CHEMINS, NULL );
    if ( NULL == modele ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en voulant créer le modèle\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -20;
    }

    do {
      /* On en définit les paramètres généraux */
      FEFF_ParametresGeneraux( modele->infos_modele.feff );

      /* On en définit les potentiels */
      FEFF_DefinirPotentiels( modele->infos_modele.feff );

      /* On en définit la structure */
      FEFF_DefinirStructure( modele->infos_modele.feff );

      /* On vérifie le modèle */
      coherent = FEFF_VerifierModele( modele->infos_modele.feff );
    } while ( FAUX == coherent );

    /* On le sauve dans un dossier temporaire */
    chemin = I_FEFF_SauverTemporaire( modele->infos_modele.feff );

    /* On lance le calcul avec FEFF */
    nom_complet = TXT_genere_chaine( "%s/feff.inp", chemin );
    reponse = attend_feff( nom_complet );
    free( nom_complet ); nom_complet = NULL;

    nom_complet = TXT_genere_chaine( "%s/paths.dat", chemin );
    modele = I_FEFF_LirePathsDat( absorbeur, chemin, nom_complet );
    free( nom_complet ); nom_complet = NULL;
    break;

  case I_ABS_MODELE_FEFF_PATHS:
    modele = I_FEFF_DemanderPathsDat( absorbeur );
    break;

  case I_ABS_MODELE_FEFF_FEFF:
    OPTION_A_FAIRE( "Lire un feff.inp pour créer le modèle" );
    break;

  default:
    AVERTISSEMENT( "Code de réponse inconnu : %d", reponse );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -25;
  }

  return 0;
}
