/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2005, Emmanuel Curis

   Interface graphique : routines gnrales indpendantes du systme

   Liste des fonctions, par noms
*/

#include <stdlib.h>
#include <stdio.h>

#include "systeme.h"
#include "fonctions.h"

#include "graphes.h"		/* Pour les codes de genre et de types de graphes */

#include "fonctions.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static const char **FonctionsPourGraphe(unsigned int champ, unsigned int genre, unsigned int type,
					unsigned long *nombre_fonctions);
static const char **FonctionsPourImage (unsigned int champ, unsigned int genre, unsigned int type,
					unsigned long *nombre_fonctions);
static const char **FonctionsPourMolecule(unsigned int champ, unsigned int genre, unsigned int type,
					  unsigned long *nombre_fonctions);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- La fonction qui renvoie la table des options -------- */

const char **FCT_ListeFonctions(unsigned int nature_objet, unsigned int champ,
				unsigned int genre, unsigned int type, unsigned long *nombre_fonctions)
{
  if ( NULL == nombre_fonctions ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de place pour mmoriser le nombre d'options !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* On appelle les fonctions spcialises par type d'objet */
  *nombre_fonctions = 0;
  switch ( nature_objet )
  {
  case FONCTION_POUR_GRAPHE:
    return FonctionsPourGraphe( champ, genre, type, nombre_fonctions );

  case FONCTION_POUR_IMAGE:
    return FonctionsPourImage( champ, genre, type, nombre_fonctions );

  case FONCTION_POUR_MOLECULE:
    return FonctionsPourMolecule( champ, genre, type, nombre_fonctions );

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Nature d'objet inconnue [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, nature_objet);
    return NULL;
  }

  /* On ne devrait jamais tre l */
  return NULL;
}

/* -------- Fonctions pour les graphes -------- */

static const char **FonctionsPourGraphe(unsigned int champ, unsigned int genre, unsigned int type,
					unsigned long *nombre_fonctions)
{
  if ( FONCTION_EST_GLOBALE == champ ) {
    *nombre_fonctions = FCT_GRF_NOMBRE_OPT_GENERIQUES;
    return GRF_options_generiques;
  }

  *nombre_fonctions = 0;

  if ( FONCTION_EST_PAR_GENRE == champ ) {
    switch ( genre )
    {
    case GRF_FCT_REELLE:
      *nombre_fonctions = FCT_GRF_NOMBRE_OPT_REEL;
      return GRF_options_reelle;

    case GRF_FCT_COMPLEXE:
      *nombre_fonctions = FCT_GRF_NOMBRE_OPT_COMPLEXE;
      return GRF_options_complexe;

    case GRF_SERIE:
      *nombre_fonctions = FCT_GRF_NOMBRE_OPT_SERIE;
      return GRF_options_serie;

    case GRF_COURBE_PLANE:
      break;

    case GRF_COURBE_ESPACE:
      break;

    case GRF_SURFACE:
      break;

    case GRF_HISTOGRAMME:
      break;

    case GRF_FAISCEAU:
      break;

    case GRF_MULTI_SERIE:
      break;

    case GRF_ERREUR_R:
      break;

    case GRF_ERREUR_C:
      break;

    default:
      (void) fprintf(stderr, "%s,%s l.%u : Genre inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, genre);
    }

    /* Si l'on est l : genre inconnu ou sans fonction particulire */
    return NULL;
  }

  if ( FONCTION_EST_PAR_TYPE == champ ) {
    switch ( type )
    {
    case GR_BRUT:
      *nombre_fonctions = FCT_GRF_NOMBRE_OPT_BRUT;
      return GRF_options_brut;

    case GR_CINETIQUE:
      *nombre_fonctions = FCT_GRF_NOMBRE_OPT_CINETIQUE;
      return GRF_options_cinetique;

    default:
      (void) fprintf(stderr, "%s,%s l.%u : Type inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, type);
    }

    /* Si l'on est l : type inconnu ou sans fonction particulire */
    return NULL;
  }

  /* Si l'on est l : champ d'application inconnu ! */
  (void) fprintf(stderr, "%s,%s l.%u : champ d'application inconnu [%u, graphe]\n",
		 __FILE__, __FUNCTION__, __LINE__, champ);
  return NULL;
}

/* -------- Fonctions pour les images -------- */

static const char **FonctionsPourImage(unsigned int champ, unsigned int genre, unsigned int type,
				       unsigned long *nombre_fonctions)
{
  if ( FONCTION_EST_GLOBALE == champ ) {
    *nombre_fonctions = FCT_IMG_NOMBRE_OPT_GENERIQUES;
    return IMG_options_generiques;
  }

  *nombre_fonctions = 0;

  if ( FONCTION_EST_PAR_GENRE == champ ) {
    switch ( genre )
    {
    default:
      (void) fprintf(stderr, "%s,%s l.%u : Genre inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, genre);
    }

    /* Si l'on est l : genre inconnu ou sans fonction particulire */
    return NULL;
  }

  if ( FONCTION_EST_PAR_TYPE == champ ) {
    switch ( type )
    {
    default:
      (void) fprintf(stderr, "%s,%s l.%u : Type inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, type);
    }

    /* Si l'on est l : type inconnu ou sans fonction particulire */
    return NULL;
  }

  /* Si l'on est l : champ d'application inconnu ! */
  (void) fprintf(stderr, "%s,%s l.%u : champ d'application inconnu [%u, graphe]\n",
		 __FILE__, __FUNCTION__, __LINE__, champ);
  return NULL;
}

/* -------- Fonctions pour les molcules -------- */

static const char **FonctionsPourMolecule(unsigned int champ, unsigned int genre, unsigned int type,
					unsigned long *nombre_fonctions)
{
  if ( FONCTION_EST_GLOBALE == champ ) {
    *nombre_fonctions = FCT_MOL_NOMBRE_OPT_GENERIQUES;
    return MOL_options_generiques;
  }

  *nombre_fonctions = 0;

  if ( FONCTION_EST_PAR_GENRE == champ ) {
    switch ( genre )
    {
    default:
      (void) fprintf(stderr, "%s,%s l.%u : Genre inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, genre);
    }

    /* Si l'on est l : genre inconnu ou sans fonction particulire */
    return NULL;
  }

  if ( FONCTION_EST_PAR_TYPE == champ ) {
    switch ( type )
    {
    default:
      (void) fprintf(stderr, "%s,%s l.%u : Type inconnu [%u, graphe]\n",
		     __FILE__, __FUNCTION__, __LINE__, type);
    }

    /* Si l'on est l : type inconnu ou sans fonction particulire */
    return NULL;
  }

  /* Si l'on est l : champ d'application inconnu ! */
  (void) fprintf(stderr, "%s,%s l.%u : champ d'application inconnu [%u, graphe]\n",
		 __FILE__, __FUNCTION__, __LINE__, champ);
  return NULL;
}
