/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 21 septembre 2010

   Interface graphique

   Tronquer un graphe, quelle que soit la méthode
    (quoi qu'il arrive, n'étend JAMAIS le graphe)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "graphes.h"		/* Pour avoir accès au nom du graphe */
#include "interface.h"


/* -------- Variables locales -------- */

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_modules.txt"

/* -------- Fonctions locales -------- */

/* ~-~-~-~-~-~-~-~-~ Les Routines d'aide au développement ~-~-~-~-~-~-~-~-~ */

/* -------- Rappeler un « module » : arbre dans une fenêtre indépendante
   (si fermé       : l'ouvre avec le graphe indiqué
    si déjà ouvert : si même graphe, le rappele ; sinon, en fonction du choix de l'utilisateur)
   Renvoie 0 si tout s'est bien passé
           un code d'erreur sinon

 numero_arbre        : numéro de l'arbre définissant ce module
 graphe              : le graphe sur lequel doit travailler le module
 nom_module          : le nom du module (utilisé comme titre de la fenêtre)
 fonction_validation : fonction qui indique si le graphe proposé convient ou non
 fonction_fermeture  : fonction à appeler pour fermer le module, s'il est déjà ouvert pour un autre graphe
*/

int I_GRF_RappelerModule(unsigned int numero_arbre, struct graphe *graphe, const char *nom_module,
			 boolean (*fonction_validation)(struct graphe *graphe, BI_Boite_Alerte **message),
			 int     (*fonction_fermeture )(Fenetre *fenetre))
{
  boolean graphe_convient;
  int reponse;
  BI_Boite_Alerte *message = NULL;
  Fenetre *fenetre;
  Graphe *graphe_actuel;

  /* Contrôles initiaux */
  if ( NULL == graphe ) return I_GRF_RAPPELMODULE_ERREUR;
  if ( FAUX == ARB_EstOuvert( numero_arbre ) ) {
    /* En même temps, contrôle que numero_arbre est une valeur admissible */
    return I_GRF_RAPPELMODULE_ERREUR;
  }

  /* Quelle est la fenêtre associée à l'arbre ? */
  fenetre = ARB_QuelleFenetre( &(arbres[ numero_arbre ]), FAUX );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre du module %s ouvert, mais pas la fenêtre !\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_module );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return I_GRF_RAPPELMODULE_ERREUR;
  }

  /* On récupère le graphe associé à la fenêtre */
  graphe_actuel = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );

  /* On vérifie si le nouveau graphe est différent de l'ancien */
  if ( graphe_actuel == graphe ) {
    /* Non : on se contente de passer au premier plan ! */
    FEN_premierplan( fenetre );
    return I_GRF_RAPPELMODULE_ANCIEN;
  }

  /* On vérifie que le nouveau graphe convient */
  graphe_convient = fonction_validation( graphe, &message );
  if ( FAUX == graphe_convient ) {
    if ( message != NULL ) {
      (void) BI_ERR_alerte( fenetre, message, VRAI, graphe->nom );
    } else {
      ERR_LeGrapheNeConvientPas.titre = nom_module;
      (void) BI_ERR_alerte( fenetre, &ERR_LeGrapheNeConvientPas, VRAI, graphe->nom );
    }

    return I_GRF_RAPPELMODULE_ERREUR;
  }

  /* On pose la question : faut-il rappeler,... ? */
  Q_ReactiverModule.titre = nom_module;
  reponse = BI_ERR_alerte( fenetre, &Q_ReactiverModule, VRAI, graphe_actuel->nom, graphe->nom );
  switch ( reponse )
  {
  case I_GRF_RAPPELERMODULE:
    FEN_premierplan( fenetre );
    return I_GRF_RAPPELMODULE_ANCIEN;

  case I_GRF_RELANCERMODULE:
    /* On termine les calculs pour l'ancien graphe : ferme la fenêtre et libère tout */
    if ( fonction_fermeture != ( int (*)(Fenetre *) ) NULL ) {
      int retour __attribute__((__unused__));

      retour = fonction_fermeture( fenetre );
#ifdef _DEVELOPPE_
      printf( "Valeur de retour  la fermeture : %d", retour );
#endif
    }

    /* On ôte le graphe ancien de la fenêtre */
    if ( graphe_actuel != NULL ) {
      graphe_actuel->drapeaux.protege = FAUX;
    }
    FEN_GRF_LierGraphe( fenetre, NULL );

    /* On rappelera donc une nouvelle fenêtre propre... */
    return I_GRF_RAPPELMODULE_NOUVEAU;

  case I_GRF_ANNULER:
    /* On ne fait rien de spécial */
    return I_GRF_RAPPELMODULE_ANNULER;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en posant la question, ou réponse inconnue [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
    return I_GRF_RAPPELMODULE_ERREUR;
  }
}
