/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2003, Emmanuel CURIS
   Cration : 5 octobre 2003

   Menu prsentant les options associes aux graphes d'un certain type
   Ici : graphes contenant une srie de valeurs relles (pas de X)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "interface.h"
#include "serie.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define OPT_SERIE_TRI       0 /* Tri des valeurs par ordre croissant */
#define OPT_SERIE_MOYENNE   1 /* Moyenne, variance, cart-type */
#define OPT_SERIE_MEDIANE   2 /* Mdiane, carts  la mdiane */
#define OPT_SERIE_AUTRES    3 /* Autres estimateurs de position & de dispersion */
#define OPT_SERIE_FREP      5 /* 4 = sparateur ; fonction de rpartition empirique */
#define OPT_SERIE_HISTON    6 /* Histogramme des effectifs */
#define OPT_SERIE_HISTOF    7 /* Histogramme des (densits de) frquences */
#define OPT_SERIE_AUTOCOR   9 /*  8 = sparateur ; fonction d'autocorrlation */
#define OPT_SERIE_ANVA_1F  11 /* 10 = sparateur ; Analyse de variance 1 facteur */
#define OPT_SERIE_ANVA_RLd 12 /* Analyse de variance 1 facteur, rgression linaire simple */

#define NOMBRE_OPTIONS_SERIE 13

#include "opt_serie.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void OPT_AppeleANVA_1F (struct graphe *graphe);
static void OPT_AppeleANVA_RLd(struct graphe *graphe);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int OPT_GraphesSerie(struct graphe *graphe)
{
  int option;

  if ( NULL == graphe ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucun graphe indiqu\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( graphe->genre_donnees != GRF_SERIE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Le graphe ne contient pas une srie de valeurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  if ( graphe->x != NULL ) {
    option = BI_ERR_alerte( principale, &MSG_XIgnores, VRAI );
    if ( 2 == option ) return +1; /* Opration annule ! */
  }

  /* On appelle le menu */
  option = SS_menu_texte( menu_series, NOMBRE_OPTIONS_SERIE, titre_menu_series, 0, ROOT );
  if ( option < 0 ) {
    return +1; /* Annul : erreur mineure... */
  }

  switch ( option )
  {
  case OPT_SERIE_TRI:
    OPTION_A_FAIRE( "Trier la srie par ordre croissant" );
    break;

  case OPT_SERIE_MOYENNE:
    (void) GRF_Serie_Parametres_Moyenne( graphe );
    break;

  case OPT_SERIE_MEDIANE:
    OPTION_A_FAIRE( "Calculer la mdiane et estimateurs associs" );
    break;

  case OPT_SERIE_AUTRES:
    OPTION_A_FAIRE( "Calculer des estimateurs exotiques de position & dispersion" );
    break;

  case OPT_SERIE_FREP:
    OPTION_A_FAIRE( "Calculer la fonction de rpartition empirique" );
    break;

  case OPT_SERIE_HISTON:
    OPTION_A_FAIRE( "Calculer l'histogramme des effectifs" );
    break;

  case OPT_SERIE_HISTOF:
    OPTION_A_FAIRE( "Calculer l'histogramme des densits de frquence" );
    break;

  case OPT_SERIE_AUTOCOR:
    OPTION_A_FAIRE( "Calculer la fonction d'autocorrlation" );
    break;

  case OPT_SERIE_ANVA_1F:
    OPT_AppeleANVA_1F( graphe );
    break;

  case OPT_SERIE_ANVA_RLd:
    OPT_AppeleANVA_RLd( graphe );
    break;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Prparer pour l'appel : analyse de variance, 1 facteur -------- */

static void OPT_AppeleANVA_1F(struct graphe *graphe)
{
  Serie_Graphes *serie;
  Graphe **graphes;
  unsigned long nbr_modalites;
  struct test_stats resultat;

  /* Dans quelle srie de graphes se trouve le graphe ? */
  serie = GRF_QuelleSerie( graphe );
  if ( NULL == serie ) {
    AVERTISSEMENT( "Graphe hors srie ! [%s]", graphe->nom );
    return;
  }

  /* On considre tous les graphes de cette srie */
  graphes = GRF_ChoisirType( serie, GR_TOUS, VRAI, &nbr_modalites );
  if ( NULL == graphes ) return;

  /* On fait l'ANVA 1 facteur */
  (void) STAT_ANVA_1facteur( graphes, nbr_modalites, &resultat );

  /* Plus besoin de la table de graphes */
  free( graphes ); graphes = NULL;
}

/* -------- Prparer pour l'appel : analyse de variance, 1 facteur, droite ajuste -------- */

static void OPT_AppeleANVA_RLd(struct graphe *graphe)
{
  Serie_Graphes *serie;
  Graphe **graphes, *x;
  unsigned long nbr_modalites;
  struct test_stats resultat;

  /* Dans quelle srie de graphes se trouve le graphe ? */
  serie = GRF_QuelleSerie( graphe );
  if ( NULL == serie ) {
    AVERTISSEMENT( "Graphe hors srie ! [%s]", graphe->nom );
    return;
  }

  /* On considre tous les graphes de cette srie */
  graphes = GRF_ChoisirType( serie, GR_TOUS, VRAI, &nbr_modalites );
  if ( NULL == graphes ) return;

  /* On demande le graphe contenant les temps [en Y] */
  x = GRF_choisir( GR_TOUS, FAUX, VRAI, NULL );
  if ( NULL == x ) {
  free( graphes ); graphes = NULL;
  }

  /* On fait l'ANVA 1 facteur, rgression linaire simple */
  (void) STAT_ANVA_droite( graphes, nbr_modalites, x, &resultat );

  /* Plus besoin de la table de graphes */
  free( graphes ); graphes = NULL;
}
