/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1996 - 2010, Emmanuel CURIS
   Cration : 14 juin 2010 (repris de N_couches/feff_modele.c)

   Cration d'un modle  partir des fichiers de sortie de FEFF

   1) Utilise le fichiers PATHS.DAT pour connatre les paramtres des diffrents chemins
   2) Cre les chemins correspondants
   3) Lit les diffrents rsultats  partir des fichiers FEFFnnnn.DAT
   4) Recalcule les paramtres importants (phase et amplitude)  partir de ces donnes
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"		/* Pour le contenu de l'absorbeur */
#include "couches.h"		/* Pour COUCHE_Effacer */
#include "feff.h"
#include "interface.h"
#include "modeles.h"		/* Pour le contenu des modles */

#include "int_paths_dat.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

extern char *chemin_calculs_feff; /* Dans variables.c */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void I_FEFF_PathsDat_PreparerArbre(void);

/* ~~~~~~~~ Ralisation pratique des fonctions ~~~~~~~~ */

/* -------- Lecture d'un paths.dat -------- */

struct modele_ajuste *I_FEFF_DemanderPathsDat(struct absorbeur *absorbeur)
{
  char *nom_fichier, **liste_noms = NULL, *chemin_travail;
  const char *masque = "paths.dat";
  int retour;
  unsigned long i, nombre_fichiers = 0;
  struct modele_ajuste *modele_cree;
  BI_InfosSelecteur infos_selecteur = {
    NULL,
    BI_SELECTEUR_AJOUTER_PREDEFINI | BI_SELECTEUR_MEMORISER_CHEMIN,
    NULL,
    1, NULL,
    0, NULL,
    (void (*)(unsigned int, unsigned int, const char *) ) NULL
  };

  /* On contrle si l'absorbeur a dj des modles de mme type */
  if ( absorbeur->nbr_modeles > 0 ) {
    struct modele_ajuste *modele;

    /* Il y a dj des modles : on regarde si le modle courant est de type FEFF */
    modele = absorbeur->modele_actif;
    if ( MODELE_EXAFS_CHEMINS == modele->type ) {
      if ( modele->infos_modele.feff->couches != NULL ) {
	/* Il y a dj un modle complet : on demande ce que l'on en fait */
	int reponse;

	reponse = BI_ERR_alerte( NULL, &Q_EffacerModeleExistant, VRAI );
	if ( I_FEFF_PATHSDAT_ANNULER == reponse ) {
	  /* On annule */
	  return NULL;
	}
	if ( I_FEFF_PATHSDAT_REMPLACER_MODELE == reponse ) {
	  /* On remplace le modle par le nouveau */
	  struct modele_feff *modele_feff;

	  modele_feff = modele->infos_modele.feff;

	  /* On efface les couches de l'ancien modle
	     ==> efface chemins, amplitudes, dphasages associs */
	  while ( modele_feff->couches != NULL ) { /* Tant qu'il y a des couches */
	    COUCHE_Effacer( modele_feff, modele_feff->couches, VRAI, VRAI );
	  }
	}
	if ( I_FEFF_PATHSDAT_NOUVEAU_MODELE == reponse ) {
	  modele = NULL;	/* On veut en crer un nouveau */
	}
      }
    } else {
      /* Actif pas du bon type : on cherche si un autre modle irait... */
      modele = NULL;
    }
    if ( NULL == modele ) {
      for ( modele = absorbeur->modeles ; modele != NULL ; modele = modele->suivant) {
	if ( ( MODELE_EXAFS_CHEMINS == modele->type ) &&
	     ( NULL == modele->infos_modele.feff->couches ) ) {
	  /* Modle FEFF vide : sera utilis */
	  break;
	}
      }
    }

    if ( NULL == modele ) {
      /* A priori, on va crer un nouveau modle... ou utiliser un vide existant ! */
      char *nom_modele;

      modele = ABS_CreerModele( absorbeur, MODELE_EXAFS_CHEMINS, NULL );
      if ( NULL == modele ) {
	(void) fprintf(stderr, "%s,%s l.%u : impossible de crer le modle !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return NULL; /* NB : ne peut pas tre une molcule juste cre */
      }

      /* On demande le nom du nouveau modle */
      nom_modele = BI_Demande_Texte( fenetre_onglets, &Q_NomModeleFEFF, "Mod"e_GRAVE"le FEFF / paths.dat" );
      (void) MODELE_Nommer( modele, nom_modele );
      free( nom_modele ); nom_modele = NULL;
    }
  }

  /* On choisit le fichier */
  infos_selecteur.titre = (char *) titre_selecteur_paths_dat;
  infos_selecteur.filtres = &masque;
  infos_selecteur.chemin = chemin_calculs_feff;

  retour = BI_SelecteurChargement( &infos_selecteur, &nombre_fichiers, &liste_noms, fenetre_onglets );
  chemin_calculs_feff = infos_selecteur.chemin;

  if ( ( BI_SELECTEUR_ANNULER == retour ) || ( NULL == liste_noms ) || ( 0 == nombre_fichiers ) ) {
    return NULL;
  }

  /* On efface les noms inutiles */
  for ( i = 1 ; i < nombre_fichiers ; i++ ) {
    free( liste_noms[ i ] ); liste_noms[ i ] = NULL;
  }

  /* On rcupre le nom complet */
  nom_fichier = liste_noms[ 0 ];
  free( liste_noms ); liste_noms = NULL;

  /* On en dduit le chemin de travail [dossier contenant le paths.dat] */
  chemin_travail = strdup( nom_fichier );
  if ( NULL == chemin_travail ) {
    ERR_Memoire( "Chemin de travail", strlen( nom_fichier ) );
    free( nom_fichier ); nom_fichier = NULL;
    return NULL;
  }
  i = strlen( chemin_travail ) - 1;
  while ( chemin_travail[ i ] != SEPARE_DOSSIERS ) {
    chemin_travail[ i-- ] = 0;
  }

  /* On construit le modle  partir de ces informations */
  modele_cree = I_FEFF_LirePathsDat( absorbeur, chemin_travail, nom_fichier );

  /* On a termin */
  free( chemin_travail ); chemin_travail = NULL;
  free( nom_fichier ); nom_fichier = NULL;

  return modele_cree;
}

struct modele_ajuste *I_FEFF_LirePathsDat(struct absorbeur *absorbeur, const char *chemin_travail, const char *nom_fichier)
{
  int retour;
  struct modele_ajuste *modele_feff;
  FEFF_PATHS_DAT infos_lecture = { PATHSDAT_FORCER_ABSORBEUR_INDIQUE, NULL,
				   0.0, 0, 0, { 0.0, 0.0, 0.0 },
				   NULL, NULL, 0, 0 };

  /* On demande les paramtres pour charger ce fichier */
  if ( FAUX == arbres[ CREER_MODELE_FEF ].drapeaux.prepare ) {
    I_FEFF_PathsDat_PreparerArbre();
  }
  retour = ARB_Bloquant( CREER_MODELE_FEF, titre_arbre_creer_modele_feff, fenetre_onglets, 0 );
  if ( CRMF_ANNULER == retour ) {
    return NULL;
  }

  /* On lit ces paramtres */
  if ( OBJ_choisit( CREER_MODELE_FEF, CRMF_LG_CHOIX, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_LIMITER_LONGUEUR;
    infos_lecture.longueur_maximale = OBJ_LireNombre( CREER_MODELE_FEF, CRMF_LG_MAX, NULL );
  }

  if ( OBJ_choisit( CREER_MODELE_FEF, CRMF_NR_MAX, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_LIMITER_NOMBRE_ATOMES;
    infos_lecture.n_maximal = OBJ_LireEntier_U( CREER_MODELE_FEF, CRMF_MAX_REFLEX );
  }

  if ( OBJ_choisit( CREER_MODELE_FEF, CRMF_I_MAX, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_LIMITER_NOMBRE_CHEMINS;
    infos_lecture.i_maximal = OBJ_LireEntier_U( CREER_MODELE_FEF, CRMF_VAL_I_MAX );
  }

  if ( OBJ_choisit( CREER_MODELE_FEF, CRMF_R_CONSTANT, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_R_CONSTANT;
  }

  if ( OBJ_choisit( CREER_MODELE_FEF, CRMF_N_CONSTANT, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_N_CONSTANT;
  }

  if ( OBJ_choisit(CREER_MODELE_FEF, CRMF_LPM_INCLURE, QUESTION ) ) {
    infos_lecture.options |= PATHSDAT_INCLURE_LPM;
  }

  infos_lecture.chemin_travail = (char *) chemin_travail;

  /* On lit le paths.dat */
  modele_feff = FEFF_LirePathsDat( &infos_lecture, absorbeur, nom_fichier );

  /* En cas d'erreur : on analyse pourquoi */
  if ( NULL == modele_feff ) {
    (void) fprintf( stderr, "%s,%s l.%u : Problme de chargement du paths.dat : code %d\n",
		    __FILE__, __FUNCTION__, __LINE__, infos_lecture.code_erreur );
  }

  /* On renvoie le modle cr */
  return modele_feff;
}

/* -------- Prparation de l'arbre -------- */

static void I_FEFF_PathsDat_PreparerArbre(void)
{
  if ( FAUX == arbres[ CREER_MODELE_FEF ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( CREER_MODELE_FEF, NULL, NULL, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return;
    }
  }

  OBJ_deplace( CREER_MODELE_FEF, CRMF_NR_MAX    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_MODELE_FEF, CRMF_MAX_REFLEX, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_MODELE_FEF, CRMF_I_MAX     , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_MODELE_FEF, CRMF_VAL_I_MAX , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( CREER_MODELE_FEF, CRMF_UTIL_CHIP, 0, +1, DEPLACE_PIXELS );

#ifndef _DEVELOPPE_
  OBJ_active( CREER_MODELE_FEF, CRMF_CREER_CHEM, FORCE_NON );
  OBJ_active( CREER_MODELE_FEF, CRMF_CH_FEFF   , FORCE_NON );
  OBJ_active( CREER_MODELE_FEF, CRMF_UTIL_CHIP , FORCE_NON );
#endif

  arbres[ CREER_MODELE_FEF ].drapeaux.prepare = VRAI;
}
