/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 2 juin 2006

   Interface graphique

   Soustraction d'un blanc à un graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "automatise.h"
#include "interface.h"
#include "lissage.h"		/* Pour interpoler le blanc */
#include "operations.h"		/* Pour les combinaisons linéaires */
#include "serie.h"
#include "tracer.h"

/* -------- Variables locales -------- */

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_xan_blanc.txt"

/* -------- Fonctions locales -------- */

static Graphe *I_XAN_OterBlanc_Unique        (const Graphe *brut, const Graphe *blanc);
static void    I_XAN_OterBlanc_PostTraitement(const Graphe *brut, const Graphe *blanc, Graphe *resultat);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Routine d'appel principale : choix des options, du blanc & calculs -------- */

Graphe *I_XAN_OterBlanc(Graphe *brut)
{
  Serie_Graphes *serie;
  Graphe *blanc = NULL, *resultat = NULL;
  int reponse;

  /* Contrôles préliminaires */
  if ( NULL == brut ) return NULL;
  serie = GRF_QuelleSerie( brut );

  /* 1. Choix du blanc à ôter */
  blanc = GRF_choisir( brut->type, FAUX, VRAI, NULL );
  if ( NULL == blanc ) {
    return NULL;
  }

  /* 2. Travail sur le graphe seul ou sur toute la série ? */
  if ( serie != NULL ) {
    reponse = BI_ERR_alerte( NULL, &Q_OterBlancSurLaSerie, VRAI, brut->nom, serie->nom );
    if ( I_XAN_BLANC_ANNULER == reponse ) {
      return NULL;
    }
  } else {
    (void) fprintf( stderr, "%s,%s l.%u : Graphe brut passé hors-série [%s]\n",
		    __FILE__, __FUNCTION__, __LINE__, brut->nom );
    reponse = I_XAN_BLANC_ISOLE;
  }

  if ( I_XAN_BLANC_ISOLE == reponse ) {
    /* On soustrait */
    resultat = I_XAN_OterBlanc_Unique( brut, blanc );

    /* On insère dans la série après le brut [et si ne marche pas : nouvelle série] */
    (void) I_GRF_InsererDansSerie( NULL, brut, resultat );

    /* On fait le post-traitement */
    I_XAN_OterBlanc_PostTraitement( brut, blanc, resultat );
  } else {
    unsigned long i, nombre_bruts;
    Graphe **bruts_a_traiter;
    Graphe *resultat_temporaire;
    Serie_Graphes *serie_resultats;

    /* On construit la liste des graphes à traiter */
    serie_resultats = GRF_QuelleSerie( brut );
    bruts_a_traiter = GRF_ChoisirMasque( serie_resultats, brut, 
					 GRF_CHOIX_GENRE|GRF_CHOIX_TYPE|GRF_CHOIX_UNITE_X|GRF_CHOIX_UNITE_Y,
					 FAUX, &nombre_bruts );

    /* On traite chacun des graphes */
    for ( i = 0 ; i < nombre_bruts ; i++ ) {
      /* On soustrait */
      resultat_temporaire = I_XAN_OterBlanc_Unique( bruts_a_traiter[ i ], blanc );

      /* On insère dans la série après le brut [et si ne marche pas : nouvelle série] */
      serie_resultats = I_GRF_InsererDansSerie( serie_resultats, bruts_a_traiter[ i ], resultat_temporaire );

      /* On fait le post-traitement */
      I_XAN_OterBlanc_PostTraitement( bruts_a_traiter[ i ], blanc, resultat_temporaire );

      /* Le premier qui marche est le résultat qui sera affiché... */
      if ( ( resultat_temporaire != NULL ) && ( NULL == resultat ) ) {
	resultat = resultat_temporaire;
      }
    }
  }

  /* On a terminé */
  return resultat;
}

/* -------- Calcul & mise en place finale pour un seul graphe -------- */

static Graphe *I_XAN_OterBlanc_Unique(const Graphe *brut, const Graphe *blanc)
{
  Graphe *resultat = NULL, *vrai_blanc = NULL;
  boolean X_compatibles, retour __attribute__((__unused__));

  /* On fait la différence */
  X_compatibles = GRF_ComparerX( brut, blanc );
  if ( VRAI == X_compatibles ) {
    vrai_blanc = (Graphe *) blanc;
  } else {
    int methode_interpolation = INTERPOLE_SPLINE, code_erreur;

    OPTION_A_FAIRE( "X différents : choix de la méthode [interpolation linéaire ?]" );

    /* On prépare le blanc : on calcule ses valeurs pour les X du brut */
    vrai_blanc = GRF_CalerSurGraphe( blanc, methode_interpolation, INTERPOLE_BORDS_Y_Y, brut, &code_erreur );
    if ( NULL == vrai_blanc ) {
      OPTION_A_FAIRE( "Affichage du code d'erreur !" );
      return NULL;
    }
  }

  /* a. Préparation du graphe résultat --- isolé pour être tranquille */
  resultat = GRF_Copier( brut, NULL, VRAI );
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la cible\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* b. Calcul proprement dit.
     Cette fonction remplace les Y : pas de souci normalement
     elle s'occupe aussi de la propagation des erreurs... */
  retour = GRF_clin( brut, 1.0, vrai_blanc, -1.0, resultat );

  /* On efface au besoin le blanc aux X corrects, temporaire */
  if ( vrai_blanc != blanc ) {
    GRF_eff_isole( vrai_blanc ); vrai_blanc = NULL;
  }

  /* Terminé : on renvoie le résultat */
  return resultat;
}

/* -------- Améliorations post-traitement en ce qui concerne les annexes du graphe --------
     Post-mise en série pour être certain d'avoir accès au titre, légende, couleur,...
 */

static void I_XAN_OterBlanc_PostTraitement(const Graphe *brut, const Graphe *blanc, Graphe *resultat)
{
  char *titre;
  if ( NULL == resultat ) return;

  /* On met un titre adapté */
  titre = TXT_genere_chaine( "%s - %s", brut->nom, blanc->nom );
  GRF_titre( resultat, titre );
  free( titre ); titre = NULL;

  /* On met les mêmes styles graphiques que l'original */
  GRF_CopierAffichage( resultat, brut,
		       GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_DESSIN );
}
