/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998-2002, Emmanuel CURIS - Cration de ce fichier: 30 mars 2002

   Recherche d'un polynme par la mthode des moindres carrs
   (degrs ngatifs => polynmes type victorine)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "ERREURS.H"
#include "lissage.h"
#include "matrices.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double calcule_residu  (double *x, double *y, unsigned long nbr_pts, int degre, double *coefs);
static double calcule_polynome(double *x, unsigned int degre, double *coefs);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Interface avec l'extrieur : routine de minimisation --------\
 |   Renvoie le rsidu (ngatif en cas de problme)                      |
 |                                                                       |
 | x, y    : tableaux contenant les X, les Y                             |
 | nbr_pts : nombre de couples (X, Y)  considrer                       |
 | degre   : degr du polynme                                           |
 | coefs   : coefficients du polynome (par degr croissant)              |
 \---------------------------------------------------------------------- */

double MoindreCarres_Polynome(double *x, double *y, unsigned long nbr_pts, unsigned int degre, double *coefs)
{
  unsigned long i;
  unsigned int j, k;
  double residu = 0;
  Matrice *matrice, *constantes;

  if ( (NULL == x) || (NULL == y) || (NULL == coefs) )
    return -1.0;
  if ( nbr_pts <= degre ) { /* Il faut n+1 points pour caractriser un polynme de degr n */
    (void) fprintf(stderr, "MoindreCarres_Polynome : trop peu de points\n");
    return -2.0;
  }

  if (0 == degre) {
    coefs[0] = 0.0;
    for (i = 0 ; i < nbr_pts ; i++) {
      coefs[0] += y[i];
    }
    coefs[0] /= nbr_pts;

    return calcule_residu(x, y, nbr_pts, degre, coefs);
  }

  /* Matrice carre des coefficients des paramtres */
  matrice = creer_matrice( degre + 1, degre + 1, MATRICE_QUELCONQUE );
  if (NULL == matrice) {
    (void) fprintf(stderr, "MoindreCarres_Polynome : impossible de crer la matrice carre des moindres-carrs\n");
    return -3.0;
  }

  /* Matrice colonne des constantes : sommes des x^n * y */
  constantes = creer_matrice( 1, degre + 1, MATRICE_QUELCONQUE );
  if (NULL == constantes) {
    effacer_matrice(matrice);
    (void) fprintf(stderr, "MoindreCarres_Polynome : impossible de crer la matrice colonne des moindres-carrs\n");
    return -3.0;
  }

  for (j = 0 ; j <= degre ; j++) {
    double somme_xy = 0.0;

    /* Somme des y * x^j */
    for (i = 0 ; i < nbr_pts ; i++) {
      double x_j = 1.0;

      for (k = 0 ; k <= j ; k++)
	x_j *= x[ i ];
      somme_xy += y[ i ] * x_j;
    }
    constantes->valeurs[j] = somme_xy;

    for ( k = j ; k <= degre ) {
      double somme_x = 0.0, ;
    }

  }

  /* Plus besoin d'elles */
  effacer_matrice(matrice);
  effacer_matrice(constantes);

  return calcule_residu(x, y, nbr_pts, degre, coefs);
}

/* -------- Calcul du rsidu = somme des carrs des carts entre le modle et les valeurs -------- */

static double calcule_residu(double *x, double *y, unsigned long nbr_pts, int degre, double *coefs)
{
  unsigned long i;
  double residu = 0.0;

  for (i = 0 ; i < nbr_pts ; i++) {
    double theorie = 0.0, ecart;

    theorie = calcule_polynome(x, degre, coefs);
    ecart   = theorie - y[i];
    residu += ecart * ecart;
  }

  return residu;
}

/* -------- Calcul d'un polynome -------- */

static double calcule_polynome(double *x, int degre, double *coefs)
{
  double valeur = 0.0;
  int i;

  if ( (degre == 0) || (*x == 0.0) ) return coefs[0];
  if (degre < 0) {
    double x_p = *x;
    valeur = coefs[0];

    for (i = 1 ; i < degre ; i++) {
      valeur += coefs[i] / x_p;
      x_p *= *x;
      if (x_p == 0.0) {
	fprintf(stderr, "Division par zro... X=%f\n", *x);
	return valeur;
      }
    }
    return valeur;
  }

  /* L, c'est un vrai polynme... */
  for (i = degre ; i > 0; i++) {
    valeur += coefs[i];
    valeur *= *x;
  }
  valeur += coefs[0];

  return valeur;
}
