/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2001, Emmanuel CURIS
   Cration              : 11 avril 2001
   Dernire modification : 

   Modules de calcul matriciel
   Inversion d'une matrice -- suppose carre, donc.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "systeme.h"
#include "erreurs.h"

#include "matrices.h"

/* ~~~~~~~~ Inversion de la matrice : utilisation de la mthode de Cholesky si possible
     Renvoie la matrice inveree, NULL en cas d'erreur
             cette matrice doit tre libre avec free() aprs usage
     Suppose que la matrice (de taille donne) est symtrique, dfinie positive
             on ne stocke donc que la triangulaire infrieure

     def_pos = VRAI : matrice symtrique, dfinie positive : on peut utiliser Cholesky
 ~~~~~~ */

Matrice *inverse_matrice(Matrice *matrice, boolean def_pos)
{
  Matrice *inverse = NULL;

  if (matrice == NULL) return NULL;

  if (matrice->n_lignes != matrice->n_colonnes) {
    (void) MAT_ERREUR(MAT_NON_CARREE);
    return NULL;
  }

  if (matrice->type == MATRICE_DIAGONALE) { /* Inversion directe */
    unsigned int i;

    inverse = creer_matrice(matrice->n_lignes, matrice->n_lignes, MATRICE_DIAGONALE);
    if (inverse == NULL) return NULL;

    for (i = 0 ; i < matrice->n_lignes ; i++) {
      if (matrice->valeurs[i] == 0.0) {
	(void) MAT_ERREUR(MAT_NON_INVERSIBLE);
	effacer_matrice(inverse);
	return NULL;
      }

      inverse->valeurs[i] = 1.0 / matrice->valeurs[i];
    }
  }

  if (matrice->n_lignes == 2) /* Inversion directe, matrice non diagonale */
  {
    double det;
    det = lit_element(matrice, 0, 0) * lit_element(matrice, 1, 1) - lit_element(matrice, 0, 1) * lit_element(matrice, 1, 0);
    if (det == 0.0)
    {
      (void) MAT_ERREUR(MAT_NON_INVERSIBLE);
      return NULL;
    }
/*    printf("Dterminant trouv : %g\n", det); */

    inverse = creer_matrice(2, 2, matrice->type);
    if (inverse == NULL)
      return NULL;

    /* ( a b )-1 = (a -c) / (ad - bc)
       ( c d )     (-b d)            */
    met_element(inverse, 0, 0,  lit_element(matrice, 1, 1) / det);
    met_element(inverse, 0, 1, -lit_element(matrice, 1, 0) / det);
    met_element(inverse, 1, 0, -lit_element(matrice, 0, 1) / det);
    met_element(inverse, 1, 1,  lit_element(matrice, 0, 0) / det);
    return inverse;
  }

  if (matrice->type == MATRICE_TRIANGULAIRE_INF) {
    unsigned int i, j;

    /* La matrice est-elle inversible ? */
    for (i = 0 ; i < matrice->n_lignes ; i++)
      if ( matrice->valeurs[ MAT_TINF_element(i, i) ] == 0.0) {
	(void) MAT_ERREUR(MAT_NON_INVERSIBLE);
	return NULL;
      }

    inverse = creer_matrice(matrice->n_lignes, matrice->n_lignes, MATRICE_TRIANGULAIRE_INF);
    if (inverse == NULL) return NULL;

    for (j = 0 ; j < inverse->n_colonnes ; j++) {/* Remplissage par colonnes */
      /* Terme diagonal */
      inverse->valeurs[ MAT_TINF_element(j, j) ] = 1/matrice->valeurs[ MAT_TINF_element(j, j) ];

      /* Termes hors_diagonale */
      for (i = j + 1 ; i < inverse->n_lignes ; i++) {
	double somme = 0.0;
	unsigned int k;

	for (k = j ; k < i ; k++)
	  somme -= matrice->valeurs[ MAT_TINF_element(i, k) ] * inverse->valeurs[ MAT_TINF_element(k, j) ];

	somme /= matrice->valeurs[ MAT_TINF_element(i, i) ];
	inverse->valeurs[ MAT_TINF_element(i, j) ] = somme;
      }
    }

    return inverse;
  }

  if (matrice->type == MATRICE_SYMETRIQUE) {
    inverse = creer_matrice(matrice->n_lignes, matrice->n_lignes, MATRICE_SYMETRIQUE);
    if (inverse == NULL) return NULL;

    if (def_pos == VRAI) { /* Factorisation de Cholesky possible */
      Matrice *L, *L_inv;
      unsigned int i, j;

      L = MAT_DecompositionCholesky(matrice);
      if (L == NULL) /* Problme de calcul */
	return NULL;

      L_inv = creer_matrice(L->n_lignes, L->n_colonnes, MATRICE_TRIANGULAIRE_INF);
      if (L_inv == NULL) {
	effacer_matrice(L);
	return NULL;
      }

      /* On inverse L */
      L_inv = inverse_matrice(L, FAUX);
#ifdef _DEVELOPPE_
/*      aff_matrice(L);
      aff_matrice(L_inv); */
#endif
      effacer_matrice(L); /* Plus besoin d'elle */

      /* On calcule la matrice inverse = produit t(L^-1) * L^-1
         Seul le triangle infrieur est intressant */
      for (i = 0 ; i < inverse->n_lignes ; i++)
	for (j = 0 ; j <= i ; j++) {
	  double somme = 0.0;
	  unsigned int k;

	  for (k = i /* car i > j */ ; k < L_inv->n_lignes ; k++)
	    somme += L_inv->valeurs[ MAT_TINF_element(k, i) ] * L_inv->valeurs[ MAT_TINF_element(k, j)];
	  inverse->valeurs[ MAT_TINF_element(i, j) ] = somme;
	}

#ifdef _DEVELOPPE_
/*      aff_matrice(inverse); */
#endif
      effacer_matrice(L_inv); /* Inutile */
      return inverse;
    }
  }

  return NULL; /* Pour le moment, inverse pas au-dessus de 2x2 */
}

/* -------- Ralise la dcomposition de Cholesky d'une matrice --------
     Renvoie la matrice L, triangulaire infrieure ; on a M = L t(L)

     La matrice pass doit tre symtrique, dfinie positive
     Seule la partie triangulaire infrieure est passe 
----------------------------------------------------------------------- */

Matrice *MAT_DecompositionCholesky(Matrice *matrice)
{
  Matrice *L = NULL;
  unsigned int i, j;

  if (matrice == NULL) return NULL;
  if (matrice->type != MATRICE_SYMETRIQUE) {
    (void) fprintf(stderr, "%s,%s l.%u : matrice non symtrique.\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if (matrice->valeurs[0] <= 0.0) { /* Terme (1, 1) */
/*     (void) MAT_ERREUR(MAT_NON_DEF_POS); */
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : M(0, 0) <= 0.0 [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, matrice->valeurs[ 0 ]);
#endif
    return NULL;
  }

  L = creer_matrice(matrice->n_lignes, matrice->n_colonnes, MATRICE_TRIANGULAIRE_INF);
  if (L == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la matrice L !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Adaptation personnelle */

  for (j = 0 ; j < matrice->n_lignes ; j++) { /* On remplit par colonne */
    double diag = 0.0;
    unsigned int k;

/*    printf("\nDiagonale : on utilise %u pour A ; pour L, ", MAT_TINF_element(j, j)); */
    /* Terme diagonal */
    for (k = 0 ; k < j ; k++) {
      double val = (L->valeurs)[ MAT_TINF_element(j, k) ];
/*      printf(" %u", MAT_TINF_element(j, k) ); */
      diag += val * val;
    }
    diag = (matrice->valeurs)[ MAT_TINF_element(j, j) ] - diag;
    if ( diag <= 0.0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Terme diagonal ngatif [%g / ligne %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, diag, j);
#endif
/*       (void) MAT_ERREUR(MAT_NON_DEF_POS); */
      effacer_matrice( L );
      return NULL;
    }
    diag = sqrt(diag);
/*    printf(" - on remplit %u\n", MAT_TINF_element(j, j)); */
    (L->valeurs)[MAT_TINF_element(j, j)] = diag;
    
    /* termes hors-diagonale */
    for (i = j + 1 ; i < matrice->n_lignes ; i++) { /* Ligne */
      double somme = (matrice->valeurs)[ MAT_TINF_element(i, j) ];
/*      printf("Ligne %u : on utilise %u pour A ; pour L", i, MAT_TINF_element(i, j)); */
      for (k = 0 ; k < j ; k++) {
/*	printf(" %u & %u - ", MAT_TINF_element(j, k), MAT_TINF_element(j, k) ); */
	somme -= (L->valeurs)[ MAT_TINF_element(i, k) ] * (L->valeurs)[ MAT_TINF_element(j, k) ];
      }

/*      printf("on remplit %u\n", MAT_TINF_element(i, j)); */
      (L->valeurs)[MAT_TINF_element(i, j)] = somme / diag;
    }
  }

  return L;
}
