/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2008, Emmanuel Curis

   Création et gestion de modèles en vue d'un ajustement

   Modèle de type « décomposition sur une base de spectres »
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "decomposition.h"
#include "graphes.h"
#include "lissage.h"
#include "matrices.h"
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#include "decomposition.txt"	/* Les traductions */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int MODELE_Decomposition_Moyenne        (Graphe *spectre, struct decomposition *infos);
static int MODELE_Decomposition_MoindresCarres (Graphe *spectre, struct decomposition *infos);
static int MODELE_Decomposition_Suivi          (Graphe *spectre, struct decomposition *infos);
static int MODELE_Decomposition_SuiviLineaire  (Graphe *spectre, struct decomposition *infos);
static int MODELE_Decomposition_Variance       (Graphe *spectre, struct decomposition *infos);
static int MODELE_Decomposition_Moindres_Somme1(Graphe *spectre, struct decomposition *infos);

/* ~~~~~~~~ Réalisation concrète ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ La gestion interne des objets... ~-~-~-~-~-~-~-~-~  */

/* ------ Créer un nouveau modèle de décomposition ------ */

struct decomposition *MODELE_NouveauDecomposition(void)
{
  struct decomposition *infos;

  /* Réservation */
  infos = (struct decomposition *) malloc( sizeof( struct decomposition ) );
  if ( NULL == infos ) {
    ERR_Memoire( "Informations de décomposition", sizeof( struct decomposition ) );
    return NULL;
  }

  /* Méthode par défaut : les moindres carrés */
  infos->methode = MODELE_DECOMPOSITION_MOINDRES_CARRES;

  /* Aucun spectre de référence initialement */
  infos->nombre_bases = 0;
  infos->base = NULL;

  /* Pas de suivi */
  infos->alpha = NULL;
  infos->beta = NULL;

  /* Terminé, c'est prêt */
  return infos;
}

/* ------ Ajouter un graphe dans la base des graphes de référence ------ */

int MODELE_Decomposition_AjouterVecteurDeBase(struct decomposition *decomposition, struct graphe *graphe_a_ajouter)
{
  unsigned long nombre_vecteurs;
  struct element_base *base;

  if ( ( NULL == decomposition ) || ( NULL == graphe_a_ajouter ) ) {
    /* Problème : modèle ou graphe NULL ! */
    return -1;
  }

  OPTION_A_FAIRE( "Vérifier que le graphe n'est pas celui décomposé..." );

  /* On vérifie que l'on n'utilise pas un graphe mémorisant les résultats... */
  if ( ( graphe_a_ajouter == decomposition->alpha ) ||
       ( graphe_a_ajouter == decomposition->beta  ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Le graphe est l'un de ceux mémorisant les résultats !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return MODELE_DECOMPOSITION_ERREUR_DECOMPOSE;
  }

  /* Combien de vecteurs de base actuellement dans le modèle ? */
  nombre_vecteurs = decomposition->nombre_bases;

  /* S'il y en a déjà : on vérifie que le graphe proposé n'est pas déjà dans la liste... */
  if ( nombre_vecteurs > 0 ) {
    unsigned long i;
    struct element_base *vecteur_base;

    for ( i = 0, vecteur_base = decomposition->base ;
	  i < nombre_vecteurs ; i++, vecteur_base++ ) {
      if ( vecteur_base->graphe == graphe_a_ajouter ) {
	(void) fprintf( stderr, "%s,%s l.%u : Graphe déjà présent dans la base, ajout annulé\n",
			__FILE__, __FUNCTION__, __LINE__);
	return MODELE_DECOMPOSITION_ERREUR_DEJA_DANS_BASE;
      }
    }
  }

  /* On redimensionne le tableau des éléments de base */
  base = ( struct element_base * ) realloc( decomposition->base,
					    sizeof( struct element_base ) * ( nombre_vecteurs + 1 ) );
  if ( NULL == base ) {
    (void) fprintf( stderr, "%s,%s l.%u : Réallocation impossible du tableau. Ajout annulé\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    ERR_Memoire( "Ajout d'un graphe dans la base de décomposition",
		 sizeof( struct element_base ) * ( decomposition->nombre_bases ) );
    return -2;
  }

  /* On remplace l'ancien tableau, au besoin */
  decomposition->base = base;

  /* On initialise les valeurs pour ce vecteur de base, qui est en dernière position */
  ( base[ nombre_vecteurs ] ).graphe = graphe_a_ajouter;
  ( base[ nombre_vecteurs ] ).erreur = 0.0;
  VAR_Initialiser( &( ( base[ nombre_vecteurs ] ).coefficient ), VAR_LIBRE,
		   0.0, 0.0, 1.0, 1e-5 );
  ( base[ nombre_vecteurs ] ).a = NULL;
  ( base[ nombre_vecteurs ] ).b = NULL;
  ( base[ nombre_vecteurs ] ).c = NULL;

  /* Il y a donc un graphe de base de plus */
  decomposition->nombre_bases++;

  /* On protège le graphe, par prudence... */
  GRF_ProtegerGraphe( graphe_a_ajouter );

  /* Tout s'est bien passé... */
  return PAS_DE_PROBLEME;
}

/* -------- Remplacer un graphe de base par un autre -------- */

int MODELE_Decomposition_RemplacerVecteurDeBase(struct decomposition *decomposition, unsigned long numero,
						struct graphe *nouveau_graphe)
{
  unsigned long i;

  if ( NULL == decomposition  ) return +1; /* Pas de modèle de décomposition ! */
  if ( NULL == nouveau_graphe ) return +2; /* Pas de nouveau graphe */

  /* On contrôle que le changement est possible */
  if ( numero >= decomposition->nombre_bases ) {
    (void) fprintf( stderr, "%s,%s l.%u : Index de numéro de vecteur de base incorrect [%lu, max : %lu]\n",
		    __FILE__, __FUNCTION__, __LINE__, numero, decomposition->nombre_bases - 1 );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* Est-ce que ce graphe est bien distinct de l'ancien ? */
  if ( nouveau_graphe == decomposition->base[ numero ].graphe ) {
    AVERTISSEMENT( "Graphe inchangé... [%s]", nouveau_graphe->nom );
    return PAS_DE_PROBLEME;
  }

  /* Est-ce que ce graphe n'est pas déjà utilisé dans cette décomposition ? */
  for ( i = 0 ; i < decomposition->nombre_bases ; i++ ) {
    if ( nouveau_graphe == decomposition->base[ i ].graphe ) {
      BI_ERR_annuler( texte_erreur_graphe_deja_utilise, nouveau_graphe->nom );
      return -10;
    }
  }

  /* On déprotège l'ancien graphe */
  GRF_DeprotegerGraphe( decomposition->base[ numero ].graphe );

  /* On le remplace par le nouveau */
  decomposition->base[ numero ].graphe = nouveau_graphe;

  /* On protège ce nouveau graphe */
  GRF_ProtegerGraphe( nouveau_graphe );

  /* Terminé */
  return PAS_DE_PROBLEME;
}

/* ------ Enlever un graphe de la base des graphes de référence ------ */

int MODELE_Decomposition_EnleverVecteurDeBase(struct decomposition *decomposition, unsigned long numero)
{
  unsigned long i;

  if ( NULL == decomposition ) return -1;

  if ( numero >= decomposition->nombre_bases ) {
    (void) fprintf( stderr, "%s,%s l.%u : Demande d'ôter le graphe n° %lu, mais seulement %lu connus !\n",
		    __FILE__, __FUNCTION__, __LINE__, numero + 1, decomposition->nombre_bases );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* On libère les informations pour le vecteur de base concerné */
  GRF_DeprotegerGraphe( decomposition->base[ numero ].graphe ); decomposition->base[ numero ].graphe = NULL;
  OPTION_A_FAIRE( "Libérer proprement la variable associée au coefficient" );
  free( decomposition->base[ numero ].a ) ; decomposition->base[ numero ].a = NULL;
  free( decomposition->base[ numero ].b ) ; decomposition->base[ numero ].b = NULL;
  free( decomposition->base[ numero ].c ) ; decomposition->base[ numero ].c = NULL;

  /* On déplace les vecteurs de base suivant sur celui-ci */
  for ( i = numero ; i < decomposition->nombre_bases - 1 ; i++ ) {
    decomposition->base[ i ] = decomposition->base[ i + 1 ];
  }

  /* Il y a donc un vecteur de base de moins... */
  decomposition->nombre_bases--;

  /* Terminé sans souci */
  return PAS_DE_PROBLEME;
}

/* ~-~-~-~-~-~-~-~-~ Les méthodes de résolution ~-~-~-~-~-~-~-~-~  */

/* -------- Routine générique de calcul [visible de l'extérieur] -------- */

int MODELE_Decomposition_Decomposer(Graphe *spectre, struct decomposition *decomposition)
{
  int reponse = -1974;

  if ( ( NULL == spectre ) || ( NULL == decomposition ) ) return -1;

  /* On distribue l'appel aux vraies fonctions */
  switch ( decomposition->methode )
  {
  case MODELE_DECOMPOSITION_MOINDRES_CARRES:  /* Méthode des moindres carrés */
    reponse = MODELE_Decomposition_MoindresCarres( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_MDR_CARR_SOMME1:  /* Idem, avec la contrainte d'une somme valant 1 */
    reponse = MODELE_Decomposition_Moindres_Somme1( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_MOYENNES:  /* Égalité des moyennes + moyenne des coefficients point par point */
    reponse = MODELE_Decomposition_Moyenne( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_MOYENNE_VARIANCE:  /* Égalité des moyennes et des variances */
    reponse = MODELE_Decomposition_Variance( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_SUIVI:  /* Suivi cinétique des coefficients */
    reponse = MODELE_Decomposition_Suivi( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_SUIVI_LINEAIRE:  /* Idem, en supposant que les coefficients varient linéairement avec le temps et qu'il y a conversion */
    reponse = MODELE_Decomposition_SuiviLineaire( spectre, decomposition );
    break;

  case MODELE_DECOMPOSITION_ACP:  /* Analyse en Composantes Principales */
    OPTION_A_FAIRE( "ACP" );
    break;

  default:
    AVERTISSEMENT( "Méthode imprévue : %d", decomposition->methode );
  }

  if ( reponse != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Problème survenu dans la décomposition --- code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
  }

  /* On a fini */
  return reponse;
}

/* -------- Recherche des coefficients : méthode des moindres carrés -\
 |  On cherche a et b de façon à minimiser
 |
 |           [                                                       ]2
 | S = Somme [ g_i - (a * x_i + b) * f1_i - (1 - a * x_i - b) * f2_i ]
 |       i   [                                                       ]
 |
 | Cela revient donc à une évolution linéaire des coefficients avec le
 | temps. Idée de Karine Provost.
 | Attention, les deux spectres ne jouent pas un rôle symétrique.
 \------------------------------------------------------------------- */

static int MODELE_Decomposition_SuiviLineaire(Graphe *spectre, struct decomposition *infos)
{
  register double *x, *y;
  unsigned long i;
  double sfg = 0.0, sxfg = 0.0, sx2g2 = 0.0, sxg2 = 0.0, sg2 = 0.0;
  double delta;

  /* Contrôle : cette méthode ne fonctionne qu'avec 2 spectres */
  if ( infos->nombre_bases != 2 ) {
    return -20;
  }

  /* Première étape : calcul des sommes */
  for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
    double f1, f2, x_lu = *x;
    double fmg, gmg;

    f1 = SPL_tt_calc( infos->base[ 0 ].graphe, x_lu, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    f2 = SPL_tt_calc( infos->base[ 1 ].graphe, x_lu, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    fmg = *y - f2; /* g - g2 */
    gmg = f1 - f2; /* f1 - f2 */

    sfg  += fmg * gmg;      /* Somme  (g - f2)(f1 - f2) */
    sxfg += fmg * gmg * *x; /* Somme x(g - f2)(f1 - f2) */

    gmg *= gmg; /* Contient (f2 - f1)² */      
    sx2g2 += (*x * *x) * gmg; /* Somme x²(f1 - f2)² */
    sxg2  += *x * gmg;        /* Somme x (f1 - f2)² */
    sg2   += gmg;             /* Somme   (f1 - f2)² */
  }

  /* Résolution du système : { Somme (g - f2)(f1 - f2)x = a * Somme x²(f1 - f2)² + b * Somme x(f1 - f2)²
                             { Somme (g - f2)(f1 - f2)  = a * Somme x (f1 - f2)² + b * Somme  (f1 - f2)²
  */

  delta = sx2g2 * sg2 - sxg2 * sxg2;
  if ( 0.0 == delta ) {
    EXF_ERREUR(DIV_ZERO, "D"e_AIGU"terminant nul", txt_err_calc_annule);
    return -21;
  }

  infos->base[ 0 ].coefficient.valeur_initiale = (sxfg  * sg2 - sfg  * sxg2) / delta;
  infos->base[ 1 ].coefficient.valeur_initiale = (sx2g2 * sfg - sxg2 * sxfg) / delta;

  /* Décomposition effectuée */
  return 0;
}

/* -------- Recherche des coefficients : méthode des moindres carrés--\
|   On cherche alpha et bêta de façon à minimiser                    |
| S = Somme_i [ g_i - alpha * f1_i - bêta * f2_i ]²                  |
\------------------------------------------------------------------- */

static int MODELE_Decomposition_MoindresCarres(Graphe *spectre, struct decomposition *infos)
{
  register double *x, *y;
  int retour;
  unsigned long i, j, k;
  double somme, y_lu_j, y_lu_k;
  Matrice *matrice, *constantes, *coefficients;

  /* On réserve pour la matrice des sommes : n×n, où n = nombre de spectres de base */
  matrice = creer_matrice( infos->nombre_bases, infos->nombre_bases, MATRICE_SYMETRIQUE );
  if ( NULL == matrice ) {
    ERR_Memoire( "Matrice des moindres carrés", sizeof( *matrice ) );
    return -5;
  }

  /* On réserve pour la matrice colonne des constantes */
  constantes = creer_matrice( infos->nombre_bases, 1, MATRICE_QUELCONQUE );
  if ( NULL == constantes ) {
    ERR_Memoire( "Matrice des constantes", sizeof( *constantes ) );
    effacer_matrice( matrice ); matrice = NULL;
    return -5;
  }

  /* On réserve pour la matrice colonne des inconnues */
  coefficients = creer_matrice( infos->nombre_bases, 1, MATRICE_QUELCONQUE );
  if ( NULL == coefficients ) {
    ERR_Memoire( "Matrice des coefficients (= inconnues)", sizeof( *coefficients ) );
    effacer_matrice( constantes ); constantes = NULL;
    effacer_matrice( matrice ); matrice = NULL;
    return -5;
  }

  /* On remplit la matrice & le vecteur des constantes */
  for ( j = 0 ; j < infos->nombre_bases ; j++ ) {
    /* Ligne de la matrice carrée : chaque colonne jusqu'à la diagonale [après, symétrie !] */
    for ( k = 0 ; k <= j ; k++ ) {
      somme = 0.0;

      for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++) {
	y_lu_j = SPL_tt_calc( infos->base[ j ].graphe, *x, infos->base[ j ].a, infos->base[ j ].b, infos->base[ j ].c );
	y_lu_k = SPL_tt_calc( infos->base[ k ].graphe, *x, infos->base[ k ].a, infos->base[ k ].b, infos->base[ k ].c );

	somme += y_lu_j * y_lu_k;
      }

      met_element( matrice, j, k, somme );
    }

    /* Matrice colonne des constantes */
    somme = 0.0;
    for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++) {
      y_lu_j = SPL_tt_calc( infos->base[ j ].graphe, *x, infos->base[ j ].a, infos->base[ j ].b, infos->base[ j ].c );
      y_lu_k = *(spectre->y + i);

      somme += y_lu_j * y_lu_k;
    }

    met_element( constantes, j, 0, somme );

    /* Matrice colonne des inconnues */
    met_element( coefficients, j, 0, 0.0 );
  }

  /* On résoud le système, par la méthode de Gauss */
  retour = MAT_ResoudSysteme( matrice, coefficients, constantes, MATRICE_LINEAIRE_GAUSS );

  /* Quoi qu'il arrive, plus besoin de ces 2 matrices */
  effacer_matrice( constantes ); constantes = NULL;
  effacer_matrice( matrice ); matrice = NULL;

  if ( retour != 0 ) {
    /* Plusieurs ou aucune solution */
    effacer_matrice( coefficients ); coefficients = NULL;
    return -10;
  }

  /* On mémorise les réponses */
  for ( j = 0 ; j < infos->nombre_bases ; j++ ) {
    infos->base[ j ].coefficient.contenu.valeur  = lit_element( coefficients, j, 0 );;
  }

  /* Terminé sans encombre */
  effacer_matrice( coefficients ); coefficients = NULL;
  return 0;
}

/* -------- Recherche des coefficients : méthode des moindres carrés somme = 1  --\
|   On cherche alpha de façon à minimiser                                         |
| S = Somme_i [ g_i - alpha * f1_i - (1-alpha) * f2_i ]²                          |
\--------------------------------------------------------------------------------*/

static int MODELE_Decomposition_Moindres_Somme1(Graphe *spectre, struct decomposition *infos)
{
  register double *x, *y;
  unsigned long i;
  double s1c = 0.0, s12 = 0.0, sg1 = 0.0, s2c = 0.0, sg2 = 0.0;
  double div;

  /* Contrôles initiaux */
  if ( 0 == infos->nombre_bases ) return -1;
  if ( 1 == infos->nombre_bases ) {
    /* Vu la contrainte, le coefficient vaut forcément 1 */
    infos->base[ 0 ].coefficient.contenu.valeur  = 1.0;

    return 0;
  }

  if ( 2 > infos->nombre_bases ) {
    OPTION_A_FAIRE( "Cas de plus de 2 spectres de référence" );
    /* Rq : revient à soustraire l'un des spectres de la base de tous les autres, y compris celui à décomposer */
    return -20;
  }

  /* Ancienne méthode, dans le cas de 2 spectres */
  for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
    double f1, f2, x_lu = *x;

    f1 = SPL_tt_calc( infos->base[ 0 ].graphe, x_lu, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    f2 = SPL_tt_calc( infos->base[ 1 ].graphe, x_lu, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    s1c += f1 * f1;
    s2c += f2 * f2;
    sg1 += *y * f1;
    sg2 += *y * f2;
    s12 += f1 * f2;
  }

  div = s1c-2*s12+s2c;
  if ( 0.0 == div ) {
    EXF_ERREUR(DIV_ZERO, "Spectres identiques", txt_err_calc_annule);
    return -10;
  }

  infos->base[ 0 ].coefficient.contenu.valeur = -( s12 - s2c - sg1 + sg2 ) / div;
  infos->base[ 1 ].coefficient.contenu.valeur = 1.0 - infos->base[ 0 ].coefficient.contenu.valeur;

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Recherche des coefficients : méthode des moyennes --------\
 |   On sait que E(g) = alpha * E(f1) + bêta * E(f2) [1]              |
 |   On résoud le système {[1], g_i = alpha * f1_i + bêta + f2_i}     |
 |       --> donne une série de (alpha_i, bêta_i)                     |
 |   On prend alpha = E(alpha_i), bêta = E(bêta_i)                    |
 \------------------------------------------------------------------- */

static int MODELE_Decomposition_Moyenne(Graphe *spectre, struct decomposition *infos)
{
  register double *x, *y;
  unsigned long i;
  double sp_moyen = 0.0, ref1_moyen = 0.0, ref2_moyen = 0.0;

  /* Contrôle : cette méthode ne fonctionne qu'avec 2 spectres */
  if ( infos->nombre_bases != 2 ) {
    return -20;
  }

  /* Première étape : calcul des sommes */
  /* Première équation : les valeurs moyennes vérifient <spectre> = coef_1 * <ref_1> + coef_2 * <ref_2>
     On calcule donc les valeurs moyennes
  */

  x = spectre->x; y = spectre->y;
  for ( i = 0 ; i < spectre->nombre_point ; i++, x++, y++ ) {
    sp_moyen   += *y; /* Le point compte */
    ref1_moyen += SPL_tt_calc( infos->base[ 0 ].graphe, *x, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    ref2_moyen += SPL_tt_calc( infos->base[ 1 ].graphe, *x, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );
  }
  sp_moyen /= spectre->nombre_point; /* On fait la moyenne */	
  ref1_moyen /= spectre->nombre_point;
  ref2_moyen /= spectre->nombre_point;

  /* Calcul effectif des coefficients - moyenne des résultats obtenus pour chaque point */
  infos->base[ 0 ].coefficient.contenu.valeur = 0.0; /* On met à zéro pour être tranquille */
  infos->base[ 1 ].coefficient.contenu.valeur = 0.0;
  x = spectre->x; y = spectre->y;
  for ( i = 0 ; i < spectre->nombre_point ; i++, x++, y++ ) {
    double delta;
    double y_1 = SPL_tt_calc( infos->base[ 0 ].graphe, *x, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    double y_2 = SPL_tt_calc( infos->base[ 1 ].graphe, *x, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    delta = y_2 * ref1_moyen - ref2_moyen * y_1;
    if ( 0.0 == delta ) {
      EXF_ERREUR(DIV_ZERO, "D"e_AIGU"terminant nul", txt_err_calc_annule);
    } else {
      double alpha, beta;

      alpha  = sp_moyen * y_2 - *y * ref2_moyen;
      alpha /= delta;
      beta   = ref1_moyen * *y - y_1 * sp_moyen;
      beta  /= delta;
	
      infos->base[ 0 ].coefficient.contenu.valeur += alpha;
      infos->base[ 1 ].coefficient.contenu.valeur += beta;
    }
  }
  infos->base[ 0 ].coefficient.contenu.valeur /= spectre->nombre_point; /* On divise par tout ça */
  infos->base[ 1 ].coefficient.contenu.valeur /= spectre->nombre_point;

  /* Terminé sans erreur */
  return 0;
}

/* -------- Recherche des coefficients, cas d'une espèce de suivi cinétique ----------\
 |  On suppose donc que alpha + bêta = 1 (deux espèces uniquement mises en oeuvres)   |
 |                  qu'en chaque point, g(t) = alpha(t) f1(t) + (1 - alpha(t)) f2(t)  |
 |  On résoud point par point et on obtient 2 graphes d'évolution de alpha et bêta    |
 \----------------------------------------------------------------------------------- */

static int MODELE_Decomposition_Suivi(Graphe *spectre, struct decomposition *infos)
{
  unsigned long i;
  double alpha;
  register double *x, *a, *b, *y, *xa, *xb;

  /* Contrôle : cette méthode ne fonctionne qu'avec 2 spectres */
  if ( infos->nombre_bases != 2 ) {
    return -20;
  }

  for (i = 0, x = spectre->x, y = spectre->y, a = infos->alpha->y, b = infos->beta->y,
	 xa = infos->alpha->x, xb = infos->beta->x;
       i < spectre->nombre_point ;
       i++, x++, y++) {
    double x_lu = *x;
    double y1, y2;

    y1 = SPL_tt_calc( infos->base[ 0 ].graphe, *x, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    y2 = SPL_tt_calc( infos->base[ 1 ].graphe, *x, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    if ( y2 == y1 ) {
      EXF_ERREUR(DIV_ZERO, "Valeurs identiques", "alpha = 0,5");
      alpha = 0.5; /* Pourquoi pas... */
    } else {
      alpha = (*y - y2) / (y1 - y2);
    }

    *xa = x_lu; xa++;
    *a = alpha; a++;
    *xb = x_lu; xb++;
    *b = 1 - alpha; b++;
  }

  GRF_bornes( infos->alpha, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( infos->beta, NE_PAS_FAIRE, DESORDONNE );

  /* Terminé sans erreur */
  return 0;
}

/* -------- Recherche des coefficients : méthode des moyennes ------------------------\
 |   On sait que E(g) = alpha * E(f1) + bêta * E(f2) [1]                              |
 |               V(g) = alpha * V(f1) + bêta * V(f2) + 2 * alpha * bêta * Cov(f1, f2) |
 |   On résoud donc ce système                                                        |
 \----------------------------------------------------------------------------------- */

static int MODELE_Decomposition_Variance(Graphe *spectre, struct decomposition *infos)
{
  double m1 = 0.0, m2 = 0.0, m = 0.0, mp = 0.0; /* Les diverses moyennes */
  double v1 = 0.0, v2 = 0.0, v = 0.0, covar = 0.0; /* Les diverses variances */
  unsigned long i;
  register double *x, *y;
  double A, B, C, delta, q, alpha;

  /* Contrôle : cette méthode ne fonctionne qu'avec 2 spectres */
  if ( infos->nombre_bases != 2 ) {
    return -20;
  }

  /* Calcul des valeurs moyennes */

  for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
    double y1, y2;

    y1 = SPL_tt_calc( infos->base[ 0 ].graphe, *x, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    y2 = SPL_tt_calc( infos->base[ 1 ].graphe, *x, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    m  += *y; /* Le point compte */
    m1 += y1;
    m2 += y2;
    mp += y1 * y2;
  }
  m  /= spectre->nombre_point; /* On fait la moyenne */	
  m1 /= spectre->nombre_point;
  m2 /= spectre->nombre_point;
  mp /= spectre->nombre_point;

  /* Calcul des variances, covariances,... */

  covar = mp - m1 * m2;
  printf("Covariance : %f", covar);

  for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
    double y1, y2;

    y1 = SPL_tt_calc( infos->base[ 0 ].graphe, *x, infos->base[ 0 ].a, infos->base[ 0 ].b, infos->base[ 0 ].c );
    y2 = SPL_tt_calc( infos->base[ 1 ].graphe, *x, infos->base[ 1 ].a, infos->base[ 1 ].b, infos->base[ 1 ].c );

    v  += (*y - m) * (*y - m); /* Le point compte */
    v1 += (y1 - m1) * (y1 - m1);
    v2 += (y2 - m2) * (y2 - m2);
  }
  v  /= spectre->nombre_point;
  v1 /= spectre->nombre_point;
  v2 /= spectre->nombre_point;

  printf(" - coefficient de corrélation : %f\n", covar / sqrt(v1 * v2));

  if ( 0.0 == m2 ) {
    EXF_ERREUR( DIV_ZERO, "Moyenne nulle", txt_err_calc_annule );
    return -15;
  }
  q = m1 / m2;

  A = v1 + q * q * v2 - 2 * covar * q;
  B = m/m2 * (covar - q * v2);
  C = (m/m2) * (m/m2) * v2 - v;

  /* Équation à résoudre : A * alpha² + 2 * B * alpha + C = 0 -> discriminant réduit */

  delta = B * B - A * C;
  if ( delta < 0.0 ) {
    printf("Aucune solution trouvée\n");
    return -16;
  }

  alpha = ( -B - sqrt(delta) ) / A;
  infos->base[ 0 ].coefficient.contenu.valeur = alpha;
  infos->base[ 1 ].coefficient.contenu.valeur = (m - alpha * m1) / m2;

  alpha = ( -B + sqrt(delta) ) / A ;
  printf( "Autre solution : a = %f, b = %f\n", alpha, (m - alpha * m1) / m2 );

  /* Terminé sans erreur */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Les routines annexes ~-~-~-~-~-~-~-~-~ */

/* -------- Contrôle de réalisation possible de l'ajustement -------- */

int MODELE_Decomposition_Controler(struct decomposition *decomposition)
{
  unsigned long i;

  if ( NULL == decomposition ) return -2;
  if ( ( NULL == decomposition->base ) || ( 0 == decomposition->nombre_bases ) ) return -3;

  for ( i = 0 ; i < decomposition->nombre_bases ; i++ ) {
    /* Est-ce que chaque vecteur de base a bien un vecteur de base associé ? */
    if ( NULL == decomposition->base[ i ].graphe ) {
      (void) fprintf( stderr, "%s,%s l.%u : Le vecteur de base %lu n'a pas de graphe associé.\n",
		      __FILE__, __FUNCTION__, __LINE__, i );
      return -4;
    }
  }

  /* Pas de problème détecté */
  return 0;
}

/* -------- Préparation de l'ajustement -------- */

int MODELE_Decomposition_PreparerAjustement(ModeleAjustement *modele)
{
  struct decomposition *decomposition;
  struct element_base *base;
  unsigned long i;

  /* Contrôles préliminaires */
  if ( NULL == modele ) return -1;
  if ( NULL == modele->experience ) {
    AVERTISSEMENT( "Appel avec un modèle sans spectre expérimental ! [%s]", modele->nom_modele );
    return -10;	 /* Ce ne devrait pas être possible, mais sait-on jamais... */
  }

  decomposition = modele->infos_modele.decomposition;
  if ( NULL == decomposition       ) return -2;
  if ( NULL == decomposition->base ) return -3;

  /* On efface d'éventuelles anciennes informations */
  GRF_eff_isole( decomposition->alpha ); decomposition->alpha = NULL;
  GRF_eff_isole( decomposition->beta  ); decomposition->beta  = NULL;

  /* Faut-il faire un suivi ? */
  if ( MODELE_DECOMPOSITION_SUIVI == decomposition->methode ) {
    decomposition->alpha = GRF_nv_TableX( NULL, GR_EPREUVE, modele->experience->genre_donnees, "Coef. a/Interne", VRAI,
					  modele->modele->x, modele->modele->nombre_point, VRAI );
    if ( NULL == decomposition->alpha ) {
      ERR_Memoire( "Coefficient a", sizeof(Graphe) );
      return -20;
    }

    decomposition->beta = GRF_nv_TableX( NULL, GR_EPREUVE, modele->experience->genre_donnees, "Coef. b/Interne", VRAI,
					 modele->modele->x, modele->modele->nombre_point, VRAI );
    if ( NULL == decomposition->beta ) {
      GRF_eff_isole( decomposition->alpha ); decomposition->alpha = NULL;
      ERR_Memoire( "Coefficient b", sizeof(Graphe) );
      return -20;
    }
  }

  /* On prépare les tables de splines */
  for ( i = 0, base = decomposition->base ; i < decomposition->nombre_bases ; i++, base++ ) {
    free( base->a ); base->a = NULL;
    free( base->b ); base->b = NULL;
    free( base->c ); base->c = NULL;

    SPL_tt_prep( base->graphe, &(base->a), &(base->b), &(base->c) );
  }

  /* Terminé sans problème : on renvoi 0 */
  return 0;
}
