/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2005, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Lecture de modles (partie gnrique)
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "systeme.h"
#include "ascii_utils.h"
#include "bin_utils.h"
#include "txt_utils.h"

#include "graphes.h"		/* Pour les genres de graphes = types de modles */
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define RACINE_DOMAINES PREFIXE_REP"/Lase/Modeles/"
#define LISTE_DOMAINES  RACINE_DOMAINES"/sommaire"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static char *MODELE_LireNomModele(FILE *fichier);

static int MODELE_GroupeGeneral  (FILE *fichier, struct modele_generique *modele);
static int MODELE_GroupeParametre(FILE *fichier, struct parametre *parametre);
static int MODELE_GroupeVariable (FILE *fichier, struct var_modele *variable);

static int ES_AtteindreValeurs(FILE *fichier);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Lecture de la liste des domaines dans lesquels existent des modles -------- */

int MODELE_ListeDomaines(unsigned long *nombre_domaines, const char ***noms_domaines)
{
  FILE *liste;
  int c;
  unsigned long i = 0;

  if ( ( NULL == nombre_domaines ) || ( NULL == noms_domaines ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramtre NULL !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* Retour en cas d'erreur */
  *nombre_domaines = 0;
  *noms_domaines = NULL;

  /* On ouvre la liste des dossiers connus */
  liste = fopen( LISTE_DOMAINES, "rb" );
  if ( NULL == liste ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'ouvrir %s\n",
		   __FILE__, __FUNCTION__, __LINE__, LISTE_DOMAINES);
    return -2;
  }

  /* Combien de domaines ? */
  *nombre_domaines = compte_lignes( liste, '#' );
  if ( 0 == *nombre_domaines ) {
    (void) fprintf(stderr, "%s, %s l.%u : Fichier %s vide !\n",
		   __FILE__, __FUNCTION__, __LINE__, LISTE_DOMAINES);
    (void) fclose( liste ); liste = NULL;
    return -2;
  }

  /* On rserve la table des domaines */
  *noms_domaines = (const char **) malloc( sizeof( char * ) * *nombre_domaines );
  if ( NULL == *noms_domaines ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( liste ); liste = NULL;
    return -3;
  }

  /* On lit ces domaines */
  c = fgetc( liste ); ungetc( c, liste );
  do {
    /* On ignore les commentaires */
    while ( '#' == c ) {
      c = ignore_ligne( liste );
      if ( EOF == c ) break;
    }

    /* Si on finit par un commentaire */
    if ( EOF == c ) break;

    /* On lit la ligne : nom du domaine */
    (*noms_domaines)[ i ] = lit_commande( liste );

    /* On passe au domaine suivant */
    i++;
    c = ignore_ligne( liste );
  } while ( c != EOF );

  /* On n'a plus besoin du fichier */
  fclose( liste ); liste = NULL;

  /* Tout s'est bien pass ! */
  return 0;
}

/* -------- Donne le nom du dossier contenant les modles du domaine indiqu [NULL si erreur] -------- */

char *MODELE_DossierDomaine(const char *nom_domaine)
{
  FILE *liste;
  char *domaine_lu;
  int c;

  if ( NULL == nom_domaine ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramtre NULL !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* On ouvre la liste des dossiers connus */
  liste = fopen( LISTE_DOMAINES, "rb" );
  if ( NULL == liste ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'ouvrir %s\n",
		   __FILE__, __FUNCTION__, __LINE__, LISTE_DOMAINES);
    return NULL;
  }

  /* On lit les domaines, un par un */
  c = fgetc( liste ); ungetc( c, liste );
  do {
    /* On ignore les commentaires */
    while ( '#' == c ) {
      c = ignore_ligne( liste );
      if ( EOF == c ) break;
    }

    /* Si on finit par un commentaire */
    if ( EOF == c ) break;

    /* On lit la ligne : nom du domaine */
    domaine_lu = lit_commande( liste );
    if ( NULL == domaine_lu ) {
      (void) fprintf(stderr, "%s,%s l.%u : Ligne vide ? Erreur de fichier ?\n",
		     __FILE__, __FUNCTION__, __LINE__);
    } else {
      /* On compare au nom de domaine cherch */
      if ( 0 == strcmp( domaine_lu, nom_domaine ) ) {
	char *nom_dossier, *nom_complet;
	size_t lg_nom = 0;

	/* On a donc trouv le bon domaine */
	nom_dossier = lit_commande( liste );
	if ( NULL == nom_dossier ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Attention, pas de dossier associ au domaine %s !\n",
			 __FILE__, __FUNCTION__, __LINE__, domaine_lu);
	}

	/* Plus besoin du nom lu */
	free( domaine_lu ); domaine_lu = NULL;
	(void) fclose( liste ); liste = NULL;

	/* On ajoute la racine */
	lg_nom = strlen( nom_dossier ) + strlen( RACINE_DOMAINES ) + 2; /* NPO le 0 final et le / sparateur */
	nom_complet = (char *) malloc( lg_nom );
	if ( NULL == nom_complet ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ajouter la racine !\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return nom_dossier;
	}

	strcpy( nom_complet, RACINE_DOMAINES"/" );
	strcat( nom_complet, nom_dossier );

	free( nom_dossier ); nom_dossier = NULL;

	return nom_complet;
      }
    }

    /* On passe au domaine suivant */
    free( domaine_lu ); domaine_lu = NULL;
    c = ignore_ligne( liste );
  } while ( c != EOF );

  /* On n'a plus besoin du fichier */
  (void) fclose( liste ); liste = NULL;

  /* Tout s'est bien pass ! */
  return 0;
}

/* -------- Lire tous les modles dans un dossier -------- */

int MODELE_ListeModeles(const char *nom_dossier, unsigned long *nombre, char ***noms_modeles, char ***noms_fichiers)
{
  char *nom_complet;
  unsigned long i;
  FILE *fichier;

  if ( NULL == nom_dossier ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de nom de dossier !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  if ( ( NULL == nombre ) || ( NULL == noms_modeles ) || ( NULL == noms_fichiers ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramtre NULL !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* On commence par la liste des fichiers */
  *noms_fichiers = DIR_ListeFichiers( nom_dossier, NULL, nombre );
  if ( ( NULL == noms_fichiers ) || ( 0 == *nombre ) ) {	/* Aucun fichier dans le dossier ! */
    *nombre = 0;
    *noms_modeles = NULL;
    return 0;
  }

  /* On rserve la table des titres */
  *noms_modeles = (char **) calloc( *nombre, sizeof( char *) );
  if ( NULL == *noms_modeles ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de rserver la table des noms de modles !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }

  /* Fichier par fichier, on ouvre et on trouve le titre du modle */
  for ( i = 0 ; i < *nombre ; i++ ) {
    nom_complet = (*noms_fichiers)[ i ];
    if ( NULL == nom_complet ) break;

    fichier = fopen( nom_complet, "rb" );
    if ( NULL == fichier ) {
      (void) fprintf(stderr, "%s,%s l.%u : Fichier absent ou inaccessible, %s\n",
		     __FILE__, __FUNCTION__, __LINE__, nom_complet);
      (*noms_modeles)[ i ] = strdup( "-* inaccessible *" );
    } else {
      (*noms_modeles)[ i ] = MODELE_LireNomModele( fichier );
      (void) fclose( fichier );
    }

/*     free( nom_complet ); nom_complet = NULL; */
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Lire le nom d'un modle dans son fichier -------- */

static char *MODELE_LireNomModele(FILE *fichier)
{
  int c;

  /* On suppose que l'on est en dbut de ligne ! */
  c = fgetc( fichier );
  while ( c != 'N' ) {
    if ( EOF == c ) {
      /* Fin de fichier atteinte sans trouver de titre ! */
      return strdup( "Sans titre !" );
    }

    c = ignore_ligne( fichier );
  }

  /* Si on est l : ligne commenant par N
     On suppose que c'est la premire --> le titre du modle [mot-clef : Nom] */
  do {
    c = fgetc( fichier );
    if ( ( EOF == c ) || ( '\n' == c ) || ( '\r' == c ) ) {
      /* Fin de fichier ou de ligne atteinte sans trouver le = : trange ! */
      (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect, mot-clef sans signe =\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return strdup( "Format incorrect" );
    }
  } while ( c != '=' );

  /* On saute les (ventuels) espaces suivants */
  c = fgetc( fichier );
  while ( isspace( c ) ) {
    c = fgetc( fichier );
  }
  ungetc( c, fichier ); /* On annule la lecture du dernier caractre, pour ne pas le rater ! */

  /* On lit la fin de la ligne : donne le titre du fichier */
  return lit_ligne( fichier, 0 );
}

/* -------- Lire une description de modle dans un fichier et en dduire le modle gnrique associ -------- */

struct modele_generique *MODELE_LireModele(const char *nom_fichier)
{
  char *nom_modele;
  int c, retour;
  unsigned long param_lu = 0;
  struct modele_generique *modele = NULL;
  FILE *fichier;

  if ( NULL == nom_fichier ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de nom de dossier !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "rb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de lire le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    return NULL;
  }

  /* On rcupre le nom du modle */
  nom_modele = MODELE_LireNomModele( fichier );
  if ( NULL == nom_modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de nom de modle (ou erreur de lecture) [fichier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    (void) fclose( fichier );
    return NULL;
  }

  /* On s'en sert pour construire le modle */
  modele = MODELE_NouveauGenerique( nom_modele );
  if ( NULL == modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer un nouveau modle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return NULL;
  }

  /* On se replace au dbut du fichier */
  rewind( fichier );

  /* On repre les dbuts de groupes */
  do {
    c = fgetc( fichier );

    if ( EOF == c ) break;	/* Fin de fichier atteinte */
    if ( '[' == c ) {
      c = fgetc( fichier );
      retour = -1;

      (void) fprintf(stderr, "%s,%s l.%u : On commence un groupe [code %c]\n",
		     __FILE__, __FUNCTION__, __LINE__, c);

      /* Dtection du groupe */
      if ( 'G' == c ) {
	c = ignore_ligne( fichier ); /* On ignore la fin de la ligne */
	retour = MODELE_GroupeGeneral( fichier, modele );
      } else if ( 'X' == c ) {
	c = ignore_ligne( fichier ); /* On ignore la fin de la ligne */
	retour = MODELE_GroupeVariable( fichier, &(modele->X) );
      } else if ( 'Y' == c ) {
	retour = MODELE_GroupeVariable( fichier, &(modele->Y) );
      } else if ( 'P' == c ) {
	/* On lit le paramtre suivant */
	if ( param_lu >= modele->nombre_parametres ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Paramtre en trop [%lu dclars, c'est le n %lu] : ignor\n",
			 __FILE__, __FUNCTION__, __LINE__, modele->nombre_parametres, param_lu + 1 );
	} else {
	  retour = MODELE_GroupeParametre( fichier, modele->parametres + param_lu );
	}

	/* On passera ensuite au paramtre suivant ! */
	param_lu++;
      } else if ( EOF == c ) {
	(void) fprintf(stderr, "%s,%s l.%u : [ en fin de fichier, format de fichier incorrect !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Groupe inconnu [%c, fichier %s] --- ignor\n",
		       __FILE__, __FUNCTION__, __LINE__, c, nom_fichier );
      }

      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Problme en lisang le groupe %c (fichier %s)\n",
		       __FILE__, __FUNCTION__, __LINE__, c, nom_fichier );
      }
    } else {
      if ( '#' == c ) { 	/* Commentaire */
      } else if ( isspace( c ) ) { /* Ligne vide */
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Commande hors-groupe, ignore ! [%c]\n",
		       __FILE__, __FUNCTION__, __LINE__, c);
      }
      c = ignore_ligne( fichier );
    }
  } while ( c != EOF );

  /* Termin */
  (void) fclose( fichier );
  return modele;
}

/* -------- Lecture des donnes du groupe gnral -------- */

static int MODELE_GroupeGeneral(FILE *fichier, struct modele_generique *modele)
{
  int c, retour;

  do {
    c = fgetc( fichier );
    /* On annule alors la lecture */
    ungetc( c, fichier );

    /* Est-ce que l'on tombe sur le groupe suivant ? */
    if ( '[' == c ) {
      /* Et on a termin, sans souci */
      return 0;
    }

    /* Est-ce que l'on tombe sur un commentaire ? */
    if ( '#' == c ) {
      c = ignore_ligne( fichier );
    } else { /* Si non : c'est que l'on est sur une commande */
      /* Analyse de la commande */
      if ( 'N' == c ) {
	/* Nom du modle : normalement, dj lu */
	c = ignore_ligne( fichier );
      } else if ( 'P' == c ) { /* Nombre de paramtres */
	unsigned long nbr_parametres;

	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Params]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Combien de paramtres dans le modle ? */
	fscanf( fichier, "%lu", &nbr_parametres );
	c = ignore_ligne( fichier );

	/* On cre ces paramtres */
	if ( nbr_parametres > 0 ) {
	  if ( 0 == modele->nombre_parametres ) {
	    retour = MODELE_Generique_AjouterParametres( modele, nbr_parametres );
	    if ( retour != 0 ) {
	      (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant crer les %lu paramtres du modle !\n",
			     __FILE__, __FUNCTION__, __LINE__, nbr_parametres );
	      return -5;
	    }
	  } else {
	    (void) fprintf(stderr, "%s,%s l.%u : Nombre de paramtres dj indiqu auparavant ! Nouvelle valeur ignore...\n",
			   __FILE__, __FUNCTION__, __LINE__);
	  }
	}
      } else if ( 'T' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* On saute les espaces ventuelles */
	c = fgetc( fichier );
	while ( isspace( c ) ) {
	  if ( ( '\r' == c ) || ('\n' == c ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Fin de ligne atteinte sans valeur aprs = [Type]\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    return -8;
	  }
	  c = fgetc( fichier );
	}

	/* On analyse la valeur --> type de modle */
	switch ( c )
	{
	case 'C': case 'c': modele->type = GRF_FCT_COMPLEXE; break; /* C = modle complexe, Z=f(X) */
	case 'P': case 'p': modele->type = GRF_COURBE_PLANE; break; /* P = paramtrique, (X,Y)=f(T) */
	case 'R': case 'r': modele->type = GRF_FCT_REELLE;   break; /* R = modle classique, Y=f(X) */
	case 'S': case 's': modele->type = GRF_SERIE;        break; /* S = srie de valeurs, Y */
	default:
	  (void) fprintf(stderr, "%s,%s l.%u : Type de modle inconnu [%c], ignor\n",
			 __FILE__, __FUNCTION__, __LINE__, c);
	  return -20;
	}

	/* On passe  la ligne suivante... */
	c = ignore_ligne( fichier );
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Commande inconnue [%c] -- ignore\n",
		       __FILE__, __FUNCTION__, __LINE__, c);
	c = ignore_ligne( fichier );
      }
    }
  } while ( c != EOF );

  /* Si on est l : fin de fichier atteinte */
  return -1;
}

static int MODELE_GroupeParametre(FILE *fichier, struct parametre *parametre)
{
  int c, retour = -5;

  if ( NULL == parametre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'endroit o stocker le paramtre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -5;
  }

  do { 
    c = fgetc( fichier );

    /* Est-ce que l'on tombe sur le groupe suivant ? */
    if ( '[' == c ) {
      ungetc( c, fichier );	/* Pour que le tour suivant dtecte le [ ! */
      /* Et on a termin, sans souci */
      return 0;
    }

    /* Est-ce que l'on tombe sur un commentaire ? */
    if ( '#' == c ) {
      c = ignore_ligne( fichier );
    } else { /* Si non : c'est que l'on est sur une commande */
      if ( 'D' == c ) {
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Derive]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Derive]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lire la formule de la drive par rapport  ce paramtre */
	parametre->derivee = lit_mot( fichier );
      } else if ( 'E' == c ) {
	double erreur;

	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Erreur]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lire la valeur numrique de l'erreur */
	retour = lit_valeurs( fichier, 1, 0, &erreur, NULL );
	parametre->erreur = erreur;
      } else if ( 'M' == c ) {
	c = fgetc( fichier );
	if ( 'a' == c ) {
	  double max;

	  retour = ES_AtteindreValeurs( fichier );
	  if ( retour != 0 ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Min]\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    return -2;
	  }

	  /* Lire la valeur numrique du minimum autoris */
	  retour = lit_valeurs( fichier, 1, 0, &max, NULL );
	  parametre->variable.max = max;
	} else if ( 'i' == c ) {
	  double min;

	  retour = ES_AtteindreValeurs( fichier );
	  if ( retour != 0 ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Max]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	    return -2;
	  }

	  /* Lire la valeur numrique du maximum autoris */
	  retour = lit_valeurs( fichier, 1, 0, &min, NULL );
	  parametre->variable.min = min;
	} else {
	  (void) fprintf(stderr, "%s,%s l.%u : Commande inconnue, M%c --- ignore\n",
			 __FILE__, __FUNCTION__, __LINE__, c);
	}
      } else if ( 'N' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Nom, paramtre]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture du nom du paramtre */
	parametre->nom = lit_mot( fichier );
      } else if ( 'S' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Symbole]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture du nom du paramtre [ne peut pas contenir d'espace] */
	parametre->symbole = lit_commande( fichier );
      } else if ( 'U' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Unit, paramtre]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture de l'unit */
	parametre->unite = lit_mot( fichier );
      }
    }

  } while ( c != EOF );

  /* Si on est l : fichier fini, pas de souci particulier */
  return 0;
}

static int MODELE_GroupeVariable (FILE *fichier, struct var_modele *variable)
{
  int c, retour;

  if ( NULL == variable ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'endroit o stocker la variable !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -5;
  }

  do {
    c = fgetc( fichier );

    /* Est-ce que l'on tombe sur le groupe suivant ? */
    if ( '[' == c ) {
      ungetc( c, fichier );	/* Pour que le tour suivant dtecte le [ ! */
      /* Et on a termin, sans souci */
      return 0;
    }

    /* Est-ce que l'on tombe sur un commentaire ? */
    if ( '#' == c ) {
      c = ignore_ligne( fichier );
    } else { /* Si non : c'est que l'on est sur une commande */
      if ( 'F' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture de la formule */
	variable->formule = lit_mot( fichier );
      } else if ( 'L' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture de la lgende */
	variable->legende = lit_mot( fichier );
      } else if ( 'T' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture du type de grandeur physique */
	variable->type = lit_mot( fichier );
      } else if ( 'U' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture de l'unit */
	variable->unite = lit_mot( fichier );
      } else if ( 'V' == c ) {
	retour = ES_AtteindreValeurs( fichier );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Format de fichier incorrect [Type]\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2;
	}

	/* Lecture de la formule pour l'incertitude */
	variable->formule_erreur = lit_mot( fichier );
      }
    }

  } while ( c != EOF );

  /* Si on est l : fichier termin ; a priori, pas d'erreur */
  return 0;
}

/* -------- Macro : atteindre la valeur associe au mot-clef, spare par un = -------- */

static int ES_AtteindreValeurs(FILE *fichier)
{
  int c;

  do {
    c = fgetc( fichier );
    if ( EOF == c ) {
      (void) fprintf(stderr, "%s,%s l.%u : Fin de fichier atteinte [mot-clef sans valeur associe ?]\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return EOF;
    }
    if ( ( '\n' == c ) || ( '\r' == c ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Fin de ligne atteinte, pas de = trouv\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return '\r';
    }
  } while ( c != '=' ) ;

  /* Ici, on est donc sur le premier caractre qui suit le = */
  return 0;
}
