/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1998, Emmanuel CURIS
   Date de cration      : 28 septembre 1998

   Module d'ajustement par un modle  n couches

   Calcul d'un signal EXAFS  partir des rsultats d'ajustement
          des drives partielles du signal EXAFS par rapport aux diffrentes variables
	  du rsidu et des erreurs absolues et relatives
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "graphes.h"
#include "lissage.h"
#include "science.h"

#include "atomes.h"
#include "couches.h"
#include "extrait.h" /* Pour la pondration : EXF_k_p_n */
#include "feff.h"
#include "modeles.h"
#include "theorie.h" /* Pour les modles de libre parcours moyen */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* #define deboguer_ajustement 1 *//* Dcommenter pour garder le modle  chaque tape */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int EXAFS_Calcul_FEFF (struct modele_feff *feff, struct graphe *exafs);

static double derivee_delta_e(struct modele_feff *feff, Couche *couche, double k, double khi);
static double derivee_gamma  (struct modele_feff *feff, Couche *couche, double k, double khi);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Calcul du spectre EXAFS pour un modle choisi -------- */

int NC_CalculeEXAFS(struct modele_ajuste *modele)
{
  if (NULL == modele) return -1;

  switch (modele->type)
  {
  case MODELE_EXAFS_CHEMINS: /* Modle FEFF simple */
    return EXAFS_Calcul_FEFF( modele->infos_modele.feff, modele->modele);

  default:
    (void) fprintf(stderr, "NC_CalculeEXAFS : pas de modle de calcul pour un modle de type %d\n", modele->type);
    return -20;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Calcul du spectre EXAFS pour un modle FEFF seulement -------- */

static int EXAFS_Calcul_FEFF(struct modele_feff *feff, struct graphe *exafs)
{
  double *y, s0_2;
  unsigned long i;
  Couche *couche;

  if (NULL == exafs) return -1;
  if (NULL == feff ) return -2;

  /* On initialise  zro les y du graphe */
  for (i = 0, y = exafs->y ; i < exafs->nombre_point ; i++, y++)
    *y = 0.0;

  /* On calcule la contribution de chacune des couches utilises du modle */
  for (couche = feff->couches ; couche != NULL ; couche = couche->suivante) {
    int retour;

    if ( FAUX == couche->utilisee ) continue;

    retour = NC_CalculeEXAFSCouche(feff, couche, exafs);
    if (retour != 0) /* Il y a eu un problme */
      return retour;
  }

  /* Pondration par S0 */
  s0_2 = VAR_LireValeur( &(feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE );
  GRF_FacteurMultiplicatif( exafs, s0_2, AXE_Y, NULL );

  /* Pondration ventuelle par k^n (sans historique) */
  EXF_k_p_n(exafs, feff->ponderation, FAUX);

  /* Remarque : on ne tient pas compte de la contribution, car peut servir parfois comme seul modle */

#ifdef deboguer_ajustement
  {
    static unsigned int appel = 0;
    Graphe *grf;
    char *titre;

    appel++;
    grf = GRF_Copier( exafs, NULL, FAUX ); /* Pour en garder une trace : analyse des rsultats */
    titre = TXT_genere_chaine("EXAFS_Calcul_FEFF, appel %u", appel);
    GRF_titre(grf, titre);
    free(titre);
  }
#endif

  /* Tout s'est bien pass */
  return 0;
}

#ifdef ancien
/* -------- Calcul de tous les spectres exafs qui interviennent dans un ajustement
  Travaille sur toute la srie des infos */

int calcule_exafs(Infos_ncouches *infos)
{
  int retour = 0;
  Infos_ncouches *bloc;

  for (bloc = infos ; bloc != NULL ; bloc = bloc->suivante)
  {
    switch (infos->nbr_absorbeurs)
    {
    case 0: /* On utilise toutes les couches disponibles, pas d'absorbeur */
      retour = calc_exafs_ncouches(premiere_couche, infos->ajuste);
      break;

    case 1: /* Un seul absorbeur -> infos->ajuste = absorbeur->exf_calcule */
/*       retour = ABS_calc_exafs(bloc->absorbeurs[0]); */
      break;

    default: /* n absorbeurs, chacun a son graphe propre associ  son signal */
      INF_calc_exafs(bloc);
    }
  }

  return retour; /* Le calcul a eu lieu - 0 si tout va bien, autre en cas de problme */
}

/* -------- Calcul du signal exafs associ  l'ensemble des couches -------\
 |  Renvoie : 0 si tout s'est bien pass, un code d'erreur sinon           |
 |                                                                         |
 | prem_couche : premire couche du modle                                 |
 | exafs       : le spectre contenant le rsultat                          |
 |               (le signal est calcul pour chacun des x=k de ce graphe)  |
 \------------------------------------------------------------------------ */

int calc_exafs_ncouches(Couche *prem_couche, Graphe *exafs)
{
  unsigned long i;
  register double *x, *y;
  Couche *couche;
  double s0_2;

  if ( (exafs == NULL) || (prem_couche == NULL) )
    return -1; /* Pas d'entre ! */

  if ( exafs->drapeaux.genre != GR_NORMAL ) /* Le spectre doit tre de type normal */
    return -2;

  /* On calcule maintenant pour les paramtres gnraux : multiplication par S0 et pondration par k^n */

  GRF_bornes(exafs, ORDONNE, DESORDONNE); /* Pour tre sr de l'affichage... */

  return 0; /* Pas de problme */
}

/* -------- Calcul du signal EXAFS associ  _un seul_ absorbeur -------- */

int ABS_calc_exafs(Graphe *exafs, Modele_FEFF *modele_feff)
{
  Couche *couche;
  double facteur;
  unsigned long i;
  register double *x, *y;

  /* ------ On met  zro le spectre ------ */
  for (i = 0, y = exafs->y ; i < exafs->nombre_point ; i++, y++)
    *y = 0.0;

  /* ------ Calcul du signal provenant de chaque couche de cet absorbeur */
  if (VRAI == modele_feff->drapeaux.DE0_general) {
    int retour;

    for (couche = modele_feff->couches ; couche != NULL ; couche = couche->suivante) {

      if (FAUX == couche->utilisee)
	continue;

      retour |= NC_CalculeEXAFSCouche(couche, exafs, &(absorbeur->Delta_E));
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "ABS_calc_exafs : erreur dans exafs_1_couche !\n");
    }
  } else { /* DE0 individuels */
    int retour;

    for (couche = modele_feff->couches ; couche != NULL ; couche = couche->suivante) {
      if (FAUX == couche->utilisee)
	continue;

      retour |= NC_CalculeEXAFSCouche(couche, exafs, &(couche->Delta_E));
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "ABS_calc_exafs : erreur dans exafs_1_couche !\n");
    }
  }

  /* ------ Pondration par les paramtres associs  l'absorbeur ------ */
  facteur  = VAR_LireValeur( &(modele_feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE );
  facteur *= VAR_LireValeur( &(absorbeur->contribution), VARIABLE_LIRE_VALEUR_AJUSTEE );
  GRF_FacteurMultiplicatif( exafs, facteur, AXE_Y);
  EXF_k_p_n(exafs, modele_feff->ponderation, FAUX); /* Pondration par k^n */

  return 0; /* Aucun problme */
}
#endif

/* -------- Calcul du signal EXAFS associ  un/plusieurs absorbeur/s de mme type ---\
 |   Renvoie 0 si tout s'est bien pass, != 0 sinon                                   |
 |                                                                                    |
 | infos = les infos d'ajustement. Travaille sur UNE SEULE famille                    |
 \----------------------------------------------------------------------------------- */

#ifdef ancien
void INF_calc_exafs(Infos_ncouches *infos)
{
  unsigned long i;
  unsigned int j;
  register double *y;
  Graphe *exafs;

  exafs = infos->ajuste; /* Le graphe que l'on calcule */

  /* ------ On met  zro les spectres ------ */
  for (i = 0, y = exafs->y ; i < exafs->nombre_point ; i++, y++)
  {
    *y = 0.0;
  }

  /* ------ Contribution de chaque absorbeur ------ */
  for (j = 0 ; j < infos->nbr_absorbeurs ; j++)
  {
    Absorbeur *absorbeur = infos->absorbeurs[j];
    register double *y_abs;

    ABS_calc_exafs(absorbeur); /* Calcul du spectre associ */

    for (i = 0, y = exafs->y, y_abs = absorbeur->exf_calcule->y ;
	 i < exafs->nombre_point ; i++, y++, y_abs++)
      *y += *y_abs; /* On l'ajoute au spectre total */
  }

  GRF_bornes(exafs, NE_PAS_FAIRE, DESORDONNE); /* Pour les redessins */
}
#endif

/* -------- Calcule la contribution d'une couche au signal EXAFS ---------\
 |   Renvoie 0 si tout s'est bien pass, != 0 sinon                       |
 |   AJOUTE la contribution  exafs, pour les X de exafs                  |
 | --> ne multiplie PAS par S0_2 (multiplication globale  la fin)        |
 | --> ne pondre PAS par k^n (car global  la fin - mais douteux :  voir |
 |                                                                        |
 | feff    : informations du modle FEFF associes                        |
 | couche  : la couche pour laquelle on veut la contribution              |
 | exafs   : le spectre dans lequel le signal EXAFS est enregistr        |
 \----------------------------------------------------------------------- */

int NC_CalculeEXAFSCouche(struct modele_feff *feff, Couche *couche, Graphe *exafs)
{
  unsigned long i;
  register double *x, *y;
  double delta_e;
  double N, longueur;

  if ( (couche->b_amp == NULL) || (couche->c_amp == NULL) || (couche->d_amp == NULL)
    || (couche->amplitude == NULL) )
    return -1; /* Pas d'informations pour les amplitudes */

  if ( (couche->b_pha == NULL) || (couche->c_pha == NULL) || (couche->d_pha == NULL)
    || (couche->phase == NULL) )
    return -2; /* Pas d'informations pour les phases */

  /* Valeurs des variables : DE0, R, N */
  if ( VRAI == feff->drapeaux.DE0_general )
    delta_e = VAR_LireValeur( &(feff->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE );
  else
    delta_e = VAR_LireValeur( &(couche->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE );
  longueur = VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_AJUSTEE );
  N = VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_AJUSTEE );

  for (i = 0, x = exafs->x, y = exafs->y ; i < exafs->nombre_point ; i++, x++, y++) {
    double amplitude, phase, debye_waller;
    double k_reel = *x, k_p_n; /* En fonction de Delta_E */

    if (delta_e != 0.0) {
      k_reel *= *x;
      k_reel += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE*H_BARRE * M_V_ANGSTROM*M_V_ANGSTROM) * delta_e;
      k_reel = sqrt(k_reel);
    }
    if (k_reel == 0.0)
      continue;

    amplitude  = SPL_tt_calc(couche->amplitude, k_reel, couche->b_amp, couche->c_amp, couche->d_amp);
    amplitude *= N;
    amplitude /= longueur * longueur;
    amplitude /= k_reel; /*  VOIR : ne faudrait-il pas mettre k ici ?
			    principe : k_reel correspond  la correspondance k rel / k FEFF
			    ==> k_reel ne servirait que pour les amplitudes et phases ??

			    D'un autre ct, les k exprimentaux sont faux : on peut dire aussi que k_reel
			    est la vraie valeur de k ==> celle  utiliser partout pour avoir un calcul correct
			    (mais ce ne sont pas forcment les valeurs des k de FEFF non plus, donc ?)

			    MTHODE ACTUELLE : deuxime solution, semble donner de bons rsultats
			 */

    phase  = SPL_tt_calc(couche->phase, k_reel, couche->b_pha, couche->c_pha, couche->d_pha);
    phase += 2 * k_reel * longueur; /*  VOIR : k_reel, mme remarque que ci-dessus */

    if (feff->type_lpm == MODELE_LPM_GRAPHE) { /* Libre parcours moyen dans un graphe spcial */
      double lpm = SPL_tt_calc(feff->lpm, k_reel, feff->b_lpm, feff->c_lpm, feff->d_lpm);

      if (lpm == 0.0) {
	EXF_ERREUR(DIV_ZERO, "Libre parcours moyen nul", txt_err_val_nulle);
	amplitude = 0.0; /* exp(-infini) */
      } else 
	amplitude *= exp ( -2.0 * longueur / lpm );
    } else if (feff->type_lpm == MODELE_LPM_GAMMA) { /* Libre parcours moyen lambda(k) = k / Gamma */
      double gamma = 0.0;

      if (VRAI == feff->drapeaux.GAM_general)
	gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
      else
	gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

      if (gamma != 0.0)
	amplitude *= exp ( -2.0 * longueur / (k_reel / gamma) ); /*  VOIR : k_reel, mme remarque */
    } else if (feff->type_lpm == MODELE_LPM_GAMMA_ETA) { /* Libre parcours moyen lambda(k) = (k + (ta/k)^4) / Gamma */
      double lpm, gamma = 0.0, e_s_k;

      if (VRAI == feff->drapeaux.GAM_general)
	gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
      else
	gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

      if ( gamma != 0.0 ) {
	e_s_k = feff->lpm_val_eta / k_reel;
	e_s_k *= e_s_k; /* (ta / k)^2 */
	e_s_k *= e_s_k; /* (ta / k)^4 */
	lpm = (k_reel + e_s_k) / gamma;

	amplitude *= exp( -2.0 * longueur / lpm);
      }
    }

    k_p_n = k_reel * k_reel; /* k */
    debye_waller  = VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_AJUSTEE ); /* Contient dj sigma */
    debye_waller *= -2.0 * k_p_n;

    k_p_n *= k_reel; /* k */
    phase -= 4.0 * VAR_LireValeur( &(couche->Cumul_2), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 3.0 * k_p_n;

    k_p_n *= k_reel; /* k^4 */
    debye_waller += 2.0 * VAR_LireValeur( &(couche->Cumul_1), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 3.0 * k_p_n;

    k_p_n *= k_reel; /* k^5 */
    phase += 4.0 * VAR_LireValeur( &(couche->Cumul_4), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 15.0 * k_p_n;

    k_p_n *= k_reel; /* k^6 */
    debye_waller -= 4.0 * VAR_LireValeur( &(couche->Cumul_3), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 45.0 * k_p_n;

    amplitude *= exp(debye_waller);
    if ( is_inf( amplitude ) ) {
      fprintf( stderr, "%s : Valeur infinie de l'amplitude : pour k = %f, debye_waller = %f\n",
	       __FUNCTION__, k_reel, debye_waller);
      return -4;
    }

    /* On _rajoute_ au signal dj existant */
    *y += amplitude * sin( phase );
  }

  return 0; /* Tout s'est bien pass */
}

/* -------- Calcul du signal EXAFS pour une couche en un point -----------\
 |  Renvoie la valeur du signal EXAFS au point considr                  |
 |                                                                        |
 | k       = le point pour lequel il faut calculer le signal EXAFS        |
 | feff    = le modle utilis (infos pour le lpm,...)                    |
 | couche  = la couche  utiliser pour faire le calcul                    |
 | delta_e = la valeur de DE0  utiliser                                  |
 |                                                                        |
 | Note : k_reel = k, cf remarques plus haut dans le code                 |
 \----------------------------------------------------------------------- */

double calc_exafs_1point(double k, Modele_FEFF *feff, Couche *couche, double delta_e)
{
  double amplitude, phase, debye_waller, amp;
  double k_reel;
  double k_p_n; /* En fonction de Delta_E */
  double longueur;

  k_reel = k;
  if ( delta_e != 0.0 ) { /* Correction d'nergie  faire */
    k_reel *= k;
    k_reel += 2.0 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE*H_BARRE * M_V_ANGSTROM*M_V_ANGSTROM) * delta_e;
    k_reel  = sqrt(k_reel);
  }
  if (k_reel == 0.0) {
#ifdef _DEVELOPPE_
    printf("calc_exafs_1point : k rel nul !\n");
#endif
    return 0.0;
  }

  k = k_reel; /* cf. plus haut, modle de correction de k retenu */ 

  longueur = VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_AJUSTEE );

  amplitude  = SPL_tt_calc(couche->amplitude, k_reel, couche->b_amp, couche->c_amp, couche->d_amp);
  amplitude *= VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_AJUSTEE );
  amplitude /= longueur * longueur;
  /*  amplitude /= k; */

  if (feff->ponderation == 0) /* Pas de pondration */
    amplitude /= k;
  else if (feff->ponderation == 2) /* Pondration par k */
    amplitude *= k;       /* car se simplifie avec le /k */

  phase  = SPL_tt_calc(couche->phase, k_reel, couche->b_pha, couche->c_pha, couche->d_pha);
  phase += 2 * k * longueur;   

  if (feff->type_lpm == MODELE_LPM_GRAPHE) { /* Libre parcours moyen dans un graphe spcial */
    double lpm = SPL_tt_calc(feff->lpm, k_reel, feff->b_lpm, feff->c_lpm, feff->d_lpm);

    printf("On est ici - exafs en 1 point\n");
    amplitude *= exp ( -2 * longueur / lpm );
  } else if (feff->type_lpm == MODELE_LPM_GAMMA) { /* Libre parcours moyen lambda(k) = k / Gamma */
    double gamma = 0.0;

    if (VRAI == feff->drapeaux.GAM_general)
      gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
    else
      gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

    if (gamma != 0)
      amplitude *= exp ( -2.0 * longueur / (k / gamma) );
  } else if (feff->type_lpm == MODELE_LPM_GAMMA_ETA) { /* Libre parcours moyen lambda(k) = (k + (ta/k)^4) / Gamma */
    double lpm, gamma = 0.0, e_s_k;

    if (VRAI == feff->drapeaux.GAM_general)
      gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
    else
      gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

    e_s_k = feff->lpm_val_eta / k;
    e_s_k *= e_s_k; /* (ta / k)^2 */
    e_s_k *= e_s_k; /* (ta / k)^4 */
    lpm = (k + e_s_k) / gamma;
    
    amplitude *= exp( -2.0 * longueur / lpm);
  }

  k_p_n = k * k; /* k */
  debye_waller  = VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_AJUSTEE ); /* Contient dj sigma */
  debye_waller *= -2.0 * k_p_n;

  if (feff->ponderation == 3) /* Pondration par k ; multiplier par k car se simplifie avec le 1/k */
    amplitude *= k_p_n;

  k_p_n *= k; /* k */
  phase -= 4 * VAR_LireValeur( &(couche->Cumul_2), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 3 * k_p_n;

  if (feff->ponderation == 4)
    amplitude *= k_p_n; /* Pondration par k^4 */

  k_p_n *= k; /* k^4 */
  debye_waller += 2 * VAR_LireValeur( &(couche->Cumul_1), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 3 * k_p_n;

  if (feff->ponderation == 5) /* Pondration par k^5 */
    amplitude *= k_p_n;

  k_p_n *= k; /* k^5 */
  phase += 4 * VAR_LireValeur( &(couche->Cumul_4), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 15 * k_p_n;

  if (feff->ponderation == 6) /* Pondration par k^6 */
    amplitude *= k_p_n;

  k_p_n *= k; /* k^6 */
  debye_waller -= 4 * VAR_LireValeur( &(couche->Cumul_3), VARIABLE_LIRE_VALEUR_AJUSTEE ) / 45 * k_p_n;

  if (feff->ponderation == 7) /* Pondration par k^7 */
    amplitude *= k_p_n;

  amp = exp(debye_waller);
  if ( is_inf(amp) ) {
    fprintf(stderr, "calc_exafs_1pts : amplitude infinie pour k = %f\n", k);
  } else {
    amplitude *= amp;
  }

  return VAR_LireValeur( &(feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE ) * amplitude * sin ( phase );
}

/* -------- Calcul de la drive partielle du signal EXAFS ----------------\
 |     par rapport  une variable donne                                   |
 |     en un point donn                                                   |
 |     pour une couche donne                                              |
 |                                                                         |
 | On utilise la diffrentielle logarithmique pour calculer                |
 |                                                                         |
 | k       = le point pour lequel on calcule                               |
 | couche  = la couche pour laquelle on fait le calcul                     |
 | feff    = le modle de chemins pour lequel on fait un calcul            |
 | type    = le type de variables (#define dans couches.h)                 |
 | valeur  = valeur de la variable par rapport  laquelle on veut driver  |
 | val_khi = valeur du signal EXAFS en ce point                            |
 | delta_e = la valeur de Delta(E_0)  utiliser pour le calcul             |
 \------------------------------------------------------------------------ */

double calc_derivee(double k, Couche *couche, Modele_FEFF *feff, int type, double valeur,
		    double val_khi, double delta_e)
{
  double r = 0.0;
  double phase;
  double val_derivee = 0.0;
  double k_p_n = 0.0;

  if ( (VAR_SO2 == type) || (VAR_CONTRIBUTION == type) || /* val_khi : signal EXAFS total */
       (VAR_N == type) ) { /* val_khi : signal EXAFS de la couche correspondante */
    /* REMARQUE : le fait que ces 3 variables aient mme formule de drivation
       impose une forte corrlation entre elles : leur ajustement simultan est quasi-impossible */
    if (0.0 == valeur) {
      EXF_ERREUR(DIV_ZERO, "S0"DEUX_EXPOSANT", contribution ou N nul", txt_err_val_nulle);
      return 0.0;
    }
    return val_khi / valeur;
  }

  /* Pour toutes les autres variables, la drive dpend de k
     ==> on fait les corrections en nergie */
  if ( delta_e != 0.0 ) { /* Correction d'nergie  faire */
    k  = k * k;
    k += 2.0 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE*H_BARRE * M_V_ANGSTROM*M_V_ANGSTROM) * delta_e;
    if (k <= 0.0) return 0.0; /* Problme... */
    k  = sqrt(k);
  }
  if (k == 0.0) return 0.0;

  switch (type)
  {
  case VAR_DELTAE_GENE:
    val_derivee = 0.0;

    for (couche = feff->couches ; couche != NULL ; couche = couche->suivante) {
      if (couche->utilisee == VRAI) { /* On utilise la couche */
	val_derivee += derivee_delta_e(feff, couche, k, calc_exafs_1point(k, feff, couche, 0.0));
	/* Rq : correction en k dj faite, inutile de la refaire dans le calcul de l'exafs => DE0 = 0 */
      }
      /* Note : on doit pouvoir optimiser en faisant moins de calculs, mais
	 c'est  voir plutt ailleurs... dans le calcul des incrments, par exemple
      */
    }
    break;

  case VAR_GAMMA_GENE:
    val_derivee = 0.0;

    for (couche = feff->couches ; couche != NULL ; couche = couche->suivante) {
      if (couche->utilisee == VRAI) {
	val_derivee += derivee_gamma(feff, couche, k, calc_exafs_1point(k, feff, couche, 0.0));
	/* Rq : correction en k dj faite, inutile de la refaire dans le calcul de l'exafs => DE0 = 0 */
      }
      /* Note : on doit pouvoir optimiser en faisant moins de calculs, mais
	 c'est  voir plutt ailleurs... dans le calcul des incrments, par exemple
	 De toute faon, intrt de cette variable limit avec les lpm calculs par FEFF
      */
    }
/*    printf("Gamma gnral - Valeur de la drive : %f\n", val_derivee); */
    break;

  case VAR_R:
    /*
       d(khi)             (        k          1      1      )
       ------ = 2 * khi * ( --------------- - - - --------  )
         dR               ( tg(2kR + phase)   R   lambda(k) )

       suppose que l'amplitude est quasi-indpendante de R,
       mme si le libre parcours moyen y est inclus !
       suppose aussi que la phase est quasi-indpendante de R
    */

    val_derivee = -1.0 / valeur;
    if (feff->type_lpm == MODELE_LPM_GAMMA) { /* Le lpm s'exprime lambda(k) = k / Gamma */
      double gamma;

      if (VRAI == feff->drapeaux.GAM_general)
	gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
      else
	gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

      val_derivee -= gamma / k;
    } else if (feff->type_lpm == MODELE_LPM_GAMMA_ETA) { /* Le lpm s'exprime lambda(k) = (k + (ta/k)^4) / Gamma */
      double gamma, kp;

      if (VRAI == feff->drapeaux.GAM_general)
	gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
      else
	gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

      kp  = feff->lpm_val_eta / k;
      kp *= kp;
      kp *= kp; /* kp = (ta / k)^4 */

      val_derivee -= gamma / (k + kp);
    } else if (feff->type_lpm == MODELE_LPM_GRAPHE) { /* Un graphe spcial contient le lpm */
      double lpm;

      lpm = SPL_tt_calc(feff->lpm, k, feff->b_lpm, feff->c_lpm, feff->d_lpm);
      if (lpm == 0.0) {
	EXF_ERREUR(DIV_ZERO, "Gamma / lpm nul", txt_err_val_nulle);
      } else
	val_derivee -= 1.0 / lpm;
    }

    k_p_n = k * k * k; /* k */

    phase  = 2.0 * k * valeur;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4.0/3.0  * k_p_n * VAR_LireValeur( &(couche->Cumul_2 ), VARIABLE_LIRE_VALEUR_AJUSTEE );
    k_p_n *= k * k;
    phase += 4.0/15.0 * k_p_n * VAR_LireValeur( &(couche->Cumul_4 ), VARIABLE_LIRE_VALEUR_AJUSTEE );

    val_derivee += k / tan(phase);
    val_derivee *= 2.0 * val_khi;
/*     printf("Drive / R ; R = %e, k corrig = %e, khi(k) = %e ==> %g\n", valeur, k, val_khi, val_derivee); */
#ifdef _DEVELOPPE_
    /* Les valeurs sont cohrentes entre la minimisation et le calcul de la matrice de covariance */
/*     printf("Drive selon R en k = %f : %g\n", k, val_derivee); */
#endif
    break;

  case VAR_DEBYE: /* La grandeur que l'on ajuste est en fait sigma */
    val_derivee = - val_khi * 2.0 * k * k;
    break;

  case VAR_CUMUL_1: /* Premire correction pour Debye-Waller */
    val_derivee  = 2.0 * val_khi * k * k * k * k;
    val_derivee /= 3.0;
    break;

  case VAR_CUMUL_2: /* Premire correction pour le dphasage */
    k_p_n = k * k * k; /* k */

    phase  = 2.0 * k * r;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4.0/3.0  * k_p_n * valeur;
    val_derivee = -4 * val_khi * k_p_n / 3.0;

    k_p_n *= k * k;
    phase += 4.0/15.0 * k_p_n * VAR_LireValeur( &(couche->Cumul_4 ), VARIABLE_LIRE_VALEUR_AJUSTEE );

    val_derivee /= tan(phase);
    break;

  case VAR_CUMUL_3: /* Deuxime correction pour Debye-Waller */
    k_p_n = k * k * k;
    k_p_n *= k_p_n; /* k_p_n = k^6 */
    val_derivee = -4.0 * val_khi * k_p_n / 45.0;
    break;

  case VAR_CUMUL_4: /* Deuxime correction pour la phase */
    k_p_n = k * k * k; /* k */

    phase  = 2.0 * k * r;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4.0/3.0  * k_p_n * VAR_LireValeur( &(couche->Cumul_2 ), VARIABLE_LIRE_VALEUR_AJUSTEE );
    k_p_n *= k * k; /* k_p_n = k^5 */
    phase += 4.0/15.0 * k_p_n * valeur;

    val_derivee  = 4.0/15.0 * val_khi * k_p_n;
    val_derivee /= tan(phase);
    break;

  case VAR_DELTAE: /* Un Delta E par couche si on est de ce type */
    val_derivee = derivee_delta_e(feff, couche, k, val_khi);
    break;

  case VAR_GAMMA:  /* Un gamma par couche si on est de ce type */
    val_derivee = derivee_gamma(feff, couche, k, val_khi);
    break;

#ifdef _DEVELOPPE_
  default:
    fprintf(stderr, "calc_derivee : calcul de la drive, varible de type inconnu : %d\n", type);
    break;
#endif
  }

  return val_derivee; /* Valeur de la drive en ce point */
}

/* -------- Sous-routines pour le calcul de la drive ------------------- */

/* -------- Calcul de la drive / delta_e pour une couche ----------------\
 |   Renvoie la valeur de la drive                                       |
 |                                                                         |
 | On utilise la diffrentielle logarithmique pour calculer                |
 |                                                                         |
 | d(khi)            [ n - 1        2   8     3    8     5   2 R (dl/dk)   |
 | ------ = khi(k) * [ ----- - 4 k s  + - C  k  - -- C  k  + ----------- + |
 | d(DE)             [   k              3  4      15  6         l(k)      |
 |                                                                         |
 |                         2   4     4            ]                        |
 |             2 R - 4 C  k  + - C  k             ]        m               |
 |                      3      3  5               ]          e             |
 |   + ------------------------------------------ ] * ------------         |
 |        (         4     3    4     5          ) ]           2            |
 |     tg ( 2 k R - - C  k  + -- C  k  + phi(k) ) ]    h_barre  k          |
 |        (         3  3      15  5             ) ]                        |
 |                                                                         |
 | Suppose donc que d(phi)/dk et d(ln(amplitude)) / dk soient ngligeables |
 |                                                                         |
 | couche  : la couche pour laquelle on calcule la drive                 |
 | k       : le point pour lequel on calcule (_dj correct_)              |
 | khi     : signal exafs pour la couche pour ce point                     |
 \------------------------------------------------------------------------ */

static double derivee_delta_e(Modele_FEFF *feff, Couche *couche, double k, double khi)
{
  double c3, c5;
  double val_derivee = 0;
  double amplitude, phase;
  double k_p_n; /* Puissances de k */
  double r;

  k_p_n = k * k; /* k^2 */
  r  = VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_AJUSTEE );

  /* Terme du libre parcours moyen : 2 * R * lambda'(k) / lambda(k)^2 */

  if (feff->type_lpm == MODELE_LPM_GAMMA) {
    double gamma;

    if (VRAI == feff->drapeaux.GAM_general)
      gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
    else
      gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

    val_derivee = 2.0 * r * gamma / k_p_n;
  } else if (feff->type_lpm == MODELE_LPM_GAMMA_ETA) {
    double e_s_k = feff->lpm_val_eta / k;
    double gamma;

    if (VRAI == feff->drapeaux.GAM_general)
      gamma = VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );
    else
      gamma = VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE );

    e_s_k *= e_s_k; /* (ta/k)^2 */
    e_s_k *= e_s_k; /* (ta/k)^4 */

    val_derivee  = (k + e_s_k);
    val_derivee *= val_derivee;
    val_derivee  = gamma / val_derivee;
    val_derivee *= 2.0 * r * (1 - 4 * e_s_k / k);

    /*
       dl/dk           1 - 4 (ta)^4 / k^5
       ----- = Gamma * -------------------
        l^2            (k + (ta / k)^4)^2
    */
  } else if (feff->type_lpm == MODELE_LPM_GRAPHE) { /* Libre parcours dans un graphe spcial */
    double lpm_derive;
    double lpm;

    lpm =  SPL_tt_derive(feff->lpm, k, feff->b_lpm, feff->c_lpm, feff->d_lpm, &lpm_derive);
    lpm *= lpm;
    if ( 0.0 == lpm ) {
      (void) fprintf(stderr, "derivee_delta_e : lpm = 0 !\n");
      return 0.0;
    }
    val_derivee = lpm_derive / (lpm * lpm);
  }

  val_derivee += -4.0 * VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_AJUSTEE ) * k;

  c3 = VAR_LireValeur( &(couche->Cumul_2), VARIABLE_LIRE_VALEUR_AJUSTEE );
  c5 = VAR_LireValeur( &(couche->Cumul_4), VARIABLE_LIRE_VALEUR_AJUSTEE );

  amplitude = 2.0 * r;
  phase = amplitude * k;

  amplitude -= 4.0 * c3 * k_p_n;

  k_p_n *= k; /* k */
  val_derivee += 8.0/3.0 * VAR_LireValeur( &(couche->Cumul_1), VARIABLE_LIRE_VALEUR_AJUSTEE ) * k_p_n;
  phase -= 4.0/3.0 * c3 * k_p_n;

  k_p_n *= k; /* k^4 */
  amplitude += 4.0/3.0 * c5 * k_p_n;

  k_p_n *= k; /* k^5 */
  phase += 4.0/15.0 * c5 * k_p_n;
  val_derivee -= 8.0/15.0 * VAR_LireValeur( &(couche->Cumul_3), VARIABLE_LIRE_VALEUR_AJUSTEE ) * k_p_n;

  phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);

  val_derivee += (feff->ponderation - 1) / k;
  val_derivee += amplitude / tan(phase);

  val_derivee *= khi;
  val_derivee *= MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM);
  val_derivee /= k;

  return val_derivee;
}

/* -------- Calcul de la drive / gamma pour une couche ------------------\
 |   Renvoie la valeur de la drive                                       |
 |                                                                         |
 | On utilise la diffrentielle logarithmique pour calculer.               |
 |                                                                         |
 | d(khi)                   -2 * R                                         |
 | ------ = khi(k) * ---------------------                                 |
 |  d(G)             ( lambda(k) / Gamma )                                 |
 |                                                                         |
 | couche  : la couche pour laquelle on calcule la drive                 |
 | k       : le point pour lequel on calcule (_dj correct_)              |
 | khi     : signal exafs pour la couche pour ce point                     |
 \------------------------------------------------------------------------ */

static double derivee_gamma(Modele_FEFF *feff, Couche *couche, double k, double khi)
{
  double derivee;

  derivee = -2.0 * VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_AJUSTEE );

  if (feff->type_lpm == MODELE_LPM_GAMMA) { /* Fonction de la forme exp (- 2 R Gamma / k ) */
    derivee /= k;
  } else if (feff->type_lpm == MODELE_LPM_GAMMA_ETA) { /* Fonction de la forme exp( -2 R Gamma / (k + (ta/k)^4) ) */
    double k_p_n = feff->lpm_val_eta / k;

    k_p_n *= k_p_n;
    k_p_n *= k_p_n; /* Contient (ta / k)^4 */

    derivee /= k + k_p_n;
  }

  derivee *= khi;
  return derivee;
}
