/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2002, Emmanuel CURIS
   Date de cration      : 19 juin 2002

   Module d'ajustement par un modle  n couches

   Calcul des contributions partielles de chaque chemin au signal EXAFS
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "atomes.h"
#include "couches.h"
#include "extrait.h" /* Pour EXF_k_p_n ;  dplacer ! */
#include "feff.h"
#include "graphes.h"
#include "lissage.h"
#include "modeles.h"
#include "science.h"
#include "serie.h"

#include "nc_contrib.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* -------- Calculer la contribution d'une couche prcise au modle total -------- */

int COUCHE_CalculerContribution(struct modele_feff *modele_feff, Couche *couche, struct graphe *modele)
{
  Graphe *contribution = NULL;

  if ( ( NULL == modele_feff ) || ( NULL == couche ) || ( NULL == modele ) ) {
    /* Problme de paramtre absurde */
    return +1;
  }

  /* Est-ce que la couche prsente dj une contribution ? Si oui, on l'efface... */
  if ( couche->signal != NULL ) {
    GRF_eff_isole( couche->signal );
    couche->signal = NULL;
  }

  /* Le graphe modle donne la table des X ; la contribution est un graphe isol a priori */
  contribution = GRF_nv_TableX( NULL, GR_EXAFS, GRF_FCT_REELLE, NULL, VRAI,
				modele->x, modele->nombre_point, VRAI );
  if ( NULL == contribution ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la contribution !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* La couche est-elle utilise dans le modle ? */
  if ( FAUX == couche->utilisee ) {
    /* Non : sa contribution est donc nulle, rien  faire ! */
    return 0;
  }

  /* La couche est utilise : on calcule sa contribution */
  /* Pralable : on vrifie que les tables des splines sont prtes */
  if ( NULL == couche->b_amp ) {
    int retour;

    retour = SPL_tt_prep( couche->amplitude, &(couche->b_amp), &(couche->c_amp), &(couche->d_amp) );
    if (retour < 0) {
      (void) fprintf(stderr, "NC_CalculeContributions : erreur dans SPL_tt_prep/amplitude, couche %s !\n", couche->nom);
      GRF_eff_isole( contribution );
      return -2;
    }
  }
  if ( NULL == couche->b_pha ) {
    int retour;

    retour = SPL_tt_prep( couche->phase, &(couche->b_pha), &(couche->c_pha), &(couche->d_pha) );
    if (retour < 0) {
      (void) fprintf(stderr, "NC_CalculeContributions : erreur dans SPL_tt_prep/phase, couche %s !\n", couche->nom);
      GRF_eff_isole( contribution );
      return -3;
    }
  }

  /* 1. On calcule le terme de la somme */
  NC_CalculeEXAFSCouche( modele_feff, couche, contribution );

  /* 2. On pondre par S0 */
  GRF_FacteurMultiplicatif( contribution, VAR_LireValeur( &(modele_feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE ), AXE_Y, NULL );

  /* 3. On pondre par k^n (sans historique) */
  EXF_k_p_n( contribution, modele_feff->ponderation, FAUX );

  /* On mmorise la contribution dans la couche */
  couche->signal = contribution;

  /* Termin sans erreur */
  return 0;
}

/* -------- Calculer la contribution (sparment) de chacune des couches --------\
 |   Utilise les infos de l'ajustement : pour chaque couche utilise dans        |
 | l'ajustement, calcule le spectre associ pour chaque point du spectre ajust. |
 |   Si la couche est associe  un chemin, ce chemin reoit comme spectre tho. |
 | le spectre ainsi calcul --> utilisable dans la vue 3D.                       |
 |   Renvoie la srie qui contient ces contributions                             |
 |                                                                               |
 | feff   : le modle  utiliser, contenant les chemins                          |
 | modele : le graphe exprimental (sert uniquement pour la grille des X)        |
 \------------------------------------------------------------------------------ */

Serie_Graphes *NC_CalculeContributions(struct modele_feff *feff, struct graphe *modele)
{
  Couche *couche;
  Graphe *exafs = NULL;
  Serie_Graphes *contributions = NULL;
  char *titre;
  unsigned long i, nbr_couches = 0;

  if ( (NULL == feff) || (NULL == modele) ) {
    (void) fprintf(stderr, "NC_CalculeContributions : un des paramtres vaut NULL\n");
    return NULL;
  }
  if ( NULL == feff->couches ) {
    (void) fprintf(stderr, "NC_CalculeContributions : modle FEFF sans couche associe\n");
    return NULL;
  }

  /* Existe-t-il dj une srie de contributions ? */
  for ( couche = feff->couches ; couche != NULL ; couche = couche->suivante ) {
    if ( FAUX == couche->utilisee ) continue;

    nbr_couches++;
    if ( (couche->signal != NULL) && (NULL == contributions) )
      contributions = GRF_QuelleSerie( couche->signal );
  }

  if ( 0 == nbr_couches ) {
    (void) fprintf(stderr, "NC_CalculeContributions : aucune couche utilise dans le modle FEFF !\n");
    return NULL;
  }

  if ( contributions != NULL ) { /* Dj des contributions : doit-on les effacer ? */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &Q_EffacerContributions, VRAI);
    if ( 1 == reponse ) { /* On efface les contributions */
      for ( couche = feff->couches ; couche != NULL ; couche = couche->suivante ) {
	GRF_efface( couche->signal, VRAI, FAUX );
      }
    }
  }

  /* On cre la srie qui contiendra les contributions */
  contributions = SERIE_GRF_CreerSerie(serie_courante, nom_serie_contributions);
  if (NULL == contributions) {
    (void) fprintf(stderr, "NC_CalculeContributions : impossible de crer la srie !\n");
    return NULL;
  }

  /* Pour chacune des couches utilise... */
  for ( couche = feff->couches, i = 0 ; couche != NULL ; couche = couche->suivante, i++ ) {
    if ( FAUX == couche->utilisee ) continue;

    /* On prpare le calcul des phases et amplitudes, si ncessaire */
    if ( NULL == couche->b_amp ) {
      int retour;

      retour = SPL_tt_prep(couche->amplitude, &(couche->b_amp), &(couche->c_amp), &(couche->d_amp));
      if (retour < 0) {
	(void) fprintf(stderr, "NC_CalculeContributions : erreur dans SPL_tt_prep/amplitude, couche %s !\n", couche->nom);
	continue;
      }
    }
    if ( NULL == couche->b_pha ) {
      int retour;

      retour = SPL_tt_prep(couche->phase, &(couche->b_pha), &(couche->c_pha), &(couche->d_pha));
      if (retour < 0) {
	(void) fprintf(stderr, "NC_CalculeContributions : erreur dans SPL_tt_prep/phase, couche %s !\n", couche->nom);
	continue;
      }
    }

    /* On prpare le graphe contenant la contribution */
    titre = TXT_genere_chaine("Contribution du chemin %s", couche->nom);
    exafs = Serie_nv_graphe(contributions, exafs, modele->nombre_point, GR_AJUSTE, GR_NORMAL, titre);
    free(titre);
    if ( NULL == exafs ) {
      (void) fprintf(stderr, "NC_CalculeContributions : impossible de crer la contribution / chemin %s !\n", couche->nom);
      continue;
    }

    /* On copie les X */
    memcpy(exafs->x, modele->x, sizeof(double) * (modele->nombre_point));
    exafs->x_min = modele->x_min;
    exafs->x_max = modele->x_max;

    /* On calcule le terme de la somme */
    NC_CalculeEXAFSCouche(feff, couche, exafs);

    /* On pondre par S0 */
    GRF_FacteurMultiplicatif( exafs, VAR_LireValeur( &(feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE ), AXE_Y, NULL );
    /* puis par k^n (sans historique) */
    EXF_k_p_n(exafs, feff->ponderation, FAUX);

    /* On l'associe  la couche (pour la suite) */
    GRF_couleur( exafs, (int) ((i % 15) + 1), (int) ((i / 15) + 1), 0 );
    couche->signal = exafs;
  }

  return contributions;
}
