/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2001, Emmanuel CURIS
   Cration du fichier : 4 fvrier 2000

   Combinaison linaire de graphes, suivant diverses mthodes.
   Transformations affines des coordonnes (oprations aussi linaires...)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "operations.h"
#include "lissage.h"

#include "lineaire.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean CLIN_reel    (const Graphe *gr_1, double coef_1, const Graphe *gr_2, double coef_2, Graphe *resultat);
static boolean CLIN_complexe(const Graphe *gr_1, double coef_1, const Graphe *gr_2, double coef_2, Graphe *resultat);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Transformation affine des X d'un graphe --------\
 |   Renvoie VRAI si tout s'est bien pass, FAUX sinon      |
 |                                                          |
 | graphe : le graphe  modifier (fait EN PLACE)            |
 | a, b, c : coefficients. X' = a * X + b (+ c *Y)          |
 \--------------------------------------------------------- */

boolean GRF_xaffine(Graphe *graphe, double a, double b, double c)
{
  Graphe *erreur;
  unsigned long i;
  register double *x;

  if (graphe == NULL) return FAUX;

  if ( (c != 0.0) && ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) ) {
    fprintf(stderr, "GRF_xaffine : combinaison des X et Y d'un graphe complexe\n");
    /* ERREUR : On ne peut pas avoir des X complexes dans un graphe ! */
    return FAUX;
  }

  /* On commence par multiplier par a */
  if ( a != 1.0 )
    GRF_FacteurMultiplicatif( graphe, a, AXE_X, NULL );

  /* On ajoute la constante - ne change pas les erreurs */
  if ( b != 0.0 ) {
    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++) {
      *x += b; /* CL elle-mme */
    }

    graphe->x_min += b;
    graphe->x_max += b;
  }

  /* On ajoute le terme C - peut changer les erreurs... */
  if ( c != 0.0 ) {
    double *y, *y_errx = NULL, *y_erry = NULL;

    if ( graphe->erreur_y == NULL ) { /* Sinon, erreurs en X inchanges */
      y_erry = graphe->erreur_y->y;

      erreur = graphe->erreur_x;
      if (erreur != NULL) {
	y_errx = erreur->y;
      }
    }

    for (i = 0 , x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; x++, y++, i++) {
      *x += c * *y;

      if ( y_errx != NULL ) {
	double Vx, Vy;

	Vx = *y_errx * *y_errx;
	Vy = *y_erry * *y_erry * c * c;
	*y_errx = sqrt( Vx + Vy ); /* Suppose X et Y indpendants => CL des variances */

	y_errx++; y_erry++;
      }
    }

    GRF_bornes(graphe, DESORDONNE, NE_PAS_FAIRE);
    GRF_bornes(graphe->erreur_x, NE_PAS_FAIRE, DESORDONNE);
  }

  /* On s'occupe des erreurs */
  erreur = graphe->erreur_x;
  if ( erreur != NULL ) {
    memcpy(erreur->x, graphe->x, sizeof(double) * graphe->nombre_point);
    erreur->x_min = graphe->x_min;
    erreur->x_max = graphe->x_max;
  }

  erreur = graphe->erreur_y;
  if ( erreur != NULL ) {
    memcpy(erreur->x, graphe->x, sizeof(double) * graphe->nombre_point);
    erreur->x_min = graphe->x_min;
    erreur->x_max = graphe->x_max;
  }

  /* Log. dans l'historique */
  HIST_Titre(4, "Transformation affine des X", "MATH");
  HIST_Texte(4, 0, "Graphe modifi"e_AIGU" : %s\n", graphe->nom);
  HIST_Texte(5, 0, "X' = %.20g * X + %.20g", a, b);
  if (c != 0.0) {
    HIST_Texte(5, 0, " + %.20g * Y", c);
  }
  HIST_Texte(5, 0, "\n");

  return VRAI;
}

/* -------- Transformation affine des Y d'un graphe --------\
 |   Renvoie VRAI si tout s'est bien pass, FAUX sinon      |
 |                                                          |
 | graphe : le graphe  modifier (fait EN PLACE)            |
 | a, b, c : coefficients. Y' = a * Y + b (+ c *X)          |
 \--------------------------------------------------------- */

boolean GRF_yaffine(Graphe *graphe, double a, double b, double c)
{
  unsigned long i;
  register double *x;
  register double *x_err = NULL;

  if (graphe == NULL) return FAUX;

  if ( a != 1.0 )
    GRF_FacteurMultiplicatif( graphe, a, AXE_Y, NULL ); /* S'occupe des erreurs et tout et tout */
  
  if (graphe->erreur_x != NULL) x_err = graphe->erreur_x->y;

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    register Complexe *y, *y_err = NULL;

    if (graphe->erreur_y != NULL) y_err = (Complexe *) graphe->erreur_y->y;

    for (i = 0, x = graphe->x, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
      y->re += b + c * *x; /* CL elle-mme */
      y->im += b + c * *x;

      if (y_err != NULL) { /* On modifie aussi les erreurs ! */
	if (x_err != NULL) { /* Il y a des erreurs en X : on en tient compte... */
	  Complexe z;

	  z.re = y_err->re * y_err->re + *x_err * *x_err;
	  z.im = y_err->im * y_err->im + *x_err * *x_err;
	  z.re = sqrt(z.re); z.im = sqrt(z.im);
	  *y_err = z;

	  x_err++;
	}
	y_err++;
      }
    }
  } else { /* Graphe 2D rel ou quivalent */
    register double *y, *y_err = NULL;

    if (graphe->erreur_y != NULL) y_err = graphe->erreur_y->y;

    for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
      *y += b + c * *x; /* CL elle-mme */

      if (y_err != NULL) {  /* On modifie aussi les erreurs ! */
	if (x_err != NULL) { /* Il y a des erreurs en X : on en tient compte... */
	  double z = *y_err * *y_err + *x_err * *x_err;
	  *y_err = sqrt(z);

	  x_err++;
	}
	y_err++;
      }
    }
  }

  /* Log. dans l'historique */
  HIST_Titre(4, "Transformation affine des Y", "MATH");
  HIST_Texte(4, 0, "Graphe modifi"e_AIGU" : %s\n", graphe->nom);
  HIST_Texte(5, 0, "Y' = %.20g * Y + %.20g", a, b);
  if (c != 0.0) {
    HIST_Texte(5, 0, " + %.20g * X", c);
  }
  HIST_Texte(5, 0, "\n");

  GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
  GRF_bornes(graphe->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* -------- Combinaison linaire de 2 graphes dans 1 troisime dj existant --------\
 |   Renvoie VRAI si tout s'est bien pass, FAUX sinon
 |   Ne fait aucune vrification sur les X et somme sur le graphe le plus petit
 |
 | gr_1, coef_1 : le premier graphe et son coefficient
 | gr_2, coef_2 : le second graphe et son coefficient
 | resultat     : le graphe qui doit contenir le rsultat
 \---------------------------------------------------------------------------------- */

boolean GRF_clin(const Graphe *gr_1, double coef_1, const Graphe *gr_2, double coef_2, Graphe *resultat)
{
  if ( (gr_1 == NULL) || (gr_2 == NULL) || (resultat == NULL) ) return FAUX;

  if ( GRF_FCT_COMPLEXE == resultat->genre_donnees ) {
    if ( ( GRF_FCT_COMPLEXE == gr_1->genre_donnees ) &&
	 ( GRF_FCT_COMPLEXE == gr_2->genre_donnees ) ) {
#ifdef _DEVELOPPE_
      printf("Combinaison de graphes complexes\n");
#endif
      return CLIN_complexe(gr_1, coef_1, gr_2, coef_2, resultat);
    } else { /* Mixte : complexe + rel, rsultat complexe bien sr */
      OPTION_A_FAIRE( "Combinaison linaire Rel + Complexe vers Complexe" );
    }
  } else { /* Rsultat rel : a priori, si gr_1 ou gr_2 est complexe, a ne va pas ! */
    if ( ( GRF_FCT_COMPLEXE == gr_1->genre_donnees ) || ( GRF_FCT_COMPLEXE == gr_2->genre_donnees ) ) {
      EXF_ERREUR( GRAPHE_COMPLEXE );
      return FAUX;
    }

    return CLIN_reel(gr_1, coef_1, gr_2, coef_2, resultat);
  }

  return FAUX; /* Si on est l, c'est qu'il y a eu un problme */
}

/* -------- Combinaison de 2 graphes  valeurs relles -------- */

static boolean CLIN_reel(const Graphe *gr_1, double coef_1, const Graphe *gr_2, double coef_2, Graphe *resultat)
{
  unsigned long i, nbr_points;
  register double *y1, *y2, *y_clin, *y_err1 = NULL, *y_err2 = NULL, *y_err_clin = NULL;

  nbr_points = gr_1->nombre_point;
  if (gr_2->nombre_point < nbr_points) {
    printf("ATTENTION, graphe 2 plus court [%lu points, contre %lu]\n", gr_2->nombre_point, nbr_points);
    nbr_points = gr_2->nombre_point;
  }
  if (resultat->nombre_point < nbr_points) {
    printf("ATTENTION, resultat trop court [%lu points, contre %lu]\n", resultat->nombre_point, nbr_points);
    nbr_points = resultat->nombre_point;
  }

  if ( (resultat->erreur_y == NULL) && /* Pas d'erreur sur le rsultat */
       ( (gr_1->erreur_y != NULL) || (gr_2->erreur_y != NULL) ) ) { /* Mais sur les intermdiaires */
    /* Donc on en cre pour le rsultat */

    (void) GRF_creer_erreurs( resultat, AXE_Y );
  }
  if ( gr_1->erreur_y != NULL ) y_err1 = gr_1->erreur_y->y;
  if ( gr_2->erreur_y != NULL ) y_err2 = gr_2->erreur_y->y;
  if ( resultat->erreur_y != NULL) y_err_clin = resultat->erreur_y->y;

  for (i = 0, y1 = gr_1->y, y2 = gr_2->y, y_clin = resultat->y ;
       i < nbr_points ; i++, y1++, y2++, y_clin++) {
    double z;

    z = coef_1 * *y1 + coef_2 * *y2;
    *y_clin = z;

    if (y_err_clin) {
      z = 0.0;

      if (y_err1) {
	z += coef_1 * coef_1 * (*y_err1 * *y_err1);
	y_err1++;
      }
      if (y_err2) {
	z += coef_2 * coef_2 * (*y_err2 * *y_err2);
	y_err2++;
      }

      z = sqrt(z); /* Les variances sont additives... */
      *y_err_clin = z;
      y_err_clin++;
    }
  }
  GRF_bornes(resultat, NE_PAS_FAIRE, DESORDONNE);
  GRF_bornes(resultat->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* -------- Combinaison de 2 graphes  valeurs complexes -------- */

static boolean CLIN_complexe(const Graphe *gr_1, double coef_1, const Graphe *gr_2, double coef_2, Graphe *resultat)
{
  unsigned long i, nbr_points;
  register Complexe *y1, *y2, *y_clin, *y_err1 = NULL, *y_err2 = NULL, *y_err_clin = NULL;

  nbr_points = gr_1->nombre_point;
  if (gr_2->nombre_point < nbr_points)
    nbr_points = gr_2->nombre_point;
  if (resultat->nombre_point < nbr_points)
    nbr_points = resultat->nombre_point;

  if ( (resultat->erreur_y == NULL) && /* Pas d'erreur sur le rsultat */
       ( (gr_1->erreur_y != NULL) || (gr_2->erreur_y != NULL) ) ) { /* Mais sur les intermdiaires */
    /* Donc on en cre pour le rsultat */

    (void) GRF_creer_erreurs(resultat, AXE_Y);
  }
  if ( gr_1->erreur_y != NULL ) y_err1 = (Complexe *) gr_1->erreur_y->y;
  if ( gr_2->erreur_y != NULL ) y_err2 = (Complexe *) gr_2->erreur_y->y;
  if ( resultat->erreur_y != NULL) y_err_clin = (Complexe *) resultat->erreur_y->y;

  for (i = 0, y1 = (Complexe *) gr_1->y, y2 = (Complexe *) gr_2->y, y_clin = (Complexe *) resultat->y ;
       i < nbr_points ; i++, y1++, y2++, y_clin++) {
    Complexe z;

    z.re = coef_1 * y1->re + coef_2 * y2->re;
    z.im = coef_1 * y1->im + coef_2 * y2->im;
    *y_clin = z;

    if (y_err_clin) {
      z.re = 0.0; z.im = 0.0;

      if (y_err1) {
	z.re += coef_1 * coef_1 * (y_err1->re * y_err1->re);
	z.im += coef_1 * coef_1 * (y_err1->im * y_err1->im);
	y_err1++;
      }
      if (y_err2) {
	z.re += coef_2 * coef_2 * (y_err2->re * y_err2->re);
	z.im += coef_2 * coef_2 * (y_err2->im * y_err2->im);
	y_err2++;
      }

      z.re = sqrt(z.re); z.im = sqrt(z.im); /* Les variances sont additives... */
      *y_err_clin = z;
      y_err_clin++;
    }
  }
  GRF_bornes(resultat, NE_PAS_FAIRE, DESORDONNE);
  GRF_bornes(resultat->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* -------- Combinaison linaire de 2 graphes dans 1 troisime dj existant --------\
 |   Renvoie VRAI si tout s'est bien pass, FAUX sinon                               |
 |   Cale sur les X du graphe rsultat                                               |
 |   Ajoute le graphe 2  la rfrence - travaille EN PLACE                          |
 |                                                                                   |
 | reference, coef_1 : la rfrence et son coefficient                               |
 | gr_2, coef_2 : le second graphe et son coefficient                                |
 | resultat     : le graphe qui doit contenir le rsultat                            |
 \---------------------------------------------------------------------------------- */

boolean GRF_Rclin(Graphe *reference, double coef_1, Graphe *gr_2, double coef_2)
{
  int retour;
  unsigned long i;
  register double *x, *y_clin, *y_err_clin = NULL;
  double *b2 = NULL, *c2 = NULL, *d2 = NULL;
  double *br2 = NULL, *cr2 = NULL, *dr2 = NULL;

  /* On prpare les splines du deuxime graphe */
  retour = SPL_tt_prep(gr_2, &b2, &c2, &d2);
  if ( retour != 0 ) {
    return FAUX;
  }

  /* On prpare les splines des erreurs, si elles existent
     En cas de problme, les tableaux sont  NULL donc pas de vrification spciale */
  if ( gr_2->erreur_y != NULL ) {
    SPL_tt_prep(gr_2->erreur_y, &br2, &cr2, &dr2);
  }

  if ( (reference->erreur_y == NULL) && (br2 != NULL) ) {
    /* Donc on en cre pour le rsultat */
    (void) GRF_creer_erreurs( reference, AXE_Y );
  }
  if (reference->erreur_y != NULL) y_err_clin = reference->erreur_y->y;

  for (i = 0, x = reference->x, y_clin = reference->y ;
       i < reference->nombre_point ; i++, x++, y_clin++) {
    double y1, y2;

    y1 = *y_clin;
    y2 = SPL_tt_calc(gr_2, *x, b2, c2, d2);

    *y_clin = coef_1 * y1 + coef_2 * y2;

    if (y_err_clin) {
      double z = 0.0;

      y1 = *y_err_clin * coef_1;
      z += y1 * y1;

      if (br2 != NULL) {
	y2 = coef_2 * SPL_tt_calc(gr_2->erreur_y, *x, br2, cr2, dr2);
	z += y2 * y2;
      }

      z = sqrt(z); /* Les _variances_ sont additives... */
      *y_err_clin = z;
      y_err_clin++;
    }
  }
  GRF_bornes(reference, NE_PAS_FAIRE, DESORDONNE);
  GRF_bornes(reference->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  free(br2); free(cr2); free(dr2);
  free(b2); free(c2); free(d2);

  return VRAI;
}
