/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2006, Emmanuel CURIS
   Création du fichier : 26 juillet 2006

   Opérations diverses sur les graphes, plus générales que l'EXAFS lui-même
   (aussi bien mathématiques que statistiques simples : tri en particulier)
   (les opérations statistiques avancées sont dans ../Stats)

   Transformations semi-logarithmiques et logarithmiques
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "operations.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_SemiLog_Y_Decimal   (Graphe *graphe);
static int GRF_SemiLog_Y_Naturel   (Graphe *graphe);
static int GRF_SemiLog_Y_Quelconque(Graphe *graphe, double base);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Transformation des valeurs en Y ; log décimal pour être cohérent avec l'échelle semi-logarithmique usuelle
    Renvoie : 0 si tout s'est bien passé

  La transformation est faite en place
 */

int GRF_SemiLog_Y(Graphe *graphe, double base)
{
  int retour;

  if ( NULL == graphe ) return +1;

  /* Contrôle de la base choisie : par défaut, log. népérien */
  if ( ( base < 0.0 ) || ( 1.0 == base ) ) {
    AVERTISSEMENT( "Base de logarithme absurde : %g", base );
    base = TRANSFORME_LOG_BASE_E;
  }

  /* Contrôle des bornes */
  if ( graphe->y_min <= 0.0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Valeurs négatives en Y, transformation impossible\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Transformation proprement dite */
  if ( TRANSFORME_LOG_BASE_E == base ) {
    retour = GRF_SemiLog_Y_Naturel( graphe );
  } else if ( 10.0 == base ) {
    retour = GRF_SemiLog_Y_Decimal( graphe );
  } else {
    retour = GRF_SemiLog_Y_Quelconque( graphe, base );
  }

  /* On a terminé */
  if ( retour != 0 ) {
    AVERTISSEMENT( "Erreur dans la transformation [code %d, base %g]", retour, base );
  }
  return retour;
}

static int GRF_SemiLog_Y_Decimal(Graphe *graphe)
{
  double *y;
  unsigned long i;

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    /* Contrôle pour être tranquille, mais ne devrait jamais arriver ! */
    if ( *y <= 0.0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Valeur Y négative [%g, i=%lu sur %lu], mais y_min > 0 [min = %g, max = %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, *y, i + 1, graphe->nombre_point, graphe->y_min, graphe->y_max );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    } else {
      *y = log10( *y );
    }
  }

  /* Transformation des bornes : le log est une fonction croissante */
  graphe->y_min = log10( graphe->y_min );
  graphe->y_max = log10( graphe->y_max );

  if ( graphe->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Propagation des erreurs" );
  }

  return 0;
}

static int GRF_SemiLog_Y_Naturel   (Graphe *graphe)
{
  double *y;
  unsigned long i;

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    /* Contrôle pour être tranquille, mais ne devrait jamais arriver ! */
    if ( *y <= 0.0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Valeur Y négative [%g, i=%lu sur %lu], mais y_min > 0 [min = %g, max = %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, *y, i + 1, graphe->nombre_point, graphe->y_min, graphe->y_max );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    } else {
      *y = log( *y );
    }
  }

  /* Transformation des bornes : le log est une fonction croissante */
  graphe->y_min = log( graphe->y_min );
  graphe->y_max = log( graphe->y_max );

  if ( graphe->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Propagation des erreurs" );
  }

  return 0;
}

static int GRF_SemiLog_Y_Quelconque(Graphe *graphe, double base)
{
  double *y, facteur_base;
  unsigned long i;

  /* Le facteur de base : log_b(x) = ln(x)/ln(b) = fb * ln(x) */
  facteur_base = 1.0 / log( base );

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    /* Contrôle pour être tranquille, mais ne devrait jamais arriver ! */
    if ( *y <= 0.0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Valeur Y négative [%g, i=%lu sur %lu], mais y_min > 0 [min = %g, max = %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, *y, i + 1, graphe->nombre_point, graphe->y_min, graphe->y_max );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    } else {
      *y = facteur_base * log( *y );
    }
  }

  /* Transformation des bornes : le log est une fonction croissante */
  graphe->y_min = facteur_base * log( graphe->y_min );
  graphe->y_max = facteur_base * log( graphe->y_max );
  if ( facteur_base < 0.0 ) {
    /* Log de base < 1.0 : fonction décroissante ! */
    double z;

    z = graphe->y_min; graphe->y_min = graphe->y_max; graphe->y_max = z;
  }

  if ( graphe->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Propagation des erreurs" );
  }

  return 0;
}

