/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996 - 2006, Emmanuel CURIS
   Création : 2 juillet 2006

   Calculs de probabilité divers

   Calculs autour de la loi de Laplace-Gauss (loi "normale")
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef AVEC_GSL_
# include <gsl/gsl_cdf.h>
#endif

#include "Utilitaires/math_utils.h"
#include "systeme.h"

#include "proba.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define H_MINIMAL_SIMPSON 1.0e-7  /* DXMIN ; Valeur originale de Dunnett. */
#define H_INITIAL_SIMPSON 0.24    /* PT24 ; Valeur originale de Dunnett. Pourquoi ?... */
#define NMAX_ITERATIONS   800	  /* NMAX ; Idem [puisque ierc=0] */
#define NMAX_SUBDIVISIONS  22	  /* Idem... */
#define PRECISION         1.0e-4  /* EPS ; idem [devrait suffire pour les stats !] */
#define QUASI_ZERO        1.0e-10 /* SMALL ; idem */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double MultiNormale_Multiplicative(unsigned int dimension, double z, double *x_max, double *facteurs, double additif, double *partiel);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~ Cas monodimensionnel ~-~-~-~-~-~-~-~ */

/* -------- Fonction de répartition pour une loi normale 1D centrée, réduite -------- */

double PROBA_FdR_Gauss(double x)
{
  double probabilite = -1.0;

  if ( 0.0 == x ) return 0.5;

#ifdef AVEC_GSL_
  probabilite = gsl_cdf_ugaussian_P( x );
#else
  OPTION_A_FAIRE("Utiliser erf pour calculer la fonction de répartition d'une variable U");
#endif

/*   printf("%s : p(U < %g) = %g\n", __FUNCTION__, x, probabilite ); */
  return probabilite;
}

/* -------- Fractiles d'une loi normale 1D centrée, réduite -------- */

double PROBA_Fractile_Gauss(double probabilite)
{
  if ( ( probabilite < 0.0 ) || ( probabilite > 1.0 ) ) {
    AVERTISSEMENT( "Valeur de probabilité absurde : %g, doit être entre 0 et 1", probabilite );
    return 0.0;
  }

#ifdef AVEC_GSL_
  return gsl_cdf_ugaussian_Pinv( probabilite );
#else
  OPTION_A_FAIRE("Utiliser erf pour calculer les fractiles d'une variable U");
  return 0.0;
#endif
}

/* ~-~-~-~-~-~-~-~ Cas multidimensionnel ~-~-~-~-~-~-~-~ */

/* -------- Fonction de répartition pour une loi normale à N dimensions avec une structure multiplicative de corrélation --------

 -- Correspondance & contrôle des variables originales de mvnprd --

 * paramètres *
   A    -> x / borne haute
   B    -> inutilisé car toujours -l'infini
   BPD  -> facteurs / termes à multiplier pour avoir les rho_i_j
   EPS  -> PRECISION / précision voulue
   N    -> dimension
   INF  -> inutilisé car toujours de -infini à x : INF(I) = 1
   IERC -> inutilisé, vaut 0
   HINC -> H_INITIAL_SIMPSON
   PROB -> probabilite [valeur renvoyée par la fonction]
   BOUND -> limite, pour contrôle de précision
   IFAULT -> inutlisée / code erreur : affichages

 * variables locales *
   NN       -> inutilisée ........................ OK
   ESTT(22) -> simpson_temp[ NMAX_SUBDIVISIONS ] . OK
   FV(5)    -> valeurs_f[ 5 ] .................... OK
   FD(5)    -> inutilisée ........................ OK
   F1T(22)  -> f0t[ NMAX_SUBDIVISIONS ] .......... OK
   F2T(22)  -> f1t[ NMAX_SUBDIVISIONS ] .......... OK
   F3T(22)  -> f2t[ NMAX_SUBDIVISIONS ] .......... OK
   G1T(22)  -> inutilisée ........................ OK
   G2T(22)  -> inutilisée ........................ OK
   G3T(22)  -> inutilisée ........................ OK
   PSUM(22) -> somme_partielle[NMAX_SUBDIVISIONS]  OK
   H(NN)    -> inutilisée/égale à x .............. OK
   HL(NN)   -> inutilisée/-infini comme B ........ OK
   BB(NN)   -> inutilisée/identique à BPD ........ OK
   INFT(NN) -> inutilisée/identique à INF ........ OK
   LDIR(22) -> sens_niveau[ NMAX_SUBDIVISIONS ] .. OK
   SMALL    -> QUASI_ZERO ........................ OK
   DXMIN    -> H_MINIMAL_SIMPSON [fixe à 10^-7] .. OK
   ERRL     -> errl (rôle : ?) ................... OK
   START    -> z_debut ........................... OK
   Z        -> z
   ADDN     -> additif ........................... OK
   EPS2     -> rendue inutile .................... OK
   EPS1     -> eps1 .............................. OK
   ZU       -> z_max ............................. OK
   Z2       -> z_milieu .......................... OK
   Z3       -> z_milieu .......................... OK
   Z4       -> z_milieu .......................... OK
   Z5       -> z_fin ............................. OK
   ZZ       -> z_Gauss ........................... OK
   ERFAC    -> inutilisée [toujours 1] ........... OK
   EL       -> el ................................ OK
   EL1      -> el_initial ........................ OK
   PART0    -> part0 ............................. OK
   PART2    -> part2 ............................. OK
   PART3    -> part3 ............................. OK
   FUNC0    -> func0 ............................. OK
   FUNC2    -> func2 ............................. OK
   FUNCN    -> funcn [inutile en réalité !] ...... OK
   WT       -> coef_simpson ...................... OK
   CONTRB   -> contrb ............................ OK
   DLG      -> dlg ............................... OK
   DX       -> dx ................................ OK
   DA       -> da ................................ OK
   ESTL     -> simpson_gauche .................... OK
   ESTR     -> simpson_droite .................... OK
   SUM      -> somme ............................. OK
   EXCESS   -> exces ............................. OK
   ERROR    -> erreur ............................ OK
   PROB1    -> proba_temp ........................ OK
   SAFE     -> inutilisée ........................ OK
   I        -> i ................................. OK
   NTM      -> NMAX .............................. OK
   LVL      -> niveau ............................
   NR       -> rendue inutile .................... OK
   NDIM     -> inutilisée/identique à dimensions . OK

 */

double PROBA_FdR_MultiNormale_Multiplicatif(unsigned int dimension, double *x, double *facteurs)
{
  double h = H_INITIAL_SIMPSON /* HINC */, da, dx, additif = -1.0;  /* ADDN */
  double z = 0.0 /* Z */, z_max = 0.0 /* ZU */, z_debut, z_milieu, z_fin;
  double probabilite = 1.0, proba_temp = 0.0;	/* PROB */
  double valeurs_f[ 5 ]	/* FV : valeurs de la fonction intégrée */, f0t[ NMAX_SUBDIVISIONS ], f1t[ NMAX_SUBDIVISIONS ], f2t[ NMAX_SUBDIVISIONS ], func0, func2, funcn;
  double simpson_droite = 0.0 /* ESTR */, simpson_gauche = 0.0 /* ESTL */, simpson_temp[ NMAX_SUBDIVISIONS ] /* ESTT */, contrb;
  double somme = 0.0 /* SUM */, somme_partielle[ NMAX_SUBDIVISIONS ] /* PSUM */;
  double limite = 0.0 /* BOUND */, erreur /* ERROR */, errl, eps1, el, el_initial, exces, dlg;
  double part0 /* PART0 */, part2 /* PART2 */, part3 = 0.0;
  double coef_simpson /* WT */;
  unsigned int i;		/* I */
  unsigned int niveau;		/* LVL : nombre de subdivisions faites */
  unsigned long iteration = 0;	/* NTM : nombre d'appels à la fonction */
  unsigned char sens_niveau[ NMAX_SUBDIVISIONS ];  /* LDIR */

  /* Variables inutiles : ERFAC [=1 tjrs ici], NR */

  /* Contrôles préliminaires */
  if ( ( NULL == x ) || ( NULL == facteurs ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramètres absurdes [x=%p, facteurs=%p]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) x, (void *) facteurs );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1.0;
  }

  if ( 1 == dimension ) {
    OPTION_A_FAIRE( "Cas monodimensionnel" );
    return -5.0;
  }

  /* On vérifie que les demi-coefficients de corrélation sont plausibles */
  for ( i = 0 ; i < dimension ; i++ ) {
    if ( fabs( facteurs[ i ] ) > 1.0 ) {
      AVERTISSEMENT( "Demi-rho absurde : %g (pour i = %u)", facteurs[ i ], i );
      return -2.0;
    }
    if ( 0.0 == facteurs[ i ] ) {
      AVERTISSEMENT( "Composante i = %u indépendante : on pourrait optimiser...", i );
    }

    /* Terme additif à ajouter à Simpson pour avoir la bonne limite [cf. A251.1].
       On sait que inft = inf = 1 partout
       -> ce terme vaut -1.0, sauf s'il existe une corrélation négative */
    if ( facteurs[ i ] * facteurs[ 0 ] < 0.0 ) {
      additif = 0.0;
    }

/*     printf("x[ %u ] = %g, facteurs[ %u ] = %g\n", i, x[ i ], i, facteurs[ i ]); */
  }

  /* Dans l'original, il y avait une préanalyse
     pour distinguer les composantes indépendantes. Ignoré ici */

  /* Ce qui suit adapte la routine venant de AS 251.1 / Charles Dunnett
     Comme la borne basse est infinie, on est dans le cas inf[ i ] = 1, a[ i ] = x_i, b[ i ] = 0
   */

  /* Préparation de la méthode de Simpson */
  /* a. Calcul de la borne supérieure pour z [ppnd7 : donne le fractile] */
  z_max = -PROBA_Fractile_Gauss( 0.05 * PRECISION ) / M_SQRT2;

  /* b. Adaptation de la précision (?) */
  el_initial = ( 0.95 * PRECISION ) / ( 1.0 + ( (int) ( z_max / H_INITIAL_SIMPSON ) ) );
  el = el_initial;

  /* c. Calcul du 1er point [en z] */
  part0 = MultiNormale_Multiplicative( dimension, z, x, facteurs, additif, &func0 );
/*   printf("Zmax = %g ; part0 = %g ; func0 = %g\n", z_max, part0, func0); */

  /* Calculs pour [z, z + h]
     Remarque : SAFE et FD ne servent que pour l'estimation de l'erreur => inutile ici
   */
  do {				/* Ligne 40 */
    erreur = 0.0;
    niveau = 0;

    valeurs_f[ 0 ] = part0;

    z_debut = z;
    da = h;
    z_milieu = z_debut + da / 2.0;
    valeurs_f[ 2 ] = MultiNormale_Multiplicative( dimension, z_milieu, x, facteurs, additif, &funcn );

    z_fin = z_debut + da;
    valeurs_f[ 4 ] = MultiNormale_Multiplicative( dimension, z_fin, x, facteurs, additif, &func2 );
    part2 = valeurs_f[ 4 ];

    coef_simpson = h / 6.0;
    contrb = coef_simpson * ( valeurs_f[ 0 ] + 4.0 * valeurs_f[ 2 ] + valeurs_f[ 4 ] );
    dlg = 0.0;

/*     printf( "Niveau 0 : contrb=%g [coef_simpson=%g]\n", contrb, coef_simpson ); */

    sens_niveau[ 0 ] = 2;
    somme_partielle[ 0 ] = 0.0;
    niveau = 1;

  L50:				/* Un goto au moins est difficile à éviter... */
    do {			/* Ligne 50 */
      dx = da / 2.0;
      coef_simpson = dx / 6.0;
      z_milieu = z_debut + dx / 2.0;
      valeurs_f[ 1 ] = MultiNormale_Multiplicative( dimension, z_milieu, x, facteurs, additif, &funcn );
      simpson_gauche = coef_simpson * ( valeurs_f[ 0 ] + 4.0 * valeurs_f[ 1 ] + valeurs_f[ 2 ] );

      z_milieu = z_debut + 1.5 * dx;
      valeurs_f[ 3 ] = MultiNormale_Multiplicative( dimension, z_milieu, x, facteurs, additif, &funcn );
      simpson_droite = coef_simpson * ( valeurs_f[ 2 ] + 4.0 * valeurs_f[ 3 ] + valeurs_f[ 4 ] );

      somme = simpson_gauche + simpson_droite;
      dlg = fabs( contrb - somme );
      eps1 = el / Puissance( 2.0, niveau - 1 );
      errl = dlg;

      /* La précision voulue est-elle atteinte ? */
      if ( ( dlg <= eps1 ) || ( dlg < QUASI_ZERO ) ) break;

      /* Contrôles pour éviter une boucle infinie ou trop longue  */
      if ( iteration > NMAX_ITERATIONS ) {
	AVERTISSEMENT( "Nombre maximal d'appels de la fonction intégrée dépassé [soit %d]", NMAX_ITERATIONS );
	break;
      }
      if ( niveau > NMAX_SUBDIVISIONS ) {
	AVERTISSEMENT( "Nombre maximal de subdivisions de l'intervalle dépassé [soit %d]", NMAX_SUBDIVISIONS );
	break;
      }
      if ( fabs( dx ) < H_MINIMAL_SIMPSON ) {
	AVERTISSEMENT( "Largeur de l'intervalle en X trop faible [%g, min. %g]", dx, H_MINIMAL_SIMPSON );
	break;
      }

      /* Ligne 60 : point d'entrée inutilisé dans la méthode choisie */
      /* Passage à la subdivision inférieure */
      sens_niveau[ niveau ] = 1;
      f0t[ niveau ] = valeurs_f[ 2 ];
      f1t[ niveau ] = valeurs_f[ 3 ];
      f2t[ niveau ] = valeurs_f[ 4 ];
      da = dx;
      valeurs_f[ 4 ] = valeurs_f[ 2 ];
      valeurs_f[ 2 ] = valeurs_f[ 1 ];

      simpson_temp[ niveau ] = simpson_droite;
      contrb = simpson_gauche;

      niveau++;

      /* Boucle tant que la précision voulue n'a pas été atteinte */
    } while ( ( dlg > eps1) && ( dlg >= QUASI_ZERO ) );

    /* Précision suffisante atteinte : on s'occupe de la partie droite */
    erreur += errl;		/* L70 */

    do {			/* L80 */
      niveau--;

      if ( 1 == sens_niveau[ niveau ] ) {
	somme_partielle[ niveau ] = somme;
	sens_niveau[ niveau ] = 2;
	dx *= 2.0;
	z_debut += dx;
	da = h / Puissance( 2.0, niveau );
	valeurs_f[ 0 ] = f0t[ niveau ];
	valeurs_f[ 2 ] = f1t[ niveau ];
	valeurs_f[ 4 ] = f2t[ niveau ];
	contrb = simpson_temp[ niveau ];

	goto L50;
      }

      somme += somme_partielle[ niveau ];
    } while ( niveau > 0 );

    contrb = somme;
    niveau = 1;
    dlg = erreur;

    /* Ligne 90 */
    proba_temp += contrb;
    limite += dlg;
    exces = el - dlg;
    el = el_initial;
    if ( exces > 0 ) el += exces;

    /* Condition d'arrêt */
/*     printf("func0 = %g, func2 = %g\n", func0, func2 ); */
    if ( ( ( func0 > 0.0 ) && ( func2 <= func0 ) ) ||
	 ( ( func0 < 0.0 ) && ( func2 >= func0 ) ) ) {
      double z_Gauss;

      z_Gauss = - z_fin * M_SQRT2;
      part3 = fabs( func2 ) * PROBA_FdR_Gauss( z_Gauss ) + limite;
/*       printf("func2 = %g ; z_Gauss = %g ; limite = %g\n", func2, z_Gauss, limite); */
      if ( part3 <= PRECISION ) {
/* 	printf("Sortie anticipée : part3=%g < %g\n", part3, PRECISION); */
	break;  /* Terminé (?) */
      }
    }
    z = z_fin;
    part0 = part2;
    func0 = func2;

    if ( iteration >  NMAX_ITERATIONS ) {
      AVERTISSEMENT( "Nombre maximal d'appels de la fonction intégrée dépassé [soit %d]", NMAX_ITERATIONS );
      break;
    }

    /* Tant que l'intervalle n'a pas été parcouru */
/*     printf(" z=%g, z_max=%g\n", z, z_max ); */
  } while ( z < z_max );

  /* On regroupe tout ! */
/*   printf( "Proba avant : %g - proba_temp = %g ; additif = %g\n", probabilite, proba_temp, additif ); */
  probabilite *= ( proba_temp - additif / 2.0 );
  limite = part3;

  if ( limite > PRECISION ) {
    AVERTISSEMENT( "Impossible d'atteindre la précision désirée [%g, on atteint %g]", PRECISION, limite );
  }
  if ( ( probabilite < 0.0 ) || ( probabilite > 1.0 ) ) {
    AVERTISSEMENT( "Valeur de probabilite absurde : %g", probabilite );
  }

  /* On renvoie le résultat */
  return probabilite;
}

/* ------ Fonction PFUNC du code original ------ */

static double MultiNormale_Multiplicative(unsigned int dimension, double z, double *x_max, double *facteurs, double additif, double *partiel)
{
  double facteur = 1.0 /* BI */, rd = 1.0 /* RSLT1 */, rg = 1.0 /* RSLT2 */;
  double borne_haute /* HI */;
  double fou = 0.0, gou = 0.0; /* Comme IERC = 0, les tableaux sont inutiles ! */
  unsigned int i;

#ifdef inutile
  /* Réservation des tableaux */
  fou = (double *) malloc( sizeof( double ) * dimension );
  if ( NULL == fou ) {
    ERR_Memoire( "Tableau fou", sizeof( double ) * dimension );
    return -1.0;		/* Une densité ne peut être négative ! */
  }
  gou = (double *) malloc( sizeof( double ) * dimension );
  if ( NULL == gou ) {
    ERR_Memoire( "Tableau gou", sizeof( double ) * dimension );
    free( fou ); fou = NULL;
    return -1.0;		/* Une densité ne peut être négative ! */
  }
#endif

  /* Initialisation des variables */
  borne_haute = x_max[ 0 ] + 1.0;  /* Pour être sûr que cela diffère ! */

  /* La boucle */
  for ( i = 0 ; i < dimension ; i++ ) {
    if ( ( facteurs[ i ] == facteur ) && ( borne_haute == x_max[ i ] ) ) {
      /* Valeurs de fou et gou identiques... Donc rien à faire ! */
    } else {
      facteur = facteurs[ i ];
      borne_haute = x_max[ i ];

      /* On est forcément dans le cas INFI = 1 */
      if ( 0.0 == facteur ) {
	fou = PROBA_FdR_Gauss( borne_haute );
	gou = 0.0;		/* En FORTRAN, tableaux init. à 0 ?? */
      } else {
	double denominateur /* DEN */, bp /* BP */;
	double u;

	denominateur = sqrt( 1.0 - facteur * facteur );
	bp = facteur * M_SQRT2 / denominateur;

	u = borne_haute / denominateur + z * bp;
	gou = PROBA_FdR_Gauss( u );

	u = borne_haute / denominateur - z * bp;
	fou = PROBA_FdR_Gauss( u );
/* 	printf( "dénominateur=%g, bp=%g, borne_haute=%g, u(fou)=%g -> fou=%g\n", */
/* 		denominateur, bp, borne_haute, u, fou ); */
      }
    }

/*     printf( "Dimension i=%u [facteur=%g] : fou=%g, gou=%g\n", i, facteurs[ i ], fou, gou ); */

    /* Mise à jour des deux côtés */
    rd *= fou;
    rg *= gou;

    /* Test de comparaison à SMALL : négligé */
  }

  /* Calcul terminé : on renvoie résultat */
  *partiel = rd + rg + additif;
/*   printf( "rd = %g, rg = %g, additif = %g ==> partiel=%g ; z=%g\n", */
/* 	  rd, rg, additif, *partiel, z ); */
  return ( *partiel ) * exp( -z * z ) / sqrt( M_PI );
}
