/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996 - 2006, Emmanuel CURIS
   Création : 2 juillet 2006

   Calculs de probabilité divers

   Calculs autour de la loi F de Fisher-Snedecor
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef AVEC_GSL_
# include <gsl/gsl_cdf.h>
# include <gsl/gsl_sf.h>
#endif

#include "math_utils.h"
#include "systeme.h"

#include "proba.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define DX_MINIMAL       4.0e-6	/* DXMIN ; valeur originale de Dunnett  */
#define NMAX_DDL_STUDENT 150	/* MAXDF ; nombre maximal de ddl, au-delà : interpolation quadratique... */
#define NMAX_NIVEAUX      30	/* Comme dans le code original */
#define NEGLIGEABLE      1.0e-8	/* SMALL ; Valeur originale de Dunnett [devrait suffire pour les stats !] */
#define PRECISION        1.0e-4 /* ERRB ; Valeur originale de Dunnett [devrait suffire pour les stats !] */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double PROBA_Student_Densite  (unsigned int dimension, unsigned long ddl, double *x, double *facteurs, double z);
static double PROBA_Student_dVariance(double x, unsigned long dimension);
static double PROBA_Student_Simpson  (unsigned int dimension, unsigned long ddl, double *x, double *facteurs);

/* -------- Fonction de répartition d'une loi de Student -------- */

double PROBA_FdR_Student(double x, double ddl_denominateur)
{
#ifdef AVEC_GSL_
  return gsl_cdf_tdist_P( x, ddl_denominateur );
#else
  AVERTISSEMENT( "Calcul de T(%g;%g) impossible sans la GSL", x, ddl_denominateur );
  return -2.0;			/* Valeur absurde pour indiquer que l'on ne sait pas [encore] faire */
#endif
}

/* -------- Fractiles d'une loi de Student -------- */

double PROBA_Fractile_Student(double probabilite, double ddl_denominateur)
{
  if ( ( probabilite <= 0.0 ) || ( probabilite >= 1.0 ) ) {
    AVERTISSEMENT( "Valeur de probabilite absurde : %g", probabilite );
    return 0.0;			/* Valeur arbitraire ! */
  }
#ifdef AVEC_GSL_
  printf( "%s : p=%g, nu=%g ==> fractile = %g\n", __FUNCTION__, probabilite, ddl_denominateur, gsl_cdf_tdist_Pinv( probabilite, ddl_denominateur ) );
  return gsl_cdf_tdist_Pinv( probabilite, ddl_denominateur );
#else
  AVERTISSEMENT( "Calcul du fractile de T(%g;%g) impossible sans la GSL", probabilite, ddl_denominateur );
  return -2.0;			/* Valeur absurde pour indiquer que l'on ne sait pas [encore] faire */
#endif
}

/* ~-~-~-~-~-~-~-~ Cas multidimensionnel ~-~-~-~-~-~-~-~ */

/* -------- Fonction de répartition pour une loi normale à N dimensions avec une structure multiplicative de corrélation --------

 -- Correspondance & contrôle des variables originales de mvstud --

 * paramètres *
   NDF  -> ddl : degrés de liberté
   A    -> x / borne haute
   B       *inutilisé* car toujours -l'infini
   BPD  -> facteurs / termes à multiplier pour avoir les rho_i_j
   ERRB -> PRECISION / précision voulue
   N    -> dimension
   INF     *inutilisé* car toujours de -infini à x : INF(I) = 1
   D       *inutilisé* [critère de non-centralité]
   IERC    *inutilisé*, vaut toujours 0
   HNC  ->
   PROB -> probabilite [valeur renvoyée par la fonction]
   BND  -> limite, pour contrôle de précision [INUTILISEE ?]
   IFLT    *inutilisé* [code erreur : affichages]

   * variables locales *
   NN      *inutilisé* [vaut 50 dans l'original, pour copie des tables]
   F    -> valeurs
   AA      *inutilisé* [simple copie de A]
   BB      *inutilisé* [simple copie de B]

   * autres variables *
   AX    -> a
   BX    -> b
   MAXDF -> NMAX_DDL_STUDENT
   NF      *inutilisé* [calculs simplistes, faits directement]
   ERB2     assimilée à ERRB = PRECISION
   XX    -> rapport
 */

double PROBA_FdR_MultiStudent_Multiplicatif(unsigned int dimension, double *x, double *facteurs, unsigned long ddl)
{
  double probabilite = 0;
  double valeurs[ 3 ];		/* Trois valeurs intermédiaires... */

  /* Si infinité de degrés de liberté : comme une gaussienne multinormale... */
  if ( 0 == ddl ) {
    return PROBA_FdR_MultiNormale_Multiplicatif( dimension, x, facteurs );
  }
  if ( ddl <= NMAX_DDL_STUDENT ) {
    probabilite = PROBA_Student_Simpson( dimension, ddl, x, facteurs );
  } else {
    double rapport, a, b;

    valeurs[ 0 ] = PROBA_FdR_MultiNormale_Multiplicatif( dimension, x, facteurs );
    valeurs[ 2 ] = PROBA_Student_Simpson( dimension, NMAX_DDL_STUDENT / 2, x, facteurs );
    valeurs[ 1 ] = PROBA_Student_Simpson( dimension, NMAX_DDL_STUDENT    , x, facteurs );

    rapport = (double) NMAX_DDL_STUDENT / (double) ddl;
    a = valeurs[ 2 ] - 2.0 * valeurs[ 1 ] + valeurs[ 0 ];
    b = 4.0 * valeurs[ 1 ] - valeurs[ 2 ] - 3.0 * valeurs[ 0 ];

    probabilite = valeurs[ 0 ] + rapport * ( a * rapport + b) / 2.0;
  }

  /* Calcul terminé */
  return probabilite;
}

/* -------- Calcul de l'intégrande, réécrit en monodimensionnel [en conditionnant par la variance]
            Trnsposition de la fonction FUN de l'original

 -- Correspondance & contrôle des variables originales de FUN --

 * paramètres *
   Z    -> z
   NDF  -> ddl
   H    -> x
   HL      *inutilisé* [correspond à B dans l'original]
   BPD  -> facteurs
   ERB2    *inutilisé* (précision pour la multi-normale)
   N    -> dimension
   INF     *inutilisé* [toujours de -infini à A]
   D       *inutilisé* [toujours pour un vecteur centré]
   F0   -> f0
   G0      *inutilisé* [pour le calcul de la précision...]
   IERC    *inutilisé*
   HNC     *inutilisé* (h initial pour la multi-normale)
   IER     *inutilisé*

 * variables locales déclarées *
   A    -> x_reel
   B       *inutilisé*

 * autres variables *
   ARG  -> arg
   TERM -> terme
   PROB -> probabilite
   -------- */

static double PROBA_Student_Densite(unsigned int dimension, unsigned long ddl, double *x, double *facteurs, double z)
{
  double f0 = 0.0, arg, terme, un_moins_z, probabilite;
  double *x_reel = NULL;
  unsigned int i;

  if ( fabs( z ) >= 1.0 ) return 0.0;

  un_moins_z = 1.0 - z;
  arg = (1.0 + z) / un_moins_z;
  terme  = arg * ( (double) ( ddl << 1 ) ) / (un_moins_z * un_moins_z);
  terme *= PROBA_Student_dVariance( arg * arg * ddl / 2.0, ddl );
  if ( terme < NEGLIGEABLE ) return 0.0;

  x_reel = (double *) malloc( sizeof( double ) * dimension );
  if ( NULL == x_reel ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver la table...\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }
  for ( i = 0 ; i < dimension ; i++ ) {
    x_reel[ i ] = arg * x[ i ];
  }
  probabilite = PROBA_FdR_MultiNormale_Multiplicatif( dimension, x_reel, facteurs );
  f0 = terme * probabilite; 

  free( x_reel ); x_reel = NULL;

  /* Terminé, calcul réussi */
  return f0;
}

/* -------- Densité de probabilité pour la variance commune --------
            Transposition de la fonction SDIST de l'original

 -- Correspondance & contrôle des variables originales de SDIST --

 * paramètres *
   Y -> x
   N -> dimension
   -------- */

static double PROBA_Student_dVariance(double x, unsigned long dimension)
{
  double ln_dp, demi_dimension;

  /* D'après le commentaire, calcule f(x) = x^(n/2 - 1) * exp(-x) / gamma(x/2)
     Plutôt que le code original, on passe en log : 
     ln f(x) = (n/2 - 1) ln x - x - ln gamma(x/2)

     REMARQUE : on doit pouvoir calculer gamma(x/2) de façon plus astucieuse
     en utilisant Gamma(n+1) = n! et Gamma(n+1/2) = ??
 */
  if ( x < 0.0 ) return 0.0;	/* Une variance ne peut pas être négative ! */

  demi_dimension = dimension / 2.0;

  ln_dp  = log( x ) * ( demi_dimension - 1.0 );
  ln_dp -= x;
#ifdef AVEC_GSL_
  ln_dp -= gsl_sf_lngamma( demi_dimension );
#else
  OPTION_A_FAIRE( "Implémenter le ln Gamma (x)" );
#endif

  return exp( ln_dp );
}

/* -------- Approximation de l'intégrale par la méthode de Simpson
            Transposition de la fonction SIMPSN de l'original

 -- Correspondance & contrôle des variables originales de SIMPSN --

 * paramètres *
   NDF  -> ddl
   A    -> x
   B       *inutilisé* [-infini]
   BPD  -> facteurs
   ERRB -> PRECISION
   N    -> dimension
   INF     *inutilisé* [toujours de -infini à x]
   D       *inutilisé*
   IERC    *inutilisé*
   HNC  ->
   PROB -> resultat [valeur renvoyée]
   BND  -> limite
   IFLT   *inutilisé* [affichage]

 * variables locales déclarées *
   DXMIN -> DX_MINIMAL
   FV    -> f
   F1T   -> f0t
   F2T   -> f2t
   F3T   -> f4t
   LDIR  -> direction
   PSUM  -> somme_partielle
   ESTT  -> estimation_tmp
   ERRR  -> erreur_tmp
   GV      *inutile* [pour calcul de la limite seulement...]
   G1T     *inutile* [pour calcul de la limite seulement...]
   G2T     *inutile* [pour calcul de la limite seulement...]
   G3T     *inutile* [pour calcul de la limite seulement...]
   GSUM    *inutile* [pour calcul de la limite seulement...]

 * autres variables *
   BOUNDA *inutile* [utilisé pour BND, juste avant return, or ici limite non-renvoyé]
   BOUNDG *inutile* [utilisé pour BOUNDA uniquement...]
   CONTRB -> contribution
   CONTRG *inutile* [utilisé pour BOUNDG uniquement...]
   DA     -> da
   DAX    -> dax
   DX     -> dx
   EPS1   -> epsilon
   ERB2   *inutile* [vaut toujours ERRB/2 = PRECISION/2.0]
   ERRL   -> errl
   ERROR  -> erreur
   ESTGL  *inutile* [pour calcul de la limite seulement...]
   ESTGR  *inutile* [pour calcul de la limite seulement...]
   IFLAG  *inutile*
   IER    *inutile*
   START  -> debut
   SUMG   *inutile* [pour calcul de la limite seulement...]
   LVL    -> niveau
   WT     -> coefficient
   Z2     *inutile* [passé en paramètre, usage unique]
   Z3     *inutile* [passé en paramètre, usage unique]
   Z4     -> z4
   -------- */

static double PROBA_Student_Simpson(unsigned int dimension, unsigned long ddl, double *x, double *facteurs)
{
  double coefficient, contribution;
  double da, dax = 1.0, dx;
  double debut = -1.0;
  double epsilon = PRECISION / 4.0;
  double erreur, errl, dlg, exces, erreur_tmp[ NMAX_NIVEAUX ];
  double estimation_droite, estimation_gauche, estimation_tmp[ NMAX_NIVEAUX ];
  double limite = 0.0;
  double f0, f[ 5 ], f0t[ NMAX_NIVEAUX ], f2t[ NMAX_NIVEAUX ], f4t[ NMAX_NIVEAUX ];
  double probabilite = 0.0;
  double somme, somme_partielle[ NMAX_NIVEAUX ];
  double z4;

  int direction[ NMAX_NIVEAUX ];
  int niveau;

  f0 = PROBA_Student_Densite( dimension, ddl, x, facteurs, 0.0 );
/*   printf( "f0 = %g\n", f0 );  */

  /* L. 10 = début de boucle */
  do {
    f[ 0 ] = 0.0;
    erreur = 0.0;
    da = dax;
    niveau = 0;

    f[ 2 ] = PROBA_Student_Densite( dimension, ddl, x, facteurs, debut + dax / 2.0 );
/*     printf("En z=%g, f[ 2 ]=%g\n", debut + dax / 2.0, f[ 2 ] ); */
    f[ 4 ] = f0;
    coefficient = fabs( dax ) / 6.0;  /* Ici, da = dax */
    contribution = coefficient * ( f[ 0 ] + 4.0 * f[ 2 ] + f[ 4 ] );

    direction[ niveau ] = 2;
    somme_partielle[ niveau ] = 0.0;

    /* L. 20 = début de boucle ? */
    do {
      dx = da / 2.0;
      coefficient = fabs( dx ) / 6.0;
      f[ 1 ] = PROBA_Student_Densite( dimension, ddl, x, facteurs, debut + dx / 2.0 );
/*       printf("En z=%g, f[ 1 ]=%g\n", debut + dx / 2.0, f[ 1 ] ); */

      z4 = debut + 1.5 * dx;
      f[ 3 ] = PROBA_Student_Densite( dimension, ddl, x, facteurs, z4 );
/*       printf("En z=%g, f[ 3 ]=%g\n", z4, f[ 3 ] ); */

      estimation_gauche = coefficient * ( f[ 0 ] + 4.0 * f[ 1 ] + f[ 2 ] );
      estimation_droite = coefficient * ( f[ 2 ] + 4.0 * f[ 3 ] + f[ 4 ] );
      somme = estimation_droite + estimation_gauche;

      dlg = fabs( contribution - somme );
/*       printf( "==> contribution=%g, somme=%g ==> dlg=%g\n", contribution, somme, dlg ); */
      errl = dlg;

      /* L. 30 */
      if ( ( dlg <= epsilon ) || /* ==> aller en L. 50 */
	   ( ( fabs( dx ) <= DX_MINIMAL ) || ( niveau >= NMAX_NIVEAUX ) ) ) {  /* ==> aller en L. 40 = 50 */
	/* L. 40 & L. 50 */
	erreur += errl;

/* 	printf("-- On est L.50, erreur=%g, niveau=%u (max : %u), dx=%g (min : %g), dlg=%g (min : %g)\n", */
/* 	       erreur, niveau, NMAX_NIVEAUX, dx, DX_MINIMAL, dlg, epsilon ); */
	/* L. 60 : début de boucle */
	do {
	  if ( 1 == direction[ niveau ] ) {
	    break;			/* En ligne 70 */
	  }

	  somme += somme_partielle[ niveau ];
	  niveau--;
	} while ( niveau > 0 );	/* Fin de boucle, L 60 */

	if ( 1 == direction[ niveau ] ) {
	  /* L. 70 */
	  somme_partielle[ niveau ] = somme;
	  direction[ niveau ] = 2;
	  da = dax / Puissance( 2.0, niveau - 1 );
	  debut += 2.0 * dx;

	  f[ 0 ] = f0t[ niveau ];
	  f[ 2 ] = f2t[ niveau ];
	  f[ 4 ] = f4t[ niveau ];

	  contribution = estimation_tmp[ niveau ];
	  exces = epsilon - dlg;
	  epsilon = erreur_tmp[ niveau ];
	  if ( exces > 0 ) epsilon += exces;
	  /* Retour en ligne 20 : on tombe sur le while ( 1 )... */
	} else {
	  contribution = somme;
	  niveau = 1;
	  dlg = erreur;

	  /* L. 80 */
	  probabilite += contribution;
	  if ( z4 > 0.0 ) {
	    /* Calcul terminé */
	    return probabilite;
	  }

	  /* L. 90 */
	  break;		/* On sort de la boucle en L. 20 */
	}
      } else {
	/* On passe à la subdivision précédente */
	niveau++;
	direction[ niveau ] = 1;
	f0t[ niveau ] = f[ 2 ];
	f2t[ niveau ] = f[ 3 ];
	f4t[ niveau ] = f[ 4 ];

	da = dx;
	f[ 4 ] = f[ 2 ];
	f[ 2 ] = f[ 1 ];
	estimation_tmp[ niveau ] = estimation_droite;
	contribution = estimation_gauche;
	epsilon /= 2.0;
	erreur_tmp[ niveau ] = epsilon;
/* 	printf( " ~ niveau %u : epsilon=%g [erreur=%g, dx=%g (min : %g), dlg=%g\n", */
/* 		niveau, epsilon, erreur, dx, DX_MINIMAL, dlg ); */
      }
    } while ( 1 );

    /* Préparation du tour suivant, boucle en L. 10 */
    epsilon = PRECISION / 4.0;
    /* À Vérifier : limite = 0.0 tjrs => epsilon = PRECISION/2.0 en pratique ?... */
    exces = epsilon - limite;
    if ( exces > 0.0 ) {
      epsilon += exces;
    }
    debut = 1;
    dax = -1.0;
  } while ( 1 );

  /* On ne devrait jamais arriver ici ! */
  (void) fprintf(stderr, "%s, %s l.%u : On ne devrait jamais être là\n",
		 __FILE__, __FUNCTION__, __LINE__);
  return -1.0;
}
