/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999, Emmanuel CURIS
   Cration de ce fichier : 26 avril 1999

   Module d'analyse statistique
   Routines de bruitage d'un spectre et de gnration d'un processus alatoire
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "systeme.h"

#include "graphes.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Bruiter un spectre ~-~-~-~-~-~-~-~-~ */

int STAT_GRF_Bruiter(struct graphe *graphe, struct infos_bruitage *parametres)
{
  int reponse = -25;

  if ( ( NULL == graphe ) || ( NULL == parametres ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramtre absurde (NULL)\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* En fonction de la mthode, on appelle la bonne fonction */
  switch ( parametres->methode )
  {
  case STAT_BRUITER_CONSTANTE:
    /* Revient donc  du bruitage simple comme construit auparavant */
    reponse = GRF_BruiterSimple( graphe, parametres->loi_proba, parametres->sigma, AXE_Y );
    break;

  case STAT_BRUITER_PROPORTIONNELLE:
    reponse = GRF_BruiterProportionnel( graphe, parametres->loi_proba, parametres->sigma, AXE_Y );
    break;

  case STAT_BRUITER_QUELCONQUE:
    OPTION_A_FAIRE( "Bruitage avec une erreur lue dans un graphe" );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Mthode de bruitage inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, parametres->methode );
    return -2;
  }

  /* Calculs finis --- On renvoie le code rponse obtenu */
  return reponse;
}

/* -------- Bruitage simple en place --------
    Renvoie 0 si tout va bien,
            != 0 en cas de problme
    Bruite par un bruit blanc centr, rduit

  graphe   : le graphe  bruiter
  loi      : le type de loi  utiliser (pas de paramtres, car elle est centre)
  erreur   : quelle erreur utiliser pour le bruit blanc = cart-type [suppose constante avec I]
  axe      : sur quel axe doit-on bruiter
  ------------------------------------------- */

int GRF_BruiterSimple(Graphe *graphe, unsigned int loi, double erreur, int axe)
{
  Graphe *bruit = NULL;
  struct processus_aleatoire processus;
  int retour = 0;

  /* Contrle prliminaire */
  if ( NULL == graphe ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun graphe  bruiter !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( erreur < 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Un cart-type doit tre positif [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, erreur);
    return -2;
  }
  if ( 0.0 == erreur ) {
    AVERTISSEMENT( "Erreur absente, cart-type nul [%g]", 0.0 );
    return 0; /* Pas d'erreur en ralit ! */
  }

  /* Processus de type bruit-blanc */
  processus.type = PA_BRUIT_BLANC;

  /* On prpare la loi du processus alatoire */
  processus.loi.famille = loi;
  switch ( loi )
  {
  case DP_GAUSSIENNE:
    processus.loi.parametres.Gauss.mu = 0.0;
    processus.loi.parametres.Gauss.sigma = erreur;
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Loi de probabilit non-gre ou inconnue [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, loi);
    return +1;
  }

  /* Bruitage sur l'axe des X ? */
  if ( axe & AXE_X ) {
    OPTION_A_FAIRE( "Bruiter l'axe des X" );
  }

  /* Bruitage sur l'axe des Y ? */
  if ( axe & AXE_Y ) {
    Graphe *erreurs;

    /* On gnre un processus alatoire de type bruit-blanc */
    bruit = PA_GenererStationnaire( &processus, graphe->nombre_point );
    if ( NULL == bruit ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de gnrer le bruit !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -20;
    }

    /* Il a forcment les mmes X que le spectre  bruiter !*/
    bruit->x = graphe->x;

    /* On gnre les erreurs en Y pour ce graphe bruit */
    erreurs = GRF_creer_erreurs( bruit, AXE_Y );
    if ( erreurs != NULL ) {	/* On a russi ! */
      GRF_MettreConstante( erreurs, erreur, AXE_Y );
    }

    /* On l'ajoute aux valeurs exprimentales */
    retour = GRF_AjouterGraphe( graphe, bruit, VRAI, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur survenue dans la somme [Y, code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }

    /* Plus besoin du processus ; graphe isol */
    GRF_eff_isole( bruit ); bruit = NULL;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Ajouter un bruit proportionnel aux valeurs -------- */

int GRF_BruiterProportionnel(Graphe *graphe, unsigned int loi, double coefficient, int axe)
{
  Graphe *bruit = NULL, *erreurs = NULL;
  double sigma = 0.0;
  int retour;
  unsigned long i;
  struct processus_aleatoire processus = { PA_BRUIT_BLANC, { { 0 } }, { DP_GAUSSIENNE, 0, { { 0, 0 } } } };

  /* On initialise les infos pour la loi de probabilit */
  processus.loi.famille    = loi;
  processus.loi.proprietes = 0;

  if ( axe & AXE_X ) {
    OPTION_A_FAIRE( "Bruit sur l'axe des X" );
  }

  if ( axe & AXE_Y ) {
    /* On cre un graphe : isol, mmes X que l'exprimental */
    bruit = GRF_nv_TableX( NULL, graphe->type, GRF_SERIE, NULL, VRAI, graphe->x, graphe->nombre_point, FAUX );
    if ( NULL == bruit ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de faire le graphe interne !\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }

    /* On cre des erreurs pour ce graphe */
    erreurs = GRF_creer_erreurs( bruit, AXE_Y );
    if ( NULL == erreurs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de faire les erreurs sur le graphe interne !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      GRF_eff_isole( bruit );
      return -20;
    }

    /* Pour chaque point, on calcule l'cart-type */
    for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
      sigma = fabs( graphe->y[ i ] * coefficient );

      erreurs->y[ i ] = sigma;
      erreurs->z.ddl[ i ] = 0;	/* Bruit thorique ! */
    }

    /* Ici, bruit est donc un spectre nul avec des barres d'erreur
       ==> on l'utilise pour gnrer dedans le bruit en chaque valeur de X */
    retour = PA_GenererSurGraphe( &processus, bruit );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur survenue en gnrant le bruit [Y, code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }

    /* On ajoute le bruit complet aux valeurs exprimentales */
    retour = GRF_AjouterGraphe( graphe, bruit, VRAI, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur survenue dans la somme [Y, code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }

    /* Plus besoin du processus ; graphe isol */
    GRF_eff_isole( bruit ); bruit = NULL;
  }

  /* Calculs termins */
  return 0;
}
