/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1999, Emmanuel CURIS
   Cration : 21 aot 2000

   Module de calculs statistiques

   Comparaison de 2 distributions par le test de Cramer-von Mises
   (actuellement, on ne fait que tester la normalit d'une distribution)
   Rfrences pour le test : Saporta, p. 336 - 338
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"

#include "graphes.h"
#include "statistiques.h"
#include "science.h" /* Pour erf sur les systmes qui ne l'ont pas */
#include "tri.h"

#include "cvm.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_GRAPHES_REF 40
#define LGMAX_DISTRIBUTION    30

extern const char *txt_distributions[]; /* Dans Stats/aleatoire.c */
extern const char *txt_aucun_choisi;    /* Dans variables.c       */

struct infos_KS
{
  Graphe *reference;
  Graphe *experience;
  int type_distrib;
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int CVM_gaussienne(double *y_exp, unsigned long nbr_pts, double m, double s, boolean theorie);

/* ---- dans k_smirnov.c : on utilise le mme arbre, seul le titre change ---- */

extern boolean GER_ks(unsigned int objet, int x, int y);
extern void prepare_arbre_ks(void);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- La fonction de rpartition indique est-elle gaussienne ? --------\
 |    Renvoie VRAI si l'hypothse est valable avec un seuil de 0,05           |
 |            FAUX dans le cas contraire                                      |
 |                                                                            |
 | y_exp   : contient les ralisations exprimentales TRIES                  |
 | nbr_pts : nombre de valeurs exprimentales                                 |
 | m       : moyenne attendue de la gaussienne                                |
 | s       : cart-type attendu de la gaussienne                              |
 | theorie : VRAI si m et s thorique, FAUX si exprimentaux                  |
 \--------------------------------------------------------------------------- */

static int CVM_gaussienne(double *y_exp, unsigned long nbr_pts, double m, double s, boolean theorie)
{
  unsigned long i;
  register double *y;
  double wn = 0.0, limite = 0.0;
  unsigned long deux_n;

  if (y_exp == NULL) return FAUX;
  if (nbr_pts == 0) {
    EXF_ERREUR(DIV_ZERO, "Aucun point", txt_err_calc_annule);
    return FAUX;
  }

  deux_n = nbr_pts << 1;
  if (deux_n < nbr_pts) {
    /* Erreur : on dpasse la capacit de la machine ! */
    return FAUX;
  }
  wn = 1.0 / ( 12.0 *  (double) nbr_pts );

  for (i = 1, y = y_exp ; i <= nbr_pts ; i++, y++) {
    double x_cr;    /* X en coordonnes centres rduites */
    double y_gauss; /* Valeur de la fonction de rpartition gaussienne attendue en ce point */
    double d;

    x_cr = (*y - m)/s;
    y_gauss = 0.5 * (1 + erf(x_cr / M_SQRT2));

    d  = (2.0 * (double) i - 1.0) / (double) deux_n;
    d -= y_gauss;
    wn += d * d;
  }
  /* Ici, on a donc bien nwn = 1/(12n) + somme ( (2i-1)/(2n) - Fgauss(x)) */

  limite = 1.0 + 1.0 / (double) deux_n;

  (void) BI_ERR_alerte(NULL, &CVM_Gauss_resultat, VRAI, wn, 0.126/limite, 0.178/limite);

  if ( VRAI == theorie ) { /* Comparaison  une distribution thorique */
    HIST_ResultatTestStatistique("Cramer -- von Mises",
				 "la distribution exp"e_AIGU"rimentale est gaussienne",  "la distribution exp"e_AIGU"rimentale n'est pas gaussienne",
				 wn, 0.0, -1.0, -1.0, -1.0 /* A remplacer par un calcul de la loi... */, FAUX,
				 "Nombre de valeurs assez grand");
    HIST_Texte(3, 0, "Param"e_GRAVE"tres de la gaussienne mod"e_GRAVE"le : "MU" = %.10g, "SIGMA" = %.10g\n", m, s);
  } else { /* Moyenne, variance estims sur l'exprience */
    HIST_ResultatTestStatistique("Cramer -- von Mises",
				 "la distribution exp"e_AIGU"rimentale est gaussienne",  "la distribution exp"e_AIGU"rimentale n'est pas gaussienne",
				 wn, 0.0, -1.0, -1.0, -1.0, FAUX,
				 "Nombre de valeurs assez grand, pour pouvoir utiliser les limites qui suivent");
    HIST_Texte(3, 0, "Valeur "a_GRAVE" ne pas d"e_AIGU"passer pour accepter H0 au seuil de 5%% : %g\n", 0.126 / limite);
    HIST_Texte(3, 0, "Valeur "a_GRAVE" ne pas d"e_AIGU"passer pour accepter H0 au seuil de 1%% : %g\n", 0.178 / limite);
    HIST_Texte(3, 0, "Param"e_GRAVE"tres estim"e_AIGU"s pour la gaussienne mod"e_GRAVE"le : m = %.10g, s* = %.10g\n", m, s);
  }

  if (wn > 0.126 / limite) {
    return FAUX;
  }

  return VRAI; /* A pass le test */
}

/* -------- Ralisation effective du test de Kolmogorov-Smirnov -------- */

void cramer_vonmises(Graphe *experience)
{
  double m, s; /* Valeur de la gaussienne */
  struct infos_KS *infos;

  if (experience == NULL)
    return;

  if (arbres[PARAM_KSMIRNOV].drapeaux.prepare == FAUX)
    prepare_arbre_ks(); /* Dans k_smirnov.c */

  infos = (struct infos_KS *) malloc( sizeof(struct infos_KS) );
  if (infos == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/c_vmises.c", "cramer_vonmises", "infos");
    return;
  }
  infos->experience = experience;
  infos->reference = NULL;
  infos->type_distrib = DP_GAUSSIENNE;
  ARB_MettreComplement( PARAM_KSMIRNOV, infos );

  GER_ks(PKS_REF_THEORIE, 0, 0);
  met_chaine(PARAM_KSMIRNOV, PKS_TYPE_DISTRIB, LGMAX_DISTRIBUTION,
	     txt_distributions[infos->type_distrib], TXT_COPIER);
  met_chaine(PARAM_KSMIRNOV, PKS_GRF_EXP, LGMAX_NOM_GRAPHES_REF,
	     experience->nom, TXT_COPIER);
  met_chaine(PARAM_KSMIRNOV, PFS_GRF_REF, LGMAX_NOM_GRAPHES_REF, txt_aucun_choisi, TXT_COPIER);

  if ( ARB_Bloquant( PARAM_KSMIRNOV, titre_arbre_cvm, principale, 0 ) == PKS_ANNULER ) {
    free(infos);
    return;
  }

  if ( OBJ_choisit(PARAM_KSMIRNOV, PKS_REF_THEORIE, QUESTION) == VRAI ) {  /* On compare avec une distribution thorique */
    double *y, *y_exp;

    if ( OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, QUESTION) == VRAI ) { /* Valeurs exprimentales */
      m = esperance(experience);
      s = sqrt(variance(experience, m));
    } else {
      m = OBJ_LireNombre( PARAM_KSMIRNOV, KS_MOYENNE, NULL );
      s = OBJ_LireNombre( PARAM_KSMIRNOV, PKS_SIGMA, NULL );
    }

    /* On trie les valeurs exprimentales */
    y = (double *) calloc(experience->nombre_point, sizeof(double));
    if (y == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Stats/k_smirnov.c", "kolmogorov_smirnov", "y");
      return;
    }
    memcpy(y, experience->y, sizeof(double) * experience->nombre_point);

    /* Tri par fusion des donnes exprimentales */
    y_exp = TRI_Tableau(y, experience->nombre_point);
    if (y_exp == NULL) { /* Le tri a chou : manque de mmoire a priori */
      free(y);
      return;
    } else
      y = y_exp;

    /* Application du test C-vM */
    switch (infos->type_distrib)
    {
    case DP_GAUSSIENNE:
      CVM_gaussienne(y_exp, experience->nombre_point, m, s,
		     OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, QUESTION));
      break;
    }

    free(infos);
    return;
  }

  /* On compare avec une autre distribution exprimentale */
}
