/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Création : 9 février 1999

   Calculs statistiques divers

   Calcul d'une analyse de variance à un facteur au sens large [donc régression linéaire & non-linéaire Y=f(X) incluses]
          des tests a posteriori de cette analyse
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "utils_tables.h"

#include "graphes.h"
#include "lissage.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double STAT_MedianeFDR    (struct graphe *valeurs);
static double STAT_MedianeUsuelle(struct graphe *valeurs);

static int STAT_QuartilesFDR   (struct graphe *valeurs, double quartiles[ 3 ]);
static int STAT_QuartilesUsuels(struct graphe *valeurs, double quartiles[ 3 ]);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~ Calcul de la médiane ~-~-~-~-~-~-~-~ */

double STAT_Mediane(struct graphe *valeurs, unsigned int methode)
{
  double mediane;

  if ( NULL == valeurs ) return 0.0;

  switch ( methode )
  {
  case STAT_MEDIANE_FDR:
    mediane = STAT_MedianeFDR( valeurs );
    break;

  case STAT_MEDIANE_USUELLE:
    mediane = STAT_MedianeUsuelle( valeurs );
    break;

  default:
    AVERTISSEMENT( "Méthode de calcul inconnue:  %d", methode );
    return 0.0;
  }

  return mediane;
}

/* -------- Calcul avec la définition probabiliste : Me, valeur telle que f(Me) = 1/2 -------- */

static double STAT_MedianeFDR(struct graphe *valeurs)
{
  double mediane;
  int retour;
  Graphe *f_repartition;

  /* 1. On estime la fonction de répartition */
  f_repartition = STAT_FdREmpirique( valeurs, STAT_FDR_EMPIRIQUE_BASE, 0 );
  if ( NULL == f_repartition ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de calculer la fonction de répartition empirique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  /* 2. On inverse X & Y de la fonction de répartition : donnera donc Y=la valeur connaissant X=le fractile */
  retour = GRF_PermuterXY( f_repartition );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en permuttant X et Y, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* 3. On l'utilise pour avoir les valeurs cherchées */
  mediane = GRF_LireValeur( f_repartition, 0.50, INTERPOLE_LINEAIRE );

  /* 4. Plus besoin de la fonction de répartition */
  GRF_eff_isole( f_repartition ); f_repartition = NULL;

  /* 5. Terminé */
  return mediane;
}

/* -------- Calcul avec la définition usuelle : coupe en deux parties égales l'échantillon --- */

static double STAT_MedianeUsuelle(struct graphe *valeurs)
{
  double *valeurs_triees;
  double mediane;

  /* 1. On fait une copie isolée des Y du graphe */
  valeurs_triees = TABLE_CopierTable( valeurs->y, valeurs->nombre_point );
  if ( NULL == valeurs_triees ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de trier la table des valeurs [réservation]\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  /* 2. On trie cette copie */
  TABLE_TrierTable( valeurs_triees, valeurs->nombre_point, 1 );

  /* 3. On récupère la médiane */
  mediane = TABLE_Triee_Mediane( valeurs_triees, valeurs->nombre_point );

  /* 4. On libère la table triée */
  free( valeurs_triees ); valeurs_triees = NULL;

  /* Fini ! */
  return mediane;
}

/* ~-~-~-~-~-~-~-~ Calcul des quartiles ~-~-~-~-~-~-~-~ */

int STAT_Quartiles(struct graphe *valeurs, unsigned int methode, double quartiles[ 3 ])
{
  int retour = -5;

  if ( NULL == valeurs ) return -1;
  if ( NULL == quartiles ) {
    (void) fprintf(stderr, "%s, %s l.%u : Rien pour mémoriser les résultats !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -25;
  }

  switch ( methode )
  {
  case STAT_MEDIANE_FDR:
    retour = STAT_QuartilesFDR( valeurs, quartiles );
    break;

  case STAT_MEDIANE_USUELLE:
    retour = STAT_QuartilesUsuels( valeurs, quartiles );
    break;

  default:
    AVERTISSEMENT( "Méthode de calcul inconnue:  %d", methode );
  }

  /* Calcul terminé */
  return retour;
}

static int STAT_QuartilesFDR(struct graphe *valeurs, double quartiles[ 3 ])
{
  int retour;
  Graphe *f_repartition;

  /* 1. On estime la fonction de répartition */
  f_repartition = STAT_FdREmpirique( valeurs, STAT_FDR_EMPIRIQUE_BASE, 0 );
  if ( NULL == f_repartition ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de calculer la fonction de répartition empirique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -10;
  }

  /* 2. On inverse X & Y de la fonction de répartition : donnera donc Y=la valeur connaissant X=le fractile */
  retour = GRF_PermuterXY( f_repartition );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en permuttant X et Y, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    GRF_eff_isole( f_repartition ); f_repartition = NULL;
    return -11;
  }

  /* 3. On l'utilise pour avoir les valeurs cherchées */
  quartiles[ 0 ] = GRF_LireValeur( f_repartition, 0.25, INTERPOLE_LINEAIRE );
  quartiles[ 1 ] = GRF_LireValeur( f_repartition, 0.50, INTERPOLE_LINEAIRE );
  quartiles[ 2 ] = GRF_LireValeur( f_repartition, 0.75, INTERPOLE_LINEAIRE );

  /* 4. Plus besoin de la fonction de répartition */
  GRF_eff_isole( f_repartition ); f_repartition = NULL;

  /* 5. Terminé */
  return 0;
}

static int STAT_QuartilesUsuels(struct graphe *valeurs, double quartiles[ 3 ])
{
  double *valeurs_triees;

  /* 1. On fait une copie isolée des Y du graphe */
  valeurs_triees = TABLE_CopierTable( valeurs->y, valeurs->nombre_point );
  if ( NULL == valeurs_triees ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de trier la table des valeurs [réservation]\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -10;
  }

  /* 2. On trie cette copie */
  TABLE_TrierTable( valeurs_triees, valeurs->nombre_point, 1 );

  /* 3. On récupère les quartiles */
  if ( TABLE_Triee_Quartiles( valeurs_triees, valeurs->nombre_point, quartiles ) < 0.0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans TABLE_Triee_Quartiles()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -11;
  }

  /* 4. On libère la table triée */
  free( valeurs_triees ); valeurs_triees = NULL;

  /* Fini ! */
  return 0;
}
