/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998, Emmanuel CURIS

   Calcul thorique de l'absorption des lments
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "chimie_utils.h"
#include "absorption.h"
#include "graphes.h"
#include "theorie.h"
#include "science.h" /* Pour la conversion E <-> longueur d'onde */

/* ~~~~~~~~ Variables locales~~~~~~~~ */

#ifdef AVEC_TRANSURANIENS_
# define Z_DERNIER_SEUIL 118      /* Dernier lment dont les seuils sont connus : aprs, seuil R1 apparat -> non prvu dans LASE */ 
#else
# define Z_DERNIER_SEUIL NOBELIUM /* Dernier lment dont les seuils sont connus : aprs, seuil P4 apparat -> il faut les transuraniens */ 
#endif

static const char *NomsSeuils[ NOMBRE_SEUILS ] = {
  "K (1s)",
  "L1 (2s)", "L2 (2p1/2)", "L3 (2p3/2)",
  "M1 (3s)", "M2 (3p1/2)", "M3 (3p3/2)", "M4 (3d3/2)", "M5 (3d5/2)",
  "N1 (4s)", "N2 (4p1/2)", "N3 (4p3/2)", "N4 (4d3/2)", "N5 (4d5/2)", "N6 (4f5/2)", "N7 (4f7/2)",
  "O1 (5s)", "O2 (5p1/2)", "O3 (5p3/2)", "O4 (5d3/2)", "O5 (5d5/2)", "O6 (5f5/2)", "O7 (5f7/2)", /* O8... jamais occups */
  "P1 (6s)", "P2 (6p1/2)", "P3 (6p3/2)",
  "Q1 (7s)"
#ifdef AVEC_TRANSURANIENS_
  , "P4 (6d3/2)", "P5 (6d5/2)", "Q2 (2p1/2)", "Q3 (2p3/2)"
#endif
};

#define FICHIER_ENERGIES_SEUILS PREFIXE_REP"/Lase/energies_seuils.asc"

static float *EnergiesSeuils[ NOMBRE_SEUILS ] = {
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL
#ifdef AVEC_TRANSURANIENS_
  , NULL, NULL, NULL, NULL,
#endif
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int          SEUIL_LireTable     (void);
static int          SEUIL_PlusProche    (unsigned int seuil, double energie, double *distance);

static int TH_SeuilEscalier  (Graphe *graphe, double position, double hauteur);
static int TH_SeuilGaussien  (Graphe *graphe, double position, double hauteur, double largeur);
static int TH_SeuilLorentzien(Graphe *graphe, double position, double hauteur, double largeur);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Lire les nergies de seuil dans le fichier ~-~-~-~-~-~-~-~-~ */

int SEUIL_InitialiserTables(void)
{
  int Z_min, Z_max = Z_DERNIER_SEUIL;
  unsigned int i;
  int retour = 0, renvoi;

  /* On libre ventuellement d'anciennes tables */
  SEUIL_LibererTables();

  /* On rserve les nouvelles tables */
  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ ) {
    Z_min = SEUIL_PremierElement( i );
    if ( Z_min > Z_max ) {
      (void) fprintf(stderr, "%s,%s l.%u : Le seuil %s n'est pas gr\n",
		     __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( i ) );
    } else {
      EnergiesSeuils[ i ] = TH_Creer_Table( Z_min , Z_max );
      if ( NULL == EnergiesSeuils[ i ] ) {
	(void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante --- le seuil %s ne sera pas gr\n",
		       __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( i ) );
	retour = +1; /* Erreur, mais mineure */
      }
    }
  }

  /* On lit les tables */
  renvoi = SEUIL_LireTable();
  if ( renvoi != 0 ) { /* Erreur grave */
    SEUIL_LibererTables();
    return -1;
  }

  /* Tout s'est bien pass */
  return retour;
}

/* -------- Libration des tables de seuils -------- */

int SEUIL_LibererTables(void)
{
  unsigned int i;

  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ ) {
    free( EnergiesSeuils[ i ] );
    EnergiesSeuils[ i ] = NULL;
  }

  return 0;
}

/* -------- Lecture des tables de coefficients -------- */

static int SEUIL_LireTable(void)
{
  FILE *fichier;
  unsigned int nbr_colonnes;
  unsigned long i, nbr_lignes;
  double *valeurs;

  /* On ouvre le fichier, compte les lignes,... */
  fichier = TH_OuvrirFichier( FICHIER_ENERGIES_SEUILS, &nbr_lignes, &nbr_colonnes, &valeurs );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de lire %s\n",
		   __FILE__, __FUNCTION__, __LINE__, FICHIER_ENERGIES_SEUILS);
    return -1;
  }

  /* On lit le fichier, ligne par ligne */
  for ( i = 0 ; i < nbr_lignes ; i++ ) {
    int Z;
    unsigned int seuil;

    Z = TH_LireLigne( fichier, nbr_colonnes, valeurs, NULL );
    if ( Z < 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_LireLigne() [ligne %lu / Z=%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, i, ATM_Z_vers_symbole( Z ) );
      (void) fclose( fichier );
      return -2;
    }

/*     printf("%s : ligne %lu [Z=%d / %s]\n", __FUNCTION__, i, Z, ATM_Z_vers_symbole( Z ) ); */

    /* Colonnes : seuils par ordre croissant */
    for ( seuil = 0 ; seuil < nbr_colonnes ; seuil++ ) {
      if ( seuil < NOMBRE_SEUILS ) {
	TH_Table_Remplir( EnergiesSeuils[ seuil ], Z, SEUIL_PremierElement( seuil ), Z_DERNIER_SEUIL, valeurs[ seuil ] );
      }
    }
  }

  /* Plus besoin du fichier */
  free( valeurs ); valeurs = NULL;
  (void) fclose( fichier );

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Utilisation des tables d'nergies de seuil ~-~-~-~-~-~-~-~-~ */

/* -------- Renvoie le premier atome pour lequel le seuil existe -------- */

unsigned int SEUIL_PremierElement(unsigned int seuil)
{
  switch ( seuil )
  {
  case SEUIL_K:    return HYDROGENE;

  case SEUIL_LI:   return LITHIUM;
  case SEUIL_LII:  return BORE;
  case SEUIL_LIII: return CARBONE;

  case SEUIL_MI:   return SODIUM;
  case SEUIL_MII:  return ALUMINIUM;
  case SEUIL_MIII: return PHOSPHORE;
  case SEUIL_MIV:  return SCANDIUM;
  case SEUIL_MV:   return CHROME;

  case SEUIL_NI:   return POTASSIUM;
  case SEUIL_NII:  return GALLIUM;
  case SEUIL_NIII: return ARSENIC;
  case SEUIL_NIV:  return YTTRIUM;
  case SEUIL_NV:   return MOLYBDENE;
  case SEUIL_NVI:  return LANTHANE;
  case SEUIL_NVII: return EUROPIUM;

  case SEUIL_OI:   return RUBIDIUM;
  case SEUIL_OII:  return INDIUM;
  case SEUIL_OIII: return ANTIMOINE;
  case SEUIL_OIV:  return LUTETIUM;
  case SEUIL_OV:   return TUNGSTENE;
  case SEUIL_OVI:  return ACTINIUM;
  case SEUIL_OVII: return AMERICIUM;

  case SEUIL_PI:   return CESIUM;
  case SEUIL_PII:  return THALLIUM;
  case SEUIL_PIII: return BISMUTH;

  case SEUIL_QI:   return FRANCIUM;

#ifdef AVEC_TRANSURANIENS_
  case SEUIL_PIV:  return LAWREMCIUM;
  case SEUIL_PV:   return SEABORGIUM;

  case SEUIL_QII:  return 113;
  case SEUIL_QIII: return 115;
#endif

  case ABSORPTION_TOTALE:
  case DIFFUSION_INCOHERENTE:
  case DIFFUSION_COHERENTE:
  case APRES_SEUIL_K:
    return HYDROGENE;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Seuil inconnu [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, seuil);
    return 2540; /* Pour etre sur d'etre au-dela de tout lment raisonnable */
  }

  return 0;
}

/* -------- Autour de l'nergie de seuil... ------ */

/* -------- Renvoie le nom du seuil -------- */

const char *TH_TexteSeuil(int seuil)
{
  if ( ( seuil >= SEUIL_K ) && ( seuil < NOMBRE_SEUILS ) ) {
    return NomsSeuils[ seuil ];
  } else if ( APRES_SEUIL_K == seuil ) {
    return "E > E(K)";
  } else if ( DIFFUSION_INCOHERENTE == seuil ) {
    return "diffusion Compton";
  } else if ( DIFFUSION_COHERENTE == seuil ) {
    return "diffusion Thompson";
  }

  /* On n'a pas trouv de seuil correct */
  (void) fprintf(stderr, "%s,%s l.%u : Seuil inconnu [%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, seuil);

  return "Seuil inconnu";
}

/* -------- Renvoie l'nergie du seuil indiqu pour l'lment Z
     -2.0 en cas d'erreur (lment incorrect, seuil inexistant,...)
     -1.0 en cas de donne inconnue ------ */

double TH_DonneesSeuil(int Z, int seuil, double *hauteur, double *largeur)
{
  int Z_min;
  double e_seuil = -1.0;

  if ( ( Z < HYDROGENE ) || ( Z > Z_DERNIER_SEUIL ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : lment inconnu [Z = %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, Z);
#endif
    ATM_ERREUR(ELEMENT_INCONNU, Z);
    return -2.0;
  }
  if ( ( seuil < SEUIL_K ) || ( seuil >= NOMBRE_SEUILS ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Seuil inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, seuil);
    return -2.0;
  }

  Z_min = SEUIL_PremierElement( seuil );
  if ( Z < Z_min ) return -1.0; /* Le seuil n'existe pas ! */

  if ( NULL == EnergiesSeuils[ seuil ] ) {
    int retour;

    retour = SEUIL_InitialiserTables();
    if ( retour != 0 ) {
      (void) fprintf(stderr, "ATTENTION, les nergies de seuil n'ont pas pu tre charges correctement\n");
      return -2.0;
    }
  }

  e_seuil = TH_Table_Lire( EnergiesSeuils[ seuil ], Z, Z_min, Z_DERNIER_SEUIL );

  if (e_seuil == -2.0) { /* Donne absente */
    (void) printf("%s : nergie du seuil %s inconnue pour l'lment Z = %s [%d]\n", __FUNCTION__, TH_TexteSeuil( seuil ), ATM_Z_vers_symbole( Z ), Z );
  }

  /* Veut-on aussi les largeurs et hauteurs de seuils ? */
  if ( NULL != largeur ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Table des largeurs de seuils\n",
		   __FILE__, __FUNCTION__, __LINE__);

    *largeur = -1.0; /* Pour l'instant, inconnu ! */
  }

  if ( NULL != hauteur ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Table des hauteurs de seuils\n",
		   __FILE__, __FUNCTION__, __LINE__);

    *hauteur = -1.0; /* Pour l'instant, inconnu ! */
  }

  return e_seuil;
}

/* ------ Z tant indiqu, renvoie l'nergie du _premier_ seuil se trouvant
     entre e_min et e_max, ainsi que le type de seuil
     Renvoie -1.0 si l'lment n'a pas de seuil (connu) dans l'intervalle considr
     dans ce cas, seuil indique le seuil aprs lequel on est

     remarque : avant un seuil = nergie plus faible, donc avant le seuil K <=> entre LI et K
                aprs un seuil = nergie plus haute, donc aprs le seuil K <=> plus de seuil aprs !
-------- */

double TH_QuelSeuil(int Z, double e_min, double e_max, int *seuil)
{
  float e_seuil;

  if ( (Z < HYDROGENE) || (Z > FERMIUM) ) {
    ATM_ERREUR(ELEMENT_INCONNU, Z);
    return -2.0;
  }
  if (e_max < e_min) { double t = e_max ; e_max = e_min ; e_min = t; }

  /* Position par rapport au seuil K */
  e_seuil = TH_DonneesSeuil( Z, SEUIL_K, NULL, NULL );
  if (seuil != NULL) *seuil = APRES_SEUIL_K;
  if (e_seuil < e_min) return -1.0; /* Seuil K de plus haute nergie possible pour un lment */
  if (seuil != NULL) *seuil = SEUIL_K; /* On contient ou est avant le seuil K */
  if (e_seuil < e_max) /* On contient donc le seuil K */
    return e_seuil;

  /* Position par rapport au seuil LI */
  e_seuil = TH_DonneesSeuil( Z, SEUIL_LI, NULL, NULL );
  if (e_seuil < e_min) return -1.0; /* On est compris entre le seuil K et le seuil LI */
  if (seuil != NULL) *seuil = SEUIL_LI; /* On contient ou est avant le seuil LI */
  if (e_seuil < e_max)
    return e_seuil;

  /* Position par rapport au seuil LII */
  e_seuil = TH_DonneesSeuil( Z, SEUIL_LII, NULL, NULL );
  if (e_seuil < e_min) return -1.0; /* On est compris entre le seuil LI et le seuil LII */
  if (seuil != NULL) *seuil = SEUIL_LII; /* On contient ou est avant le seuil LII */
  if (e_seuil < e_max)
    return e_seuil;

  /* Position par rapport au seuil LIII */
  e_seuil = TH_DonneesSeuil( Z, SEUIL_LIII, NULL, NULL );
  if (e_seuil < e_min) return -1.0; /* On est compris entre le seuil LI et le seuil LIII */
  if (seuil != NULL) *seuil = SEUIL_LIII; /* On contient ou est avant le seuil LIII */
  if (e_seuil < e_max)
    return e_seuil;

  return -1.0; /* On est au-del du dernier seuil tudi */
}

/* Z et E tant connus, renvoie la rgion dans laquelle on est */

int TH_QuelleRegion(int Z, double energie)
{
  double e_seuil;
  unsigned int i;

  /* Remarque : suppose les seuils par nergie croissante. Ce n'est pas rigoureusement vrai
     quand on est dans les seuils de valence, mais de toute faon les nergies n'ont alors pas grand sens ! */

  for ( i = SEUIL_K ; i < NOMBRE_SEUILS ; i++ ) {
    e_seuil = TH_DonneesSeuil( Z, i, NULL, NULL );
    if ( -1.0 == e_seuil ) { /* On peut avoir des seuils manquants, puisque le remplissage n'est pas dans l'ordre ! */
      unsigned int j;

      j = i;
      while ( -1.0 == e_seuil ) {
	j++;
	if ( j >= NOMBRE_SEUILS ) /* Il n'y a donc pas de seuil aprs */
	  return i; /* On est donc avant le seuil indiqu */
	e_seuil = TH_DonneesSeuil( Z, j, NULL, NULL );
      }
      i = j; /* Ce seuil compte */
    }

    if ( energie >= e_seuil ) return ( i - 1 ); /* On est entre le seuil (i-1) et le seuil (i) */
  }

  return (NOMBRE_SEUILS - 1);
}

/* -------- Rechercher quel seuil est le plus proche de l'nergie indique -------\
 |   Renvoie le numro atomique de l'lment possdant ce seuil                   |
 |           SYMBOLE_CHIMIQUE_INCONNU en cas d'erreur ou de problme              |
 |                                                                                |
 | energie    : l'nergie pour laquelle on fait la recherche (en eV)              |
 | marge      : marge autorise entre l'nergie donne et l'nergie du seuil      |
 | seuil      : le type de seuil trouv (K, LI,...)                               |
 | historique : VRAI, on mentionne tous les seuils acceptables dans l'historique  |
 \------------------------------------------------------------------------------- */

int TH_IdentifierSeuil(double energie, double marge, int *seuil, boolean historique)
{
  double distance = -1.0, meilleure_distance = -1.0;
  int type_seuil = SEUIL_K;
  int element = SYMBOLE_CHIMIQUE_INCONNU, Z;
  unsigned int i, nbr_trouves = 0;

  if (marge < 0.0) {
    (void) fprintf(stderr, "TH_IdentifierSeuil : marge ngative, recherche annule !\n");
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return SYMBOLE_CHIMIQUE_INCONNU;
  }

  /* On vrifie que les nergies de seuil sont connues */
  if ( NULL == EnergiesSeuils[ SEUIL_K ] ) {
    int retour;

    retour = SEUIL_InitialiserTables();
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s l.%u : Impossible de charger les nergies de seuil\n\n", __FUNCTION__, __LINE__);
      return SYMBOLE_CHIMIQUE_INCONNU;
    }
  }

  /* On commence par chercher le seuil le plus proche */
  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ ) {
    Z = SEUIL_PlusProche( i, energie, &distance );
    if ( SYMBOLE_CHIMIQUE_INCONNU == Z ) continue; /* Problme... */

    if ( distance <= marge ) { /* On est dans la rgion acceptable */
      nbr_trouves++;

      if ( VRAI == historique ) {
	HIST_Texte( 3, 2, "Seuil acceptable : %s [%d], seuil %s, E=%g eV [d=%g eV]\n",
		    ATM_Z_vers_symbole( Z ), Z, TH_TexteSeuil( i ), TH_DonneesSeuil( Z, i, NULL, NULL ), distance );
      }
#ifdef _DEVELOPPE_
      printf("Seuil acceptable : %s, seuil %s, E=%g eV [d=%g eV]\n", ATM_Z_vers_symbole( Z ), TH_TexteSeuil( i ), TH_DonneesSeuil( Z, i, NULL, NULL ), distance );
#endif

      if ( SYMBOLE_CHIMIQUE_INCONNU == element ) { /* On n'a rien trouv d'autre jusqu' prsent */
	element = Z;
	meilleure_distance = distance;
	type_seuil = i;
      } else if ( meilleure_distance > distance ) { /* On est plus proche que l'essai prcdent */
	element = Z;
	meilleure_distance = distance;
	type_seuil = i;
      }
    }
  }

  /* A-t-on trouv quelque chose ? */
  if ( nbr_trouves == 0 ) {
#ifdef _DEVELOPPE_
    (void) printf("TH_IdentifierSeuil : aucun seuil trouv - E = %g, plus proche : Z = %d / seuil %d, E = %g soit une diffrence de %g\n",
		  energie, element, type_seuil, energie + distance, distance);
#endif
    return SYMBOLE_CHIMIQUE_INCONNU;
  }

  *seuil = type_seuil;
  return element;
}

static int SEUIL_PlusProche(unsigned int seuil, double energie, double *distance)
{
  int Z_min, Z_max = Z_DERNIER_SEUIL;
  int Z_avant, Z_apres, Z_milieu;
  unsigned int nbr_Z;
  float *table;

  /* Quelle table utiliser ? */
  Z_min = SEUIL_PremierElement( seuil );
  if ( ( Z_max < Z_min ) || ( Z_min < HYDROGENE ) ) {
    return SYMBOLE_CHIMIQUE_INCONNU;
  }
  nbr_Z = Z_max - Z_min + 1; /* Nombre de valeurs dans la table */

  table = EnergiesSeuils[ seuil ];
  if ( NULL == table ) return SYMBOLE_CHIMIQUE_INCONNU;

  /* Les valeurs les plus hautes peuvent tre inconnues -> on les passe ! */
  while ( table[ nbr_Z - 1 ] <= 0.0 ) {
    nbr_Z--;
    if ( 0 == nbr_Z ) return SYMBOLE_CHIMIQUE_INCONNU;
  }

  /* On peut supposer les seuils classs par nergie croissante
     1. On vrifie que l'on n'est pas en-dehors */
  if ( energie <= table[ 0 ] ) {
    *distance = table[ 0 ] - energie;
    return Z_min;
  } else if ( energie >= table[ nbr_Z - 1 ] ) {
    *distance = energie - table[ nbr_Z - 1 ];
    return Z_max;
  }

  /* 2. On est dedans -> on cherche donc par dichotomie */
  Z_avant = 0;
  Z_apres = nbr_Z;
  do {
    Z_milieu = ( Z_avant + Z_apres ) / 2;
    if ( energie == table[ Z_milieu ] ) { /* On est pile dessus */
      *distance = 0.0;
      return Z_min + Z_milieu;
    } else if ( table[ Z_milieu ] < energie ) {
      Z_avant = Z_milieu;
    } else {
      Z_apres = Z_milieu;
    }
  } while ( Z_avant + 1 < Z_apres );

  /* On estime la distance entre le plus proche et l'nergie indique */
  if ( ( energie - table[ Z_avant ] ) < ( table[ Z_apres ] - energie ) ) {
    *distance = energie - table[ Z_avant ];
    return Z_min + Z_avant;
  } else {
    *distance = table[ Z_apres ] - energie;
    return Z_min + Z_apres;
  }

  /* On ne devrait jamais tre l */
  (void) fprintf(stderr, "%s,%s l.%u : On ne devrait jamais arriver l\n", __FILE__, __FUNCTION__, __LINE__);
  return SYMBOLE_CHIMIQUE_INCONNU;
}

/* -~-~-~-~-~-~-~-~ Modlisation de l'absorption due au seuil seul -~-~-~-~-~-~-~-~ */

/* -------- Routine gnraliste, visible de l'extrieur --------
   Le seuil simul est AJOUT au graphe indiqu, dont les X sont supposs des nergies en eV
                                                      les Y sont supposs tre des mu/rho
   position, largeur : en eV
   hauteur           : compatible avec mu/rho en cm/g
 */

int TH_SimulerSeuil(Graphe *graphe, double position, double hauteur, double largeur, int methode)
{
  int retour;

  if ( NULL == graphe ) return -1;

  switch ( methode )
  {
  case SEUIL_MODELE_ESCALIER:
    retour = TH_SeuilEscalier( graphe, position, hauteur );
    break;

  case SEUIL_MODELE_LORENTZ:
    retour = TH_SeuilLorentzien( graphe, position, hauteur, largeur );
    break;

  case SEUIL_MODELE_GAUSS:
    retour = TH_SeuilGaussien( graphe, position, hauteur, largeur );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Modle de seuil inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode);
    return -5;
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul du seuil [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* Tout s'est bien pass */
  GRF_bornes( graphe, NE_PAS_FAIRE, DESORDONNE );
  return 0;
}

/* -------- Modle idal : marche d'escalier -------- */

static int TH_SeuilEscalier(Graphe *graphe, double position, double hauteur)
{
  unsigned long i;
  double *x, *y;

  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( *x >= position ) {
      *y += hauteur;
    }
  }

  return 0;
}

static int TH_SeuilGaussien(Graphe *graphe, double position, double hauteur, double largeur)
{
  unsigned long i;
  double *x, *y;

  if ( 0.0 == largeur ) {
    (void) fprintf(stderr, "%s,%s l.%u : largeur nulle => modle en escalier !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return TH_SeuilEscalier( graphe, position, hauteur );
  }

  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( *x >= position ) {
    }
  }

  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [E0=%g, L=%g, h=%g\n",
		 __FILE__, __FUNCTION__, __LINE__, position, hauteur, largeur);

  return -1;
}

static int TH_SeuilLorentzien(Graphe *graphe, double position, double hauteur, double largeur)
{
  unsigned long i;
  double *x, *y;
  double mu_seuil;

  if ( 0.0 == largeur ) {
    (void) fprintf(stderr, "%s,%s l.%u : largeur nulle => modle en escalier !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return TH_SeuilEscalier( graphe, position, hauteur );
  }

  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    mu_seuil  = ( *x - position ) / largeur;
    mu_seuil  = atan( mu_seuil ) / M_PI;
    mu_seuil += 0.5;
    mu_seuil *= hauteur;

    *y += mu_seuil;
  }

  return 0;
}

/* -~-~-~-~-~-~-~-~ Modlisation de l'absorption  hors seuil  (seuil discontinu) -~-~-~-~-~-~-~-~ */

/* -------- Interface : calcul pour une nergie donne --------\
 | Renvoie mu si tout s'est bien pass, <0 sinon               |
 |                                                             |
 | Z       : l'lment pour lequel on veut faire le calcul     |
 | energie : l'nergie  laquelle on veut calculer [en eV]     |
 | methode : modle de mu  utiliser [cf. theorie.h]           |
 | seuil   : si >=0, que la contribution du seuil indiqu,     |
 |           sinon, absorption totale                          |
 \------------------------------------------------------------ */

double TH_CalculerMu(int Z, double energie, int methode, int seuil)
{
  double mu;

  /* Vrification prliminaire */
  if ( ( energie <= 0.0 ) || ( Z < HYDROGENE ) || ( Z > PLUTONIUM ) ) {
    (void) fprintf(stderr, "TH_CalculerMu : paramtre absurde [Z=%d, E=%g]\n", Z, energie);
    return -1.0;
  }

  if ( seuil >= SEUIL_K ) {
    double e_seuil;

    e_seuil = (double) TH_DonneesSeuil( Z, seuil, NULL, NULL );
    if ( e_seuil > energie ) {
      /* On est avant le seuil qui nous intresse => pas d'absorption
	 [Rq. pas rigoureusement vrai si on est juste avant...]*/
      return 0.0;
    }

    /* On est aprs le seuil */
    switch ( methode )
    {
    case MUM_SELON_VICTOREEN:
      {
	double C_av, D_av, C_ap, D_ap;

	TH_coef_Victoreen( Z, seuil, &C_av, &D_av ); /* Modle avant le seuil considr */
	TH_coef_Victoreen( Z, seuil - 1, &C_ap, &D_ap ); /* Modle aprs le seuil considr */
	mu = TH_Victoreen( energie, Z, C_ap, D_ap ) - TH_Victoreen( energie, Z, C_av, D_av );
      }
      break;

    case MUM_SELON_MCMASTER:
      {
	double C_av[4], C_ap[4];

	TH_coef_McMaster( Z, seuil, C_av, C_av + 1, C_av + 2, C_av + 3);
	TH_coef_McMaster( Z, seuil - 1, C_ap, C_ap + 1, C_ap + 2, C_ap + 3);
	mu = TH_McMaster( energie, Z, C_ap[ 0 ], C_ap[ 1 ], C_ap[ 2 ], C_ap[ 3 ] ) - TH_McMaster( energie, Z, C_av[ 0 ], C_av[ 1 ], C_av[ 2 ], C_av[ 3 ] );
      }
      break;

    case MUM_SELON_EBEL:
      {
	double c[ 6 ];

	TH_coef_Ebel( Z, seuil, FAUX, c, NULL );
	mu = TH_Ebel( energie, Z, c );
      }
      break;

    default:
      (void) fprintf(stderr, "TH_CalculerMu : Mthode inconnue [%d]\n", methode);
      return 0.0;
    }

#ifdef _DEVELOPPE_
    if ( ( VRAI == isinf( mu ) ) || ( VRAI == isnan( mu ) ) || ( mu < 0.0 ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Section efficace absurde [%g] : E=%g, Z=%d (%s), seuil = %d (%s)\n",
		     __FILE__, __FUNCTION__, __LINE__, mu, energie, Z, ATM_Z_vers_symbole( Z ), seuil, TH_TexteSeuil( seuil ));
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      return -20;
    }
#endif
    return mu;
  }

  /* Si on est l : on veut tous les seuils -> plus simple ! */

  /* On commence par trouver dans quelle rgion on est */
  seuil = TH_QuelleRegion( Z, energie );

  /* Puis on calcule */
  switch ( methode )
  {
  case MUM_SELON_VICTOREEN:
    {
      double C, D;

      TH_coef_Victoreen( Z, seuil, &C, &D );
      mu = TH_coef_Victoreen( Z, seuil, &C, &D );
    }
    break;

  case MUM_SELON_MCMASTER:
    {
      double C[4];

      TH_coef_McMaster( Z, seuil, C, C + 1, C + 2, C + 3);
      mu = TH_McMaster( energie, Z, C[ 0 ], C[ 1 ], C[ 2 ], C[ 3 ] );
    }
    break;

  default:
    (void) fprintf(stderr, "TH_CalculerMu : Mthode inconnue [%d]\n", methode);
    return 0.0;
  }

  /* Calcul termin */
  return mu;
}

/* -------- Calcul selon le modle de Victoreen, avec (au moins) un seuil au milieu ------ */

int victorine_seuil(Graphe *graphe, int Z, float proportion, int seuil, double e_seuil)
{
  unsigned long i;
  register double *x, *y;
  double C = 0.0, D = 0.0;

  /* On commence par ce qui est avant seuil */
  if (e_seuil > 0.0) /* On contient le seuil indiqu dans seuil */
    seuil--; /* Donc on dmarre avant ce seuil */

  TH_coef_Victoreen(Z, seuil, &C, &D);

#ifdef _DEVELOPPE_
  printf("Valeurs de C et D : %f & %f\n", C, D);
#endif
  if ( (C < 0.0) || (D < 0.0) ) /* Valeurs inconnues ou seuil absurde */
    return -1;

  /* Travail en nergies dcroissantes pour avoir les seuils dans l'ordre... */
  for (i = 0, x = graphe->x + graphe->nombre_point - 1, y = graphe->y + graphe->nombre_point - 1 ;
       i < graphe->nombre_point ; i++, x--, y--) {
    double nrj = *x;
    double mu;

    if ( (e_seuil > 0.0) && (nrj < e_seuil) ) { /* On change de seuil */
      double mum_avant, mum_apres;

      nrj = e_seuil;
      printf("Pour l'lment Z=%d, on change de seuil  E=%f\n", Z, nrj);

      /* Absorption sans l'lment */
      mum_avant = TH_Victoreen(e_seuil, Z, C, D);

      seuil++; /* On passe au seuil suivant */
      e_seuil = (double) TH_DonneesSeuil(Z, seuil + 1, NULL, NULL);
      TH_coef_Victoreen(Z, seuil, &C, &D);
      printf(" --- Nouvelle nergie de seuil critique : %f\n", e_seuil);
      printf("Nouvelles valeurs de C et D : %f & %f\n", C, D);
      if ( (C < 0.0) || (D < 0.0) ) /* Valeurs inconnues ou seuil absurde */
	return -1;

      /* Absorption avec l'lment */
      mum_apres = TH_Victoreen(nrj, Z, C, D);

      /* Hauteur de seuil attendue */
      printf(" ==> hauteur de seuil attendue : %f\n", (mum_apres - mum_avant) * proportion);
    }

    /* Calcul de la valeur correspondante */
    mu = TH_Victoreen( *x, Z, C, D);
    mu *= proportion;
    *y += mu; /* On _ajoute_ au graphe existant */
  }

  return 0; /* Tout s'est bien pass */
}

int McMaster_seuil(Graphe *graphe, int Z, float proportion, int seuil, double e_seuil)
{
  unsigned long i;
  register double *x, *y;
  double A, B, C, D; /* Coefficients de McMaster */
  double Ae, Be, Ce, De;
  double Ai, Bi, Ci, Di;

  /* On commence par ce qui est avant seuil */
  if (e_seuil > 0.0) /* On contient le seuil indiqu dans seuil */
    seuil--; /* Donc on dmarre avant ce seuil */
  if ( TH_coef_McMaster(Z, seuil, &A, &B, &C, &D) == FAUX )
    return -1;
  TH_coef_McMaster(Z, DIFFUSION_COHERENTE  , &Ae, &Be, &Ce, &De);
  TH_coef_McMaster(Z, DIFFUSION_INCOHERENTE, &Ai, &Bi, &Ci, &Di);

  /* Travail en nergies dcroissantes pour avoir les seuils dans l'ordre... */
  for (i = 0, x = graphe->x + graphe->nombre_point - 1, y = graphe->y + graphe->nombre_point - 1 ;
       i < graphe->nombre_point ; i++, x--, y--) {
    double nrj = *x;

    if ( (e_seuil > 0.0) && (nrj < e_seuil) ) { /* On change de seuil */
      printf("Pour l'lment Z=%d, on change de seuil  E=%f\n", Z, nrj);

      seuil++; /* On passe au seuil suivant */
      e_seuil = (double) TH_DonneesSeuil(Z, seuil + 1, NULL, NULL);
      printf(" --- Nouvelle nergie de seuil critique : %f\n", e_seuil);

      if ( TH_coef_McMaster(Z, seuil, &A, &B, &C, &D) == FAUX )
	return -1;
    }

    /* Calcul de la valeur correspondante */
    if (*x <= 0.0 ) {
      EXF_ERREUR(DIV_ZERO, E_AIGU"nergie nulle", txt_err_val_nulle);
      *y = 0.0;
#ifdef _DEVELOPPE_
      *(char *)NULL=0;
#endif
    } else {
      double coef_McM;

      coef_McM = TH_McMaster(*x, Z, A, B, C, D); /*  Coefficient d'absorption  massique */
/*      coef_McM += TH_McMaster(*x, Z, Ae, Be, Ce, De);*/ /* Contribution de la diffusion lastique */
/*      coef_McM += TH_McMaster(*x, Z, Ai, Bi, Ci, Di);*/ /* Contribution de la diffusion inlastique */
      coef_McM *= proportion;
      *y += coef_McM;
    }
  }

  return 0; /* Tout s'est bien pass */
}
