/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Modlisation des dtecteurs de rayonnement

   Remarque : fonctions finies par _I = dtecteurs non rsolus en nergie
                                   _R = dtecteurs rsolus en nergie
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "detecteur.h"
#include "echantillon.h" /* Pour la composition */
#include "faisceau.h"
#include "graphes.h"
#include "images.h"
#include "lissage.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double DETECTEUR_Monochromatique_I(Detecteur *detecteur, Faisceau *faisceau, double energie, Faisceau *f_sortie);
static double DETECTEUR_Polychromatique_I(Detecteur *detecteur, Faisceau *faisceau, double energie, Faisceau *f_sortie);

/* -------- Briques individuelles ---- */

static double DET_Mono_Ponctuel(Detecteur *detecteur, double energie, double intensite, double *attenuation);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* Gestion des structures dcrivant les dtecteurs */

Detecteur *DET_CreerDetecteur(int type, int proprietes)
{
  Detecteur *detecteur;

  /* On rserve la mmoire pour le dtecteur */
  detecteur = (Detecteur *) calloc( 1, sizeof( Detecteur ) );
  if ( NULL == detecteur ) {
    (void) fprintf(stderr, "DET_CreerDetecteur : mmoire insuffisante\n");
    return NULL;
  }

  /* Position, caractrisation,... */
  detecteur->orientation.position.x = 0.0; /* Plac  l'origine par dfaut */ 
  detecteur->orientation.position.y = 0.0;
  detecteur->orientation.position.z = 0.0;
  detecteur->orientation.theta = 0.0; /* avec des axes compatibles avec le repre absolu */ 
  detecteur->orientation.phi = 0.0;
  detecteur->orientation.alpha = 0.0;

  /* Les proprits physiques */
  detecteur->type = type;
  detecteur->proprietes = proprietes;

  return detecteur;
}

/* -------- Donner un nom au dtecteur -------- */

int DET_NommerDetecteur(Detecteur *detecteur, const char *nom)
{
  if ( NULL == detecteur ) return -1;
  if ( NULL == nom ) return -2;

  free( detecteur->nom );

  detecteur->nom = strdup( nom );
  if ( NULL == detecteur->nom ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    return +1;
  }

  return 0;
}

/* -------- Effacement -------- */

int DET_EffacerDetecteur(Detecteur *detecteur, boolean forcer)
{
  if ( NULL == detecteur ) return -1;

  if ( FAUX == forcer ) { /* Poser la question */
    (void) fprintf(stderr, "%s,%s l.%u : Poser la question\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  /* On le retire du chanage */
  DET_DechainerDetecteur( detecteur );

  /* Effacer les informations associes */
  if ( detecteur->r_signal != NULL ) {
    if ( VRAI == detecteur->r_signal->drapeaux.isole ) {
      GRF_eff_isole( detecteur->r_signal );
    }
  }
  if ( detecteur->r_energie != NULL ) {
    if ( VRAI == detecteur->r_energie->drapeaux.isole ) {
      GRF_eff_isole( detecteur->r_energie );
    }
  }

  if ( detecteur->r_surface != NULL ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Effacer l'image\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  free( detecteur->nom );

  /* Plus besoin de la structure */
  free( detecteur );

  return 0;
}

int DET_ControlerOptions(Detecteur *detecteur)
{
  if ( NULL == detecteur ) return -1;

  /* Spectre d'absorption */
  if ( detecteur->proprietes & DETECTEUR_ABSORBANT ) {
    int retour;

    if ( 0.0 >= detecteur->longueur ) {
      (void) fprintf(stderr, "DET_ControlerOptions : Dtecteur absorbant, longueur absurde [%g]\n", detecteur->longueur);
      return -1;
    }
    retour = Matiere_ControlerComposition( detecteur->composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "DET_ControlerOptions : Dtecteur absorbant, composition incorrecte [code %d]\n", retour);
      return -1;
    }
  }

  /* Non-linarit de la rponse */
  if ( detecteur->proprietes & DETECTEUR_NON_LINEAIRE ) {
    if ( NULL == detecteur->r_signal ) {
      (void) fprintf(stderr, "DET_ControlerOptions : Dtecteur non-linaire , pas de courbe de rponse dfinie\n");
      return -2;
    }
  }

  /* Rponse variable sur la surface */
  if ( detecteur->proprietes & DETECTEUR_INHOMOGENE ) {
    if ( NULL == detecteur->r_surface ) {
      (void) fprintf(stderr, "DET_ControlerOptions : Dtecteur inhomogne, pas de surface de rponse dfinie\n");
      return -3;
    }
  }

  /* Rponse variable avec l'nergie */
  if ( detecteur->proprietes & DETECTEUR_NRJ_VARIABLE ) {
    if ( NULL == detecteur->r_energie ) {
      (void) fprintf(stderr, "DET_ControlerOptions : Dtecteur d'efficacit variable, pas de courbe d'efficacit dfinie\n");
      return -4;
    }
  }

  /* Aucun problme dtect */
  return 0;
}

/* Calculs sur les dtecteurs */

/* -------- Calculer le spectre vu par un dtecteur --------
   ATTENTION, le spectre cr est un spectre ISOL, sans X rels [type GRF_SERIE]
    charge de l'appelant de le modifier s'il faut le conserver !
*/

Graphe *DETECTEUR_CalculerSpectre(Detecteur *detecteur, Faisceau *faisceau)
{
  Graphe *spectre = NULL;

  if ( ( NULL == detecteur ) || ( NULL == faisceau ) ) return NULL;
  if ( FAUX == ( detecteur->proprietes & DETECTEUR_RESOLU ) ) {
    (void) fprintf(stderr, "DETECTEUR_CalculerSpectre : ATTENTION, dtecteur non-marqu comme rsolu en nergie !\n");
  }
  if ( FAUX == ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) ) {
    (void) fprintf(stderr, "DETECTEUR_CalculerSpectre : Faisceau monochromatique !\n");
    return NULL;
  }

  /* L'absorption par le dtecteur ne gene pas. Seuls effets  prendre en compte :
     - rponse avec la surface
     - largissement des raies  cause de la rsolution
     - rponse variable avec l'nergie
  */
  /* 1. Spectre global vu = effets de surface */
  if ( detecteur->proprietes & DETECTEUR_INHOMOGENE ) {
    (void) fprintf(stderr, "DETECTEUR_CalculerSpectre :  FAIRE --- dtecteur inhomogne spatialement\n");
  } else {
    spectre = GRF_Copier( faisceau->g_E, NULL, VRAI );
    if ( NULL == spectre ) {
      (void) fprintf(stderr, "DETECTEUR_CalculerSpectre : ERREUR dans GRF_Copier()\n");
      return NULL;
    }

    GRF_FacteurMultiplicatif( spectre, faisceau->intensite, AXE_Y, NULL );
  }

  /* 2. Rponse avec l'nergie */
  if ( detecteur->proprietes & DETECTEUR_NRJ_VARIABLE ) {
    (void) fprintf(stderr, "DETECTEUR_CalculerSpectre :  FAIRE --- dtecteur  rponse variable avec l'nergie\n");
  }

  /* 3. largissement d  la rsolution */

  /* 4. Pertes de linarit */
  if ( detecteur->proprietes & DETECTEUR_NON_LINEAIRE ) {
    (void) fprintf(stderr, "DETECTEUR_CalculerSpectre :  FAIRE --- dtecteur non-linaire\n");
  }

  /* 5. Codage en entiers */
  if ( detecteur->proprietes & DETECTEUR_ENTIER ) {
    unsigned long i;
    double *y;

    for ( i = 0, y = spectre->y ; i < spectre->nombre_point ; i++, y++ ) {
      *y = rint( *y );
    }
    spectre->y_min = rint( spectre->y_min );
    spectre->y_max = rint( spectre->y_max );
  }

  /* Calcul du spectre termin */
  return spectre;
}

/* --------  quoi ressemble le faisceau aprs avoir travers le dtecteur ? --------
   Renvoie le signal mesur par le dtecteur
           si f_sortie != NULL, l'allure du faisceau aprs avoir travers le dtecteur

    energie : l'nergie du faisceau incident [nergie nominale si polychromatique]
 */

double DETECTEUR_FaisceauTraverse_I(Detecteur *detecteur, Faisceau *faisceau, double energie, Faisceau *f_sortie)
{
  double signal = 0.0;

  ENTREE_FONCTION("DETECTEUR_FaisceauTraverse_I", "detecteur=%p faisceau=%p energie=%g f_sortie=%p", detecteur, faisceau, energie, f_sortie);

  /* Contrles pralables */
  if ( NULL == detecteur ) {
    SORTIE_FONCTION("DETECTEUR_FaisceauTraverse_I", "%g [ERREUR : pas de dtecteur]", 0.0);
    return 0.0;
  }

  if ( detecteur->proprietes & DETECTEUR_A_VIDE ) {
    signal = detecteur->coups_a_vide;
  }

  if ( NULL == faisceau ) {
    (void) fprintf(stderr, "DETECTEUR_FaisceauTraverse_I : Pas de faisceau [signal  vide ?]\n");
    SORTIE_FONCTION("DETECTEUR_FaisceauTraverse_I", "%g [ERREUR : pas de faisceau]", signal);
    return signal;
  }

#ifdef _DEVELOPPE_
  if ( f_sortie == faisceau ) {
    (void) fprintf(stderr, "DETECTEUR_FaisceauTraverse_I : faisceau = f_sortie, rsultat imprvisible\n");
  }
#endif
  if ( 0.0 == faisceau->intensite ) { /* Pas de faisceau -> pas de signal, sauf signal  vide ventuel ! */
    SORTIE_FONCTION("DETECTEUR_FaisceauTraverse_I", "%g [faisceau d'intensit nulle]", signal);
    return signal;
  }

  if ( detecteur->proprietes & DETECTEUR_RESOLU ) {
    (void) fprintf(stderr, "DETECTEUR_FaisceauTraverse_I : Dtecteur rsolu utilis en intgration!\n");
  }

  /* Influence du fondamental */
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    signal += DETECTEUR_Polychromatique_I( detecteur, faisceau, energie, f_sortie );
  } else {
    signal += DETECTEUR_Monochromatique_I( detecteur, faisceau, energie, f_sortie );
#ifdef _DEVELOPPE_
    if ( is_inf( signal ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : INTENSIT INFINIE [E=%g eV]\n",
		     __FILE__, __FUNCTION__, __LINE__, energie);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }
#endif
  }

  /* Influence des harmoniques */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    unsigned int i_h;
    Faisceau pseudo_faisceau, ps_sortie;

    pseudo_faisceau = *faisceau;
    pseudo_faisceau.proprietes &= ~FAISCEAU_HARMONIQUES;

    /* Pour chaque harmonique... */
    for ( i_h = 0 ; i_h < pseudo_faisceau.nbr_harmoniques ; i_h++ ) {
      pseudo_faisceau.intensite = faisceau->intensite * faisceau->i_harmoniques[ i_h ];
      if ( 0.0 == pseudo_faisceau.intensite ) /* Harmonique teinte */
	continue;

      if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
	pseudo_faisceau.g_E = faisceau->g_harmoniques[ i_h ];
	ps_sortie.g_E = GRF_Copier( faisceau->g_harmoniques[ i_h ], NULL, VRAI );
	if ( NULL == ps_sortie.g_E ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Copier, harmonique %u\n",
			 __FILE__, __FUNCTION__, __LINE__, i_h);
	  continue;
	}

	signal += DETECTEUR_Polychromatique_I( detecteur, &pseudo_faisceau, energie * ( i_h + 2 ), &ps_sortie );

	if ( NULL != f_sortie ) {
	  f_sortie->g_harmoniques[ i_h ] = ps_sortie.g_E;
	}
      } else {
	signal += DETECTEUR_Monochromatique_I( detecteur, &pseudo_faisceau, energie * ( i_h + 2 ), &ps_sortie );
      }

      if ( f_sortie->intensite > 0.0 )
	f_sortie->i_harmoniques[ i_h ] = ps_sortie.intensite / f_sortie->intensite;
      else
	f_sortie->i_harmoniques[ i_h ] = ps_sortie.intensite;
    }
  }

  /* Le dtecteur est-il linaire ? */
  if ( detecteur->proprietes & DETECTEUR_NON_LINEAIRE ) {
#ifdef _DEVELOPPE_
/*     (void) fprintf(stderr, "%s,%s l.%u : DETECTEUR_NON_LINEAIRE\n", */
/* 		   __FILE__, __FUNCTION__, __LINE__); */
#endif
    signal = GRF_LireValeur( detecteur->r_signal, signal, INTERPOLE_LINEAIRE );
  }

  /* Le dtecteur est-il  codage entier ? */
  if ( detecteur->proprietes & DETECTEUR_ENTIER ) {
#ifdef _DEVELOPPE_
/*     (void) fprintf(stderr, "%s,%s l.%u : DETECTEUR_ENTIER\n", */
/* 		   __FILE__, __FUNCTION__, __LINE__); */
#endif
    signal = rint( signal );
  }

  /* On a fini, on renvoie le signal */
  SORTIE_FONCTION("DETECTEUR_FaisceauTraverse_I", "%g", signal);
  return signal;
}

/* -------- Transformation pour un faisceau monochromatique -------- */ 

static double DETECTEUR_Monochromatique_I(Detecteur *detecteur, Faisceau *faisceau, double energie, Faisceau *f_sortie)
{
  double signal = 0.0;
  double r_surface = 1.0; /* Inhomognit de surface du dtecteur */
  double f_restante; /* Fraction restante aprs traverse du dtecteur */
  double i_originale;

  ENTREE_FONCTION("DETECTEUR_Monochromatique_I", "detecteur=%p faisceau=%p energie=%g f_sortie=%p", detecteur, faisceau, energie, f_sortie);

  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) { /* Faisceau ponctuel */
    DANS_FONCTION("Faisceau ponctuel [i=%g]", faisceau->intensite);

    /* Faut-il prendre en compte l'inhomognit de surface du dtecteur ? */
    if ( detecteur->proprietes & DETECTEUR_INHOMOGENE ) {
      printf("DETECTEUR_Monochromatique_I : Faisceau ponctuel, dtecteur inhomogne ---  FAIRE, effet du point d'impact\n");
      /* 1. Trouver  quel endroit du dtecteur passe le faisceau */
      /* 2. En dduire la rponse du dtecteur  cet endroit */
    }

    i_originale = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie );
    signal = r_surface * DET_Mono_Ponctuel( detecteur, energie, i_originale, &f_restante );
    if ( f_sortie != NULL ) {
      f_sortie->intensite = i_originale * f_restante;
    }
  } else {
    if ( detecteur->proprietes & DETECTEUR_INHOMOGENE ) {
      printf(" FAIRE --- DETECTEUR_Monochromatique_I, faisceau non-ponctuel, dtecteur inhomogne\n");
    } else {
      if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
	unsigned int i, j;
	double x, y;
	double correction;

	DANS_FONCTION("Faisceau inhomogne [i=%g, g_espace=%p]", faisceau->intensite, faisceau->g_espace);

	for ( i = 0 ; i < faisceau->g_espace->larg ; i++) {
	  x = faisceau->g_espace->x_min + i * faisceau->g_espace->l_reel;

	  for ( j = 0 ; j < faisceau->g_espace->haut ; j++) {
	    y = faisceau->g_espace->y_min + j * faisceau->g_espace->h_reel;

	    /* Le fondamental */
	    i_originale = FAISCEAU_LireIntensite( faisceau, x, y, energie );
	    if ( 0.0 == i_originale ) {
#ifdef _DEVELOPPE_
	      /* Peut arriver pour des faisceaux projets avec un angle particulier,
		  cause des incohrences de pas */
/* 	      (void) fprintf(stderr, "%s,%s l.%u : Intensit de faisceau nulle x=%g y=%g E=%g\n", */
/* 			     __FILE__, __FUNCTION__, __LINE__, x, y, energie); */
/* 	      FIN_ANTICIPEE( FIN_ERREUR_MINEURE ); */
#endif
	      continue;
	    }
#ifdef _DEVELOPPE_
	    if ( is_inf( i_originale ) ) {
	      (void) fprintf(stderr, "%s,%s l.%u : INTENSIT INFINIE [E=%g eV]\n",
			     __FILE__, __FUNCTION__, __LINE__, energie);
	      abort();
	    }
#endif

	    signal += DET_Mono_Ponctuel( detecteur, energie, i_originale, &f_restante );
#ifdef _DEVELOPPE_
	    if ( is_inf( signal ) ) {
	      (void) fprintf(stderr, "%s,%s l.%u : INTENSIT INFINIE [E=%g eV ; i=%u,j=%u]\n",
			     __FILE__, __FUNCTION__, __LINE__, energie, i, j);
	      abort();
	    }
#endif
	    if ( ( f_sortie != NULL ) && ( f_sortie->g_espace != NULL ) ) {
	      /* On peut avoit f_sortie mais pas de coupe, si la coupe en sortie ne nous intresse pas mais l'intensit si
		 Exemple : traverse de I1 avec un faisceau polychromatique... */
	      unsigned long n_case;
#ifdef _DEVELOPPE_
	      printf("%s : I(%u,%u) = %g, reste %g%% => I = %g / %u%u\n", __FUNCTION__, i, j, i_originale, f_restante * 100.0, i_originale * f_restante,
		     f_sortie->g_espace->larg, f_sortie->g_espace->haut );
#endif

	      IMG_IJ_v_N( f_sortie->g_espace, i, j, &n_case );
	      f_sortie->g_espace->valeurs.R[ n_case ] = i_originale * f_restante;
	    }
	  }
	}

	/* Terme correctif = de renormalisation pour pouvoir comparer au cas d'un faisceau homogne
	   Si le faisceau est homogne, I = i(0,0) * L * h
	   S'il est dit inhomogne, mais homogne, I = i(0,0) * N_L * N_h

	   => terme correctif : L * h / N_L * N_h [revient au prfacteur de l'intgrale dans les trapzes !]
	*/
	correction = (double) faisceau->g_espace->larg * (double) faisceau->g_espace->haut;
	if ( 0.0 == correction ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Division par zro\n",
			 __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
	  abort();
#endif
	  return -5.0;
	}
	correction = faisceau->largeur * faisceau->hauteur / correction;

	signal *= correction;
	if ( f_sortie != NULL ) {
	  f_sortie->intensite = 1.0; /* Les vraies valeurs sont dans la carte ou la table des harmoniques */
	}
      } else {
	/* Faisceau homogne : on peut faire comme s'il s'agissait d'un ponctuel et multiplier par la surface ! */
	i_originale = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie );
	signal = DET_Mono_Ponctuel( detecteur, energie, i_originale, &f_restante );
	if ( f_sortie != NULL ) {
	  f_sortie->intensite = i_originale * f_restante;
	}

	/* Effet de surface */
	signal *= faisceau->largeur * faisceau->hauteur;
      }
    }
  }

  /* On a fini, on renvoie le signal obtenu */
  SORTIE_FONCTION("DETECTEUR_Monochromatique_I", "%g", signal);
  return signal;
}

/* ---- Faisceau monochromatique ponctuel : signal  une nergie donne ---- */

static double DET_Mono_Ponctuel(Detecteur *detecteur, double energie, double intensite, double *attenuation)
{
  double signal;

  ENTREE_FONCTION("DET_Mono_Ponctuel", "detecteur = %p energie=%g intensite=%g attenuation=%p", detecteur, energie, intensite, attenuation);

  /* Si le dtecteur est idal, il compte tout */
  signal = intensite;

  /* Si on connat l'absorbance : le signal est proportionnel  la fraction absorbe */
  if ( detecteur->proprietes & DETECTEUR_ABSORBANT ) {
    double f_absorbee, mu_detecteur;

 /* Partie absorbe donnant du signal : photolectrique seul
    La longueur du dtecteur est code en cm, mu contient (mu/rho) en cm^-1 */
    mu_detecteur  = GRF_LireValeur( detecteur->composition->mu_total, energie, INTERPOLE_LINEAIRE );
    mu_detecteur *= detecteur->composition->densite; /* On a ici mu linique, en cm^-1 */
    f_absorbee = 1.0 - exp( -mu_detecteur * detecteur->longueur );
    signal *= f_absorbee;
#ifdef _DEVELOPPE_
/*     printf("%s : E=%g mu=%g : exp = %g f_absorbee = %g signal=%g\n", __FUNCTION__, */
/* 	   energie, mu_detecteur, exp( -mu_detecteur * detecteur->longueur ), f_absorbee, signal ); */
#endif

    /* Le faisceau diffus est aussi perdu, mais ne donne pas de signal ! */
    mu_detecteur += GRF_LireValeur( detecteur->composition->te_total, energie, INTERPOLE_LINEAIRE );
    mu_detecteur += GRF_LireValeur( detecteur->composition->ti_total, energie, INTERPOLE_LINEAIRE );
    *attenuation = exp( - mu_detecteur * detecteur->longueur ); /* Fraction de faisceau restant aprs traverse */
  } else {
    /* Dtecteur idal : mme absorbance quelle que soit l'nergie */
    *attenuation = 1.0;
  }

  /* Si la rponse du dtecteur aussi varie avec l'nergie */
  if ( detecteur->proprietes & DETECTEUR_NRJ_VARIABLE ) {
    double reponse;

    reponse = GRF_LireValeur( detecteur->r_energie, energie, INTERPOLE_LINEAIRE );
#ifdef _DEVELOPPE_
    printf("%s : E=%g signal=%g attenuation=%g\n", __FUNCTION__,
	   energie, signal, reponse );
#endif
    signal *= reponse;
  }

  /* On a calcul le signal  cette nergie, pour ce point */
  SORTIE_FONCTION("DET_Mono_Ponctuel", "signal=%g / attenuation = %g", signal, *attenuation);
  return signal;
}

/* -------- Faisceau polychromatique -------- */

static double DETECTEUR_Polychromatique_I(Detecteur *detecteur, Faisceau *faisceau, double energie, Faisceau *f_sortie)
{
  double signal = 0.0;
  double e_vraie, *de, I_vraie, *gE;
  unsigned long i;
  Faisceau pseudo_faisceau, ps_sortie;

  if ( 0.0 == faisceau->intensite ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Intensit nulle\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return 0.0;
  }

  /* Le faisceau est trait comme un monochromatique ! */
  pseudo_faisceau = *faisceau;
  pseudo_faisceau.proprietes &= ~( FAISCEAU_POLYCHROMATIQUE | FAISCEAU_HARMONIQUES );
  pseudo_faisceau.g_E = NULL;

  if ( f_sortie != NULL ) {
    ps_sortie = *f_sortie;
  } else { /* Pointeurs  zro pour contrler que l'on ne les utilise pas par erreur ! */
    ps_sortie.g_energie = NULL;
    ps_sortie.g_E = NULL;
    ps_sortie.i_harmoniques = NULL;
    ps_sortie.g_harmoniques = NULL;
    ps_sortie.g_espace = NULL;
  }

  /* Faisceau ponctuel ou homogne : il suffit de cumuler pour toutes les nergies */
  if ( ( faisceau->proprietes & FAISCEAU_PONCTUEL ) || ( NULL == faisceau->g_espace ) ) {
    /* Pour chacune des nergies enregistres */
    for ( i = 0, de = faisceau->g_E->x, gE = faisceau->g_E->y ; i < faisceau->g_E->nombre_point ; i++, de++, gE++ ) {
      /* nergie prise en compte */
      e_vraie = energie * ( 1.0 + *de );

      /* Intensit pour cette nergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
      I_vraie = faisceau->intensite * *gE;
      if ( 0.0 == I_vraie ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : Faisceau nul [E=%g, de=%g, I=%g, g(E)=%g\n",
		       __FILE__, __FUNCTION__, __LINE__, energie, *de, faisceau->intensite, *gE);
	abort();
#endif
	continue;
      }
      pseudo_faisceau.intensite = I_vraie;

      /* Contribution de cette nergie */
      signal += DETECTEUR_Monochromatique_I( detecteur, &pseudo_faisceau, e_vraie, &ps_sortie );
      if ( f_sortie != NULL ) {
	f_sortie->g_E->y[ i ] = ps_sortie.intensite / faisceau->intensite;
#ifdef _DEVELOPPE_
	(void) printf("%s,%s l.%u : g(E) sortie = %g [E=%g, de=%g, I=%g]\n",
		      __FILE__, __FUNCTION__, __LINE__, ps_sortie.intensite, energie, *de, faisceau->intensite);
#endif
      }
    }
  } else if ( IMG_GENRE_REELLE == faisceau->g_espace->genre ) { /* Idem si g(E) est le mme en tout point */
    /* Pour chacune des nergies enregistres */
    for ( i = 0, de = faisceau->g_E->x, gE = faisceau->g_E->y ; i < faisceau->g_E->nombre_point ; i++, de++, gE++ ) {
      /* nergie prise en compte */
      e_vraie = energie * ( 1.0 + *de );

      /* Intensit pour cette nergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
      I_vraie = faisceau->intensite * *gE;
      if ( 0.0 == I_vraie ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : Faisceau nul [E=%g, de=%g, I=%g, g(E)=%g\n",
		       __FILE__, __FUNCTION__, __LINE__, energie, *de, faisceau->intensite, *gE);
	abort();
#endif
	continue;
      }
      pseudo_faisceau.intensite = I_vraie;

      /* Contribution de cette nergie
	 On suppose que g(E) reste le mme en tout point en traversant le dtecteur
	 ==> on peut le calculer  partir de n'importe-quel point */
      signal += DETECTEUR_Monochromatique_I( detecteur, &pseudo_faisceau, e_vraie, &ps_sortie );
    }

    /* Pour un faisceau rectangulaire inhomogne, les valeurs sont dans la carte ->
       f_sortie->g_espace contient la dernire carte, qui contient les vraies valeurs * g(E) pour la dernire nergie
       De plus, par hypothse g(E) n'est pas chang */
    if ( f_sortie != NULL ) {
      f_sortie->intensite = 1.0 / faisceau->g_E->y[ faisceau->g_E->nombre_point - 1 ];
    }
  } else {
    Image *coupe_mono;
    unsigned long i_case, nbr_valeurs;
    struct graphe **grf_gE;

    /* On prpare une pseudo-carte pour le faisceau monochromatique */
    coupe_mono = IMG_CreerImageIsolee( IMG_GENRE_REELLE, IMG_TYPE_INTENSITES, faisceau->g_espace->larg, faisceau->g_espace->haut );
    if ( NULL == coupe_mono ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la pseudo-coupe\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -487;
    }
    coupe_mono->l_reel = faisceau->g_espace->l_reel; coupe_mono->h_reel = faisceau->g_espace->h_reel;
    coupe_mono->x_min = faisceau->g_espace->x_min; coupe_mono->y_min = faisceau->g_espace->y_min;

    nbr_valeurs = coupe_mono->larg * (unsigned long) coupe_mono->haut;

    /* C'est donc elle que l'on utilisera ! */
    pseudo_faisceau.g_espace = coupe_mono;

    /* Pour chacune des nergies enregistres */
    for ( i = 0, de = faisceau->g_espace->valeurs.G[ 0 ]->x ; i < faisceau->g_espace->valeurs.G[ 0 ]->nombre_point ; i++, de++ ) {
      /* nergie prise en compte */
      e_vraie = energie * ( 1.0 + *de );

      /* On cre la coupe pour cette nergie  partir des g(E) de l'image relle.
	 Suppose que les g(E) en chaque point existent : c'est FAISCEAU_GeometrieApresEchantillon() qui s'en occupe */
      for ( i_case = 0, grf_gE = faisceau->g_espace->valeurs.G ; i_case < nbr_valeurs ; i_case++, grf_gE++ ) {
	coupe_mono->valeurs.R[ i_case ] = (*grf_gE)->y[ i ];
      }

      /* Contribution de cette nergie */
      signal += DETECTEUR_Monochromatique_I( detecteur, &pseudo_faisceau, e_vraie, &ps_sortie );
      if ( f_sortie != NULL ) {
	(void) printf("%s,%s l.%u :  FAIRE, mmorisation du faisceau sortant pour g(E) variable en chaque point\n",
		      __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
	abort();
#endif
      }
    }

    /* Plus besoin de la pseudo-coupe */
    IMG_EffacerImage( coupe_mono, FAUX );
  }

  /* Calcul de la contribution termin */
  return signal;
}
