/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Calcul du signal obtenu en fluorescence
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "experience.h"
#include "graphes.h"
#include "lissage.h" /* pour GRF_LireValeur */

#ifdef _DEVELOPPE_
# include "serie.h"
#endif

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FLUO_DetecteurPonctuel(Experience *experience, Faisceau *f_detecteur);
static int FLUO_DectecteurPonctuel_Infini(Experience *experience, Faisceau *f_detecteur, Graphe *spectre);

static int FLUO_ContributionCase(Experience *experience, Faisceau *f_detecteur, Graphe *spectre, unsigned long n_case, Vecteur3D *propagation);
static int FLUO_ContributionCase_H(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, double distance, double N_absorbe);
static int FLUO_ContributionCase_G(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, unsigned long n_case, Vecteur3D *p_depart, Vecteur3D *direction);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Génération du spectre ~-~-~-~-~-~-~-~- */

Graphe *FLUO_CalculerSpectre(Experience *experience, Faisceau *f_detecteur, double energie __attribute__((__unused__)) )
{
  int retour = -1;
  Graphe *spectre;

#ifdef _DEVELOPPE_
/*   printf("%s : Calcul pour E=%g eV\n", __FUNCTION__, energie ); */
#endif

  /* On génère le faisceau fluorescé, tel qu'il sera vu par le détecteur */
  if ( f_detecteur->proprietes & FAISCEAU_PONCTUEL ) {
    unsigned long i;

    /* On commence par mettre à zéro le spectre du détecteur */
    for ( i = 0 ; i < f_detecteur->g_E->nombre_point ; i++ ) {
      f_detecteur->g_E->y[ i ] = 0.0;
    }

    /* On calcule le nouveau signal */
    retour = FLUO_DetecteurPonctuel( experience, f_detecteur );
  } else {
    (void) fprintf(stderr, "FLUO_CalculerSpectre : À FAIRE --- Faisceau de sortie non-ponctuel\n");
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_CalculerSpectre : ERREUR dans la génération du faisceau fluorescé [code %d]\n", retour);
  }

  /* On applique le détecteur */
  spectre = DETECTEUR_CalculerSpectre( experience->i1, f_detecteur );
  if ( NULL == spectre ) {
    (void) fprintf(stderr, "FLUO_CalculerSpectre : ERREUR dans DETECTEUR_CalculerSpectre()\n");
    return NULL;
  }

  return spectre;
}

/* ~-~-~-~-~-~-~-~- Cas d'un détecteur ponctuel [ou assimilable] ~-~-~-~-~-~-~-~- */

static int FLUO_DetecteurPonctuel(Experience *experience, Faisceau *f_detecteur)
{
  Echantillon *echantillon;
  unsigned long nbr_cases;
  Graphe *spectre_genere;

  ENTREE_FONCTION( __FUNCTION__, "experience=%p f_detecteur=%p", experience, f_detecteur );

  /* Préparation */
  echantillon = experience->echantillon;
  nbr_cases = echantillon->nbr_x * echantillon->nbr_y * echantillon->nbr_z ;

  /* Spectre reçu par le détecteur en provenance d'un seul point */
  spectre_genere = GRF_nv_isole( f_detecteur->g_E->nombre_point, GR_FLUOX, GRF_SERIE );
  if ( NULL == spectre_genere ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_nv_isole()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR dans GRF_nv_isole]", -1);
    return -1;
  }
  spectre_genere->x = f_detecteur->g_E->x;
#ifdef _DEVELOPPE_
  GRF_titre( spectre_genere, "spectre_genere, FLUO_DInfini_Homogene");
#endif

  if ( experience->i1->proprietes & DETECTEUR_A_L_INFINI ) {
    int retour;

    retour = FLUO_DectecteurPonctuel_Infini( experience, f_detecteur, spectre_genere );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_DectecteurPonctuel_Infini() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return -28;
    }
  } else {
    Vecteur3D u_faisceau, point_emetteur;
    unsigned long i;

    /* On calcule la contribution de chaque point de l'échantillon */
    for ( i = 0 ; i < nbr_cases ; i++ ) {
      if ( echantillon->nombre[ i ] > 0.0 ) { /* Le point contribue */
	/* Position réelle du point émetteur */
	ECH_Case_v_XYZ( echantillon, i, &(point_emetteur.x), &(point_emetteur.y), &(point_emetteur.z) );

	/* Propagation : du point vers le détecteur => change pour chaque case */
	u_faisceau.x = point_emetteur.x - experience->i1->orientation.position.x;
	u_faisceau.y = point_emetteur.y - experience->i1->orientation.position.y;
	u_faisceau.z = point_emetteur.z - experience->i1->orientation.position.z;

	/* Normalisation de ce vecteur */
	U3D_Normalise( &u_faisceau, NULL );

	/* Contribution de la case */
	FLUO_ContributionCase( experience, f_detecteur, spectre_genere, i, &u_faisceau );
      }
    }
  }

#ifdef _DEVELOPPE_
/*   exporter( spectre_genere, "fgenere.asc" ); */
/*   GRF_bornes( spectre_genere, ORDONNE, DESORDONNE ); / En cas d'affichage... */
/*   GRF_InsererDansSerie( spectre_genere, serie_courante, NULL ); */
#endif

  /* Plus besoin du graphe généré */
  GRF_eff_isole( spectre_genere );

  /* Tout s'est bien passé */
  GRF_bornes( f_detecteur->g_E, NE_PAS_FAIRE, DESORDONNE ); /* En cas d'affichage... */
  SORTIE_FONCTION( __FUNCTION__, "%d", 0);
  return 0;
}

/* -------- Détecteur ponctuel placé à l'infini (ou assimilé) -------- */

static int FLUO_DectecteurPonctuel_Infini(Experience *experience, Faisceau *f_detecteur, Graphe *spectre)
{
  int direction_propagation;
  unsigned long n_case;
  Vecteur3D u_faisceau, point_sortie;

  /* Direction de propagation du faisceau vers le détecteur
     Comme le détecteur est à l'infini, le faisceau fluorescé est parallèle */
  u_faisceau.x = -sin( f_detecteur->orientation.theta );
  u_faisceau.y = -cos( f_detecteur->orientation.theta ) * cos( f_detecteur->orientation.phi );
  u_faisceau.z =  cos( f_detecteur->orientation.theta ) * sin( f_detecteur->orientation.phi );

 /* On cherche le point de sortie, sachant que le faisceau ponctuel vu passe par le point visé */
  ECH_TrouverSortieFaisceau( experience->echantillon, &(experience->i1->orientation.position), &u_faisceau, &point_sortie, VRAI );

  /* Selon quelle direction avance-t-on dans les tranches ? */
  direction_propagation = U3D_DirectionPreferentielle( &u_faisceau );
  if ( U3D_PROPAGATION_SELON_X == direction_propagation ) {
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Propagation selon X\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 458;
  } else if ( U3D_PROPAGATION_SELON_Y == direction_propagation ) {
    unsigned int idx_y, j_debut;
    Vecteur3D point_source;
    double delta_t;

    /* Index de la case de départ */
    ECH_XYZ_v_IJK( experience->echantillon, point_sortie.x, point_sortie.y, point_sortie.z, NULL, &j_debut, NULL );

    /* Incrément pour passer d'une tranche à l'autre */
    delta_t = experience->echantillon->pas_y / u_faisceau.y;

    for ( idx_y = j_debut ; idx_y < experience->echantillon->nbr_y ; idx_y++ ) {
      /* Coordonnées de la case correspondante */
      point_source = point_sortie;
      point_source.x += u_faisceau.x * delta_t * (idx_y - j_debut);
      point_source.y += u_faisceau.y * delta_t * (idx_y - j_debut);
      point_source.z += u_faisceau.z * delta_t * (idx_y - j_debut);

      /* Numéro associé */
      n_case = ECH_XYZ_v_Case( experience->echantillon, point_source.x, point_source.y, point_source.z );

      /* Contribution de la case correspondante */
      FLUO_ContributionCase( experience, f_detecteur, spectre, n_case, &u_faisceau );
    }
  } else if ( U3D_PROPAGATION_SELON_Z == direction_propagation ) {
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Propagation selon Z\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 458;
  } else {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la direction de propagation !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* Contribution terminée */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Contribution d'une case ~-~-~-~-~-~-~-~-~ */

/* -------- Routine générique qui appelle les autres --------
   Renvoie 0 si tout va bien, != 0 sinon

   experience  : l'expérience en cours de modélisation
   f_detecteur : le faisceau vu par le détecteur
   spectre     : le spectre qui contiendra le faisceau fluorescé après réabsorption
   n_case      : le numéro de la case à considérer
   propagation : la direction de propagation du faisceau

   La contribution de la case est _ajoutée_ au g(E) du faisceau
   Les valeurs du spectre seront remplacées par celle de fluo idéales corrigées
                  spectre doit avoir les mêmes X que exp->fluo
 */

static int FLUO_ContributionCase(Experience *experience, Faisceau *f_detecteur, Graphe *spectre, unsigned long n_case, Vecteur3D *propagation)
{
  double N_absorbe;
  unsigned long i;
  Echantillon *echantillon;
  Graphe *fluo_ideale;
  Vecteur3D point_emetteur, point_sortie;

  echantillon = experience->echantillon;

  /* Combien de photons ont été absorbés par cette case ? */
  N_absorbe = echantillon->nombre[ n_case ];
  if ( 0.0 == N_absorbe ) return 0; /* La case ne contribue pas */

  /* Position réelle du point émetteur */
  ECH_Case_v_XYZ( echantillon, n_case, &(point_emetteur.x), &(point_emetteur.y), &(point_emetteur.z) );
  
  /* Point de sortie */
  ECH_TrouverSortieFaisceau( echantillon, &point_emetteur, propagation, &point_sortie, VRAI );

  /* Réabsorption par l'échantillon, en fonction de son type */
  if ( experience->echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    double distance_dans_echantillon;

    fluo_ideale = echantillon->valeurs.homogene.composition->fluo;

    /* Distance parcourue dans l'échantillon [en cm, pour compatibilité avec mu/rho] */
    distance_dans_echantillon = U3D_Distance( &point_emetteur, &point_sortie );
    distance_dans_echantillon /= 10000.0; /* Coordonnées en microns, conversion en cm */

    FLUO_ContributionCase_H( echantillon, spectre, fluo_ideale, distance_dans_echantillon, N_absorbe );
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    fluo_ideale = echantillon->valeurs.homogene.composition->fluo;

    FLUO_ContributionCase_G( echantillon, spectre, fluo_ideale, n_case, &point_emetteur, propagation );
  } else {
    FLUO_ContributionCase_G( echantillon, spectre, NULL, n_case, &point_emetteur, propagation );
  }

  /* Correction éventuelle : expérience absorbante */
  if ( experience->proprietes & EXPERIENCE_ABSORBANTE ) {
    double distance, *x, *y, mu;

    /* Distance parcourue dans l'air [en cm, pour compatibilité avec mu/rho] */
    distance = U3D_Distance( &point_sortie, &(f_detecteur->orientation.position) );
    distance /= 10000.0; /* Coordonnées en microns, conversion en cm */

    for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
      mu  = GRF_LireValeur( experience->composition->mu_total, *x, INTERPOLE_LINEAIRE );
      mu += GRF_LireValeur( experience->composition->te_total, *x, INTERPOLE_LINEAIRE );
      mu += GRF_LireValeur( experience->composition->ti_total, *x, INTERPOLE_LINEAIRE );
      mu *= experience->composition->densite;

      *y *= exp( -mu * distance );
    }
  }

  /* On ajoute cette contribution au spectre vu par le faisceau */
  for ( i = 0  ; i < spectre->nombre_point ; i++ ) {
    f_detecteur->g_E->y[ i ] += spectre->y[ i ];
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Contribution d'une case : cas d'un échantillon homogène -------- */

static int FLUO_ContributionCase_H(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, double distance, double N_absorbe)
{
  double mu = 0.0;
  double *x, *y;
  unsigned long i;

  /* Atténuation du spectre */
  for ( i = 0, x = spectre->x, y = spectre->y ; i < spectre->nombre_point ; i++, x++, y++ ) {
    mu = ECH_CalculerMuTotal( echantillon, NULL, *x, NULL );

    /* Assez peu optimal, mais en attendant mieux... */
    *y = N_absorbe * fluo->y[ i ] * exp( -mu * distance );

/*     printf(" %g * %g * %g [exp(-%g * %g)] = %g\n", N_absorbe, GRF_LireValeur( fluo, *x, INTERPOLE_LINEAIRE ), exp( -mu * distance ), mu, distance, *y ); */
  }
/*   printf("%s : d = %g ==> att = %g\n", __FUNCTION__, distance, exp( -mu * distance ) ); */

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Contribution d'une case : cas d'un échantillon de type gradient -------- */

static int FLUO_ContributionCase_G(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, unsigned long n_case,
				   Vecteur3D *point_emetteur, Vecteur3D *direction)
{
  double distance, mu_total, densite __attribute__((__unused__)) = 1.0, N_absorbe;
  double *x, *y_fluo, *y_spectre;
  unsigned int i_x, i_y, i_z;
  unsigned long i_energie;
  boolean sortie = FAUX;

  /* Nombre de photons absorbés */
  N_absorbe = echantillon->nombre[ n_case ];
  if ( NULL == fluo ) { /* Échantillon inhomogène */
    struct matiere *composition;

    composition = echantillon->valeurs.inhomogene[ n_case ];
    if ( NULL == composition ) return 0; /* Trou : ne contribue pas à l'émission ! */
    fluo = composition->fluo;
  }

  /* On copie le spectre de fluo idéale : résultat attendu sans correction */
  for ( i_energie = 0, x = fluo->x, y_fluo = fluo->y, y_spectre = spectre->y ;
	i_energie < fluo->nombre_point ; i_energie++, x++, y_fluo++, y_spectre++ ) {
    *y_spectre = *y_fluo * N_absorbe;
  }

  /* On part de la case initiale */
  ECH_Case_v_IJK( echantillon, n_case, &i_x, &i_y, &i_z );
#ifdef _DEVELOPPE_
/*   printf("%g en (%g, %g, %g)=(%u, %u, %u)=%lu\n", echantillon->nombre[ n_case ], */
/* 	 point_emetteur->x, point_emetteur->y, point_emetteur->z, i_x, i_y, i_z, n_case); */
#endif

  while ( FAUX == sortie ) {
    /* Mu pour la case considérée */
    if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      densite = echantillon->valeurs.homogene.concentration[ n_case ];
    }

    /* Distance parcourue dans cette case */
    distance = ECH_TrouverCaseSuivante( echantillon, direction, point_emetteur, &i_x, &i_y, &i_z, &sortie );
    distance /= 10000.0; /* Travail en microns, on veut des cm pour les µ/rho */

    /* Calcul de l'absorption pour chaque énergie */
    for ( i_energie = 0, x = fluo->x, y_spectre = spectre->y ;
	  i_energie < fluo->nombre_point ; i_energie++, x++, y_spectre++ ) {
      mu_total = ECH_CalculerMuTotal( echantillon, point_emetteur, *x, NULL );

      *y_spectre *= exp( -mu_total * distance );
    }

    if ( VRAI == sortie ) break;
    /* On passe à la case suivante... [sauf si l'on est sorti] */
    n_case = ECH_IJK_v_Case( echantillon, i_x, i_y, i_z );
  }

  /* Tout s'est bien passé */
  return 0;
}

#ifdef ancien
/* ---- Propagation avec comme direction principale l'axe des X ---- */

static int FLUO_CaseG_X(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, unsigned long n_case, Vecteur3D *u_faisceau, Vecteur3D *point_sortie)
{
  double dd, delta_t;
  double N_absorbe, attenuation = -181.168, mu_total;
  double *x, *y_fluo, *y_spectre;
  unsigned long i_energie;
  unsigned int i, i_debut, i_fin;
  Vecteur3D point_entree, p_entree, p_sortie;

  /* Dans quelle tranche commence-t-on ? */
  ECH_Case_v_XYZ( echantillon, n_case, &(point_entree.x), &(point_entree.y), &(point_entree.z) );
  ECH_Case_v_IJK( echantillon, n_case, &i_debut, NULL, NULL );

  /* Dans quelle tranche finit-on ? */
  ECH_XYZ_v_IJK( echantillon, point_sortie->x, point_sortie->y, point_sortie->z, &i_fin, NULL, NULL );

  /* Nombre de photons absorbés */
  N_absorbe = echantillon->nombre[ n_case ];

  /* Incrément pour passer d'une tranche à l'autre */
  delta_t = echantillon->pas_z / u_faisceau->z;
  if  ( i_debut > i_fin ) delta_t = -delta_t; /* Pour bien aller dans le bon sens ! */

  /* Distance parcourue en passant d'une tranche à l'autre
     C'est aussi l'incrément en t pour cela, car u_faisceau est normalisé */
  dd = fabs( delta_t );
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* Travail pour chaque énergie
     Remarque : par construction, spectre et fluo ont forcément les mêmes X [?]
  */
#ifdef _DEVELOPPE_
  if ( ( fluo->x != spectre->x ) || ( fluo->nombre_point != spectre->nombre_point ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : X différents\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
  }
#endif
  for ( i_energie = 0, x = fluo->x, y_fluo = fluo->y, y_spectre = spectre->y ; i_energie < fluo->nombre_point ; i_energie++, x++, y_fluo++, y_spectre++ ) {
    attenuation = 1.0;

    /* Le point d'entrée */
    p_entree = point_entree;

    /* Calcul de l'atténuation totale */
    /* 1. Section efficace d'absorption (totale) [seule la densité varie] */
    mu_total = ECH_CalculerMuTotal( echantillon, NULL, *x, NULL );

    /* 2. Boucle jusqu'à la sortie */
    for ( i = 0 ; i < (unsigned int) ( abs( i_fin - i_debut ) ) ; i++ ) {
      double t, att_tmp;

      /* Passage d'un point à l'autre ; on démarre sur un point de la grille -> les Y varient de pas_y à chaque fois */
      p_sortie = point_entree;
      t = delta_t * (double) (i + 1);
      p_sortie.x += t * u_faisceau->x;
      p_sortie.y += t * u_faisceau->y;
      p_sortie.z += t * u_faisceau->z;

      /* Erreurs d'arrondi possibles en Y, risque de sortie de l'échantillon => d fausse :
	 on corrige donc à l'avance */
      if ( p_sortie.x < -(echantillon->larg / 2.0) ) p_sortie.x = -echantillon->larg / 2.0;
      if ( p_sortie.x >  (echantillon->larg / 2.0) ) p_sortie.x =  echantillon->larg / 2.0;

      /* Absorption entre ces deux points */
      (void) FLUO_PhotonsAbsorbes( echantillon, &p_entree, &p_sortie, dd, 0.0, mu_total, *x, &att_tmp );

      attenuation *= att_tmp;

      /* On passe au suivant */
      p_entree = p_sortie;
    }

    /* On mémorise le nombre final */
    *y_spectre += N_absorbe * attenuation * *y_fluo;
  }

  /* Tout s'est bien passé, calcul fini */
  return 0;
}

/* ---- Propagation avec comme direction principale l'axe des Y ---- */

static int FLUO_CaseG_Y(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, unsigned long n_case, Vecteur3D *u_faisceau, Vecteur3D *point_sortie)
{
  double dd, delta_t;
  double N_absorbe, attenuation = -181.168, mu_total;
  double *x, *y_fluo, *y_spectre;
  unsigned long i_energie;
  unsigned int j, j_debut, j_fin;
  Vecteur3D point_entree, p_entree, p_sortie;

  /* Dans quelle tranche commence-t-on ? */
  ECH_Case_v_XYZ( echantillon, n_case, &(point_entree.x), &(point_entree.y), &(point_entree.z) );
  ECH_Case_v_IJK( echantillon, n_case, NULL, &j_debut, NULL );

  /* Dans quelle tranche finit-on ? */
  ECH_XYZ_v_IJK( echantillon, point_sortie->x, point_sortie->y, point_sortie->z, NULL, &j_fin, NULL );

#ifdef _DEVELOPPE_
/*   printf("%s : --- Case %lu : tranches %u (%g, %g, %g) à %u (%g, %g, %g), faisceau (%g, %g, %g) ---\n", __FUNCTION__, n_case, */
/* 	 j_debut, p_entree.x, p_entree.y, p_entree.z, j_fin, point_sortie->x, point_sortie->y, point_sortie->z, u_faisceau->x, u_faisceau->y, u_faisceau->z ); */
#endif

 /* On est en surface -> on néglige la réabsorption ? */
/*   if ( j_debut == j_fin ) { */
/*     return 0; */
/*   } */

  /* Nombre de photons absorbés */
  N_absorbe = echantillon->nombre[ n_case ];

  /* Incrément pour passer d'une tranche à l'autre */
  delta_t = echantillon->pas_y / u_faisceau->y;
  if  ( j_debut > j_fin ) delta_t = -delta_t; /* Pour bien aller dans le bon sens ! */

  /* Distance parcourue en passant d'une tranche à l'autre
     C'est aussi l'incrément en t pour cela, car u_faisceau est normalisé */
  dd = fabs( delta_t );
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* Travail pour chaque énergie
     Remarque : par construction, spectre et fluo ont forcément les mêmes X [?]
  */
  for ( i_energie = 0, x = fluo->x, y_fluo = fluo->y, y_spectre = spectre->y ; i_energie < fluo->nombre_point ; i_energie++, x++, y_fluo++, y_spectre++ ) {
    attenuation = 1.0;

    /* Le point d'entrée */
    p_entree = point_entree;

    /* Calcul de l'atténuation totale */
    /* 1. Section efficace d'absorption (totale) [seule la densité varie] */
    mu_total = ECH_CalculerMuTotal( echantillon, NULL, *x, NULL );

    /* 2. Boucle jusqu'à la sortie */
    for ( j = 0 ; j < (unsigned int) ( abs( j_fin - j_debut ) ) ; j++ ) {
      double t, att_tmp;

      /* Passage d'un point à l'autre ; on démarre sur un point de la grille -> les Y varient de pas_y à chaque fois */
      p_sortie = point_entree;
      t = delta_t * (double) (j + 1);
      p_sortie.x += t * u_faisceau->x;
      p_sortie.y += t * u_faisceau->y;
      p_sortie.z += t * u_faisceau->z;

      /* Erreurs d'arrondi possibles en Y, risque de sortie du faisceau => d fausse : on corrige */
      if ( p_sortie.y < 0.0 ) p_sortie.y = 0.0;
      if ( p_sortie.y > echantillon->epaisseur ) p_sortie.y = echantillon->epaisseur;

      /* Absorption entre ces deux points */
/*       printf("%s : Calcul, tranche %u (%u, [%u-%u]) (%g, %g, %g) -> (%g, %g, %g)\n", __FUNCTION__, j, j_debut + j, j_debut, j_fin, */
/* 	      p_entree.x, p_entree.y, p_entree.z, p_sortie.x, p_sortie.y, p_sortie.z); */
      (void) FLUO_PhotonsAbsorbes( echantillon, &p_entree, &p_sortie, dd, 0.0, mu_total, *x, &att_tmp );

      attenuation *= att_tmp;

      /* On passe au suivant */
      p_entree = p_sortie;
    }

    /* On mémorise le nombre final */
    *y_spectre += N_absorbe * attenuation * *y_fluo;
  }
/*   printf("%s : d = %g [%g] ==> att = %g\n", __FUNCTION__, dd * abs( j_fin - j_debut ), tst_d, attenuation ); */

  /* Tout s'est bien passé, calcul fini */
  return 0;
}

/* ---- Propagation avec comme direction principale l'axe des Z ---- */

static int FLUO_CaseG_Z(Echantillon *echantillon, Graphe *spectre, Graphe *fluo, unsigned long n_case, Vecteur3D *u_faisceau, Vecteur3D *point_sortie)
{
  double dd, delta_t;
  double N_absorbe, attenuation = -181.168, mu_total;
  double *x, *y_fluo, *y_spectre;
  unsigned long i_energie;
  unsigned int j, k_debut, k_fin;
  Vecteur3D point_entree, p_entree, p_sortie;

  /* Dans quelle tranche commence-t-on ? */
  ECH_Case_v_XYZ( echantillon, n_case, &(point_entree.x), &(point_entree.y), &(point_entree.z) );
  ECH_Case_v_IJK( echantillon, n_case, NULL, NULL, &k_debut );

  /* Dans quelle tranche finit-on ? */
  ECH_XYZ_v_IJK( echantillon, point_sortie->x, point_sortie->y, point_sortie->z, NULL, NULL, &k_fin );

  /* Nombre de photons absorbés */
  N_absorbe = echantillon->nombre[ n_case ];

  /* Incrément pour passer d'une tranche à l'autre */
  delta_t = echantillon->pas_z / u_faisceau->z;
  if  ( k_debut > k_fin ) delta_t = -delta_t; /* Pour bien aller dans le bon sens ! */

  /* Distance parcourue en passant d'une tranche à l'autre
     C'est aussi l'incrément en t pour cela, car u_faisceau est normalisé */
  dd = fabs( delta_t );
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* Travail pour chaque énergie
     Remarque : par construction, spectre et fluo ont forcément les mêmes X [?]
  */
  for ( i_energie = 0, x = fluo->x, y_fluo = fluo->y, y_spectre = spectre->y ; i_energie < fluo->nombre_point ; i_energie++, x++, y_fluo++, y_spectre++ ) {
    attenuation = 1.0;

    /* Le point d'entrée */
    p_entree = point_entree;

    /* Calcul de l'atténuation totale */
    /* 1. Section efficace d'absorption (totale) [seule la densité varie] */
    mu_total = ECH_CalculerMuTotal( echantillon, NULL, *x, NULL );

    /* 2. Boucle jusqu'à la sortie */
    for ( j = 0 ; j < (unsigned int) ( abs( k_fin - k_debut ) ) ; j++ ) {
      double t, att_tmp;

      /* Passage d'un point à l'autre ; on démarre sur un point de la grille -> les Y varient de pas_y à chaque fois */
      p_sortie = point_entree;
      t = delta_t * (double) (j + 1);
      p_sortie.x += t * u_faisceau->x;
      p_sortie.y += t * u_faisceau->y;
      p_sortie.z += t * u_faisceau->z;

      /* Erreurs d'arrondi possibles en Y, risque de sortie de l'échantillon => d fausse :
	 on corrige donc à l'avance */
      if ( p_sortie.z < -(echantillon->haut / 2.0) ) p_sortie.z = -echantillon->haut / 2.0;
      if ( p_sortie.z >  (echantillon->haut / 2.0) ) p_sortie.z =  echantillon->haut / 2.0;

      /* Absorption entre ces deux points */
      (void) FLUO_PhotonsAbsorbes( echantillon, &p_entree, &p_sortie, dd, 0.0, mu_total, *x, &att_tmp );

      attenuation *= att_tmp;

      /* On passe au suivant */
      p_entree = p_sortie;
    }

    /* On mémorise le nombre final */
    *y_spectre += N_absorbe * attenuation * *y_fluo;
  }

  /* Tout s'est bien passé, calcul fini */
  return 0;
}
#endif
