/* Traitement de l'E.X.A.F.S. sur Atari
   Tout ce qui concerne le X.A.N.E.S.
   (c) 1996-1998, Emmanuel CURIS - Création : 22 mai 1998

   Soustraction de la contribution de l'échantillon (mu_matrice)
   Récupéré de ../XANES.C et de ../EXF_GEM.C ( GER_matrice() )
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "serie.h"

#include "automatise.h"		/* Pour le traitement par lots... */
#include "interface.h" /* Pour RCT_pointeur */
#include "lissage.h"   /* Pour l'interpolation et les splines */
#include "matrices.h"   /* Pour le calcul matriciel / MDC linéaires */
#include "operations.h" /* Pour les dérivées */
#include "tracer.h"
#include "xanes.h"

/* ************ Variables spécifiques de ce module *************************************** */

#include "mu_matrice.txt"

/* ~~~~~~~~ Fonctions spécifiques à ce module **************************************** */

static boolean XAN_ModeleEstLineaire(int methode);

static double XAN_InitMatriceLineaire(unsigned long nbr_pts, double *X, double *Y, MAT_infos *infos, Matrice *mat_MDC, Matrice *v_constant);
static double XAN_TrouveLineaireReference(Graphe *brut, MAT_infos *infos, Matrice *mat_MDC, Matrice *v_constant);
static double XAN_TrouveModeleLineaire   (Graphe *brut, MAT_infos *infos, Matrice *mat_MDC, Matrice *v_constant);

static int XAN_BornesIntervalle       (Graphe *brut, MAT_infos *infos, double *x_min, double *x_max);
static int XAN_CalculerModele         (Graphe *brut, MAT_infos *infos, boolean avec_erreurs);

static int XAN_MatriceVictorine_1(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *mat_MDC, Matrice *v_constant);
static int XAN_MatriceVictorine_2(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *mat_MDC, Matrice *v_constant);
static int XAN_MatriceVictorine_3(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *mat_MDC, Matrice *v_constant);

static int XAN_CalculeVictorine(Graphe *brut, double cte, double coef_X, double coef_X3, double coef_X4);

static unsigned int XAN_NombreParametres(int modele, boolean sur_reference, unsigned int degre);

/* ************ Définitions effectives des opérations à effectuer ************************ */

/* ~~~~~~~~~~~~ Partie mathématique : on effectue les calculs ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ~-~-~-~-~-~-~-~- Aides à la construction du modèle ~-~-~-~-~-~-~-~- */

/* -------- Sur quel intervalle en X ajuste-t-on ? -------- */

static int XAN_BornesIntervalle(Graphe *brut, MAT_infos *infos, double *x_min, double *x_max)
{
  if ( infos->mode == DEUX_POINTS ) { /* On vérifie que les deux points sont sensés */
    if (infos->e_1p > infos->e_1) { /* Par convention, graphe->x_min <= infos->e_1p < infos->e_1 <= graphe->x_max */
      double x = infos->e_1 ; infos->e_1 = infos->e_1p ; infos->e_1p = x;
    }

    if ( infos->e_1p < brut->x_min )
      infos->e_1p = brut->x_min;
    
    *x_min = infos->e_1p;
  } else {
    *x_min = brut->x_min;
  }
  if ( infos->e_1 > brut->x_max) /* On vérifie que l'on ne déborde pas */
    infos->e_1 = brut->x_max;
  *x_max = infos->e_1;
  if ( *x_max < *x_min ) { double x = *x_min; *x_min = *x_max; *x_max = x; } /* Probablement, il y a eu un problème... */

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Le modèle est-il linéaire ? -------- */

static inline boolean XAN_ModeleEstLineaire(int methode)
{
  switch ( methode )
  {
  case MUM_DROITE:
  case MUM_VICTORINE_1:
  case MUM_VICTORINE_2:
  case MUM_POLYNOME:
  case MUM_THEORIQUE:
  case MUM_VICTORINE_3:
    return VRAI;

  case MUM_CONSTANTE:
    /* Linéaire, mais c'est plus rapide de faire directement la moyenne des points qui conviennent ! */
    return FAUX;

  case MUM_GAUSSIENNE:
    return FAUX;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Modèle inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode);
  }

  return FAUX;
}

/* -------- Combien de paramètres dans le modèle ? -------- */

static unsigned int XAN_NombreParametres(int modele, boolean sur_reference, unsigned int degre)
{
  unsigned int nbr_parametres;

  switch ( modele )
  {
  case MUM_CONSTANTE:		/* a0 */
    nbr_parametres = 1;
    break;

  case MUM_DROITE: /* a0 + a1 * X */
    nbr_parametres = 2;
    break;

  case MUM_VICTORINE_1: /* a0 + a1 / X³ */
    nbr_parametres = 2;
    break;

  case MUM_VICTORINE_2: /* a0 + a1 / X³ + a2 / X4 */
    nbr_parametres = 3;
    break;

  case MUM_POLYNOME: /* a0 + a1 * X +... */
    nbr_parametres = degre + 1;
    break;

  case MUM_GAUSSIENNE: /* a0 * exp( - ( X - a1 )² / a2 ) */
    nbr_parametres = 3;
    break;

  case MUM_THEORIQUE: /* a0 + a1 * modele_theorique */
    nbr_parametres = 2;
    break;

  case MUM_VICTORINE_3: /* a0 + a1 * X + a2 / X³ */
    nbr_parametres = 3;
    break;

  default:
    nbr_parametres = 0;
    (void) fprintf(stderr, "%s,%s l.%u : Modèle de µ(matrice) inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, modele);
    abort();
  }

  /* Si sur une référence, ajouter pour la normalisation automatique */
  if ( VRAI == sur_reference ) {
    nbr_parametres++;
  }

  return nbr_parametres;
}

/* ~-~-~-~-~-~-~-~- Recherche du modèle ~-~-~-~-~-~-~-~- */

/* -------- Initialisation de la matrice des MC / fonction µ(m), cas linéaire -------- */

static double XAN_InitMatriceLineaire(unsigned long nombre_points, double *table_X, double *table_Y, MAT_infos *infos,
				   Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  double norme = 1.0;
  int retour;

  switch ( infos->methode )
  {
  case MUM_DROITE:
    norme = MC_Polynome_PreparationMatrices( 1, table_X, table_Y, nombre_points, matrice_MDC, vecteur_constant );
    break;

  case MUM_VICTORINE_1:
    retour = XAN_MatriceVictorine_1( nombre_points, table_X, table_Y, matrice_MDC, vecteur_constant );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_MatriceVictorine_1() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return 0.0;
    }
    break;

  case MUM_VICTORINE_2:
    retour = XAN_MatriceVictorine_2( nombre_points, table_X, table_Y, matrice_MDC, vecteur_constant );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_MatriceVictorine_2() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return 0.0;
    }
    break;

  case MUM_POLYNOME:
    norme = MC_Polynome_PreparationMatrices( infos->degre, table_X, table_Y, nombre_points, matrice_MDC, vecteur_constant );
    break;

  case MUM_VICTORINE_3:
    retour = XAN_MatriceVictorine_3( nombre_points, table_X, table_Y, matrice_MDC, vecteur_constant );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_MatriceVictorine_3() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return 0.0;
    }
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Méthode inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->methode );
    return 0.0;
  }

  /* Norme pour les coefficients des moindres carrés, après */
  return norme;
}

/* -------- Résolution dans le cas d'un modèle linéaire sur une référence -------- */

static double XAN_TrouveLineaireReference(Graphe *brut, MAT_infos *infos, Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  double norme = 1.0, y_reference;
  double *b, *c, *d; /* Table des coefficients des splines / référence */
  double *x_brut, *y_brut;
  int retour;
  unsigned long i, j;

/*   printf("%s : référence = %s\n", __FUNCTION__, infos->reference->nom ); */

  /* Coefficients de la spline / référence */
  retour = SPL_tt_prep( infos->reference, &b, &c, &d );
  if ( retour < 0 ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Le lissage de la référence a échoué\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    LIS_ERREUR( retour, infos->reference );
    return 0.0;
  }

  /* Remplissage de la matrice
     Ce n'est pas le plus efficace en temps de calcul (deux passes), mais plus simple...
  */
  /* 1. Partie du modèle µ(m) */
  norme = XAN_InitMatriceLineaire( brut->nombre_point, brut->x, brut->y, infos, matrice_MDC, vecteur_constant );
  if ( 0.0 == norme ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_InitMatriceLineaire()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( b ); free( c ); free( d );
    return 0.0;
  }

  /* 2. Attention, la matrice ici n'est PAS symétrique ! */
  retour = MAT_ChangerType( matrice_MDC, MATRICE_QUELCONQUE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans MAT_ChangerType()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( b ); free( c ); free( d );
    return 0.0;
  }

  /* 3. Initialisation à zéro du vecteur constant */
  for ( i = 0 ; i < vecteur_constant->n_lignes ; i++ ) {
    vecteur_constant->valeurs[ i ] = 0.0;
  }

  switch ( infos->methode )
  {
  case MUM_CONSTANTE:
    OPTION_A_FAIRE( "Ôter une constante calée sur une référence" );
    break;

  case MUM_DROITE:
    for ( i = 0, x_brut = brut->x, y_brut = brut->y ; i < brut->nombre_point ; i++, x_brut++, y_brut++ ) {
      double x/*, x_carre*/;

      x = norme * *x_brut;
      /* x_carre = x * x; */

      /* Valeur de la référence en ce point ; - pour avoir le bon signe dans les sommes */
      y_reference = - SPL_tt_calc( infos->reference, *x_brut, b, c, d );

      /* Troisième ligne */
      matrice_MDC->valeurs[ 6 ] += *y_brut;
      matrice_MDC->valeurs[ 7 ] += *y_brut * x;
      matrice_MDC->valeurs[ 8 ] += -( *y_brut * *y_brut );

      /* Vecteur constant */
      vecteur_constant->valeurs[ 0 ] += y_reference;
      vecteur_constant->valeurs[ 1 ] += y_reference * x;
      vecteur_constant->valeurs[ 2 ] += y_reference * *y_brut;
    }

    /* Complétion par symétrie */
    matrice_MDC->valeurs[ 2 ] = - matrice_MDC->valeurs[ 6 ];
    matrice_MDC->valeurs[ 5 ] = - matrice_MDC->valeurs[ 7 ];
    break;

  case MUM_VICTORINE_1:
    for ( i = 0, x_brut = brut->x, y_brut = brut->y ; i < brut->nombre_point ; i++, x_brut++, y_brut++ ) {
      double x, x_cube;

      x = *x_brut;
      x_cube = x * x * x;
      if ( 0.0 == x_cube ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : X nul [point %lu/%lu, Y=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, i + 1, brut->nombre_point, *y_brut);
#endif
	continue;
      }
      x_cube = 1.0 / x_cube;

      /* Valeur de la référence en ce point ; - pour avoir le bon signe dans les sommes */
      y_reference = - SPL_tt_calc( infos->reference, x, b, c, d );

      /* Troisième ligne */
      matrice_MDC->valeurs[ 6 ] += *y_brut;
      matrice_MDC->valeurs[ 7 ] += *y_brut * x_cube;
      matrice_MDC->valeurs[ 8 ] += -( *y_brut * *y_brut );

      /* Vecteur constant */
      vecteur_constant->valeurs[ 0 ] += y_reference;
      vecteur_constant->valeurs[ 1 ] += y_reference * x_cube;
      vecteur_constant->valeurs[ 2 ] += y_reference * *y_brut;
    }

    /* Complétion par symétrie */
    matrice_MDC->valeurs[ 2 ] = - matrice_MDC->valeurs[ 6 ];
    matrice_MDC->valeurs[ 5 ] = - matrice_MDC->valeurs[ 7 ];
    break;

  case MUM_VICTORINE_2:
    for ( i = 0, x_brut = brut->x, y_brut = brut->y ; i < brut->nombre_point ; i++, x_brut++, y_brut++ ) {
      double x, x_cube, x_4;

      x = *x_brut;
      x_cube = x * x * x;
      if ( 0.0 == x_cube ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : X nul [point %lu/%lu, Y=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, i + 1, brut->nombre_point, *y_brut);
#endif
	continue;
      }
      x_cube = 1.0 / x_cube;
      x_4 = x_cube / x;

      /* Valeur de la référence en ce point ; - pour avoir le bon signe dans les sommes */
      y_reference = - SPL_tt_calc( infos->reference, x, b, c, d );

      /* Troisième ligne */
      matrice_MDC->valeurs[ 12 ] += *y_brut;
      matrice_MDC->valeurs[ 13 ] += *y_brut * x_cube;
      matrice_MDC->valeurs[ 14 ] += *y_brut * x_4;
      matrice_MDC->valeurs[ 15 ] += -( *y_brut * *y_brut );

      /* Vecteur constant */
      vecteur_constant->valeurs[ 0 ] += y_reference;
      vecteur_constant->valeurs[ 1 ] += y_reference * x_cube;
      vecteur_constant->valeurs[ 2 ] += y_reference * x_4;
      vecteur_constant->valeurs[ 3 ] += y_reference * *y_brut;
    }

    /* Complétion par symétrie */
    matrice_MDC->valeurs[  3 ] = - matrice_MDC->valeurs[ 12 ];
    matrice_MDC->valeurs[  7 ] = - matrice_MDC->valeurs[ 13 ];
    matrice_MDC->valeurs[ 11 ] = - matrice_MDC->valeurs[ 14 ];
    break;

  case MUM_VICTORINE_3:
    for ( i = 0, x_brut = brut->x, y_brut = brut->y ; i < brut->nombre_point ; i++, x_brut++, y_brut++ ) {
      double x, x_cube;

      x = *x_brut;
      x_cube = x * x * x;
      if ( 0.0 == x_cube ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : X nul [point %lu/%lu, Y=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, i + 1, brut->nombre_point, *y_brut);
#endif
	continue;
      }
      x_cube = 1.0 / x_cube;

      /* Valeur de la référence en ce point ; - pour avoir le bon signe dans les sommes */
      y_reference = - SPL_tt_calc( infos->reference, x, b, c, d );

      /* Troisième ligne */
      matrice_MDC->valeurs[ 12 ] += *y_brut;
      matrice_MDC->valeurs[ 13 ] += *y_brut * x;
      matrice_MDC->valeurs[ 14 ] += *y_brut * x_cube;
      matrice_MDC->valeurs[ 15 ] += -( *y_brut * *y_brut );

      /* Vecteur constant */
      vecteur_constant->valeurs[ 0 ] += y_reference;
      vecteur_constant->valeurs[ 1 ] += y_reference * x;
      vecteur_constant->valeurs[ 2 ] += y_reference * x_cube;
      vecteur_constant->valeurs[ 3 ] += y_reference * *y_brut;
    }

    /* Complétion par symétrie */
    matrice_MDC->valeurs[  3 ] = - matrice_MDC->valeurs[ 12 ];
    matrice_MDC->valeurs[  7 ] = - matrice_MDC->valeurs[ 13 ];
    matrice_MDC->valeurs[ 11 ] = - matrice_MDC->valeurs[ 14 ];
    break;

  case MUM_POLYNOME:
    for ( i = 0, x_brut = brut->x, y_brut = brut->y ; i < brut->nombre_point ; i++, x_brut++, y_brut++ ) {
      double x;

      x = norme * *x_brut;

      /* Valeur de la référence en ce point ; - pour avoir le bon signe dans les sommes */
      y_reference = - SPL_tt_calc( infos->reference, *x_brut, b, c, d );

      matrice_MDC->valeurs[ matrice_MDC->n_colonnes * (infos->degre + 1) ] += *y_brut; /* Terme en µ */
      vecteur_constant->valeurs[ 0 ] += y_reference;
      for ( j = 1 ; j <= (unsigned int) infos->degre ; j++ ) {
	matrice_MDC->valeurs[ matrice_MDC->n_colonnes * (infos->degre + 1) + j ] += *y_brut * x; /* Termes en µ X^n */
	vecteur_constant->valeurs[ j ] += y_reference * x;
	x *= (norme * *x_brut);
      }
      matrice_MDC->valeurs[ matrice_MDC->n_colonnes * (infos->degre + 1) + (infos->degre + 1) ] += *y_brut * *y_brut; /* Terme en µ² */
      vecteur_constant->valeurs[ infos->degre + 1 ] += y_reference * *y_brut;
    }

    /* Complétion par symétrie */
    for ( j = 0 ; j <= (unsigned int) infos->degre ; j++ ) {
      met_element( matrice_MDC, j, infos->degre + 1, -lit_element( matrice_MDC, infos->degre + 1, j ) );
    }
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Méthode inconnue / non implémentée [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->methode );
    norme = 0.0;
  }

  free( b ); free( c ); free( d );

  /* Calcul réussi, pas de problème particulier */
  return norme;
}

/* -------- Résolution dans le cas d'un modèle linéaire -------- */

static double XAN_TrouveModeleLineaire(Graphe *brut, MAT_infos *infos, Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  double *debut_x, *debut_y;
  double x_min, x_max, norme = 1.0;
  unsigned long nbr_pts, premier_point;
  /*  Matrice v_parametres; */

  /* Bornes en X pour réaliser l'ajustement */
  XAN_BornesIntervalle( brut, infos, &x_min, &x_max );

  /* On repère les points qui conviennent [Suppose le graphe trié par X croissants] */
  nbr_pts = GRF_intervalle( brut, x_min, x_max, &premier_point );
  if ( nbr_pts < matrice_MDC->n_lignes ) { /* Aucun point ne convient (ou trop peu) ! */
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : trop peu de points (intervalle : [%f, %f], %lu points mais %u paramètres)\n",
		   __FILE__, __FUNCTION__, __LINE__, x_min, x_max, nbr_pts, matrice_MDC->n_lignes);
#endif
    (void) BI_ERR_alerte(NULL, &ERR_TropPeuPoints, VRAI, x_min, x_max);
    return 0.0;
  }
  debut_x = brut->x + premier_point;
  debut_y = brut->y + premier_point;

  /* Matrice des paramètres : la table des coefficients est supposée connue */
  /* v_parametres.type = MATRICE_QUELCONQUE; */
  /* v_parametres.n_lignes = matrice_MDC->n_lignes; */
  /* v_parametres.n_colonnes = 1; */
  /* v_parametres.valeurs = infos->coefficients; */

  /* Remplissage de la matrice des moindres carrés */
  norme = XAN_InitMatriceLineaire( nbr_pts, debut_x, debut_y, infos, matrice_MDC, vecteur_constant );
  if ( 0.0 == norme ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_InitMatriceLineaire()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  /* Tout s'est bien passé */
  return norme;
}

/* -------- Calcul de la matrice, dans le cas d'une victorine simple -------- */

static int XAN_MatriceVictorine_1(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  unsigned long i;
  boolean erreur = FAUX;

  met_element( matrice_MDC, 0, 0, (double) nbr_points );

  for ( i = 0 ; i < nbr_points ; i++ ) {
    double x, x_cube, y;

    x = table_X[ i ];
    x_cube = x * x * x;
    if ( x_cube == 0.0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : X = 0 [victorine 1] pour i =%lu [Y=%g]\n",
		     __FILE__, __FUNCTION__, __LINE__, i, table_Y[ i ]);
#endif
      if ( FAUX == erreur ) { /* On n'affiche le message qu'une fois... */
	EXF_ERREUR(DIV_ZERO, "x nul", txt_err_calc_annule);
	erreur = VRAI;
      }
      continue;
    }
    x_cube = 1.0 / x_cube;

    matrice_MDC->valeurs[ 1 ] += x_cube;
    matrice_MDC->valeurs[ 2 ] += x_cube * x_cube;

    y = table_Y[ i ];
    vecteur_constant->valeurs[ 0 ] += y;
    vecteur_constant->valeurs[ 1 ] += y * x_cube;
  }

  /* Tout s'est bien passé */
  return 0;
}

static int XAN_MatriceVictorine_2(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  unsigned long i;
  boolean erreur = FAUX;

  matrice_MDC->valeurs[ 0 ] = (double) nbr_points;

  for ( i = 0 ; i < nbr_points ; i++ ) {
    double x, x_cube, x_4, y;

    x = table_X[ i ];
    x_cube = x * x * x;
    if ( x_cube == 0.0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : X = 0 [victorine 1] pour i =%lu [Y=%g]\n",
		     __FILE__, __FUNCTION__, __LINE__, i, table_Y[ i ]);
#endif
      if ( FAUX == erreur ) { /* On n'affiche le message qu'une fois... */
	EXF_ERREUR(DIV_ZERO, "x nul", txt_err_calc_annule);
	erreur = VRAI;
      }
      continue;
    }
    x_cube = 1.0 / x_cube; /* 1/x³ */
    x_4 = x_cube / x;      /* 1/x^4 */

    matrice_MDC->valeurs[ 1 ] += x_cube;
    matrice_MDC->valeurs[ 2 ] += x_cube * x_cube; /* 1/x^6 */
    matrice_MDC->valeurs[ 3 ] += x_4;
    matrice_MDC->valeurs[ 4 ] += x_cube * x_4 ; /* 1/x^7 */
    matrice_MDC->valeurs[ 5 ] += x_4 * x_4; /* 1/x^8 */

    y = table_Y[ i ];
    vecteur_constant->valeurs[ 0 ] += y;
    vecteur_constant->valeurs[ 1 ] += y * x_cube;
    vecteur_constant->valeurs[ 2 ] += y * x_4;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Victorine mixte A + B*x + C/x^3 -------- */

static int XAN_MatriceVictorine_3(unsigned long nbr_points, double *table_X, double *table_Y, Matrice *matrice_MDC, Matrice *vecteur_constant)
{
  unsigned long i;
  boolean erreur = FAUX;

  matrice_MDC->valeurs[ 0 ] = (double) nbr_points;

  for ( i = 0 ; i < nbr_points ; i++ ) {
    double x, x_carre, x_cube, y;

    x = table_X[ i ];
    x_carre = x * x;
    x_cube = x_carre * x;
    if ( x_cube == 0.0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : X = 0 [victorine 1] pour i =%lu [Y=%g]\n",
		     __FILE__, __FUNCTION__, __LINE__, i, table_Y[ i ]);
#endif
      if ( FAUX == erreur ) { /* On n'affiche le message qu'une fois... */
	EXF_ERREUR(DIV_ZERO, "x nul", txt_err_calc_annule);
	erreur = VRAI;
      }
      continue;
    }
    x_cube = 1.0 / x_cube; /* 1/x³ */

    matrice_MDC->valeurs[ 1 ] += x;
    matrice_MDC->valeurs[ 2 ] += x_carre ;
    matrice_MDC->valeurs[ 3 ] += x_cube; /* 1/x³ */
    matrice_MDC->valeurs[ 4 ] += 1.0 / x_carre; /* 1/x², suppose que si x_carre = 0.0, alors x_cube = 0.0, sûr avec le calcul précédent */
    matrice_MDC->valeurs[ 5 ] += x_cube * x_cube; /* 1/x^6 */

    y = table_Y[ i ];
    vecteur_constant->valeurs[ 0 ] += y;
    vecteur_constant->valeurs[ 1 ] += y * x;
    vecteur_constant->valeurs[ 2 ] += y * x_cube;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Calcul du modèle ~-~-~-~-~-~-~-~- */

/* -------- Les coefficients étant connus, on calcule le modèle de µ(m) -------- */

static int XAN_CalculerModele(Graphe *brut, MAT_infos *infos, boolean avec_erreurs __attribute__((__unused__)))
{
  int retour = -1;

  /* On vérifie que le modèle existe bien */
  if ( NULL == infos->fond_continu ) {
    /* Sinon, on le crée : graphe isolé avec les X du brut */
    infos->fond_continu = GRF_nv_TableX( NULL, GR_MUMATRICE, GRF_SERIE, NULL, VRAI, brut->x, brut->nombre_point, FAUX );
    if ( NULL == infos->fond_continu ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer le modèle\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }
    infos->fond_continu->x_min = brut->x_min;
    infos->fond_continu->x_max = brut->x_max;
  }

  /* On remplit */
  switch ( infos->methode )
  {
  case MUM_CONSTANTE:
    retour = GRF_GenerePolynome( infos->fond_continu, FAUX, 0, infos->coefficients, NULL );
    break;

  case MUM_DROITE:
    retour = GRF_GenerePolynome( infos->fond_continu, FAUX, 1, infos->coefficients, NULL );
    break;

  case MUM_VICTORINE_1:
    retour = XAN_CalculeVictorine( infos->fond_continu, infos->coefficients[ 0 ], 0.0, infos->coefficients[ 1 ], 0.0 );
    break;

  case MUM_VICTORINE_2:
    retour = XAN_CalculeVictorine( infos->fond_continu, infos->coefficients[ 0 ], 0.0, infos->coefficients[ 1 ], infos->coefficients[ 2 ] );
    break;

  case MUM_POLYNOME:
    retour = GRF_GenerePolynome( infos->fond_continu, FAUX, infos->degre, infos->coefficients, NULL );
    break;

  case MUM_GAUSSIENNE:
    OPTION_A_FAIRE( "Fond continu gaussien" );
    break;

  case MUM_THEORIQUE:
    OPTION_A_FAIRE( "Fond continu calculé exterieurement" );
    break;

  case MUM_VICTORINE_3:
    retour = XAN_CalculeVictorine( infos->fond_continu, infos->coefficients[ 0 ], infos->coefficients[ 1 ], infos->coefficients[ 2 ], 0.0 );
    break;
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul du modèle\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }
  return retour;
}

static int XAN_CalculeVictorine(Graphe *modele, double cte, double coef_X, double coef_X3, double coef_X4)
{
  double *x, *y;
  double x_cube;
  unsigned long i;

  for ( i = 0, x = modele->x, y = modele->y ; i < modele->nombre_point ; i++, x++, y++ ) {
    *y = cte;
    *y += coef_X * *x;

    x_cube = *x * *x * *x;
    if ( x_cube != 0.0 ) {
      x_cube = 1.0 / x_cube;
      *y += coef_X3 * x_cube;

      x_cube *= 1.0 / *x;
      *y += coef_X4 * x_cube;
    }
  }

  GRF_bornes( modele, NE_PAS_FAIRE, DESORDONNE );
  return 0;
}

/* ------------- Soustraction effective de la contribution de la matrice ------------------\
 |   Ne renvoie rien
 |   Suppose que les coefficients de la fonction ont déjà été définis                      |
 |           que les X sont déjà installés dans traite et sont classés par ordre croissant |
 |
 | brut : le spectre brut
 | traite : le spectre contenant (brut - mu de la matrice)
 | infos  : les paramètres de calcul de la chose
 \---------------------------------------------------------------------------------------- */

void MAT_soustrait(Graphe *brut, Graphe *traite, MAT_infos *infos)
{
  int retour;

  if (brut   == NULL) return;
  if (traite == NULL) return;
  if (infos  == NULL) return;

  if ( brut->nombre_point != traite->nombre_point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Les deux graphes n'ont pas le même nombre de points [%lu / %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, brut->nombre_point, traite->nombre_point );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }

  /* On copie le brut dans le resultat */
  memcpy( traite->y, brut->y, sizeof( double ) * brut->nombre_point );
  traite->y_min = brut->y_min;
  traite->y_max = brut->y_max;

  /* Si référence : coefficient pour la normalisation en plus */
  if ( infos->reference != NULL ) {
    unsigned int nbr_parametres;
    double coef;

    nbr_parametres = XAN_NombreParametres( infos->methode, VRAI, infos->degre );
    coef = infos->coefficients[ nbr_parametres - 1 ]; /* Le dernier paramètre est le coefficient multiplicatif */

    GRF_FacteurMultiplicatif( traite, coef, AXE_Y, NULL );
  }

  /* On fait la soustraction */
  retour = GRF_SoustraireGraphe( traite, infos->fond_continu, VRAI, NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_SoustraireGraphe()\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }
}

/* -------- Calcul du fond continu, compte-tenu du graphe expérimental [brut] et du modèle choisi [infos] -------- *
 |   Suppose que le graphe est trié selon les X croissants                                 |
 \---------------------------------------------------------------------------------------- */

int XAN_MUM_CalculerFond(struct graphe *brut, MAT_infos *infos)
{
  unsigned int nbr_parametres;

  if ( ( NULL == brut ) || ( NULL == infos ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramètres passés inutilisables [NULL]\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -15;
  }

  /* On libère les anciens coefficients */
  free( infos->coefficients ); infos->coefficients = NULL;

  /* Combien de paramètres sont utilisés dans le modèle ? */
  nbr_parametres = XAN_NombreParametres( infos->methode, infos->reference != NULL, infos->degre );

  /* On crée les nouveaux coefficients */
  infos->coefficients = (double *) malloc( sizeof( double ) * nbr_parametres );
  if ( NULL == infos->coefficients ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante / table des coefficients\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  if ( XAN_ModeleEstLineaire( infos->methode ) == VRAI ) {
    double norme = 1.0;
    Matrice *matrice_MDC, *vecteur_constant;
    Matrice v_parametres;

    /* Création du vecteur constant et de la matrice des moindres carrés */
    vecteur_constant = creer_matrice( nbr_parametres, 1, MATRICE_QUELCONQUE );
    if ( NULL == vecteur_constant ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante / matrice des coefficients constants\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -3;
    }
    matrice_MDC = creer_matrice( nbr_parametres, nbr_parametres, MATRICE_SYMETRIQUE );
    if ( NULL == matrice_MDC ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante / matrice des moindres carrés\n",
		     __FILE__, __FUNCTION__, __LINE__);
      effacer_matrice( vecteur_constant );
      return -2;
    }

    /* Initialisation des matrices */
    if ( NULL == infos->reference ) {
      norme = XAN_TrouveModeleLineaire( brut, infos, matrice_MDC, vecteur_constant );
    } else {
      norme = XAN_TrouveLineaireReference( brut, infos, matrice_MDC, vecteur_constant );
    }
    if ( 0.0 == norme ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans XAN_TrouveModeleLineaire()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      effacer_matrice( matrice_MDC );
      effacer_matrice( vecteur_constant );
      return -10;
    }

    /* Vecteur des paramètres : la table des coefficients est supposée connue */
    v_parametres.type = MATRICE_QUELCONQUE;
    v_parametres.n_lignes = nbr_parametres;
    v_parametres.n_colonnes = 1;
    v_parametres.valeurs = infos->coefficients;

    /* Résolution de l'équation des moindres carrés (cas linéaire) */
    MC_Lineaire_Resolution( matrice_MDC, vecteur_constant, &v_parametres );
    if ( norme != 1.0 ) { /* On tient compte d'une normalisation éventuelle */
      double X;
      unsigned int i, i_max;

      i_max = v_parametres.n_lignes;
      if ( infos->reference != NULL ) /* Attention, le coef. de normalisation ne doit pas être corrigé */
	i_max--;
      X = 1.0;
      for ( i = 1 ; i < i_max ; i++ ) {
	X *= norme;
	v_parametres.valeurs[ i ] *= X;
      }
    }
#ifdef _DEVELOPPE_
    {
      unsigned int i;
      printf("%s : Coefficients trouvés = ( %g", __FUNCTION__, v_parametres.valeurs[ 0 ] );
      for ( i = 1 ; i < v_parametres.n_lignes ; i++ ) {
	printf(" , %g", v_parametres.valeurs[ i ]);
      }
      printf(" )\n");
    }
#endif

    /* Plus besoin des matrices */
    effacer_matrice( matrice_MDC );
    effacer_matrice( vecteur_constant );
  } else {
    if ( MUM_CONSTANTE == infos->methode ) {
      unsigned long i, nombre = 0;
      double moyenne = 0.0;
      double x_min, x_max;

      /* Bornes en X pour réaliser l'ajustement */
      XAN_BornesIntervalle( brut, infos, &x_min, &x_max );

      for ( i = 0 ; i < brut->nombre_point ; i++ ) {
	if ( ( brut->x[ i ] >= x_min ) && ( brut->x[ i ] <= x_max ) ) {
	  moyenne += brut->y[ i ];
	  nombre++;
	}
      }

      if ( nombre > 0 ) {
	infos->coefficients[ 0 ] = moyenne / (double) nombre;
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Aucun point ne convient ! {E1=%g, intervalle [%g, %g], graphe de %g à %g}\n",
		       __FILE__, __FUNCTION__, __LINE__, infos->e_1, x_min, x_max, brut->x_min, brut->x_max);
	infos->coefficients[ 0 ] = 0.0;
	return -20;
      }
      printf("%s : constante trouvée = %g\n", __FUNCTION__, infos->coefficients[ 0 ] );
    } else {
      OPTION_A_FAIRE( "Modèle non-linéaire" );
      return -1; /* Problème ! */
    }
  }

  /* On a l'opposé des coefficients [?] si on travaille sur une référence */
/*   if ( NULL != infos->reference ) { */
/*   } */

  /* Maintenant que l'on a les paramètres du modèle, on peut calculer le modèle de µ(m) */
  XAN_CalculerModele( brut, infos, FAUX );

  /* Le calcul s'est bien passé */
  return 0;
}

/* -------- Calcul du résidu : différence entre le fond continu calculé et le spectre expérimental -------- */

double XAN_MUM_CalculerResidu(const Graphe *brut, const Graphe *resultat, const MAT_infos *infos, unsigned long *ddl)
{
  unsigned long i, nombre_points_valables = 0;
  double x_min, x_max, *x;
  double SCE = 01.0;

  if ( NULL == infos ) return -1;

  /* L'intervalle des valeurs autorisées */
  x_max = infos->e_1;
  if ( UN_POINT == infos->mode ) {
    x_min = brut->x_min;
  } else {
    x_min = infos->e_1p;
  }

  if ( NULL == resultat ) {
    /* On fait la comparaison brut/modèle */
    if ( ( NULL == brut ) || ( NULL == infos->fond_continu ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Paramètres transmis insuffisants\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1.0;
    }

    for ( i = 0, x = brut->x ; i < brut->nombre_point ; i++, x++ ) {
      if ( ( *x >= x_min ) && ( *x <= x_max ) ) {
	double ecart;

	ecart = brut->y[ i ] - infos->fond_continu->y[ i ];
	SCE += ecart * ecart;
	nombre_points_valables++;
      }
    }
  } else {
    /* Le résultat étant la soustraction, autant l'utiliser directement ! */
    for ( i = 0, x = resultat->x ; i < resultat->nombre_point ; i++, x++ ) {
      if ( ( *x >= x_min ) && ( *x <= x_max ) ) {
	SCE += resultat->y[ i ] * resultat->y[ i ];
	nombre_points_valables++;
      }
    }
  }

  /* On renvoie le résultat */
  if ( ddl != NULL ) *ddl = ( nombre_points_valables - XAN_NombreParametres( infos->methode, FAUX, infos->degre ) );
  return SCE;
}

/* -------- Amélioration du graphe résultat, après soustraction du fond continu -------- */

int XAN_MUM_PostSoustraction(struct graphe *brut, struct graphe *resultat, MAT_infos *infos __attribute__((__unused__)) )
{
  char *titre = NULL;

  if ( NULL == resultat ) return -1; /* Pas de graphe à améliorer ! */
  if ( NULL == brut ) return 0; /* Pas d'amélioration possible ! */

  /* On construit un titre adapté */
  if ( GR_BRUT == brut->type ) {
    /* On ôte le fond continu à un spectre XAS */
    titre = TXT_genere_chaine( "%s - "MU"m", brut->nom );
  } else {
    /* Autre cas : titre plus générique */
    titre = TXT_genere_chaine( txt_titre_generique, brut->nom );
  }

  /* On installe ce titre */
  GRF_titre( resultat, titre );
  free( titre ); titre = NULL;

  /* On adapte si nécessaire le type */
  if ( GR_BRUT == brut->type ) {
    /* On ôte le fond continu à un spectre XAS */
    GRF_ChangerType( resultat, GR_MOINSMAT );
  } else if ( GR_MOINSMAT == brut->type ) {
  } else {
    /* On garde le même type que l'original ! */
  }

  /* On propage les légendes, unités, couleurs,... : inchangées */
  GRF_CopierAffichage( resultat, brut, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_DESSIN );

  /* On propage les erreurs [en X]*/
  if ( brut->erreur_x != NULL ) { /* Il y a des erreurs en X */
    resultat->erreur_x = GRF_Copier( brut->erreur_x, NULL, VRAI ); /* Il n'y a pas eu de traitement sur les X */
    resultat->drapeaux.erreur_x = VRAI;
  }

  if ( ( brut->erreur_y != NULL ) && ( resultat->erreur_y != NULL ) ) { /* Il y a des erreurs en Y, qui n'ont pas encore été propagées */
    AVERTISSEMENT( "Erreurs en Y de %s non propagées dans le calcul !", brut->nom );
    /* Soustraction d'une constante = les écarts-types ne sont pas modifiés */
    resultat->erreur_y = GRF_Copier( brut->erreur_y, NULL, VRAI ); 
    resultat->drapeaux.erreur_y = VRAI;
  }

  /* Terminé */
  return 0;
}

/* -------- Proposer une valeur de E1 pour un spectre -------- */

double XAN_MUM_ProposerE1(struct graphe *brut)
{
  unsigned long i;

  if ( NULL == brut ) return -1.0;
  if ( 1 == brut->nombre_point ) return brut->x[ 0 ];

  /* Pour l'instant : méthode grossière, on choisit le premier X tel que la fonction soit croissante... */
  i = 1;
  while ( brut->y[ i ] <= brut->y[ i - 1 ] ) {
    i++;
    if ( i == brut->nombre_point ) {
      i--;
      break;
    }
  }

  /* Si on est là : on a trouvé le i */
  if ( i > 0 ) {
    return brut->x[ i ];
  }

  /* Si on est là : on essaye le seuil - 50 eV */
  return cherche_e0( brut, FAUX ) - 50.0;
}

/* -------- Soustraction du fond continu [déjà calculé] --------
   Suppose que brut & resultat aient le même nombre de points & les mêmes X
           que infos->fond_continu ait aussi ces X et ait été calculé

   Tous les graphes doivent être de type réel
 */

int XAN_MUM_SoustraireFond(struct graphe *brut, struct graphe *resultat, MAT_infos *infos)
{
  int retour;

  /* Contrôles préliminaires */
  if ( ( NULL == brut ) || ( NULL == resultat ) || ( NULL == infos ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramètre passé absurde [NULL]\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( brut->nombre_point != resultat->nombre_point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de points différent entre le brut [%s : %lu] et le résultat [%s : %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, brut->nom, brut->nombre_point, resultat->nom, resultat->nombre_point );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* Le résultat contient initialement les mêmes Y que le brut */
  memcpy( resultat->y, brut->y, sizeof( double ) * resultat->nombre_point );
  resultat->y_min = brut->y_min;
  resultat->y_max = brut->y_max;
  if ( brut->erreur_y != NULL ) {
    Graphe *erreurs_y;

    /* On copie aussi les incertitudes */
    if ( NULL == resultat->erreur_y ) {
      GRF_creer_erreurs( resultat, AXE_Y );
    }

    erreurs_y = resultat->erreur_y;
    if ( erreurs_y != NULL ) {
      memcpy( erreurs_y->y, brut->erreur_y->y, sizeof( double ) * resultat->nombre_point );
      memcpy( erreurs_y->z.ddl, brut->erreur_y->z.ddl, sizeof( int ) * resultat->nombre_point );
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer les erreurs en Y sur le résultat : elles seront ignorées !\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }
  } else {
    /* Il n'y a pas d'erreurs : on fait sauter celles qui pourraient déjà exister sur le résultat */
    GRF_SupprimerErreurs( resultat, AXE_Y );
  }

  /* S'il y a une référence : coefficient pour la normalisation en plus */
  if ( infos->reference != NULL ) {
    unsigned int nbr_parametres;
    double coef;

    nbr_parametres = XAN_NombreParametres( infos->methode, VRAI, infos->degre );
    coef = infos->coefficients[ nbr_parametres - 1 ]; /* Le dernier paramètre est le coefficient multiplicatif */

    GRF_FacteurMultiplicatif( resultat, coef, AXE_Y, NULL );
  }

  /* On fait la soustraction */
  retour = GRF_SoustraireGraphe( resultat, infos->fond_continu, VRAI, NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans la soustraction [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -5;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Soustraction du µ(matrice) pour un graphe, en propageant les erreurs, couleurs,... -------- */

Graphe *XAN_SoustraireMatrice(Graphe *brut, MAT_infos *infos, boolean recalcule)
{
  Graphe *traite = NULL;
  char *titre;

  if ( (brut == NULL) || (NULL == infos) ) return NULL;

  titre = TXT_genere_chaine("%s - "MU"m", brut->nom);
  traite = GRF_nouveau(brut, brut->nombre_point, GR_MOINSMAT, GR_NORMAL, titre);
  free(titre); /* Plus besoin */
  if ( NULL == traite ) {
    (void) fprintf(stderr, "XAN_SoustraireMatrice(%s, %p, %d) : impossible de créer le graphe résultat !",
		   brut->nom, (void *) infos, recalcule);
    return NULL; /* Problème */
  }

  /* On prépare le graphe qui contiendra le modèle */
  GRF_eff_isole( infos->fond_continu ); infos->fond_continu = NULL;

  traite->x_min = brut->x_min;        /* On prépare les X */
  traite->x_max = brut->x_max;
  memcpy(traite->x, brut->x, (size_t) sizeof(double) * brut->nombre_point);
  if ( VRAI == recalcule ) { /* On recalcule les coefficients pour ce spectre */
    int retour;

    retour = XAN_MUM_CalculerFond(brut, infos);
    if ( retour != 0 ) { /* Il y a eu un problème */
      GRF_efface(traite, VRAI, VRAI);
      (void) fprintf(stderr, "XAN_SoustraireMatrice(%s, %p, %d) : erreur dans MAT_calc, renvoie %d !",
		       brut->nom, (void *) infos, recalcule, retour);
      (void) BI_ERR_alerte(NULL, &ERR_ProblemeModele, FAUX, brut->nom);
      return NULL;
    }
  } else { /* On se contente de recalculer la fonction à soustraire */
    int retour;

    retour = XAN_CalculerModele( brut, infos, VRAI );
    if ( retour != 0 ) {
      GRF_efface(traite, VRAI, VRAI);
      (void) fprintf(stderr, "XAN_SoustraireMatrice(%s, %p, %d) : erreur dans MAT_calc, renvoie %d !",
		     brut->nom, (void *) infos, recalcule, retour);
      (void) BI_ERR_alerte(NULL, &ERR_ProblemeModele, FAUX, brut->nom);
      return NULL;
    }
  }
  MAT_soustrait(brut, traite, infos); /* On fait la soustraction */

  /* On s'occupe de l'historique */
  HIST_Titre(1, "Soustraction de la contribution de la matrice", "MUM");
  HIST_Texte(1, 0, "Spectre initial  : %s\n", brut->nom);
  HIST_Texte(1, 0, "Spectre obtenu   : %s\n", traite->nom);
  { /* On sauve les informations d'extraction */
    double residu = 0.0;
    unsigned long ddl;
    residu = XAN_MUM_CalculerResidu( brut, traite, infos, &ddl );
    HIST_Texte(3, 0, "R"e_AIGU"sidu : %g - soit une erreur statistique d'environ %g\n", residu, sqrt( residu / (double) ddl) );
  }
  HIST_Texte(2, 0, "M"e_AIGU"thode utilis"e_AIGU"e : ");

  switch ( infos->methode ) {
  case MUM_CONSTANTE:
    HIST_Texte(2, 0, " soustraction d'une constante, C\n");
    HIST_Table(4, 5, infos->coefficients, 1, 'C');
    break;

  case MUM_DROITE:
    HIST_Texte(2, 0, " soustraction d'une droite, A + B * X\n");
    HIST_Table(4, 5, infos->coefficients, 2, 'A');
    break;

  case MUM_VICTORINE_1:
    HIST_Texte(2, 0, " soustraction d'une victorine, A + B / X"TROIS_EXPOSANT"\n");
    HIST_Table(4, 5, infos->coefficients, 2, 'A');
    break;

  case MUM_VICTORINE_2:
    HIST_Texte(2, 0, " soustraction d'une victorine, A + B / X"TROIS_EXPOSANT" + C / X^4\n");
    HIST_Table(4, 5, infos->coefficients, 3, 'A');
    break;

  case MUM_POLYNOME:
    HIST_Texte(2, 0, " soustraction d'un polyn"o_CIRC"me, \n");
    HIST_Polynome(2, 4, 5, infos->coefficients, infos->degre, 'a');
    break;

  case MUM_THEORIQUE:
    break;

  case MUM_VICTORINE_3:
    HIST_Texte(2, 0, " soustraction d'une victorine, A + B * X + C / X"TROIS_EXPOSANT"\n");
    HIST_Table(4, 5, infos->coefficients, 3, 'A');
    break;
  }
  if ( infos->reference != NULL ) {
    HIST_Texte(2, 2, "  Optimis"e_AIGU" pour la r"e_AIGU"f"e_AIGU"rence %s\n", infos->reference->nom);
  } else {
    HIST_Texte(2, 2, "  Premier point de contr"o_CIRC"le, E1 : %f\n", infos->e_1);
    if (infos->mode == DEUX_POINTS)
      HIST_Texte(2, 2, "  Second point de contr"o_CIRC"le, E1' : %f\n", infos->e_1p);
  }

  return traite;
}

/* ~-~-~-~-~-~-~-~- Automatisation ~-~-~-~-~-~-~-~- */

int XAN_MUM_Automatisation(struct automatiser *lot)
{
  int retour;
  unsigned long i;
  Graphe *brut = NULL, *resultat = NULL;
  MAT_infos *infos_mum = NULL;

  if ( NULL == lot ) return 0;

  /* On récupère les conditions d'extraction */
  infos_mum = (MAT_infos *) lot->infos_necessaires;
  if ( NULL == infos_mum ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas d'infos d'extraction !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On traite chaque graphe de la série */
  for ( i = 0 ; i < lot->nombre_objets ; i++ ) {
    brut = lot->objets[ i ].graphe;
    if ( NULL == brut ) {
      AVERTISSEMENT( "Graphe %lu [0..%lu] du lot NULL", i, lot->nombre_objets - 1 );
      continue;
    }

    /* On efface le fond continu, car les X sont a priori différents */
    GRF_eff_isole( infos_mum->fond_continu ); infos_mum->fond_continu = NULL;

    /* On prépare le résultat [graphe isolé] */
    resultat = GRF_nv_TableX( NULL, GR_MOINSMAT, GRF_FCT_REELLE, NULL, VRAI, brut->x, brut->nombre_point, VRAI );
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de créer le graphe résultat [graphe %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
      continue;
    }

    /* En fonction de la méthode de calcul choisie */
    switch ( lot->methode )
    {
    case AUTOMATISE_METHODE_ROBOT:
      /* Même méthode, avec les mêmes coefficients */
      /* Pas de référence */
      infos_mum->reference = NULL;

      /* On calcule le fond continu avec les coefficients indiqués */
      retour = XAN_CalculerModele( brut, infos_mum, FAUX );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul du modèle [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }

      /* Puis on le soustrait */
      retour = XAN_MUM_SoustraireFond( brut, resultat, infos_mum );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans la soustraction du fond [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }
      break;

    case AUTOMATISE_METHODE_ADAPTE:
      /* Même méthode, avec les coefficients recalculés */
      /* Pas de référence */
      infos_mum->reference = NULL;

      /* On recalcule donc le fond en cherchant les coefficients */
      retour = XAN_MUM_CalculerFond( brut, infos_mum );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul du fond [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }

      /* Puis on le soustrait */
      retour = XAN_MUM_SoustraireFond( brut, resultat, infos_mum );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans la soustraction du fond [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }
      break;

    case AUTOMATISE_METHODE_OPTIMISE:
      /* La référence à utiliser */
      infos_mum->reference = lot->reference.graphe;

      /* Le reste est comme la méthode "adapte" */
      retour = XAN_MUM_CalculerFond( brut, infos_mum );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul du fond [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }
      retour = XAN_MUM_SoustraireFond( brut, resultat, infos_mum );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans la soustraction du fond [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( resultat ); resultat = NULL;
      }
      break;

    case AUTOMATISE_METHODE_REDEMANDE:
      GRF_eff_isole( resultat ); resultat = NULL;
      break;

    default:
      AVERTISSEMENT( "Méthode inconnue [%u]", lot->methode );
      GRF_eff_isole( resultat ); resultat = NULL;
    }

    /* Dans tous les cas, on réalise les adaptations post-soustraction */
    XAN_MUM_PostSoustraction( brut, resultat, infos_mum );

    /* On mémorise le résultat dans la table des résultats */
    lot->resultats[ i ].graphe = resultat;
    resultat = NULL;
  }

  /* On efface le dernier fond continu, inutile */
  GRF_eff_isole( infos_mum->fond_continu ); infos_mum->fond_continu = NULL;

  /* Terminé sans erreur */
  return 0;
}
